/*
 * -----------------------------------------------------------------
 * EDFScheduler.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package com.timesys;
import java.util.*;

public class EDFScheduler extends Scheduler
{
    static EDFScheduler defaultEDFScheduler = new EDFScheduler(); //null;
    static HashSet feasibilitySet = new HashSet();

    public EDFScheduler()
    {
    }

    private float getSchedulableLoad(Schedulable currentSchedulable)
    {
	float load = 0f;

	if (currentSchedulable.getReleaseParameters() != null)
	    {
		if (currentSchedulable.getReleaseParameters() instanceof PeriodicParameters)
		    {
			PeriodicParameters pp = (PeriodicParameters) currentSchedulable.getReleaseParameters();
			long cost, period;

			if ( (pp.getCost() != null) && (pp.getPeriod() != null))
			    {
				cost = pp.getCost().getMilliseconds() * 1000 * 1000;
				cost += pp.getCost().getNanoseconds();

				period = pp.getPeriod().getMilliseconds() * 1000 * 1000;
				period += pp.getPeriod().getNanoseconds();

				load += ( (cost * 1.0f) / (period * 1.0f) );
			    }

		    }
		else if (currentSchedulable.getReleaseParameters() instanceof SporadicParameters)
		    {
			SporadicParameters p = (SporadicParameters) currentSchedulable.getReleaseParameters();
			long  cost, period;

			if ((p.getCost() != null) && (p.getMinimumInterarrival() != null))
			    {
				cost = p.getCost().getMilliseconds() * 1000 * 1000;
				cost += p.getCost().getNanoseconds();

				period = p.getMinimumInterarrival().getMilliseconds() * 1000 * 1000;
				period += p.getMinimumInterarrival().getNanoseconds();

				load += ( (cost * 1.0f) / (period * 1.0f) );
			    }
		    }
		else
		    {
			/* Aperiodic not sporadic */
			AperiodicParameters p = (AperiodicParameters) currentSchedulable.getReleaseParameters();
			long  cost=0, period=1;

			if (p.getCost() != null)
			    {
				cost = p.getCost().getMilliseconds() * 1000 * 1000;
				cost += p.getCost().getNanoseconds();
			    }

			if (currentSchedulable.getProcessingGroupParameters() != null)
			    {
				ProcessingGroupParameters pg = currentSchedulable.getProcessingGroupParameters();
				if (pg.getPeriod() != null)
				    {
					period = pg.getPeriod().getMilliseconds() * 1000 * 1000;
					period += pg.getPeriod().getNanoseconds();
				    }
			    }
			else
			    {
				/* period is consider equal to deadline */
				if (p.getDeadline() != null)
				    {
					period = p.getDeadline().getMilliseconds() * 1000 * 1000;
					period += p.getDeadline().getNanoseconds();
				    }
			    }
			load += ((cost * 0.1f) / (period * 0.1f) );
		    }
	    }
	return load;
    }

    private float getLoad()
    {
	/* Simple capacity test */
	float load = 0f;

	Iterator iter = feasibilitySet.iterator();
	Schedulable currentSchedulable;

	while (iter.hasNext())
	    {
		currentSchedulable = (Schedulable)iter.next();
		load += getSchedulableLoad(currentSchedulable);
	    }

	return load;
    }

    public boolean isFeasible()
    {
	return (getLoad() <= 1.0f);
    }

    /**
      Return a pointer to an instance of <code>PriorityScheduler</code>.
      */
    public static EDFScheduler instance()
    {
	if (defaultEDFScheduler == null)
	    defaultEDFScheduler = new EDFScheduler();
	return defaultEDFScheduler;
    }

    protected boolean addToFeasibility(Schedulable schedulable)
    {
      RealtimeThread t = null;
      PriorityParameters p = null;
      synchronized (this)
	  {
	      if (feasibilitySet == null)
		  {
		      feasibilitySet = new HashSet();
		  }

	      if (!feasibilitySet.contains(schedulable))
		  {
		      feasibilitySet.add(schedulable);
		  }
	      return isFeasible();
	  }
    }

    boolean inFeasibility(Schedulable schedulable)
    {
	return feasibilitySet.contains(schedulable);
    }

    protected boolean removeFromFeasibility(Schedulable schedulable)
    {
	synchronized (this)
	    {
		if (feasibilitySet != null)
		    {
			feasibilitySet.remove(schedulable);
		    }
		return isFeasible();
	    }
    }

    public boolean setIfFeasible(Schedulable schedulable,
				    ReleaseParameters release,
				    MemoryParameters memory)
    {
	return setIfFeasible(schedulable, release, memory, schedulable.getProcessingGroupParameters());
    }

    public boolean setIfFeasible(Schedulable schedulable,
				    ReleaseParameters release,
				    MemoryParameters memory,
				    ProcessingGroupParameters group)
    {
	ReleaseParameters rp= schedulable.getReleaseParameters();
	MemoryParameters mp = schedulable.getMemoryParameters();
	ProcessingGroupParameters gp = schedulable.getProcessingGroupParameters();

//	float prevLoad = getSchedulableLoad(schedulable);
	float prevLoad = 0f;

	synchronized(Scheduler.class)
	    {
		synchronized(this)
		    {
			float load = getLoad();
			removeFromFeasibility(schedulable);
			if (release != null)
			    schedulable.setReleaseParameters(release);
			if (memory != null)
			    schedulable.setMemoryParameters(memory);

			if (group != null)
			    schedulable.setProcessingGroupParameters(group);

			System.out.println ("      PriorityScheduler: load " + load +
				", new load " + getSchedulableLoad(schedulable) +
				", prev load " + prevLoad);

			if ( (load + getSchedulableLoad(schedulable) - prevLoad) <= 1.0f)
			    {
				/* addToFeasibility will not add it twice */
				addToFeasibility(schedulable);
				return true;
			    }
			schedulable.setReleaseParameters(rp);
			schedulable.setMemoryParameters(mp);
			schedulable.setProcessingGroupParameters(gp);
			addToFeasibility(schedulable);
		    }
	    }
	return false;
    }

    public void fireSchedulable(Schedulable schedulable)
    {
	if (schedulable instanceof NoHeapRealtimeThread)
	    ((NoHeapRealtimeThread)schedulable).start();
	else if (schedulable instanceof RealtimeThread)
	    ((RealtimeThread)schedulable).start();
	else
	    {
//		System.out.println("EDFScheduler.java:FiringSchedulable");
		RealtimeThread t = new RealtimeThread(schedulable.getSchedulingParameters(),
						      schedulable.getReleaseParameters(),
						      schedulable.getMemoryParameters(),
						      null, // memory area
						      schedulable.getProcessingGroupParameters(),
						      schedulable);
		schedulable.getScheduler().addToFeasibility(t);
		//if (schedulable.getScheduler().isFeasible())
		t.start();
		// else
		// What should I do?
	    }
    }

    public String getPolicyName()
    {
	return new String("EDF (Earliest Deadline First)");
    }

}
