/*
 * -----------------------------------------------------------------
 * AsyncEvent.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;
import java.util.*;
import com.timesys.*;

/**
  An asynchronous event represents something that can happen, like a
  light turning red.  It can have a set of handlers associated with it, and
  when the event occurs, the handler is scheduled by the scheduler to which
  it holds a reference (see {@link AsyncEventHandler} and {@link Scheduler}).


  <p> A major motivator for this style of building events is that we expect
  to have lots of events and lots of event handlers.  An event handler is
  logically very similar to a thread, but it is intended to have
  a much lower cost (in both time and 
  space) -- assuming that a relatively small number of events are fired and in the
  process of being handled at once.
  <code>AsyncEvent.fire()</code> differs from a method call
  because the handler (a) has scheduling parameters and (b) is executed
   asynchronously.
*/
public class AsyncEvent
{
  // The set of handlers associated with this event 
  private Vector handlerSet = new Vector(1,1);
  
  // An 64 element array of Vectors (of AsyncEvent's)
  // for signals #'s 1-63 (index 0 is unused and there for clarity)
  // NOTE: the value of 64 should really be dependant upon _NSIG in
  // /usr/include/bits/signum.h
  private static final int NUM_HAPPENINGS = 64;
  private static Vector happenings[] = new Vector[NUM_HAPPENINGS];

  // Reserved Happenings. This is a mirror of the list in Dispatcher.h
  // Pratik: Added SIGUSR1 to reserved signals (17th Dec 2001)
  private static final int HAPPENING_RESERVED[] = {10, 32, 33, 34, 35, 36, 63};
    
  private int numBindings = 0;
  
  static
  { 
    // initialize the happenings array
    for(int i=0;i<NUM_HAPPENINGS;i++){
      happenings[i] = new Vector(1,1);
    }    
  }
  
  public AsyncEvent()
  {
  } 

  /**
     Add a handler to the set of handlers associated with this
     event. An <code>AsyncEvent</code> may have more than one associated handler.
     @param handler The new handler to add to the list of handlers already
     associated with this. If <code>handler</code> is null then nothing happens.
     <p>Since this affects the
     constraints expressed in the release parameters of the
     existing schedulable objects, this may change the
     feasibility of the current schedule.
  */   
  public void addHandler(AsyncEventHandler handler)
  {
      if(numBindings > 0){
	  if(MemoryArea.getMemoryArea(handler) != ImmortalMemory.instance()){
	      throw new MemoryAccessError("Cannot add a handler in non-immortal memory to a Timer or bindTo'd AsyncEvent");
	  }
      }
    handlerSet.addElement(handler);
  }

  /**
     Remove a handler from the set associated with
     this event.
     @param handler The handler to be disassociated from this. If null nothing happens. If not already
     associated with this then nothing happens.
  */
  public void removeHandler(AsyncEventHandler handler)
  {
    if(handler==null)
      return;
    else
      handlerSet.removeElement(handler);
  }

  /**
     Returns true if and only if this event is handled by this handler.
     @param target The handler to be tested to determine if it is
     associated with this. Returns false if <code>target</code> is null.
  */
  public boolean handledBy(AsyncEventHandler handler)
  {  
      if(numBindings > 0){
	  if(MemoryArea.getMemoryArea(handler) != ImmortalMemory.instance()){
	      throw new MemoryAccessError("Cannot add a handler in non-immortal memory to a Timer or bindTo'd AsyncEvent");
	  }
      }

    return(handlerSet.contains(handler));
  }
    
  /**
     Associate a new handler with this event, removing all existing
     handlers.
     <p>Since this affects the
     constraints expressed in the release parameters of the
     existing schedulable objects, this may change the
     feasibility of the current schedule.
     
     @param handler The new and only handler to be associated with this.
     If <code>handler</code> is null then no handler will be associated with this 
     (i.e., remove all handlers).
  */
  public void setHandler(AsyncEventHandler handler)
  {
      if(numBindings > 0){
	  if(MemoryArea.getMemoryArea(handler) != ImmortalMemory.instance()){
	      throw new MemoryAccessError("Cannot add a handler in non-immortal memory to a Timer or bindTo'd AsyncEvent");
	  }
      }
      
    handlerSet.clear();
    addHandler(handler);
  }
  
  
  /** 
      Create a {@link ReleaseParameters} block appropriate to the timing characteristics
      of this event.  The default is the most pessimistic: {@link AperiodicParameters}.
      This is typically called by code that is setting up a handler for this event
      that will fill in the parts of the release parameters that it knows
      the values for, like cost. 
  */
  public ReleaseParameters createReleaseParameters()
  {
    return new AperiodicParameters(null, null, null, null);
  }
  
  // does this really need to be synchronized??
  /**
     Binds this to an external event (a happening).  The meaningful values of 
     <code>happening</code> are implementation dependent. This 
     <code>AsyncEvent</code> is considered to have occurred whenever 
     the external event occurs.
     @param happening An implementation dependent value that binds
     this <code>AsyncEvent</code> to some external event.
     @exception UnknownHappeningException (a runtime exception) if
     if this event is not bound to the <code>happening</code> or if
     the happening string is not supported by the system.
  */
  public synchronized void bindTo(String happening)
  {
    // The string happening represents the signal number
    // that this event will be bound to. Linux signals range
    // from 1 to 63 inclusive (1-31 are the basic posix signals, 32-
    // 63 are the real-time signals). Of which 32 and 63 are reserved
    // for the system and cannot be bound to.

    int happnum;
    
    if(MemoryArea.getMemoryArea(this) != ImmortalMemory.instance()){
	throw new MemoryAccessError("Cannot bindTo an AsyncEvent in non-immortal memory");
    }
    else{
	for(int i=handlerSet.size()-1; i>=0; i--){
	    if(MemoryArea.getMemoryArea(handlerSet.elementAt(i)) != ImmortalMemory.instance()){
		throw new MemoryAccessError("Cannot bindTo an AsyncEvent with handlers in non-immortal memory");
	    }
	}
    }
    
    try{
      happnum = Integer.parseInt(happening);
    }
    catch(NumberFormatException e){
      //happening not happnin'
      throw new UnknownHappeningException("Unsupported happening");
    }
  
    if(happnum<1 || happnum >= NUM_HAPPENINGS){
      throw new UnknownHappeningException("Unsupported happening");
    }

    if(isReserved(happnum)){
      throw new UnknownHappeningException("System reserved happening");
    }

    //bind this to the event (only once).
    if(!happenings[happnum].contains(this)){
      happenings[happnum].addElement(this);
    }
    //install the event if it is the only one bound to this happening.
    //if(happenings[happnum].size()<=1)
    //installEvent(happnum);    

    numBindings++;

  }
  
  /** Removes a binding to an external event (a happening). The meaningful
      values of  <code>happening</code> are implementation dependent. 
      @param happening An implementation dependent value representing some
      external event to which this <code>AsyncEvent</code> is bound.
      @exception UnknownHappeningException if this <code>AsyncEvent</code> is
      not bound to the given <code>happening</code> or the given <code>happening</code>
      string is not supported by the system.
  */
  
  public synchronized void unbindTo(String happening) throws UnknownHappeningException
  {
    int happnum;
    try{
      happnum = Integer.parseInt(happening);
    }
    catch(NumberFormatException e){
      throw new UnknownHappeningException("Unsupported happening");
    }
    
    if(happnum<1 || happnum >= NUM_HAPPENINGS){
      throw new UnknownHappeningException("Unsupported happening");
    }

    if(isReserved(happnum)){
      throw new UnknownHappeningException("System reserved happening");
    }
    
    if(!happenings[happnum].contains(this)){
      throw new UnknownHappeningException("AsyncEvent not bound to this happening");
    }
    else{
      happenings[happnum].remove(this);
    }

    numBindings--;
  }

  private static boolean isReserved(int happnum)
  {
    for(int i=0; i<HAPPENING_RESERVED.length; i++){
      if(HAPPENING_RESERVED[i]==happnum){
	return true;
      }
    }
    return false;
  }

    /**
       For subclasses, Timer, in particular, since they are always "bindTo'd"
       in a sense, to the dispatcher, to call at construction time. This way, 
       all subsequent handlers will have to be in immortal memory. 
    */
    // package protection
    void indexNumBindings()
    {
	numBindings++;
    }
    /** 
	I don't really see this being used, but it's consistent
    */
    void decrementNumBindings()
    {
	numBindings--;
    }

  //private static native void initEvents();
  //private static native void installEvent(int signum);
  //private static native void uninstallEvent(int signum);
      
  /**
     Fire (schedule the <code>run()</code> methods of) the handlers associated with
     this event.
  */
  public synchronized void fire()
  {
    AsyncEventHandler h;
    int fc;

    for(int i=0;i<handlerSet.size();i++)
    {
      h = (AsyncEventHandler) handlerSet.elementAt(i);
      h.handle();
    }
  }
  
  // This function is called by native signal handling code
  // and schedules the associated handlers
  private static synchronized void eventHappened(int eventNum)
  {
    Vector eventSet;
    
    // itterate through the AsyncEvent vector associated with
    // the happening eventNum and run their fire methods.
    eventSet = happenings[eventNum];
        
    for(int i=0;i<eventSet.size();i++)
            ((AsyncEvent)eventSet.elementAt(i)).fire();
  }

  /* ROBBINS SEZ: temporary debugging test function: */
  //public static native void testreserved(int pid);
  
}
