/*
 * -----------------------------------------------------------------
 *  AsyncEventHandler.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;

import java.lang.*;
/**
  An asynchronous event handler encapsulates code that gets run at some 
  time after an {@link AsyncEvent} occurs.
  <p> It is essentially a {@link Runnable}
  with a set of parameter objects, making it very much like a
  {@link RealtimeThread}. The expectation is that
  there may be thousands of events, with corresponding handlers, averaging
  about one handler per event. The number of unblocked (i.e., scheduled) 
  handlers is expected to be relatively small.

  <p>It is guaranteed that multiple firings of an event handler will be
  serialized.  It is also guaranteed that (unless the handler explicitly
  chooses otherwise) for each firing of the handler, there will be one
  execution of the <code>handleAsyncEvent()</code> method.

  <p>There is no restriction on what handlers may do.  They may run for
  a long or short time, and they may block. (Note: blocked handlers may 
  hold system resources.)

  <p> Normally, handlers are bound to an execution context dynamically, 
  when their {@link AsyncEvent} occurs.
  This can introduce a (small) time penalty.  For critical handlers that can
  not afford the expense, and where this penalty is a problem, use a
  {@link BoundAsyncEventHandler}.
  */
public class AsyncEventHandler implements Schedulable
{
    // package protection
  Object countLock = new Object();
  int fireCount=0;

  // The thread that runs "this", if nonheap is true, then it is
  // a NoHeapRealtimeThread, if false then it is a RealtimeThread
  // this is package because BoundAsyncEventHandler needs to putz with it
  RealtimeThread HandlerThread;
  
  // The thread object needs to be allocated within a memory scope so that
  // the dispatcher does not leak memory
  // private ScopedMemory scope = new CTMemory(1024);

  // True means that the handler thread is noheap.
  private boolean nonheap = false;

  // Args to the RealtimeThread
  private SchedulingParameters scheduling = null;
  private ReleaseParameters release = null;
  private MemoryParameters memory = null;
  private MemoryArea area = null;
  private ProcessingGroupParameters group = null;
  
  // if set, this will be executed in lieu of 
  private Runnable logic = null;
  
  /**
    Create a handler whose {@link SchedulingParameters} are
    inherited from the current thread and does not have
    either {@link ReleaseParameters} or {@link MemoryParameters}.
  */
  public AsyncEventHandler()
  {
    // This thread sits around and does nothing except provide
    // the functionality for the Schedulable interface.
    try
      {
	RealtimeThread currt = RealtimeThread.currentRealtimeThread();
	HandlerThread = new RealtimeThread(currt.getSchedulingParameters(),
					   currt.getReleaseParameters(),
					   currt.getMemoryParameters(),
					   currt.getMemoryArea(),
					   currt.getProcessingGroupParameters(),
					   this);
      }
    catch(ClassCastException e)
      {
	HandlerThread = new RealtimeThread(null, null, null, null, null, this);	      
      }
  }
  
  /**
    Create a handler whose {@link SchedulingParameters} are
    inherited from the current thread and does not have
    either {@link ReleaseParameters} or {@link MemoryParameters}.
    @param logic The {@link Runnable} object whose <code>run</code>
    is executed by handleAsyncEvent.
  */
  public AsyncEventHandler(Runnable logic)
  {
      this.logic = logic;
      
    // This thread sits around and does nothing except provide
    // the functionality for the Schedulable interface.
    try
      {
	RealtimeThread currt = RealtimeThread.currentRealtimeThread();
	HandlerThread = new RealtimeThread(currt.getSchedulingParameters(),
					   currt.getReleaseParameters(),
					   currt.getMemoryParameters(),
					   currt.getMemoryArea(),
					   currt.getProcessingGroupParameters(),
					   this);
      }
    catch(ClassCastException e)
      {
	HandlerThread = new RealtimeThread(null, null, null, null, null, this);	      
      }
  }

    /**
      Create a handler with the specified parameters.
      @param release A {@link ReleaseParameters} object which will be associated with the
      constructed instance of this. If null this will have no {@link ReleaseParameters}.
      @param scheduling A {@link SchedulingParameters} object which will be associated with the
      constructed instance of this. If null this will be assigned the reference to the
      {@link SchedulingParameters} of the current thread.
      @param memory A {@link MemoryParameters} object which will be associated with the
      constructed instance of this. If null this will have no {@link MemoryParameters}.
      @param area The {@link MemoryArea} for this.  If null the memory area will be that of the
             current thread.
      @param group A {@link ProcessingGroupParameters} object to which this will be associated.  If
      null this will not be associated with any processing group.
      @param logic The {@link Runnable} object whose <code>run</code>
      is executed by handleAsyncEvent.
      */
  public AsyncEventHandler(SchedulingParameters scheduling,
			   ReleaseParameters release,
			   MemoryParameters memory,
			   MemoryArea area,
			   ProcessingGroupParameters group,
			   Runnable logic)
  {
    this.scheduling = scheduling;
    this.release = release;
    this.memory = memory;
    this.area = area;
    this.group = group;
    this.logic = logic;

    // This thread sits around and does nothing except provide
    // the functionality for the Schedulable interface.
    HandlerThread = new RealtimeThread(scheduling, release, memory,
				       area, group, this);
  }
  
    /**
      Create a handler whose parameters are
      inherited from the current thread, if it is a {@link RealtimeThread},
      or null otherwise.
      @param nonheap A flag meaning, when true, that this will have characteristics
      identical to a {@link NoHeapRealtimeThread}.  A false value means this will have
      characteristics identical to a {@link RealtimeThread}. If true and the current
      thread is <em>not</em> a {@link NoHeapRealtimeThread} or a {@link RealtimeThread} 
      executing within a {@link ScopedMemory} or {@link ImmortalMemory} 
      scope then an IllegalArgumentException is thrown.
      */
  public AsyncEventHandler(boolean nonheap)
  {
    Thread currentThread;
    RealtimeThread currRTThread;
        
    if(nonheap){
      currentThread = Thread.currentThread();

      // from rtsj:
      // if (nonheap) true and the current thread is not a NoHeapRealtimeThread or
      // a RealtimeThread executing within a ScopedMemory or ImmortalMemory
      // scope then an IlleagalArgumentException is thrown.
      

      if(currentThread instanceof RealtimeThread)
	{
	  currRTThread = (RealtimeThread) Thread.currentThread();
	  if((currRTThread instanceof NoHeapRealtimeThread) ||
	     (currRTThread.getMemoryArea() instanceof ScopedMemory ||
	      currRTThread.getMemoryArea() instanceof ImmortalMemory))
	    {
	      // were okay;
	      this.nonheap = nonheap;
	      scheduling = currRTThread.getSchedulingParameters();
	      area = currRTThread.getMemoryArea();
	      HandlerThread = new NoHeapRealtimeThread(scheduling, area);
	    }
	  else{
	    throw new IllegalArgumentException();
	  }
	}
      else{
	throw new IllegalArgumentException();
      }
    }
    else{
      HandlerThread = new RealtimeThread();
    }
  }
  
    /**
      Create a handler whose parameters are
      inherited from the current thread, if it is a {@link RealtimeThread},
      or null otherwise.
      @param nonheap A flag meaning, when true, that this will have characteristics
      identical to a {@link NoHeapRealtimeThread}.  A false value means this will have
      characteristics identical to a {@link RealtimeThread}. If true and the current
      thread is <em>not</em> a {@link NoHeapRealtimeThread} or a {@link RealtimeThread} 
      executing within a {@link ScopedMemory} or {@link ImmortalMemory} 
      scope then an IllegalArgumentException is thrown.
      @param logic The {@link Runnable} object whose <code>run</code>
      is executed by handleAsyncEvent.      
    */
  public AsyncEventHandler(boolean nonheap, Runnable logic)
  {
    Thread currentThread;
    RealtimeThread currRTThread;
        
    this.logic = logic;
    
    if(nonheap){
      currentThread = Thread.currentThread();

      // from rtsj:
      // if (nonheap) true and the current thread is not a NoHeapRealtimeThread or
      // a RealtimeThread executing within a ScopedMemory or ImmortalMemory
      // scope then an IlleagalArgumentException is thrown.
      

      if(currentThread instanceof RealtimeThread)
	{
	  currRTThread = (RealtimeThread) Thread.currentThread();
	  if((currRTThread instanceof NoHeapRealtimeThread) ||
	     (currRTThread.getMemoryArea() instanceof ScopedMemory ||
	      currRTThread.getMemoryArea() instanceof ImmortalMemory))
	    {
	      // were okay;
	      this.nonheap = nonheap;
	      scheduling = currRTThread.getSchedulingParameters();
	      area = currRTThread.getMemoryArea();
	      HandlerThread = new NoHeapRealtimeThread(scheduling, area);
	    }
	  else{
	    throw new IllegalArgumentException();
	  }
	}
      else{
	throw new IllegalArgumentException();
      }
    }
    else{
      HandlerThread = new RealtimeThread();
    }
  }

    /**
      Create a handler with the specified parameters.
      @param scheduling A {@link SchedulingParameters} object which will be associated with the
      constructed instance of this. If null this will be assigned the reference to the
      {@link SchedulingParameters} of the current thread.
      @param release A {@link ReleaseParameters} object which will be associated with the
      constructed instance of this.  If null this will have no {@link ReleaseParameters}.
      @param memory A {@link MemoryParameters} object which will be associated with the
      constructed instance of this. If null this will have no {@link MemoryParameters}.
      @param area The {@link MemoryArea} for this. Must be a reference to 
               a {@link ScopedMemory} or {@link ImmortalMemory} object if <code>noheap</code> is true.
      @param group A {@link ProcessingGroupParameters} object to which this will be associated.  If
      null this will not be associated with any processing group.
      @param nonheap A flag meaning, when true, that this will have characteristics
      identical to a {@link NoHeapRealtimeThread}.
      */
  public AsyncEventHandler(SchedulingParameters scheduling,
			   ReleaseParameters release,
			   MemoryParameters memory,
			   MemoryArea area,
			   ProcessingGroupParameters group,
			   boolean nonheap,
			   Runnable logic)
  {
    this.scheduling = scheduling;
    this.release = release;
    this.memory = memory;
    this.area = area;
    this.group = group;
    this.logic = logic;
    
    if(nonheap && ((area instanceof ScopedMemory) || (area instanceof ImmortalMemory))){
      this.nonheap = nonheap;
      HandlerThread = new NoHeapRealtimeThread(scheduling, release, memory,
					       area, group, this);
    }
    else{
      HandlerThread = new RealtimeThread(scheduling, release, memory,
					 area, group, this);
    }
  }


  // handler logic gets implemented by the user by overriding this method.
  /** 
      If this handler was constructed using a seperate <code>Runnable</code>
      logic object, then that <code>Runnable</code> object's <code>run</code>
      method is called; This method will be invoked repeatedly while
      fireCount is greater than zero.
  */
    public void handleAsyncEvent(){
	if(logic != null){
	    logic.run();
	}
    }
  
  // gets called from AsyncEvent to start the handler
  // package level protection, user code should not futz with this.
  void handle()
  {

    if(getAndIncrementPendingFireCount()==0)
    {

      if(nonheap){
	HandlerThread = new NoHeapRealtimeThread(scheduling, release, memory,
						 area, group, this);
      }
      else{
	HandlerThread = new RealtimeThread(scheduling, release, memory,
					   area, group, this);
      }

      HandlerThread.setDaemon(false);

      HandlerThread.start();
    }    
  }
  

    /** Return the number of pending executions of this handler 
	@return The pending fire count.
    */
    protected final synchronized int getPendingFireCount()
    {
	return(fireCount);
    }

  /** 
      Atomically set to zero the number of pending executions
      of this handler and returns the value from before it was
      cleared.  This is used in handlers that can handle multiple
      firings and that want to collapse them together.  The
      general form for using this is:<pre>
      public void handleAsyncEvent() {
      int fireCount = getAndClearPendingFireCount();
      &lt;handle the events>
      }</pre>
      @return The pending fire count.
  */
  protected final synchronized int getAndClearPendingFireCount()
  {
    int count;
    synchronized(countLock){
      count = fireCount;
      fireCount = 0;
    }
    return(count);
  }
  
    /**
      Atomically decrements the number of pending executions of
      this handler (if it was non-zero) and returns the value
      from before the decrement.  This can be used in the <code>handleAsyncEvent()</code>
      method in this form to handle multiple firings:<pre>
      public void handleAsyncEvent() {
      &lt;setup>
      do {
      &lt;handle the event>
      } while(getAndDecrementPendingFireCount()>0);
      }</pre>
      This construction is necessary only in the case where one wishes to
      avoid the setup costs since the framework guarantees that 
      <code>handleAsyncEvent()</code>
      will be invoked the appropriate number of times.
      @return The pending fire count.
      */
  protected synchronized int getAndDecrementPendingFireCount()
  {
    int count;
    synchronized(countLock){
      count = fireCount;
      if(fireCount>0)
	fireCount--;
    }
    return(count);
  }
  
  /** 
      Atomically increments the number of pending executions of
      this handler and returns the value
      from before the increment.

      The <code>handleAsyncEvent()</code> method does not need to do this, 
      since the surrounding
      framework guarantees that the handler will be re-executed the
      appropriate number of times.  It is only of value when there
      is common setup code that is expensive.
      @return The pending fire count.
  */
  protected synchronized int getAndIncrementPendingFireCount()
  {
    int count;
    synchronized(countLock){
      count = fireCount;
      fireCount++;
    }
    return(count);
  }

    /**
       Used by the asynchronous event mechanism, see {@link AsyncEvent}. 
       This method invokes <code>handleAsyncEvent()</code> repeatedly while
       the fire count is greater than zero. Applications cannot override this
       method and should thus override <code>handleAsyncEvent()</code> in
       subclasses with the logic of the handler.
    */
  public final void run()
  {
    runHandler();
  }
  
  // this is the code that run runs.
  void runHandler()
  {
    // If an unhandled exception occures during handleAsyncEvent, we're done
    // here and the thread is on it's way to destruciton :), we need
    // to clear the pending fire count so the next time the AE happens, 
    // a new thread is properly created.
    try{
      do{
	handleAsyncEvent();
      }// the > 1 is because it returns the PRE decrement value
      while(getAndDecrementPendingFireCount() > 1 );
    }
    finally{
      getAndClearPendingFireCount();
    }
  }

    /** 
	Get the current memory area.
	@return The current memory area in which allocations occur.
    */
    public MemoryArea getMemoryArea()
    {
	return(HandlerThread.getMemoryArea());
    }

    // Methods to implement Schedulable, accessor methods for RealtimeThread

    /** 
      Inform the scheduler and cooperating facilities that this 
      thread's feasibility parameters should be considered in 
      feasibility analysis until further notified.
      */
    public boolean addToFeasibility()
    {
	return HandlerThread.addToFeasibility();
    }

    /**
       add the the feasibility of the already set scheduler if the resulting feasibility set is
       schedulable. If successful return true, if not return false. If there is not assigned 
       scheduler it will return false
    */   
    public boolean addIfFeasible()
    {
	return HandlerThread.addIfFeasible();
    }


    public boolean setIfFeasible(ReleaseParameters release,
				    MemoryParameters memory)
    {
	if(HandlerThread.setIfFeasible(release, memory)){
	    this.release = release;
	    this.memory = memory;
	    return true;
	}
	else{
	    return false;
	}	
    }

   public boolean setIfFeasible(ReleaseParameters release,
				    MemoryParameters memory,
				    ProcessingGroupParameters group)
    {
	if(HandlerThread.setIfFeasible(release, memory, group)){
	    this.release = release;
	    this.memory = memory;
	    this.group = group;
	    return true;
	}
	else{
	    return false;
	}
    }

    public boolean setReleaseParametersIfFeasible(ReleaseParameters release)
    {
	if(HandlerThread.setReleaseParametersIfFeasible(release)){
	    this.release = release;
	    return true;
	}
	else{
	    return false;
	}	
    }
    
    public boolean setProcessingGroupParametersIfFeasible(ProcessingGroupParameters group)
    {
	if(HandlerThread.setProcessingGroupParametersIfFeasible(group)){
	    this.group = group;
	    return true;
	}
	else{
	    return false;
	}
    }

    public boolean setIfFeasible(ReleaseParameters release,
				    ProcessingGroupParameters group)
    {
	if(HandlerThread.setIfFeasible(release, group)){
	    this.release = release;
	    this.group = group;
	    return true;
	}
	else{
	    return false;
	}
    }


    public boolean setMemoryParametersIfFeasible(MemoryParameters memory)
    {
	if(HandlerThread.setMemoryParametersIfFeasible(memory)){
	    this.memory = memory;
	    return true;
	}
	else{
	    return false;
	}	
    }

    /**
      Get the memory parameters associated with this handler.
      @return The {@link MemoryParameters} object associated with this.
      */
  public MemoryParameters getMemoryParameters()
  {
    return(HandlerThread.getMemoryParameters());
  }

    /**
       Get the release parameters associated with this handler.
       @return The {@link ReleaseParameters} object associated with this.
    */
    public ReleaseParameters getReleaseParameters()
    {
	return(HandlerThread.getReleaseParameters());
    }

    /**
      Return the {@link Scheduler} for this handler.
      @return The instance of the scheduler managing this.
      */
  public Scheduler getScheduler()
  {
    return(HandlerThread.getScheduler());
  }
  
    /**
      Returns a reference to the scheduling parameters object.
      @return The {@link SchedulingParameters} object associated with this.
      */
  public SchedulingParameters getSchedulingParameters()
  {
    return(HandlerThread.getSchedulingParameters());
  }

    /** Returns a reference to the {@link ProcessingGroupParameters} object.
     */
  public ProcessingGroupParameters getProcessingGroupParameters()
  {
    return(HandlerThread.getProcessingGroupParameters());
  }

    /** 
      Inform the scheduler and cooperating facilities that this thread's 
      feasibility parameters should not be considered in
      feasibility analysis until further notified.
      */
  public boolean removeFromFeasibility()
  {
    return HandlerThread.removeFromFeasibility();
  }
  
    /**
      Set the memory parameters associated with this handler.
      When it is next fired, the executing thread will use these
      parameters to control memory allocation.
      Does not affect the current invocation of the <code>run()</code> of this handler.
      @param memory A {@link MemoryParameters} object which will become the {@link MemoryParameters}
      associated with this after the method call.
      */
  public void setMemoryParameters(MemoryParameters memory)
  {
    HandlerThread.setMemoryParameters(memory);
    this.memory = memory;
  }

    /**
      Set the release parameters associated with this handler.
      When it is next fired, the executing thread will use these
      parameters to control scheduling.
      If the scheduling parameters of a handler is set to null,
      the handler will be executed immediately when it is fired, in the
      thread of the firer.
      Does not affect the current invocation of the <code>run()</code> of this handler.
      <p>Since this affects the
        constraints expressed in the release parameters of the
        existing schedulable objects, this may change the
        feasibility of the current schedule.
      @param parameters A {@link ReleaseParameters} object which will become the {@link ReleaseParameters}
      associated with this after the method call.
      */
    public void setReleaseParameters(ReleaseParameters release)
    {
	HandlerThread.setReleaseParameters(release);
	this.release = release;
    }

    /**
      Set the scheduler for this handler.
      A reference to the scheduler which will manage the execution of this thread.
      @param scheduler An instance of {@link Scheduler} (or subclasses) which will manage the
      execution of this thread.  If <code>scheduler</code> is null nothing happens.
      */
    public void setScheduler(Scheduler scheduler) throws IllegalThreadStateException
    {
	HandlerThread.setScheduler(scheduler);
    }

    /**
      Set the scheduler for this handler.
      A reference to the scheduler which will manage the execution of this thread.
      @param scheduler An instance of {@link Scheduler} (or subclasses) which will manage the
      execution of this thread.  If <code>scheduler</code> is null nothing happens.
      */
    public void setScheduler(Scheduler scheduler, SchedulingParameters scheduling, ReleaseParameters release, MemoryParameters memoryParameters, ProcessingGroupParameters processingGroup) throws IllegalThreadStateException
    {
	HandlerThread.setScheduler(scheduler, scheduling, release, memoryParameters, processingGroup);
	this.scheduling = scheduling;
	this.release = release;
    }

    /**
      Set the scheduling parameters associated with this handler.
      When it is next fired, the executing thread will use these
      parameters to control scheduling.
      Does not affect the current invocation of the <code>run()</code> of this handler.
      @param parameters A {@link SchedulingParameters} object which will become the {@link SchedulingParameters}
      object associated with this after the method call.
      */
  public void setSchedulingParameters(SchedulingParameters scheduling)
  {
    HandlerThread.setSchedulingParameters(scheduling);
    this.scheduling = scheduling;
  }

    /** Sets the reference to the {@link ProcessingGroupParameters} object.
     */  
  public void setProcessingGroupParameters(ProcessingGroupParameters group)
  {
    HandlerThread.setProcessingGroupParameters(group);
    this.group = group;
  }
  
    
    /** 
	Set the {@link SchedulingParameters} of this scheduable object only if the resulting task
	set is feasible.
	@param  scheduling The {@link SchedulingParameters} object. If null nothing happens. 
    */

    public boolean setSchedulingParametersIfFeasible(SchedulingParameters sched)
    {
	if (sched == null)
	    return false;
	if (HandlerThread.getScheduler() ==null)
	    return false;

	synchronized(HandlerThread.getScheduler())
	    {
		SchedulingParameters tmpsched = this.scheduling;
		removeFromFeasibility();
		setSchedulingParameters(scheduling);
		if (!addIfFeasible())
		    {
			setSchedulingParameters(tmpsched);
			addToFeasibility();
			return false;
		    }
	    }
	return true;
    }

}
