package javax.realtime;

/** 
 * An instance of <code>ImmortalPhysicalMemory</code> allows objects to be allocated 
 * from a range of physical memory with particular attributes, determined by
 * their memory type. This memory area has the same restrictive set
 * of assignment rules as {@link ImmortalMemory} memory areas, and may be used in any
 * context where <code>ImmortalMemory</code> is appropriate.
 * Objects allocated in immortal physical memory hava a lifetime greater than the application as
 * do objects allocated in immortal memory.
 */

public class ImmortalPhysicalMemory extends MemoryArea
{
    private long base = 0;
    private long size = 0;
    private long virtualAddress = 0;

    /**
      @param	type	An Object representing the type of
      memory required (e.g., <em>dma, shared</em>) - used to define the base address
      and control the mapping.
      @param	size	The size of the area in bytes.
      @exception	SecurityException	The application doesn't have
      permissions to access physical memory or the given type of memory.
      @exception	SizeOutOfBoundsException	The size is negative or
      extends into an invalid range of memory.
      @exception UnsupportedPhysicalMemoryException Thrown if the underlying hardware does not
      support the given type.
      */
    public ImmortalPhysicalMemory(Object type, long size)
	        throws SecurityException, SizeOutOfBoundsException, 
	        UnsupportedPhysicalMemoryException, MemoryTypeConflictException 
    {
	    super(size);
	    this.virtualAddress = PhysicalMemoryManager.create(type, size);
        this.base = 0;
        this.size = size;
        // Call MemoryArea.registerMemoryArea to register
        registerMemoryArea(this.virtualAddress, size);
    }

    /**
      @param	type	An Object representing the type of
      memory required (e.g., <em>dma, shared</em>)
      @param	base	The physical memory address of the area.
      @param	size	The size of the area in bytes.
      @exception	SecurityException	The application doesn't have
      permissions to access physical memory or the given range of memory.
      @exception	OffsetOutOfBoundsException	The address is invalid.
      @exception	SizeOutOfBoundsException	The size is negative or
      extends into an invalid range of memory.
      @exception UnsupportedPhysicalMemoryException Thrown if the underlying hardware does not
      support the given type.
      @exception MemoryTypeConflictException The specified base does not point to
        memory that matches the request type, or if <code>type</code> specifies
        attributes with a conflict.
      @exception MemoryInUseException The specified memory is already in use.
      */
    public ImmortalPhysicalMemory(Object type, long base, long size)
	        throws SecurityException, SizeOutOfBoundsException,
            OffsetOutOfBoundsException, UnsupportedPhysicalMemoryException,
            MemoryTypeConflictException, MemoryInUseException 
    {
        super(size);
	    this.virtualAddress = PhysicalMemoryManager.create(type, base, size);
        this.base = base;
        this.size = size;
        // Call MemoryArea.registerMemoryArea to register
        registerMemoryArea(this.virtualAddress, size);
    }
    
    /**
      @param	type	An Object representing the type of
      memory required (e.g., <em>dma, shared</em>) - used to define the base address
      and control the mapping.
      @param	size	A size estimator for this area.
      @exception	SecurityException	The application doesn't have
      permissions to access physical memory or the given type of memory.
      @exception	SizeOutOfBoundsException	The size is negative or
      extends into an invalid range of memory.
      @exception UnsupportedPhysicalMemoryException Thrown if the underlying hardware does not
      support the given type.
      @exception MemoryTypeConflictException The specified base does not point to
        memory that matches the request type, or if <code>type</code> specifies
        attributes with a conflict.
      */
    public ImmortalPhysicalMemory(Object type, SizeEstimator size)
	        throws SecurityException, SizeOutOfBoundsException, 
	        UnsupportedPhysicalMemoryException, MemoryTypeConflictException 
    {
	    super(size);
	    virtualAddress = PhysicalMemoryManager.create(type, size.getEstimate());
        this.base = 0;
        this.size = size.getEstimate();
        // Call MemoryArea.registerMemoryArea to register
        registerMemoryArea(this.virtualAddress, this.size);
    }

    /**
      @param	type	An Object representing the type of
      memory required (e.g., <em>dma, shared</em>)
      @param	base	The physical memory address of the area.
      @param	size	A size estimator for this memory area.
      @exception	SecurityException	The application doesn't have
      permissions to access physical memory or the given range of memory.
      @exception	OffsetOutOfBoundsException	The address is invalid.
      @exception	SizeOutOfBoundsException	The size is negative or
      extends into an invalid range of memory.
      @exception UnsupportedPhysicalMemoryException Thrown if the underlying hardware does not
      support the given type.
      @exception MemoryTypeConflictException The specified base does not point to
        memory that matches the request type, or if <code>type</code> specifies
        attributes with a conflict.
      @exception MemoryInUseException The specified memory is already in use.
      */
    public ImmortalPhysicalMemory(Object type, long base, SizeEstimator size)
	        throws SecurityException, SizeOutOfBoundsException,
            OffsetOutOfBoundsException, UnsupportedPhysicalMemoryException,
            MemoryTypeConflictException, MemoryInUseException  
    {
        super(size);
	    this.virtualAddress = PhysicalMemoryManager.create(type, base, size.getEstimate());
        this.base = base;
        this.size = size.getEstimate();
        // Call MemoryArea.registerMemoryArea to register
        registerMemoryArea(this.virtualAddress, this.size);
    }
    
    /**
      @param	type	An Object representing the type of
      memory required (e.g., <em>dma, shared</em>) - used to define the base address
      and control the mapping.
      @param	size	The size of the area in bytes.
      @param logic The <code>run()</code> method of this object will be 
      called whenever {@link MemoryArea#enter()} is called.
      @exception	SecurityException	The application doesn't have
      permissions to access physical memory or the given type of memory.
      @exception	SizeOutOfBoundsException	The size is negative or
      extends into an invalid range of memory.
      @exception UnsupportedPhysicalMemoryException Thrown if the underlying hardware does not
      support the given type.
      @exception MemoryTypeConflictException The specified base does not point to
        memory that matches the request type, or if <code>type</code> specifies
        attributes with a conflict.
      */
    public ImmortalPhysicalMemory(Object type, long size, Runnable logic)
	        throws SecurityException, SizeOutOfBoundsException,
	        UnsupportedPhysicalMemoryException, MemoryTypeConflictException 
    {
    	super(size, logic);
	    virtualAddress = PhysicalMemoryManager.create(type, size);
        this.base = 0;
        this.size = size;
        // Call MemoryArea.registerMemoryArea to register
        registerMemoryArea(this.virtualAddress, size);
    }

    /**
      @param	type	An Object representing the type of
      memory required (e.g., <em>dma, shared</em>)
      @param	base	The physical memory address of the area.
      @param	size	The size of the area in bytes.
      @param logic The <code>run()</code> method of this object will be 
      called whenever {@link MemoryArea#enter()} is called.
      @exception	SecurityException	The application doesn't have
      permissions to access physical memory or the given range of memory.
      @exception	OffsetOutOfBoundsException	The address is invalid.
      @exception	SizeOutOfBoundsException	The size is negative or
      extends into an invalid range of memory.
      @exception UnsupportedPhysicalMemoryException Thrown if the underlying hardware does not
      support the given type.
      @exception MemoryTypeConflictException The specified base does not point to
        memory that matches the request type, or if <code>type</code> specifies
        attributes with a conflict.
      @exception MemoryInUseException The specified memory is already in use.
      */
    public ImmortalPhysicalMemory(Object type, long base, long size, Runnable logic)
	        throws SecurityException, SizeOutOfBoundsException,
            OffsetOutOfBoundsException, UnsupportedPhysicalMemoryException,
            MemoryTypeConflictException, MemoryInUseException  {
        super(size, logic);
        this.virtualAddress = PhysicalMemoryManager.create(type, base, size);
        this.base = base;
        this.size = size;
        // Call MemoryArea.registerMemoryArea to register
        registerMemoryArea(this.virtualAddress, size);
/*	try {
	enter(logic);
	} catch (ScopedCycleException e){
	    throw new Error("Error in implementation. ScopedCycleException for new memory area");
	}*/
    }
    
    /**
      @param	type	An Object representing the type of
      memory required (e.g., <em>dma, shared</em>) - used to define the base address
      and control the mapping.
      @param	size	A size estimator for this area.
      @param logic The <code>run()</code> method of this object will be 
      called whenever {@link MemoryArea#enter()} is called.
      @exception	SecurityException	The application doesn't have
      permissions to access physical memory or the given type of memory.
      @exception	SizeOutOfBoundsException	The size 
      extends into an invalid range of memory.
      @exception UnsupportedPhysicalMemoryException Thrown if the underlying hardware does not
      support the given type.
      @exception MemoryTypeConflictException The specified base does not point to
        memory that matches the request type, or if <code>type</code> specifies
        attributes with a conflict.
      */
    public ImmortalPhysicalMemory(Object type, SizeEstimator size, Runnable logic)
	        throws SecurityException, SizeOutOfBoundsException,
	        UnsupportedPhysicalMemoryException, MemoryTypeConflictException 
    {
    	super(size.getEstimate(), logic);
	    virtualAddress = PhysicalMemoryManager.create(type, size.getEstimate());
        this.base = 0;
        this.size = size.getEstimate();
        // Call MemoryArea.registerMemoryArea to register
        registerMemoryArea(this.virtualAddress, this.size);
    }

    /**
      @param	type	An Object representing the type of
      memory required (e.g., <em>dma, shared</em>)
      @param	base	The physical memory address of the area.
      @param	size	A size estimator for this memory area.
      @exception	SecurityException	The application doesn't have
      permissions to access physical memory or the given range of memory.
      @exception	OffsetOutOfBoundsException	The address is invalid.
      @exception	SizeOutOfBoundsException	The size 
      extends into an invalid range of memory.
      @exception UnsupportedPhysicalMemoryException Thrown if the underlying hardware does not
      support the given type.
      @exception MemoryTypeConflictException The specified base does not point to
        memory that matches the request type, or if <code>type</code> specifies
        attributes with a conflict.
      @exception MemoryInUseException The specified memory is already in use.
      */
    public ImmortalPhysicalMemory(Object type, long base, SizeEstimator size,
        Runnable logic)
	        throws SecurityException, SizeOutOfBoundsException,
		    OffsetOutOfBoundsException, UnsupportedPhysicalMemoryException,
		    MemoryTypeConflictException, MemoryInUseException  
    {
        super(size.getEstimate(), logic);
	    this.virtualAddress = PhysicalMemoryManager.create(type, base, size.getEstimate());
        this.base = base;
        this.size = size.getEstimate();
        // Call MemoryArea.registerMemoryArea to register
        registerMemoryArea(this.virtualAddress, this.size);
/*    	try {
	    enter(logic);
    	} catch (ScopedCycleException e){
	        throw new Error("Error in implementation. ScopedCycleException for new memory area");
    	}   */
    }

    protected void finalize()
    {
        runAllFinalizers();
        PhysicalMemoryManager.free(virtualAddress, size);
    }
}
