/*
 * -----------------------------------------------------------------
 * LTMemory.java : LTMemory represents a memory area guaranteed to 
 * have linear time allocation.
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;

/**
  <code>LTMemory</code> represents a memory area, allocated per 
  {@link RealtimeThread}, or for a group of real-time threads, 
  guaranteed by the system to have linear time allocation. The memory 
  area described by a <code>LTMemory</code> instance does not exist in 
  the Java heap, and is not subject to garbage collection. Thus, it is
  safe to use a <code>LTMemory</code> object as the memory area 
  associated with a {@link NoHeapRealtimeThread}, or to enter the 
  memory area using the {@link ScopedMemory#enter} method within a 
  {@link NoHeapRealtimeThread}.  An <code>LTMemory</code> area has an 
  initial size. Enough memory must be committed by the completion of 
  the constructor to satisfy this initial requirement. (Committed means
  that this memory must always be available for allocation). The 
  initial memory allocation must behave, with respect to successful 
  allocation, as if it were contiguous; i.e., a correct implementation
  must guarantee that any sequence of object allocations that could 
  ever succeed without exceeding a specified initial memory size will
  always succeed without exceeding that initial memory size and succeed
  for any instance of <code>LTMemory</code> with that initial memory 
  size. <em>(Note: It is important to understand that the above statement
  does <b>not</b> require that if the initial memory size is N and
  (sizeof(object1) + sizeof(object2) + ... + sizeof(objectn) = N) 
  the allocations of objects 1 through n will necessarily succeed.)</em>
  Execution time of an allocator allocating from this initial area must 
  be linear in the size of the allocated object. Execution time of an 
  allocator allocating from memory between initial and maximum is 
  allowed to vary.  Furthermore, the underlying system is not required 
  to guarantee that memory between initial and maximum will always be 
  available. (Note: to ensure that all requested memory is available 
  set initial and maximum to the same value)
 
  @see MemoryArea
  @see ScopedMemory
  @see RealtimeThread
  @see NoHeapRealtimeThread

  @author Pratik Solanki
 */
public class LTMemory extends ScopedMemory
{
    long initialSize;
    long maximumSize;
    
    /**
     * Create an <code>LTMemory</code> of the given size.
     *
     * @param initialSizeInBytes The size in bytes of the memory to 
     *               allocate for this area. This memory must be committed 
     *               before the completion of the constructor.
     * @param maxSizeInBytes     The size in bytes of the memory to 
     *               allocate for this area.
     */              
    public LTMemory(long initialSizeInBytes, long maxSizeInBytes)
    {
        this(initialSizeInBytes, maxSizeInBytes, null);
    }

    
    public LTMemory(long initialSizeInBytes, long maxSizeInBytes, Runnable logic)
    {
        super(initialSizeInBytes, logic);
        
        if(initialSizeInBytes > maxSizeInBytes)
            System.out.println("LTMemory - initial size is greater than maximum size. Ignoring maximum size.");
            
        memorySize = (maxSizeInBytes>initialSizeInBytes)?maxSizeInBytes:initialSizeInBytes;

        initialSize = initialSizeInBytes;
        maximumSize = memorySize;
        
		// The following is a ScopedMemory function that I am calling.
		// This should ideally be in LTMemory but I did this so that I didn't
		// have to duplicate the function in LT and VT Memory.
		// - Pratik
		initScoped(memorySize);
    }

    public LTMemory(SizeEstimator initial, SizeEstimator maximum)
    {
        this(initial.getEstimate(), maximum.getEstimate());
    }
    
    public LTMemory(SizeEstimator initial, SizeEstimator maximum, Runnable logic)
    {
        this(initial.getEstimate(), maximum.getEstimate(), logic);
    }

    public long getMaximumSize()
    {
        return (int)maximumSize;
    }

    public String toString()
    {
        return new String("(LTMemory) " + super.toString());
    }
}

