package javax.realtime;

/** 
 * An instance of <code>LTPhysicalMemory</code> allows objects to be allocated from
 * a range of physical memory with particular attributes, determined by
 * their memory type.  This memory area has the same restrictive set
 * of assignment rules as {@link ScopedMemory} memory areas, and the same performance
 * restrictions as <code>LTMemory</code>.
 *
 * @see MemoryArea
 * @see ScopedMemory
 * @see VTMemory
 * @see LTMemory
 * @see VTPhysicalMemory
 * @see ImmortalPhysicalMemory
 * @see RealtimeThread
 * @see NoHeapRealtimeThread
 * 
 */

public class LTPhysicalMemory extends ScopedMemory 
{
    private long virtualAddress;  
    private long base;
    private long size;

    /**
     * @param type An Object representing the type of
     * memory required (e.g., <em>dma, shared</em>) - used to define the base address
     * and control the mapping.
     * @param size The size of the area in bytes.
     * 
     * @throws SecurityException The application doesn't have
     * permissions to access physical memory or the given type of memory.
     * @throws SizeOutOfBoundsException The size is negative or
     * extends into an invalid range of memory.
     * @throws UnsupportedPhysicalMemoryException Thrown if the underlying 
     * hardware does not support the given type.
     * @throws MemoryTypeConflictException The specified base does not point to
     *   memory that matches the request type, or if <code>type</code> specifies
     *   attributes with a conflict.
     * 
     */
    public LTPhysicalMemory(Object type, long size)
            throws SecurityException, SizeOutOfBoundsException, 
            UnsupportedPhysicalMemoryException, MemoryTypeConflictException 
    {
        super(size);
        virtualAddress = PhysicalMemoryManager.create(type, size);
        this.size = size;
        this.base = 0;
        // Call MemoryArea.registerMemoryArea to register
        registerMemoryArea(this.virtualAddress, this.size);
    }

    /**
     * @param type An Object representing the type of
     * memory required (e.g., <em>dma, shared</em>)
     * @param base The physical memory address of the area.
     * @param size The size of the area in bytes.
     *
     * @throws SecurityException The application doesn't have
     * permissions to access physical memory or the given range of memory.
     * @throws OffsetOutOfBoundsException The address is invalid.
     * @throws SizeOutOfBoundsException The size is negative or
     * extends into an invalid range of memory.
     * @throws UnsupportedPhysicalMemoryException Thrown if the underlying
     * hardware does not support the given type.
     * @throws MemoryTypeConflictException The specified base does not point to
     *   memory that matches the request type, or if <code>type</code> specifies
     *   attributes with a conflict.
     * @throws MemoryInUseException The specified memory is already in use.
     */
    public LTPhysicalMemory(Object type, long base, long size)
            throws SecurityException, SizeOutOfBoundsException,
                OffsetOutOfBoundsException, UnsupportedPhysicalMemoryException,
                MemoryTypeConflictException, MemoryInUseException 
    {
        super(size);
        this.virtualAddress = PhysicalMemoryManager.create(type, base, size);
        this.size = size;
        this.base = base;
        // Call MemoryArea.registerMemoryArea to register
        registerMemoryArea(this.virtualAddress, this.size);
    }
    
    /**
     * @param type An Object representing the type of
     * memory required (e.g., <em>dma, shared</em>) - used to define the base address
     * and control the mapping.
     * @param size A size estimator for this area.
     *
     * @throws SecurityException The application doesn't have
     * permissions to access physical memory or the given type of memory.
     * @throws SizeOutOfBoundsException The size
     * extends into an invalid range of memory.
     * @throws UnsupportedPhysicalMemoryException Thrown if the underlying 
     * hardware does not support the given type.
     * @throws MemoryTypeConflictException The specified base does not point to
     * memory that matches the request type, or if <code>type</code> specifies
     * attributes with a conflict.
     */
    public LTPhysicalMemory(Object type, SizeEstimator size)
            throws SecurityException, SizeOutOfBoundsException, 
            UnsupportedPhysicalMemoryException, MemoryTypeConflictException 
    {
        super(size);
        virtualAddress = PhysicalMemoryManager.create(type, size.getEstimate());
        this.base = 0;
        this.size = size.getEstimate(); 
        // Call MemoryArea.registerMemoryArea to register
        registerMemoryArea(this.virtualAddress, this.size);
    }

    /**
     * @param type An Object representing the type of
     * memory required (e.g., <em>dma, shared</em>)
     * @param base The physical memory address of the area.
     * @param size A size estimator for this memory area.
     *
     * @throws SecurityException The application doesn't have
     * permissions to access physical memory or the given range of memory.
     * @throws OffsetOutOfBoundsException The address is invalid.
     * @throws SizeOutOfBoundsException    The size
     * extends into an invalid range of memory.
     * @throws UnsupportedPhysicalMemoryException Thrown if the underlying
     * hardware does not support the given type.
     * @throws MemoryTypeConflictException The specified base does not point to
     * memory that matches the request type, or if <code>type</code> specifies
     * attributes with a conflict.
     * @throws MemoryInUseException The specified memory is already in use.
     */
    public LTPhysicalMemory(Object type, long base, SizeEstimator size)
        throws SecurityException, SizeOutOfBoundsException,
            OffsetOutOfBoundsException, UnsupportedPhysicalMemoryException,
            MemoryTypeConflictException, MemoryInUseException  
    {
        super(size);
        this.virtualAddress = PhysicalMemoryManager.create(type, base, size.getEstimate());
        this.size = size.getEstimate();
        this.base = base;
        // Call MemoryArea.registerMemoryArea to register
        registerMemoryArea(this.virtualAddress, this.size);
    }
    
    /**
     * @param type An Object representing the type of
     * memory required (e.g., <em>dma, shared</em>) - used to define the base address
     * and control the mapping.
     * @param size The size of the area in bytes.
     * @param logic <code>enter</code> this memory area with this <code>Runnable</code>
     * after the memory area is created.
     *
     * @throws    SecurityException    The application doesn't have
     * permissions to access physical memory or the given type of memory.
     * @throws    SizeOutOfBoundsException    The size is negative or
     * extends into an invalid range of memory.
     * @throws UnsupportedPhysicalMemoryException Thrown if the underlying 
     * hardware does not support the given type.
     * @throws MemoryTypeConflictException The specified base does not point to
     * memory that matches the request type, or if <code>type</code> specifies
     * attributes with a conflict.
     */
    public LTPhysicalMemory(Object type, long size, Runnable logic)
            throws SecurityException, SizeOutOfBoundsException, 
            UnsupportedPhysicalMemoryException, MemoryTypeConflictException
    {
        super(size, logic);
        virtualAddress = PhysicalMemoryManager.create(type, size);
        this.size = size;
        this.base = 0;
        // Call MemoryArea.registerMemoryArea to register
        registerMemoryArea(this.virtualAddress, this.size);
        /*try 
        {
            enter(logic);
        } 
        catch (ScopedCycleException e)
        {
            throw new Error("Error in implementation. ScopedCycleException for new memory area");
        }*/
    }

    /**
     * @param type An Object representing the type of
     * memory required (e.g., <em>dma, shared</em>)
     * @param base The physical memory address of the area.
     * @param size The size of the area in bytes.
     * @param logic <code>enter</code> this memory area with this <code>Runnable</code>
     * after the memory area is created.
     *
     * @throws SecurityException The application doesn't have
     * permissions to access physical memory or the given range of memory.
     * @throws OffsetOutOfBoundsException The address is invalid.
     * @throws SizeOutOfBoundsException The size is negative or
     * extends into an invalid range of memory.
     * @throws UnsupportedPhysicalMemoryException Thrown if the underlying 
     * hardware does not support the given type.
     * @throws MemoryTypeConflictException The specified base does not point to
     * memory that matches the request type, or if <code>type</code> specifies
     * attributes with a conflict.
     * @throws MemoryInUseException The specified memory is already in use.
     */
    public LTPhysicalMemory(Object type, long base, long size, Runnable logic)
            throws SecurityException, SizeOutOfBoundsException,
              OffsetOutOfBoundsException, UnsupportedPhysicalMemoryException,
              MemoryTypeConflictException, MemoryInUseException  
    {
        super(size, logic);
        this.virtualAddress = PhysicalMemoryManager.create(type, base, size);
        this.base = base;
        this.size = size;
        // Call MemoryArea.registerMemoryArea to register
        registerMemoryArea(this.virtualAddress, this.size);
        /*try 
        {
            enter(logic);
        } catch (ScopedCycleException e)
        {
            throw new Error("Error in implementation. ScopedCycleException for new memory area");
        }*/
    }
    
    /**
     * @param type An Object representing the type of
     * memory required (e.g., <em>dma, shared</em>) - used to define the base address
     * and control the mapping.
     * @param size A size estimator for this area.
     * @param logic <code>enter</code> this memory area with this <code>Runnable</code>
     * after the memory area is created.
     *
     * @throws SecurityException    The application doesn't have
     * permissions to access physical memory or the given type of memory.
     * @throws SizeOutOfBoundsException The size
     * extends into an invalid range of memory.
     * @throws UnsupportedPhysicalMemoryException Thrown if the underlying 
     * hardware does not support the given type.
     * @throws MemoryTypeConflictException The specified base does not point to
     * memory that matches the request type, or if <code>type</code> specifies
     * attributes with a conflict.
     */
    public LTPhysicalMemory(Object type, SizeEstimator size, Runnable logic)
            throws SecurityException, SizeOutOfBoundsException, 
                UnsupportedPhysicalMemoryException, MemoryTypeConflictException 
    {
        super(size, logic);
        virtualAddress = PhysicalMemoryManager.create(type, size.getEstimate());
        this.size = size.getEstimate();
        this.base = 0;
        // Call MemoryArea.registerMemoryArea to register
        registerMemoryArea(this.virtualAddress, this.size);
        /*try 
        {
            enter(logic);
        } 
        catch (ScopedCycleException e)
        {
            throw new Error("Error in implementation. ScopedCycleException for new memory area");
        }*/
    }

    /**
     * @param type An Object representing the type of
     * memory required (e.g., <em>dma, shared</em>)
     * @param base The physical memory address of the area.
     * @param size A size estimator for this memory area.
     * @param logic <code>enter</code> this memory area with this <code>Runnable</code>
     * after the memory area is created.
     *
     * @throws SecurityException The application doesn't have
     * permissions to access physical memory or the given range of memory.
     * @throws OffsetOutOfBoundsException The address is invalid.
     * @throws SizeOutOfBoundsException The size
     * extends into an invalid range of memory.
     * @throws UnsupportedPhysicalMemoryException Thrown if the underlying 
     * hardware does not support the given type.
     * @throws MemoryTypeConflictException The specified base does not point to
     * memory that matches the request type, or if <code>type</code> specifies
     * attributes with a conflict.
     * @throws MemoryInUseException The specified memory is already in use.
     */
    public LTPhysicalMemory(Object type, long base, SizeEstimator size,
                                Runnable logic)
            throws SecurityException, SizeOutOfBoundsException,
               OffsetOutOfBoundsException, UnsupportedPhysicalMemoryException,
               MemoryTypeConflictException, MemoryInUseException  
    {
        super(size, logic);
        this.virtualAddress = PhysicalMemoryManager.create(type, base, size.getEstimate());
        this.size = size.getEstimate();
        this.base = base;
        // Call MemoryArea.registerMemoryArea to register
        registerMemoryArea(this.virtualAddress, this.size);
        /*try 
        {
            enter(logic);
        } 
        catch (ScopedCycleException e)
        {
            throw new Error("Error in implementation. ScopedCycleException for new memory area");
        }*/
    }

    public String toString()
    {
        return new String("(LTPhysicalMemory) " + super.toString());
    }

    protected void finalize()
    {
        runAllFinalizers();
        PhysicalMemoryManager.free(virtualAddress, size);
    }
}
