/*
 * -----------------------------------------------------------------
 * MemoryParameters.java : This class is used for pacing the memory 
 * allocation by a RealtimeThread.
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;

import java.util.Vector;

/**
 * Memory parameters can be given on the constructor of 
 * <code>RealtimeThread</code> and <code>AsyncEventHandler</code>. 
 * These can be used both for the purposes of admission control by
 * the scheduler and for the purposes of pacing the garbage collector
 * to satisfy all of the thread allocation rates.
 * <p>
 * When a reference to a <code>MemoryParameters</code> object is given 
 * as a parameter to a constructor, the <code>MemoryParameters</code> 
 * object becomes bound to the object being created. Changes to the 
 * values in the <code>MemoryParameters</code> object affect the 
 * constructed object. If given to more than one constructor, then 
 * changes to the values in the <code>MemoryParameters</code> object 
 * affect <em>all</em> of the associated objects. Note that this is a
 * one-to-many relationship and <em>not</em> a many-to-many.
 *
 * <p><b>Caution:</b> This class is explicitly unsafe in multithreaded
 * situations when it is being changed.  No synchronization is done.  It
 * is assumed that users of this class who are mutating instances will be
 * doing their own synchronization at a higher level.
 */
public class MemoryParameters
{

    private long maxMemoryArea = NO_MAX;
    private long maxImmortal = NO_MAX;
    private long allocationRate = NO_MAX;

    /** 
     * Specifies no maximum limit. 
     */
    public static final long NO_MAX = -1;

    private Vector associatedThreads = null;
    
    /*
     * This is a temporary variable that is used for traversing the 
     * associatedThreads vector. Reason I am declaring it here
     * instead of using a local variable is because I am afraid that
     * using a local variable might create an IllegalAssignmentError
     * depening on which memory area I am in.
     * 
     * - Pratik.
     */
    
    private RealtimeThread traverser;
    
   /** 
    * Create a <code>MemoryParameters</code> object with the given values.
    * @param  maxMemoryArea  A limit on the amount of memory the thread
    *        may allocate in the memory area. Units are in bytes. If 
    *        zero, no allocation allowed in the memory area. To specify
    *        no limit, use NO_MAX or a value less than zero.
    * @param  maxImmortal  A limit on the amount of memory the thread 
    *        may allocate in the immortal area. Units are in bytes. If 
    *        zero, no allocation allowed in immortal. To specify no limit,
    *        use NO_MAX or a value less than zero.
    */   
    public MemoryParameters(long maxMemoryArea, long maxImmortal)
                    throws IllegalArgumentException
    {
        this.maxMemoryArea = maxMemoryArea;
	    this.maxImmortal = maxImmortal;
        //dummy try-catch block
        try
        {
            associatedThreads = (Vector) ImmortalMemory.instance().newInstance(Vector.class);
        }
        catch(Exception e)
        {
            e.printStackTrace(); 
        }
    }

   /** 
    * Create a <code>MemoryParameters</code> object with the given values.
    * @param  maxMemoryArea  A limit on the amount of memory the thread
    *        may allocate in the memory area. Units are in bytes. If 
    *        zero, no allocation allowed in the memory area. To specify
    *        no limit, use NO_MAX or a value less than zero.
    * @param  maxImmortal  A limit on the amount of memory the thread 
    *        may allocate in the immortal area. Units are in bytes. If 
    *        zero, no allocation allowed in immortal. To specify no limit,
    *        use NO_MAX or a value less than zero.
    * @param  allocationRate  A limit on the rate of allocation in the
    *        heap. Units are in bytes per second. If zero, no allocation
    *        is allowed in the heap. To specify no limit, use NO_MAX or 
    *        a value less than zero.
    */
    public MemoryParameters(long maxMemoryArea, long maxImmortal,
			    long allocationRate)
                        throws IllegalArgumentException
    {
	    this(maxMemoryArea, maxImmortal);
	    this.allocationRate = allocationRate;
    }

    /**
     * Get the allocation rate. Units are in bytes per second.
     */
    public long getAllocationRate()
    {
    	return allocationRate;
    }

    /**
     * Get the limit on the amount of memory the thread may allocate in
     * the immortal area. Units are in bytes.
     */
    public long getMaxImmortal()
    {
        return maxImmortal;
    }

    /**
     * Get the limit on the amount of memory the thread may allocate in
     * the memory area. Units are in bytes.
     */
    public long  getMaxMemoryArea()
    {
        return maxMemoryArea;
    }
    
    /**
     * A limit on the rate of allocation in the heap. 
     * @param  allocationRate  Units are in bytes per second. If zero,
     *          no allocation is allowed in the heap. To specify no 
     *          limit, use NO_MAX or a value less than zero.
     */
    public void setAllocationRate(long allocationRate)
    {
        this.allocationRate  = allocationRate;
    }

    /**
     * A limit on the amount of memory the thread may allocate in the 
     * immortal area. 
     * @param  maxImmortal Units are in bytes. If zero, no allocation 
     *          allowed in immortal. To specify no limit, use NO_MAX 
     *          or a value less than zero.
     * @return False if any of the threads have already allocated more 
     *         than the given value. In this case the call has no effect.
     */
    public boolean setMaxImmortal(long maximum)
    {
        // First check to see if any thread has allocated beyond the 
        // maximum that I am about to set
        int size = associatedThreads.size();
        for(int i=0; i< size; i++)
        {
            traverser = (RealtimeThread) associatedThreads.elementAt(i);
            if(traverser.immortalAllocated > maximum)
                return false;
        }

        this.maxImmortal = maximum;
        return true;
    }
    
    /**
     * A limit on the amount of memory the thread may allocate in the 
     * memory area.
     * @param maximum Units are in bytes. If zero, no allocation allowed
     *          in the memory area. To specify no limit, use NO_MAX or
     *          a value less than zero.
     * @return False if any of the threads have already allocated more 
     *         than the given value. In this case the call has no effect.
     */
    public boolean setMaxMemoryArea(long maximum)
    {
        // First check to see if any thread has allocated beyond the 
        // maximum that I am about to set
        int size = associatedThreads.size();
        for(int i=0; i< size; i++)
        {
            traverser = (RealtimeThread) associatedThreads.elementAt(i);
            if(traverser.memAreaAllocated > maximum)
                return false;
        }


        this.maxMemoryArea = maximum;
        return true;
    }

    /**
     * Change the limit on the rate of allocation in the heap. If this 
     * <code>MemoryParameters</code> object is currently
     * associated with one or more realtime threads that have been passed
     * admission control, this change in allocation rate will be submitted 
     * to admission control. The scheduler (in conjunction with the garbage
     * collector) will either admit all the effected threads with the
     * new allocation rate, or leave the allocation rate unchanged
     * and cause <code>setAllocationRateIfFeasible</code> to return <code>false</code>.
     *
     * @param allocationRate Units are in bytes per second. If zero, no
     * allocation is allowed in the heap. To specify no limit,
     * use NO_MAX or a value less than zero.
     *
     * @return <code>true</code> if the request was fulfilled.
     */
    public boolean setAllocationRateIfFeasible(int allocationRate)
    {
        return false;
    }

    /**
     * Associates the passed <code>RealtimeThread</code> object with
     * this <code>MemoryParameters</code> object. This is to implement
     * the many-to-one functionality of <code>MemoryParameters</code>,
     * specifically this is to facilitate the requirements of
     * setMaxMemoryArea and setMaxImmortal.
     * 
     * @param t The <code>RealtimeThread</code> that will use this 
     *          <code>MemoryParameters</code> object for budgeted 
     *          allocation.
     */
    void associateThread(RealtimeThread t)
    {
        if(associatedThreads.contains(t) == false)
        {
            associatedThreads.add(t);
        }
    }

    /**
     * De-associates or removes the <code>RealtimeThread</code> object
     * passed with this <code>MemoryParameters</code> object. This is 
     * to implement the many-to-one functionality of 
     * <code>MemoryParameters</code>, specifically this is to facilitate
     * the requirements of setMaxMemoryArea and setMaxImmortal.
     * 
     * @param t <code>RealtimeThread</code> object that will no longer
     *          be using this <code>MemoryParameters</code> object for 
     *          budgeted allocation.
     */
    
    void deassociateThread(RealtimeThread t)
    {
        associatedThreads.remove(t);
    }
}
