/*
 * -----------------------------------------------------------------
 *  PeriodicTimer.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;

/**
 * An AsyncEvent whose fire method is executed periodically according to the
 * given parameters. If a clock is given, calculation of the period uses the
 * increments of the clock. If an interval is given or set the system
 * guarantees that the fire method will  execute interval time units after the
 * last execution or its given start time as appropriate. If one of the
 * HighResolutionTime argument types is  RationalTime then the system guarantees
 * that the fire method will be executed  exactly frequency times every unit
 * time (see RationalTime constructors) by  adjusting the interval between
 * executions of fire().  This is similar to a thread with PeriodicParameters
 * except that it is lighter weight.
 * If a PeriodicTimer is disabled, it still counts, and if enabled at some later
 * time, it will fire at its next scheduled fire time.
 */


public class PeriodicTimer extends Timer{

	private RelativeTime interval;
	private long inter_milli;
	private int inter_nano;

	/**
	* Create an instance of AsyncEvent that executes its fire method periodiacally
	* @param  start     The time when the first interval begins
	* @param  interval  The time between successive executions of the fire method
	* @param  handler   The instance of AsyncEventHandler that will be scheduled each
	*                   time the fire method is exceuted
	*/
	public PeriodicTimer(HighResolutionTime start, RelativeTime interval,AsyncEventHandler handler){
		
		super(start,null,handler);
		if(start==null)
			time = clock.getTime();

		this.interval = interval;
		RelativeTime rt = interval.getInterarrivalTime(null);
		inter_milli = rt.getMilliseconds();
		inter_nano = rt.getNanoseconds();

	}

	/**
	* Create an instance of AsyncEvent that executes its fire method periodiacally
	* @param  start     The time when the first interval begins
	* @param  interval  The time between successive executions of the fire method
	* @param  clock     The clock whose increments are used to calculate the interval
	* @param  handler   The instance of AsyncEventHandler that will be scheduled each
	*                   time the fire method is exceuted
	*/
	public PeriodicTimer(HighResolutionTime start, RelativeTime interval, Clock clock,AsyncEventHandler handler){
		super(start,clock,handler);
		if(start==null)
			time = clock.getTime();

		this.interval = interval;
		RelativeTime rt = interval.getInterarrivalTime(null);
		inter_milli = rt.getMilliseconds();
		inter_nano = rt.getNanoseconds();
		
	}

	/**
	* Create a {@link ReleaseParameters} object with the next fire time as the
	* start time and the interval of <code>this</code> as the period.
	*
	* @return  an instance of ReleaseParameters object
	*/
	public ReleaseParameters createReleaseParameters()
	{
		return new PeriodicParameters(time, interval, new RelativeTime(0,0), interval, null, null);
	}

	public void fire(){//Why this method is public, no idea..

		//call super's fire method at each interval//
		do{

			//fire here and wait
			if(isEnabled){
				super.fire();
			}
			try{
				waitLock.wait(inter_milli,inter_nano);			
			}
			catch(Exception e){
				e.printStackTrace();
			}
		}while(true);
		
	}

	public AbsoluteTime getFireTime(){

		AbsoluteTime at = time.absolute(clock,null);
		return at.add(interval.getInterarrivalTime(null));
	}

	/**
	* Return the interval of this Timer
	* @return  a RelativeTime object which is the current interval of this
	*/
	public RelativeTime getInterval(){
		
		return interval;
	}

	/**
	* Reset the interval of this Timer
	* @param interval  A RelativeTime object which is the interval to reset this Timer
	*/
	public void setInterval(RelativeTime interval){
		
		synchronized(this.interval){
			this.interval = interval;
			
			RelativeTime rt = interval.getInterarrivalTime(null);
			inter_milli = rt.getMilliseconds();
			inter_nano = rt.getNanoseconds();
		
		}
	}

	protected void Init(){//OneShot

	    while(true){


		//sleep until action time..!
		at = time.absolute(clock,at);
		AbsoluteTime now = clock.getTime();
	
		//System.out.println(at.toString());
		//System.out.println(now.toString());

		RelativeTime rt = at.subtract(now);
		long milli = rt.getMilliseconds();
		int nano = rt.getNanoseconds();
		//System.out.println(rt.toString());


		if((milli<0 || nano<0 || (milli==0 && nano==0)) && isEnabled){
			fire(); //AsyncEvent's method
			return;
		}

		if(isEnabled){
			try{
				waitLock.wait(milli,nano);
				if(rescheduled){
					rescheduled = false;
					continue;
				}
			}
			catch(Exception e){
				e.printStackTrace();
			}
			fire(); //AsyncEvent's method
			return;
		}
             }//endof while loop.
	
	}

}




