/*
 * -----------------------------------------------------------------
 * PriorityParameters.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;
import java.util.*;

/** 
  Instances of this class should 
  be assigned to threads that are managed by schedulers which use
  a single integer to determine execution order. The base scheduler required
  by this specification and represented by the class {@link PriorityScheduler} is
  such a scheduler.
  */

public class PriorityParameters extends SchedulingParameters
{
    int priority;
    
    /** 
      Create an instance of {@link SchedulingParameters} with the given priority.
      @param priority    The priority assigned to a thread.
                         This value is used in place of the value returned by 
			 {@link java.lang.Thread#setPriority}.
      */
    public PriorityParameters(int priority)
    {
	schedSet = new Vector();
	this.priority = priority;
    }
    
    /** 
      Get the priority.
     */
    public int getPriority()
    {
	return priority;
    }

    private Vector schedSet;

    void bind(Schedulable sch)
    {
	if (!schedSet.contains(sch))
	    {
		schedSet.addElement(sch);
	    }
    }

    void unbind(Schedulable sch)
    {
	if (schedSet.contains(sch))
	    {
		schedSet.remove(sch);
	    }
    }


    private boolean validRange(int priority)
    {
	Schedulable s;

	for (int i = 0 ; i < schedSet.size(); i++)
	    {
		s = (Schedulable) schedSet.elementAt(i);
		try
		    {
			PriorityScheduler ps = (PriorityScheduler)s.getScheduler();
			if ((priority < ps.getMinPriority()) ||
			    (priority > ps.getMaxPriority()))
			    return false;
		    }
		catch(Exception e)
		    {
			return false;
		    }
	    }
	return true;
    }

    /**
      Set the priority.
      @param priority The new value of priority.  
      @exception IllegalArgumentException Thrown if the given priority value is less than 
      the minimum priority of the scheduler of any of the associated threads or greater
      then the maximum priority of the scheduler of any of the associated threads.
      */

    public void setPriority(int priority) throws IllegalArgumentException
    {
	if (!validRange(priority))
	    {
		throw new IllegalArgumentException("Priority out of range");
	    }

	this.priority = priority;

	Schedulable s;

	for (int i = 0 ; i < schedSet.size(); i++)
	    {
		s = (Schedulable) schedSet.elementAt(i);
		s.setSchedulingParameters((SchedulingParameters)this);
	    }	
    }

    public java.lang.String toString()
    {
	return Integer.toString(priority);
    }
}
