/*
 * -----------------------------------------------------------------
 * ProcessingGroupParameters.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;
import java.util.*;

/**
  This is associated with one or more schedulable objects for which the system guarantees that the associated
  objects will not be given more time per period than indicated by cost.
  For all threads with a reference to an instance of <code>ProcessingGroupParameters</code> p and a reference to
  an instance of {@link AperiodicParameters} no more than p.cost will be allocated to the execution of
  these threads in each interval of time given by p.period after the time indicated by p.start.

  When a reference to a <code>ProcessingGroupParameters</code> object is given
  as a parameter to a constructor the <code>ProcessingGroupParameters</code> object becomes bound
  to the object being created. Changes to the values in the <code>ProcessingGroupParameters</code>
  object affect the constructed object. If given to more than one constructor,
  then changes to the values in the <code>ProcessingGroupParameters</code> object affect <em>all</em>
  of the associated objects. Note that this is a one-to-many relationship and
  <em>not</em> a many-to-many.

  <p><b>Caution:</b> This class is explicitly unsafe in multithreaded
  situations when it is being changed.  No synchronization is done.  It
  is assumed that users of this class who are mutating instances will be
  doing their own synchronization at a higher level.

  <p><b>Caution:</b> The <code>cost</code> parameter time should be considered to be 
  measured against the target platform.
  */

public class ProcessingGroupParameters
{
    HighResolutionTime start;
    RelativeTime period;
    RelativeTime cost;
    RelativeTime deadline;
    AsyncEventHandler overrunHandler;
    AsyncEventHandler missHandler;
    
    /** 
     Create a <code>ProcessingGroupParameters</code> object.
     @param start          Time at which the first period begins.
     @param period         The period is the interval between successive 
                           unblocks of <code>waitForNextPeriod()</code>.
     @param cost           Processing time per period.
     @param deadline       The latest permissible completion time measured from the start of the current period.
                           Changing the deadline might not take effect after the expiration 
                           of the current deadline.
     @param overrunHandler This handler is invoked if the <code>run()</code> method of the schedulable
                           object of the previous period is still executing at the start
                           of the current period.
     @param missHandler    This handler is invoked if the <code>run()</code> method of the schedulable
                           object is still executing after the deadline has passed.
     */

    public ProcessingGroupParameters(HighResolutionTime start,
				     RelativeTime period,
				     RelativeTime cost,
				     RelativeTime deadline,
				     AsyncEventHandler overrunHandler,
				     AsyncEventHandler missHandler)
    {
	if (start == null)
	    this.start = new AbsoluteTime(0,0);
	else
	    this.start = start;

	this.period = period;
	this.cost = cost;
	this.deadline = deadline;
	this.overrunHandler = overrunHandler;
	this.missHandler = missHandler;
    }

    protected Vector schedSet= new Vector();

    void bind(Schedulable sch)
    {
	if (!schedSet.contains(sch))
	    {
		schedSet.addElement(sch);
	    }
    }

    void unbind(Schedulable sch)
    {
	if (schedSet.contains(sch))
	    {
		schedSet.remove(sch);
	    }
    }


    public boolean setIfFeasible(RelativeTime period, RelativeTime cost, RelativeTime deadline)
    {
	boolean retval = true;
	Schedulable s;
	RelativeTime oldCost, oldDeadline, oldPeriod;

	oldPeriod = getPeriod();
	oldCost = getCost();
	oldDeadline = getDeadline();

	setPeriod(period);
	setCost(cost);
	setDeadline(deadline);

	// get globalLock to lock all schedulers
	
	synchronized(Scheduler.class)
	    {
		for (int i=0; i< schedSet.size() ; i++)
		    {
			s = (Schedulable) schedSet.elementAt(i);
			if (!s.setProcessingGroupParametersIfFeasible(this))
			    {
				// undo the changes

				setPeriod(oldPeriod);
				setCost(oldCost);
				setDeadline(oldDeadline);
				for (int j = i -1; i> 0 ; i--)
				    {
					s = (Schedulable) schedSet.elementAt(i);
					s.setProcessingGroupParametersIfFeasible(this);
				    }
				retval = false;
				break;
			    }
		    }
	    }
	return retval;
    }




    
    /**
      Get the cost value.
      */
    public RelativeTime getCost()
    {
	return cost;
    }
    
    /**
      Get the cost overrun handler.
      @return An {@link AsyncEventHandler} object that is cost overrun handler of this.
      */
    public AsyncEventHandler getCostOverrunHandler()
    {
	return overrunHandler;
    }
    
    /**
      Get the deadline value.
      @return A {@link RelativeTime} object that represents the deadline of this.
      */
    public RelativeTime getDeadline()
    {
	return deadline;
    }
    
    /**
      Get the deadline missed handler.
      @return An {@link AsyncEventHandler} object that is deadline miss handler of this.
      */
    public AsyncEventHandler getDeadlineMissHandler()
    {
	return missHandler;
    }
    
    /**
      Get the period.
      @return A {@link RelativeTime} object that represents the period of time of this.
      */
    public RelativeTime getPeriod()
    {
	return period;
    }
    
    /**
      Get the start time.
      @return A {@link HighResolutionTime} object that represents the start time of this.
      */
    public HighResolutionTime getStart()
    {
	return start;
    }
    
    /**
      Set the cost value.
      @param cost The schedulable objects with a reference to this receive cumulatively
      no more than cost time per period on implementations that can collect execution time
      per thread.
      */
    public void setCost(RelativeTime cost)
    {
	this.cost = cost;
    }
    
    /**
      Set the cost overrun handler.
      @param handler  This handler is invoked if the <code>run()</code> method of the schedulable
      object of the previous period is still executing at the start of the current period.
      */
    public void setCostOverrunHandler(AsyncEventHandler handler)
    {
	this.overrunHandler = handler;
    }
    
    /**
      Set the deadline value.
      @param deadline The latest permissible completion time measured from the start of the current period.
      Not used in a minimum implementation. Other implmentations may use this parameter
      to compute execution eligibility.  The default value is the same as <code>period</code>.
      */
    public void setDeadline(RelativeTime deadline)
    {
	this.deadline = deadline;
    }
    
    /**
      Set the deadline miss handler.
      @param handler  This handler is invoked if the <code>run()</code> method of the schedulable
      object is still executing after the deadline has passed.
      */
    public void setDeadlineMissHandler(AsyncEventHandler handler)
    {
	this.missHandler = handler;
    }
    
    /**
      Set the period.
      @param period Interval used to enforce allocation of processing resources to the
      associated schedulable objects.  Also used in the feasibility analysis
      and admission control algorithms.
      */
    public void setPeriod(RelativeTime period)
    {
	this.period = period;
    }
    
    /**
      Set the start time.
      @param start Time at which the first period begins.
      */
    public void setStart(HighResolutionTime start)
    {
	if (start == null)
	    this.start = new AbsoluteTime(0,0);
	else
	    this.start = start;
    }
}
