/*
 * -----------------------------------------------------------------
 * RationalTime.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


package javax.realtime;

/**
 * An object that represents a time interval millis/1E3+nanos/1E9 seconds long that
 * is divided into subintervals by some frequency. This is generally used in periodic
 * events, threads, and feasibility analysis to specify periods where there is a basic
 * period that must be adhered to strictly (the interval), but within that interval the
 * periodic events are supposed to happen frequency times, as uniformly spaced as
 * possible, but clock and scheduling jitter is moderately acceptable.
 * Caution: This class is explicitly unsafe in multithreaded situations when it is
 * being changed. No synchronization is done. It is assumed that users of this class who
 * are mutating instances will be doing their own synchronization at a higher level.
 *
 * All Implemented Interfaces: java.lang.Comparable
 */


public class RationalTime extends RelativeTime{

	int freq;
		
	/**
	* Construct a new Object of RationalTime
	* Equivalent to new RationalTime(1000, 0, frequency) -- essentially a cycles -per-second value
	*/
	public RationalTime(int frequency){
		this(frequency,1000,0);
	}

	/**
	* Construct a new Object of RationalTime.
	* All arguments must be >= 0.
	* @param  frequency The frequency value of this
	* @param  millis  The milliseonds value of this
	* @param  nanos   The nanoseconds value of this
	*/
	public RationalTime(int frequency, long millis, int nanos) throws IllegalArgumentException {
		super(millis,nanos);
		if((frequency<0 || millis<0 || nanos<0) || (millis==0 && nanos==0))
			throw new IllegalArgumentException("Frequency or Interval is -ve");

		freq = frequency;
	}

	/**
	* Construct a new Object of RationalTime from the given RelativeTime
	* @param frequency   The frequency value of this
	* @param interval    The relativeTime object used as the source for the copy
	*/
	public RationalTime(int frequency, RelativeTime interval) throws IllegalArgumentException {
		super(interval);
		if(frequency<0)
			throw new IllegalArgumentException("Frequency is -ve");

		freq = frequency;
	}

	/** Convert this time to an absolute time. For a RelativeTime, this invovled
	* adding the clocks notion of now to this interval and constructing a new
	* AbsoluteTime based on the sum
	*
	* @param clock if null, Clock.getRealTimeClock() is used
	* @param destination
	*/
	public AbsoluteTime absolute(Clock clock, AbsoluteTime destination){
		if(clock==null)
			clock = Clock.getRealtimeClock();
		AbsoluteTime at = clock.getTime();
		if(destination!=null){
			destination.set(milliseconds+at.getMilliseconds(),
					nanoseconds+at.getNanoseconds());
			return destination;
		}

		return new AbsoluteTime(milliseconds+at.getMilliseconds(),
					nanoseconds+at.getNanoseconds());
	}

	public AbsoluteTime absolute(Clock clock)
    {
        return this.absolute(clock, null);
	}

	/**
	* Add this time to an AbsoluteTime. It is almost the same dest.add(this, dest) except
	* that it accounts for(ie. divides by) the frequency.
	* @param dest
	*/
	public void addInterarrivalTo(AbsoluteTime destination){

		long totalnano = milliseconds*1000000 + nanoseconds;
		long intinnano = (long)(totalnano/freq);
		long intmillipart = (long)(intinnano/1000000);
		int intnanopart = (int)(intinnano - (intmillipart*1000000));
		
		if(destination != null){
			destination.add(intmillipart,intnanopart);
		}
	}

	public int getFrequency(){
		return freq;
	}
	
       /**
	*  Gets the time duration between two consecutive ticks using frequency
	*
	*  @param   dest    Result is stored in dest and returned, if null new object is returned.
	*/
	public RelativeTime getInterarrivalTime(RelativeTime dest){

		long totalnano = milliseconds*1000000 + nanoseconds;
		long intinnano = (long)(totalnano/freq);
		long intmillipart = (long)(intinnano/1000000);
		int intnanopart = (int)(intinnano - (intmillipart*1000000));

		if(dest != null){
			dest.set(intmillipart,intnanopart);
			return dest;
		}

		return new RelativeTime(intmillipart,intnanopart);
	}

       /**
	*  Gets the time duration between two consecutive ticks using frequency
	*/
	public RelativeTime getInterarrivalTime(){
		long totalnano = milliseconds*1000000 + nanoseconds;
		long intinnano = (long)(totalnano/freq);
		long intmillipart = (long)(intinnano/1000000);
		int intnanopart = (int)(intinnano - (intmillipart*1000000));

		return new RelativeTime(intmillipart,intnanopart);
	}



	/**
	* Change the indicated interval of this to the sum of the values of the arguments
	* @param   millis   Millisecond part
	* @param   nanos    Nanosecond part
	*/
	public void set(long millis, int nanos) throws IllegalArgumentException{
		if((millis<0 || nanos<0) || (millis==0 && nanos==0))
			throw new IllegalArgumentException("Interval is less than zero");
		
		milliseconds = millis;
		nanoseconds = nanos; 
	}

	/**
	* set the frequency of this
	* @param f the frequency to be set for this
	*/
	public void setFrequency(int frequency) throws ArithmeticException{
		if(frequency<0)
			throw new ArithmeticException("Frequency is -ve");
		freq = frequency;
	}

    public RelativeTime relative(Clock clock)
    {
        if(clock == null)
            clock = Clock.getRealtimeClock();
        AbsoluteTime dest = null;
        dest = this.absolute(clock, dest);
        return dest.relative(clock);
    }

    public RelativeTime relative(Clock clock, RelativeTime destination)
    {
        if(clock == null)
            clock = Clock.getRealtimeClock();
        RelativeTime rt = this.relative(clock);
        if(destination != null)
        {
            destination.set(rt);
            return destination;
        }
        return rt;
    }

}

