/*
 * -----------------------------------------------------------------
 * RawMemoryAccess.java : RawMemoryAccess models a range of physical
 *  memory as a fixed sequence of bytes.
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;


/**
 * An instance of <code>RawMemoryAccess</code> models a range of physical
 * memory as a fixed sequence of bytes. A full complement of accessor 
 * methods allow the contents of the physicalarea to be accessed through 
 * offsets from the base, interpreted as byte, short, int, or long data 
 * values or as arrays of these types.
 * <p> 
 * Whether the offset addresses the high-order or low-order byte is based on the
 * value of the BYTE_ORDER static boolean variable in class {@link RealtimeSystem}.
 * <p> 
 * The <code>RawMemoryAccess</code> class allows a real-time program to implement device
 * drivers, memory-mapped I/O, flash memory, battery-backed RAM, and similar
 * low-level software.
 * <p> 
 * A raw memory area cannot contain references to Java objects.
 * Such a capability would be unsafe (since it could be used to defeat
 * Java's type checking) and error-prone (since it is sensitive to the
 * specific representational choices made by the Java compiler).
 * <p> 
 * Many of the constructors and methods
 * in this class throw {@link OffsetOutOfBoundsException}. This
 * exception means that the value given in the offset parameter is either
 * negative or outside the memory area.
 * <p> 
 * Many of the constructors and methods
 * in this class throw {@link SizeOutOfBoundsException}. This
 * exception means that the value given in the size parameter is either
 * negative, larger than an allowable range, or would cause an accessor
 * method to access an address outside of the memory area.
 *
 * @author Pratik Solanki
 */
public class RawMemoryAccess
{
    // The virtual address of the memory area mapped.
    long virtAddress = -1;
    
    long base;
    long size;
    boolean mapped;

    
    private static int SIZEOF_BYTE = 1;
    private static int SIZEOF_SHORT = 2;
    private static int SIZEOF_INT = 4;
    private static int SIZEOF_LONG = 8;

//    protected static PhysicalMemoryFactory factory = DefaultPhysicalMemoryFactory.instance();
    
    /*
     * Constructor reserved for use by the memory object factory.
     */
    public RawMemoryAccess(Object type, long base, long size) 
        throws SecurityException, OffsetOutOfBoundsException, 
            SizeOutOfBoundsException,UnsupportedPhysicalMemoryException,
            MemoryTypeConflictException 
    {
        this.base = base;
        this.size = size;
        //map();
        virtAddress = PhysicalMemoryManager.create(type, base, size);
        mapped = true;
    }
  
    
    /*
     * Constructor reserved for use by the memory object factory.
     */
    public RawMemoryAccess(Object type, long size) 
        throws SecurityException, OffsetOutOfBoundsException, 
            SizeOutOfBoundsException,UnsupportedPhysicalMemoryException,
            MemoryTypeConflictException 
                            
    {
        //this(type, PhysicalMemoryManager.create(type,size), size);
        this.base = 0;
        this.size = size;
        //map();
        virtAddress = PhysicalMemoryManager.create(type, size);
        mapped = true;

    }
    
    /**
     * Creates a <code>RawMemoryAccess</code> object based on the 
     * parameters passed.
     *
     * @param type An Object representing the type of memory required (
     * e.g., dma, shared, etc) - used to define the base 
     * address and control the mapping.
     * @param size The size of the area in bytes.
     *
     * @exception SecurityException The application doesn't have
     * permissions to access physical memory or the given type of memory.
     * @exception OffsetOutOfBoundsException The address is invalid.
     * @exception SizeOutOfBoundsException The size is negative or
     * extends into an invalid range of memory.
     * @exception UnsupportedPhysicalMemoryException Thrown if the underlying 
     * hardware does not support the given type.
     * @exception MemoryTypeConflictException The specified base does not point to
     * memory that matches the request type, or if <code>type</code> specifies
     * attributes with a conflict.
     */
    /*public RawMemoryAccess(Object type, long size)
            throws SecurityException, OffsetOutOfBoundsException, 
                    SizeOutOfBoundsException, UnsupportedPhysicalMemoryException,
                    MemoryTypeConflictException
    {
        //return create(type, factory.getTypedMemoryBase(type,size), size);
        this.base = PhysicalMemoryManager.create(type, size);
        this.size = size;
        mapped = false;
        
    }*/
  
    /**
     * Creates a RawMemoryAccess object based on the parameters passed.
     *
     * @param type An Object representing the type of memory required.
     * Used to define the base address and control the mapping.
     * @param base The physical memory address of the region.
     * @param size The size of the area in bytes.
     *
     * @exception SecurityException The application doesn't have
     * permissions to access physical memory or the given type of memory.
     * @exception OffsetOutOfBoundsException The address is invalid.
     * @exception SizeOutOfBoundsException The size is negative or
     * extends into an invalid range of memory.
     * @exception UnsupportedPhysicalMemoryException Thrown if the underlying 
     * hardware does not support the given type.
     * @exception MemoryTypeConflictException The specified base does not point to
     * memory that matches the request type, or if <code>type</code> specifies
     * attributes with a conflict.
     * @exception MemoryInUseException The specified memory is already in use.
 
     */
    /*public RawMemoryAccess(Object type, long base, long size)
            throws SecurityException, OffsetOutOfBoundsException, 
                    SizeOutOfBoundsException, UnsupportedPhysicalMemoryException,
                    MemoryTypeConflictException
    {
        //return new RawMemoryAccess(base, size);
        this.base = PhysicalMemoryManager.create(type, base, size);
        this.size = size;
        this.mapped = false;
    }*/
    
    /**
     * Get the <code>byte</code> at the given offset.
     * @param offset The offset at which to read the byte.
     * @return The byte read.
     */
    public byte getByte(long offset) 
            throws OffsetOutOfBoundsException, SizeOutOfBoundsException
    {
        standardChecks(offset, SIZEOF_BYTE);
        return getByte0(virtAddress+offset);
    }
    
    /**
     * Get <code>number</code> bytes starting at the given offset and assign 
     * them to the byte array passed starting at position <code>low</code>.
     */
    public void getBytes(long offset, byte[] bytes, int low, int number) 
            throws OffsetOutOfBoundsException, SizeOutOfBoundsException
    {
        arrayChecks(offset, SIZEOF_BYTE, bytes.length, low, number);
        getBytes0(virtAddress+offset, bytes, low, number);
    }
    
    
    /**
     * Get the <code>int</code> at the given offset.
     * @param offset The offset at which to read the integer.
     * @return The int read.
     */
    public int getInt(long offset) 
            throws OffsetOutOfBoundsException, SizeOutOfBoundsException
    {
    	standardChecks(offset, SIZEOF_INT);
	    return getInt0(virtAddress+offset);
    }
    
    /**
     * Get <code>number</code> ints starting at the given offset and assign 
     * them to the int array passed starting at position <code>low</code>.
     */
    public void getInts(long offset, int[] ints, int low, int number) 
            throws OffsetOutOfBoundsException, SizeOutOfBoundsException
    {
        arrayChecks(offset, SIZEOF_INT, ints.length, low, number);
        getInts0(virtAddress+offset, ints, low, number);
    }
    
    /**
     * Get the <code>long</code> at the given offset.
     * @param offset The offset at which to read the long.
     * @return The long read.
     */
    public long getLong(long offset) 
            throws OffsetOutOfBoundsException, SizeOutOfBoundsException
    {
        standardChecks(offset, SIZEOF_LONG);
        return getLong0(virtAddress+offset);
    }
  
    /**
     * Get <code>number</code> longs starting at the given offset and assign 
     * them to the long array passed starting at position <code>low</code>.
     */
    public void getLongs(long offset, long[] longs, int low, int number)
            throws OffsetOutOfBoundsException, SizeOutOfBoundsException
    {
        arrayChecks(offset, SIZEOF_LONG, longs.length, low, number);
        getLongs0(virtAddress+offset, longs, low, number);
    }
    
    /**
     * Get the <code>short</code> at the given offset.
     * @param offset The offset at which to read the short.
     * @return The short read.
     */
    public short getShort(long offset) 
            throws OffsetOutOfBoundsException, SizeOutOfBoundsException
    {
        standardChecks(offset, SIZEOF_SHORT);
        return getShort0(virtAddress+offset);
    }
  
    /**
     * Get <code>number</code> shorts starting at the given offset and assign 
     * them to the short array passed starting at position <code>low</code>.
     */
    public void getShorts(long offset, short[] shorts, int low, int number)
      throws OffsetOutOfBoundsException, SizeOutOfBoundsException
    {
      arrayChecks(offset, SIZEOF_SHORT, shorts.length, low, number);
      getShorts0(virtAddress+offset, shorts, low, number);
    }
    
    /**
     * Set the <code>byte</code> at the given offset.
     * @param offset The offset at which to write the byte.
     * @param value The byte to write.
     */
    public void setByte(long offset, byte value)
	throws OffsetOutOfBoundsException, SizeOutOfBoundsException
    {	
	standardChecks(offset, 1);
	setByte0(virtAddress+offset, value);
	return;
	
    }
    
    /**
     * Set <code>number</code> bytes starting at the given offset from the 
     * byte array passed starting at position <code>low</code>.
     */
    public void setBytes(long offset, byte[] bytes, int low, int number)
	throws OffsetOutOfBoundsException, SizeOutOfBoundsException
    {
      arrayChecks(offset, SIZEOF_BYTE, bytes.length, low, number);
      setBytes0(virtAddress+offset, bytes, low, number);
    }    
  
  
  /**
   * Set the <code>int</code> at the given offset.
   * @param offset The offset at which to write the int.
   * @param value The integer to write.
   */
  public void setInt(long offset, int value) 
    throws OffsetOutOfBoundsException, SizeOutOfBoundsException
  {
    standardChecks(offset, SIZEOF_INT);
    setInt0(virtAddress+offset, value);
  }
    
  /**
   * Set <code>number</code> ints starting at the given offset from the 
   * int array passed starting at position <code>low</code>.
   */
  public void setInts(long offset, int[] ints, int low, int number)
    throws OffsetOutOfBoundsException, SizeOutOfBoundsException
  {
    arrayChecks(offset, SIZEOF_INT, ints.length, low, number);
    setInts0(virtAddress+offset, ints, low, number);
  }
    
  
    /**
     * Set the <code>long</code> at the given offset.
     * @param offset The offset at which to write the long.
     * @param value The long to write.
     */
    public void setLong(long offset, long value) 
      throws OffsetOutOfBoundsException, SizeOutOfBoundsException
    {
	standardChecks(offset, SIZEOF_LONG);
	setLong0(virtAddress+offset, value);
    }
    
  /**
   * Set <code>number</code> longs starting at the given offset from the 
   * long array passed starting at position <code>low</code>.
   */
  public void setLongs(long offset, long[] longs, int low, int number)
    throws OffsetOutOfBoundsException, SizeOutOfBoundsException
  {
    arrayChecks(offset, SIZEOF_LONG, longs.length, low, number);
    setLongs0(virtAddress+offset, longs, low, number);
  }
  
  /**
   * Set the <code>short</code> at the given offset.
   * @param offset The offset at which to write the short.
   * @param value The short to write.
   */
  public void setShort(long offset, short value) 
    throws OffsetOutOfBoundsException, SizeOutOfBoundsException
  {
    standardChecks(offset, 2);
    setShort0(virtAddress+offset, value);
  }
  
  /**
   * Set <code>number</code> shorts starting at the given offset from the 
   * short array passed starting at position <code>low</code>.
   */
  public void setShorts(long offset, short[] shorts, int low, int number)
    throws OffsetOutOfBoundsException, SizeOutOfBoundsException
  {
    arrayChecks(offset, SIZEOF_SHORT, shorts.length, low, number);
    setShorts0(virtAddress+offset, shorts, low, number);
  }
  
  
  /**
   * Returns the virtual memory location at which the memory region is mapped.
   * @return The virtual address to which this is mapped (for reference 
   * purposes). Same as the base address if virtual memory is not supported.
   */
  public long getMappedAddress()
  {
    return virtAddress;
  }
  
  /**
   * Maps the physical memory range into virtual memory. No-op if the system
   * doesn't support virtual memory.
   */
  public long map()
  {
    return map(base, size);
  }
    
  /**
   * Maps the physical memory range into virtual memory at the specified 
   * location. No-op if the system doesn't support virtual memory.
   * @param base The location to map at the virtual memory space.
   */
  public long map(long base)
  {
    return map(base,size);
  }
  
  /**
   * Maps the physical memory range into virtual memory. No-op if the system
   * doesn't support virtual memory.
   * @param base The location to map at the virtual memory space.
   * @param size Teh size of the block to map in.
   */
  public long map(long base, long size)
  {
    try
      {
          if(mapped)
            return -1;
          virtAddress = map0(base, size);
          return virtAddress;
      }
    catch(OffsetOutOfBoundsException obe)
      {
        //System.out.println("The system did not like the offset passed :(");
        obe.printStackTrace();
      }
    catch(SizeOutOfBoundsException sbe)
      {
        //System.out.println("The system did not like the size passed :(");
        sbe.printStackTrace();
      }
    catch(UnsupportedPhysicalMemoryException uspe)
      {
        //System.out.println("The system does not support physical memory");
        uspe.printStackTrace();
      }
    return -1;
  }
  
  /**
   * Unmap the physical memory range from virtual memory. No-op if the system
   * doesn't support virtual memory.
   */
  public void unmap()
  {
      if(mapped)
        PhysicalMemoryManager.free(virtAddress, size);
  }

  protected void finalize()
  {
      unmap();
      //super.finalize();
  }
  
  
  protected void standardChecks(long offset, int datatypeSize) throws OffsetOutOfBoundsException, SizeOutOfBoundsException
  {
    if(/*fd < 0 ||*/ virtAddress == -1)
      {
	throw new OffsetOutOfBoundsException("No memory mapping was done! (I know I should throw another exception)");
      }	  
    if(offset >= size || 
       offset < 0 || 
       (offset+datatypeSize > size)
       )
      {
	throw new OffsetOutOfBoundsException("Offset passed is greater than memory size or less than 0");
      }
  }

  protected void arrayChecks(long offset, int datatypeSize, int arrayLength, int low, int number)
    throws OffsetOutOfBoundsException, SizeOutOfBoundsException
  {
    if(number+low > arrayLength)
      throw new OffsetOutOfBoundsException("Array passed is not big enough");
    if(low < 0 || number < 0)
      throw new OffsetOutOfBoundsException("The starting position (low) or 'number' is negative");
    standardChecks(offset, datatypeSize * number);
  }
    
  private native long map0(long base, long offset) 
            throws OffsetOutOfBoundsException, SizeOutOfBoundsException, UnsupportedPhysicalMemoryException;
  private native void unmap0();
  
  private native byte getByte0(long address);
  private native void setByte0(long address, byte value);
  
  private native int getInt0(long address);
  private native void setInt0(long address, int value);
  
  private native long getLong0(long address);
  private native void setLong0(long address, long value);
  
  private native short getShort0(long address);
  private native void setShort0(long address, short value);
  
  private native void getBytes0(long address, byte[] byteArray, int low, int number);
  private native void setBytes0(long address, byte[] byteArray, int low, int number);

  private native void getInts0(long address, int[] intArray, int low, int number);
  private native void setInts0(long address, int[] intArray, int low, int number);

  private native void getLongs0(long address, long[] longArray, int low, int number);
  private native void setLongs0(long address, long[] longArray, int low, int number); 

  private native void getShorts0(long address, short[] shortArray, int low, int number);
  private native void setShorts0(long address, short[] shortArray, int low, int number);


}
