/*
 * -----------------------------------------------------------------
 * RealtimeSystem.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;

import java.lang.Object;

/**
  <code>RealtimeSystem</code> provides a means for tuning the behavior of the
  implementation by specifying parameters such as the maximum number of
  locks that can be in use concurrently, and the monitor control policy.
  In addition, <code>RealtimeSystem</code> provides a mechanism for obtaining access to
  the security manager, garbage collector and scheduler, to make queries
  from them or to set parameters.
  */
public final class RealtimeSystem
{

    public static final byte BIG_ENDIAN = 1;
    public static final byte LITTLE_ENDIAN = 0;
    public static final byte BYTE_ORDER = LITTLE_ENDIAN;

    private static GarbageCollector gc = new MarkAndSweepCollector();
    private static native void setMaximumConcurrentLocks0(int number);
    private static native int getMaximumConcurrentLocks0();
    private static native int getConcurrentLocksUsed0();
    private static RealtimeSecurity rs = new RealtimeSecurity();

    /**
     Return a reference to the currently active garbage collector
     for the heap.
     @return A {@link GarbageCollector} object which is the current collector collecting
     objects on the traditional Java heap.
     */
    public static GarbageCollector currentGC()
    {
        return gc;
    }
  
    /**
      Get the maximum number of locks that have been used concurrently.
      This value can be used for tuning the concurrent locks parameter, which
      is used as a hint by systems that use a monitor cache.
      @return An int whose value is the number of locks in use at the
      time of the invocation of the method.
      */
    public static int getConcurrentLocksUsed()
    {
        return getConcurrentLocksUsed0();
    }

   /**
     Get the maximum number of locks that can be used concurrently without
     incurring an execution time increase
     as set by the <code>setMaximumConcurrentLocks()</code> methods.
     @return An int whose value is the maximum number of locks that can be in simultaneous use.
     */
    public static int getMaximumConcurrentLocks() 
    {
        return getMaximumConcurrentLocks0();
    }
  
    /**
      Get a reference to the security manager used to control
      access to real-time system features such as access to
      physical memory.
      @return A {@link RealtimeSecurity} object representing the default real-time security manager.
      */
    public static RealtimeSecurity getSecurityManager()
    {
        return rs;
    }
    
    /**
     Set the anticipated maximum number of locks that may be held
     or waited on concurrently.  Provide a hint to systems that
     use a monitor cache as to how much space to dedicate to the cache.

     @param number An integer whose value becomes the number of locks that
     can be in simultaneous use without incurring an execution time increase.
     If <code>number</code> is less than or equal to zero nothing happens.
    */
    public static void setMaximumConcurrentLocks(int numLocks)
    {
        setMaximumConcurrentLocks0(numLocks);
    }
    
    /**
     Set the anticipated maximum number of locks that may be held
     or waited on concurrently.  Provide a limit for the size of
     the monitor cache on systems that provide one if hard is true.
     
     @param number The maximum number of locks that can be in simultaneous
     use without incurring an execution time increase.
     If <code>number</code> is less than or equal to zero nothing happens.
     @param hard If true, <code>number</code> sets a limit.  If a lock is attempted which would cause the
     number of locks to exceed <code>number</code> then a {@link ResourceLimitError} is thrown.
    */
    public static void setMaximumConcurrentLocks(int number, boolean hard)
    {
        if (hard && (number > 0))
            setMaximumConcurrentLocks0(number);
    }
  
    /**
      Set a new real-time security manager.
      @param manager A {@link RealtimeSecurity} object which will become the new security manager.
      @exception SecurityException Thrown if security manager has already been set.
     */
    public static void setSecurityManager(RealtimeSecurity manager)
    {
        rs = manager;
    }
}
