/*
 * -----------------------------------------------------------------
 * ReleaseParameters.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;

import java.util.*;

/** 
  The abstract top-level class for release characteristics of threads.

  When a reference to a <code>ReleaseParameters</code> object is given
  as a parameter to a constructor, the <code>ReleaseParameters</code> object becomes bound
  to the object being created. Changes to the values in the <code>ReleaseParameters</code>
  object affect the constructed object. If given to more than one constructor,
  then changes to the values in the <code>ReleaseParameters</code> object affect <em>all</em>
  of the associated objects. Note that this is a one-to-many relationship and
  <em>not</em> a many-to-many.

  <p><b>Caution:</b> This class is explicitly unsafe in multithreaded
  situations when it is being changed.  No synchronization is done.  It
  is assumed that users of this class who are mutating instances will be
  doing their own synchronization at a higher level.

  <p><b>Caution:</b> The <code>cost</code> parameter time should be considered to be 
  measured against the target platform.
  */

public class ReleaseParameters
{
    protected RelativeTime cost;
    protected RelativeTime deadline;
    protected AsyncEventHandler overrunHandler;
    protected AsyncEventHandler missHandler;

    protected Vector schedSet = new Vector();

    void bind(Schedulable sch)
    {
	if (!schedSet.contains(sch))
	    {
		schedSet.addElement(sch);
	    }
    }

    void unbind(Schedulable sch)
    {
	if (schedSet.contains(sch))
	    {
		schedSet.remove(sch);
	    }
    }


    public boolean setIfFeasible(RelativeTime cost, RelativeTime deadline)
    {
	boolean retval = true;
	Schedulable s;
	RelativeTime oldCost, oldDeadline;

	oldCost = getCost();
	oldDeadline = getDeadline();

	setCost(cost);
	setDeadline(deadline);

	// get globalLock to lock all schedulers
	
	synchronized(Scheduler.class)
	    {
		for (int i=0; i< schedSet.size() ; i++)
		    {
			s = (Schedulable) schedSet.elementAt(i);
			if (!s.setReleaseParametersIfFeasible(this))
			    {
				// undo the changes

				setCost(oldCost);
				setDeadline(oldDeadline);
				for (int j = i -1; i> 0 ; i--)
				    {
					s = (Schedulable) schedSet.elementAt(i);
					s.setReleaseParametersIfFeasible(this);
				    }
				retval = false;
				break;
			    }
		    }
	    }
	return retval;
    }

    protected ReleaseParameters()
    {
    }
    
    /** Subclasses use this constructor to create a <code>ReleaseParameters</code> type object.
     *
     *  @param cost Processing time units per interval.
     *              On implementations which can measure the amount of time
     *              a schedulable object is executed, this value is the maximum amount of time
     *              a schedulable object receives per interval.
     *              On implementations which cannot measure execution time, this value is used
     *              as a hint to the feasibility algorithm.  On such systems it is not possible
     *              to determine when any particular object exceeds cost.
     *              Equivalent to <code>RelativeTime(0,0)</code> if null.
     *  @param  deadline  The latest permissible completion time measured from the
     *                    release time of the associated invocation of the schedulable object.
     *                    Changing the deadline might not take effect after the expiration 
     *                    of the current deadline.
     *                    More detail provided in the subclasses.
     *
     *  @param overrunHandler  This handler is invoked if an invocation of the schedulable
     *                         object exceeds cost.  Not required for minimum implementation.
     *                         If null, nothing happens on the overrun condition, and
     *                         waitForNextPeriod returns false immediately and updates
     *                         the start time for the next period.
     *  @param missHandler  This handler is invoked if the <code>run()</code> method of the schedulable
     *                      object is still executing after the deadline has passed.
     *                      Although minimum implementations do not consider deadlines in
     *                      feasibility calculations, they must recognize variable deadlines
     *                      and invoke the miss handler as appropriate.
     *                      If null, nothing happens on the miss deadline condition.
     */
    protected ReleaseParameters(RelativeTime cost,
				RelativeTime deadline,
				AsyncEventHandler overrunHandler,
				AsyncEventHandler missHandler)
    {
	this.cost = cost;
	this.deadline = deadline;
	this.overrunHandler = overrunHandler;
	this.missHandler = missHandler;
    }
    
    /** Get the cost value.
     */
    public RelativeTime getCost()
    {
	return cost;
    }
  
   /** Get the cost overrun handler.
     */
    public AsyncEventHandler getCostOverrunHandler()
    {
	return overrunHandler;
    }
  
   /** Get the deadline.
     */
    public RelativeTime getDeadline()
    {
	return deadline;
    }
  
   /** Get the deadline miss handler.
     */
    public AsyncEventHandler getDeadlineMissHandler()
    {
	return missHandler;
    }
    
    /** Set the cost value.
     *  @param cost Processing time units per period or per minimum interarrival interval.
     *              On implementations which can measure the amount of time
     *              a schedulable object is executed, this value is the maximum amount of time
     *              a schedulable object
     *              receives per period or per minimum interarrival interval.
     *              On implementations which cannot measure execution time, this value is used
     *              as a hint to the feasibility algorithm.  On such systems it is not possible
     *              to determine when
     *              any particular object exceeds or will exceed cost time units in a period
     *              or interval.
     *              Equivalent to <code>RelativeTime(0,0)</code> if null.
     */

    public void setCost(RelativeTime cost)
    {
	this.cost = cost;
    }
    
    /** Set the cost overrun handler.
     *  @param handler  This handler is invoked if an invocation of the schedulable
     *                         object exceeds cost.  Not required for minimum implementation.
     *                         See comments in <code>setCost()</code>.
     */
    public void setCostOverrunHandler(AsyncEventHandler handler)
    {
	this.overrunHandler = handler;
    }
    
    /** Set the deadline value.
     *  @param  deadline  The latest permissible completion time measured from the
     *                    release time of the associated invocation of the schedulable object.
     *                    For a minimum implementation for purposes of feasibility analysis,
     *                    the deadline is equal to the period
     *                    or minimum interarrival interval.
     *                    Other implementations may use this parameter
     *                    to compute execution eligibility.
     */
    public void setDeadline(RelativeTime deadline)
    {
	this.deadline = deadline;

	Schedulable s;

	for (int i = 0 ; i < schedSet.size(); i++)
	    {
		s = (Schedulable) schedSet.elementAt(i);
		s.setReleaseParameters((ReleaseParameters)this);
	    }
    }


    /** Set the deadline miss handler.
      *  @param handler  This handler is invoked if the <code>run()</code> method of the schedulable
      *                      object is still executing after the deadline has passed.
      *                      Although minimum implementations do not consider deadlines in
      *                      feasibility calculations, they must recognize variable deadlines
      *                      and invoke the miss handler as appropriate.
      */
    public void setDeadlineMissHandler(AsyncEventHandler handler)
    {
	this.missHandler = handler;
    }
}
