/*
 * -----------------------------------------------------------------
 * Scheduler.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;

/**
  An instance of <code>Scheduler</code> manages
  the execution of schedulable objects and may implement a feasibility
  algorithm. The feasibility algorithm may determine if the
  known set of schedulable objects, given their
  particular execution ordering (or priority assignment), is a feasible schedule.

  Subclasses of <code>Scheduler</code> are used for alternative scheduling policies
  and should define an <code>instance()</code> class method to return the
  default instance of the subclass.  The name of the subclass should be descriptive
  of the policy, allowing applications to deduce the policy available for the scheduler
  obtained via {@link Scheduler#getDefaultScheduler} (e.g., <code>EDFScheduler</code>).
  */

public abstract class Scheduler
{

    private static Scheduler defaultScheduler = PriorityScheduler.instance();

    /* ISHIN */
    private final static int PRIORITY = 1;
    private final static int EDF      = 2;

    private native static void setScheduler (int s);

    private static int nScheduler     = PRIORITY;
    public static int getScheduler()
    {
	return nScheduler;
    }
    /**
      Set the default scheduler.  This is the scheduler given to instances of
      {@link RealtimeThread} when they are constructed.
      The default scheduler is set to the required {@link PriorityScheduler} at startup.
      @param scheduler The <code>Scheduler</code> that becomes the default scheduler
                assigned to new threads.
		If null nothing happens.
      */
    public static void setDefaultScheduler(Scheduler scheduler)
    {
	defaultScheduler = scheduler;
	if (scheduler instanceof PriorityScheduler) nScheduler = PRIORITY;
	else if (scheduler instanceof EDFScheduler) nScheduler = EDF;
	setScheduler (nScheduler);
    }
    /* ISHIN */

    /* TimeSys internal... should be overloaded */

    boolean inFeasibility(Schedulable sched)
    {
	return false;
    }


    protected Scheduler()
    {
    }
    
    /** 
      Inform the scheduler that this thread's {@link ReleaseParameters} should be considered in 
      feasibility analysis until further notified.
      */
    protected abstract boolean addToFeasibility(Schedulable schedulable);
    
    /**
      Returns true if, after changing the
      {@link Schedulable}'s release and GC parameters
      isFeasible would return true.  The parameters wil
      be changed.  If the resulting system would not be
      feasible, this method returns false and no changes
      are made.
      @param schedulable The {@link Schedulable} object for which to check admittance.
         If null nothing happens.
      @param release The proposed {@link ReleaseParameters}.  If null, no
          change is made.
      @param memory The proposed {@link MemoryParameters}.  If null, no
          change is made.
      */
    public boolean setIfFeasible(Schedulable schedulable,
				    ReleaseParameters release,
				    MemoryParameters memory)
    {
	return false;
    }
    
    public boolean setIfFeasible(Schedulable schedulable,
				    ReleaseParameters release,
				    MemoryParameters memory,
				    ProcessingGroupParameters group)
    {
	return false;
    }

    /**
      Return a reference to the default scheduler.
     */
    public static Scheduler getDefaultScheduler()
    {
	return defaultScheduler;
    }
    
    /**
      Used to determine the policy of the <code>Scheduler</code>.
      @return A String object which is the name of the scheduling policy used by this.
     */
    public abstract java.lang.String getPolicyName();
    
    /** 
      Returns true if and only if the system is able to satisfy the
      constraints expressed in the release parameters of the
      existing schedulable objects.
      */
    public abstract boolean isFeasible();
    
    /**
      Inform the scheduler that this thread's {@link ReleaseParameters} should not be considered in
      feasibility analysis until further notified.
      */
    protected abstract boolean removeFromFeasibility(Schedulable schedulable);


    /**
      Trigger the execution of a schedulable object (like an {@link AsyncEventHandler}).
      @param h The schedulable object to make active.
     */
    public abstract void fireSchedulable(Schedulable schedulable);
}
