/*
 * -----------------------------------------------------------------
 * SizeEstimator.java - This class helps in determining the size of 
 * a MemoryArea given the objects that will be created in it.
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;

/**
 * This is a convenient class to help people figure out how much memory
 * they need. Instead of passing actual numbers to the 
 * <code>MemoryArea</code> constructors, one can pass <code>SizeEstimator
 * </code> objects with which you can have a better feel of how big a 
 * memory area you require.
 *
 * @author Pratik Solanki
 * @see MemoryArea#MemoryArea(SizeEstimator)
 * @see LTMemory#LTMemory(SizeEstimator, SizeEstimator)
 * @see VTMemory#VTMemory(SizeEstimator, SizeEstimator)
 *
 */
public final class SizeEstimator 
{
    private long estimate = 0; 
    private long numberOfObjects = 0;

    /**
     * Take into account additional <code>n</code> instances of 
     * Class <code>c</code> when estimating the size of the 
     * {@link javax.realtime.MemoryArea}.
     *
     * @param c The class to take into account.
     * @param n The number of instances of <code>c</code> to estimate.
     */
    public void reserve(Class c, int n)
    {
        // reserve room for n instances of class c
        if(c == null)
            return;
        estimate += n*sizeOf(c);
        numberOfObjects += n;
    }

    /**
     * Take into account additional <code>n</code> instances of 
     * SizeEstimator <code>s</code> when estimating the size of the 
     * {@link javax.realtime.MemoryArea}.
     *
     * @param c The class to take into account.
     * @param n The number of instances of <code>c</code> to estimate.
     */
    public void reserve(SizeEstimator s, int n)
    {
        if(s == null)
            return;
        estimate += s.getEstimate()*n;
    }

    /**
     * Take into account an additional instance of 
     * SizeEstimator <code>s</code> when estimating the size of the 
     * {@link javax.realtime.MemoryArea}.
     *
     * @param c The class to take into account.
     * @param n The number of instances of <code>c</code> to estimate.
     */
    public void reserve(SizeEstimator s)
    {
        reserve(s, 1);
    }

    /**
     * Returns an estimate of the nummber of bytes needed to store all the 
     * all the objects reserved.
     *
     * @returns The estimated size in bytes.
     */
    public long getEstimate()
    {
        return estimate;
    }
    
    private native int sizeOf(Class c);
}

