/*
 * -----------------------------------------------------------------
 * SporadicParameters.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;
import java.util.*;

/** 
  A notice to the scheduler that the associated schedulable object's 
  run method will be released aperiodically but with a
  minimum time between releases.

  When a reference to a <code>SporadicParameters</code> object is given
  as a parameter to a constructor, the <code>SporadicParameters</code> object becomes bound
  to the object being created. Changes to the values in the <code>SporadicParameters</code>
  object affect the constructed object. If given to more than one constructor,
  then changes to the values in the <code>SporadicParameters</code> object affect <em>all</em>
  of the associated objects. Note that this is a one-to-many relationship and
  <em>not</em> a many-to-many.

  <p><b>Caution:</b> This class is explicitly unsafe in multithreaded
  situations when it is being changed.  No synchronization is done.  It
  is assumed that users of this class who are mutating instances will be
  doing their own synchronization at a higher level.
  */

public class SporadicParameters extends AperiodicParameters
{

    public static final java.lang.String arrivalTimeQueueOverflowExcept="arrivalTimeQueueOverflowExcept";
    public static final java.lang.String arrivalTimeQueueOverflowIgnore="arrivalTimeQueueOverflowIgnore";
    public static final java.lang.String arrivalTimeQueueOverflowReplace="arrivalTimeQueueOverflowReplace";
    public static final java.lang.String arrivalTimeQueueOverflowSave="arrivalTimeQueueOverflowSave";
    public static final java.lang.String mitViolationExcept="mitViolationExcept";
    public static final java.lang.String mitViolationIgnore="mitViolationIgnore";
    public static final java.lang.String mitViolationReplace="mitViolationReplace";
    public static final java.lang.String mitViolationSave="mitViolationSave";
    

    RelativeTime minInterarrival;
     
     /** Create a <code>SporadicParameters</code> object.
     *
     *  @param minInterarrival The release times of the schedulable object will
     *                         occur no closer than this interval.
     *                         Must be greater than zero when entering feasibility analysis.
     *
     *  @param cost            Processing time per minimum interarrival interval.
     *                         On implementations which can measure the amount of time
     *                         a schedulable object is executed, this value is the maximum amount of time
     *                         a schedulable object receives per interval.
     *                         On implementations which cannot measure execution time, this value is used
     *                         as a hint to the feasibility algorithm. On such systems it is not possible
     *                         to determine when any particular object exceeds cost.
     *                         Equivalent to <code>RelativeTime(0,0)</code> if null.
     *
     *  @param  deadline       The latest permissible completion time measured from the
     *                         release time of the associated invocation of the schedulable object.
     *                         For a minimum implementation for purposes of feasibility analysis,
     *                         the deadline is equal to the minimum interarrival interval.
     *                         Other implementations may use this parameter
     *                         to compute execution eligibility.
     *                         If null, deadline will equal the minimum interarrival time.
     *
     *  @param overrunHandler  This handler is invoked if an invocation of the schedulable
     *                         object exceeds cost.  Not required for minimum implementation.
     *                         If null, nothing happens on the overrun condition.
     *
     *  @param missHandler     This handler is invoked if the <code>run()</code> method of the schedulable
     *                         object is still executing after the deadline has passed.
     *                         Although minimum implementations do not consider deadlines in
     *                         feasibility calculations, they must recognize variable deadlines
     *                         and invoke the miss handler as appropriate.
     *                         If null, nothing happens on the miss deadline condition.
     */
    public SporadicParameters(RelativeTime minInterarrival,
			      RelativeTime cost,
			      RelativeTime deadline,
			      AsyncEventHandler overrunHandler,
			      AsyncEventHandler missHandler)
    {
	super(cost, deadline, overrunHandler, missHandler);
	if (minInterarrival == null)
	    minInterarrival =  new RelativeTime(0,0);

	this.minInterarrival = minInterarrival;
    }
    


    public boolean setIfFeasible(RelativeTime interarrival, RelativeTime cost, RelativeTime deadline)
    {
	boolean retval = true;
	Schedulable s;
	RelativeTime oldCost, oldDeadline, oldInterarrival;

	oldInterarrival = getMinimumInterarrival();
	oldCost = getCost();
	oldDeadline = getDeadline();

	setMinimumInterarrival(interarrival);
	setCost(cost);
	setDeadline(deadline);

	// get globalLock to lock all schedulers
	
	synchronized(Scheduler.class)
	    {
		for (int i=0; i< schedSet.size() ; i++)
		    {
			s = (Schedulable) schedSet.elementAt(i);
			if (!s.setReleaseParametersIfFeasible(this))
			    {
				// undo the changes

				setMinimumInterarrival(oldInterarrival);
				setCost(oldCost);
				setDeadline(oldDeadline);
				for (int j = i -1; i> 0 ; i--)
				    {
					s = (Schedulable) schedSet.elementAt(i);
					s.setReleaseParametersIfFeasible(this);
				    }
				retval = false;
				break;
			    }
		    }
	    }
	return retval;
    }




    /** 
      Get the minimum interarrival time.
     */
    public RelativeTime getMinimumInterarrival()
    {
	return minInterarrival;
    }
    
    /** 
      Set the minimum interarrival time.
      @param minimum The release times of the schedulable object will
                             occur no closer than this interval.
                             Must be greater than zero when entering feasibility analysis.
     */
    public void setMinimumInterarrival(RelativeTime minimum)
    {
	if (minimum == null)
	    minimum = new RelativeTime(0,0);

	this.minInterarrival = minimum;
    }

    private String arrivalTimeQueueOverflowBehavior;
    private int initialArrivalTimeQueueLength;
    private java.lang.String mitViolationBehavior;

    public java.lang.String getArrivalTimeQueueOverflowBehavior()    
    {
	return arrivalTimeQueueOverflowBehavior;
    }

   public int getInitialArrivalTimeQueueLength() 
    {
	return initialArrivalTimeQueueLength;
    }
    
    public java.lang.String getMitViolationBehavior()
    {
	return mitViolationBehavior;
    }

    public void setArrivalTimeQueueOverflowBehavior(java.lang.String behavior)
    {
	arrivalTimeQueueOverflowBehavior = behavior;
    }

    public void setInitialArrivalTimeQueueLength(int initial)
    {
	initialArrivalTimeQueueLength = initial;
    }

    public void setMitViolationBehavior(java.lang.String behavior)   
    {
	mitViolationBehavior = behavior;
    }
}
