/*
 * -----------------------------------------------------------------
 *  WaitFreeDequeue.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;

/**
  The wait-free queue classes facilitate communication and synchronization
  between instances of {@link RealtimeThread} and 
  {@link Thread}. See {@link WaitFreeWriteQueue} or
  {@link WaitFreeReadQueue} for more details.  Instances of this class create
  a {@link WaitFreeWriteQueue} and a {@link WaitFreeReadQueue} and make calls on the
  respective <code>read()</code> and <code>write()</code> methods.
 */

public class WaitFreeDequeue
{
    WaitFreeWriteQueue writeQ;
    WaitFreeReadQueue  readQ;

    /**
      A queue with unsynchronized and nonblocking 
      <code>read()</code> and <code>write()</code> methods and
      synchronized and blocking <code>read()</code>and <code>write()</code>  methods.
      @param writer An instance of Thread.
      @param reader An instance of Thread.
      @param maximum Then maximum number of elements in the both the
      {@link WaitFreeReadQueue} and the {@link WaitFreeWriteQueue}.
      @param area The {@link MemoryArea} in which this object and internal elements are
      allocated.
      */
    public WaitFreeDequeue (Thread writer, Thread reader, int maximum, MemoryArea area)
	throws IllegalArgumentException, IllegalAccessException, ClassNotFoundException, InstantiationException
    {
	writeQ = new WaitFreeWriteQueue(writer, reader, maximum, area);
	readQ = new WaitFreeReadQueue(writer, reader, maximum, area);
    }
    
    /**
      An unsynchronized call of the <code>read()</code> method of the underlying {@link WaitFreeReadQueue}.
      @return A {@link Object} object read from this.  If there are no elements
      in this then null is returned.
      */
    public Object nonBlockingRead() 
    {
	return readQ.read();
    }

    /**
      A synchronized call of the <code>write()</code> method of the underlying {@link WaitFreeReadQueue}.
      This call blocks on queue full and waits until there is space in this.
      @param object The {@link Object} to place in this.
      @return True if <code>object</code> is now in this.
      */
    public boolean blockingWrite(Object object)
	throws MemoryScopeException
    {
	readQ.write(object);
	return true;
    }

    /**
      An unsynchronized call of the <code>write()</code> method of the underlying {@link WaitFreeWriteQueue}.
      This call does not block on queue full.
      @param object The {@link Object} to attempt to place in this.
      @return True if the <code>object</code> is now in this, otherwise returns false.
      */
    public boolean nonBlockingWrite(Object object)
	throws MemoryScopeException
    {
	return writeQ.write(object);
    }

    /**
      A synchronized call of the <code>read()</code> method of the underlying {@link WaitFreeWriteQueue}.
      This call blocks on queue empty and will wait until there is an element in
      the queue to return.
      @return An {@link Object} from this.
      */
    public Object blockingRead() 
    {
	return writeQ.read();
    }

    /**
      If this is full then this call overwrites the last object
      written to this with the given object.  If this is not full this call
      is equivalent to the  <code>nonBlockingWrite()</code> call.
      @param object The {@link Object} which will overwrite the last object if this is full.
                 Otherwise <code>object</code> will be placed in this.
     */
    public boolean force(Object object) 
    {
	try
	    {
		return writeQ.force(object);
	    }
	catch(Exception e)
	    {
		return false;
	    }
    }
}
