/*
 * -----------------------------------------------------------------
 * WaitFreeReadQueue.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;
/**
  The wait-free queue classes facilitate communication and synchronization
  between instances of {@link RealtimeThread} and {@link Thread}.  
  The problem is that synchronized access
  objects shared between real-time threads and threads might cause the real-time threads to
  incur delays due to execution of the garbage collector.  
  
  <p>The <code>read()</code> method of
  this class does not block on an imagined queue-empty condition variable.  If the
  <code>read()</code> is called on an empty queue null is returned.  If two
  real-time threads intend to read from this queue they must provide their own
  synchronization. 

  <p>The write method of this queue is synchronized and may be
  called by more than one writer and will block on queue empty.
  */

public class WaitFreeReadQueue
{

    Object [] theQueue;
    int head=0, tail=0, queueSize=0;
    MemoryArea memArea, allowableMemory;
    Object ReaderWaitingLock;
    boolean ReaderWaiting = false;
    Object WriterWaitingLock;
    boolean WriterWaiting = false;
    boolean notify = false;
    AsyncEvent addedEvent = null;

/*    class AddEventHandler extends AsyncEventHandler 
    {
        AsyncEvent event;

        public AddEventHandler() 
	{
	    // Non-realtime parameters to avoid priority boosting of regular thread
	    super(null,null,null,null,null,true,null); 
        }
	
        public void handleAsyncEvent() 
	{
	    synchronized (event) 
		{
		    event.notify();
		}
        }
	public void setEvent(AsyncEvent event) 
	{
	    this.event = event;
	}
    }*/
 
    /**
      A queue with an unsynchronized and nonblocking <code>read()</code> method and a 
      synchronized and blocking <code>write()</code> method.  
      The memory areas of the given threads are found.  If these memory areas are 
      the same the queue is created in that memory area.  If these memory areas are 
      different the queue is created in the memory area accessible by the most 
      restricted thread type.
      @param writer An instance of {@link Thread}.
      @param reader An instance of {@link Thread}.
      @param maximum The maximum number of elements in the queue.
      @param memory The {@link MemoryArea} in which this object and internal elements are stored.
      */
   
    public WaitFreeReadQueue(java.lang.Thread writer,
			     java.lang.Thread reader,
			     int maximum,
			     MemoryArea memory,
			     boolean notify) throws 
			          IllegalArgumentException, 
				  InstantiationException, 
				  ClassNotFoundException, 
				  IllegalAccessException
    {
	this(writer,reader,maximum,memory);
	this.notify=notify;
	if (notify)
	    {
		addedEvent = (AsyncEvent) memArea.newInstance(AsyncEvent.class);
		AddEventHandler handler = (AddEventHandler)memArea.newInstance(AddEventHandler.class);
		handler.setEvent(addedEvent);
		addedEvent.addHandler(handler);
	    }
    }

    /**
      A queue with an unsynchronized and nonblocking <code>read()</code> method and a
      synchronized and blocking <code>write()</code> method.
      @param writer An instance of {@link Thread}.
      @param reader An instance of {@link Thread}.
      @param maximum The maximum number of elements in the queue.
      @param memory The {@link MemoryArea} in which this object and internal elements are stored.
      @param notify Whether or not the reader is notified when data is added.
      */
    public WaitFreeReadQueue(java.lang.Thread writer,
			     java.lang.Thread reader,
			     int maximum,
			     MemoryArea memory) throws 
			          IllegalArgumentException, 
				  InstantiationException, 
				  ClassNotFoundException, 
				  IllegalAccessException
    {
	MemoryArea writerMemoryArea=null, readerMemoryArea=null;
	/* find the memoryArea where to create it */

	/* should check that object being enqueued are from
	 * this area or from
	 * immortal memory
	 */

	allowableMemory = null;

	if (writer instanceof RealtimeThread)
	    {
		writerMemoryArea = ((RealtimeThread)writer).getMemoryArea();
	    }
	else
	    {
		// simple Thread
		writerMemoryArea = HeapMemory.instance();
	    }		
	
	if (reader instanceof RealtimeThread)
	    {
		readerMemoryArea = ((RealtimeThread)reader).getMemoryArea();
	    }
	else
	    {
		// simple Thread
		readerMemoryArea = HeapMemory.instance();
	    }
	
	/* Common area? */
	if ((writerMemoryArea != null) && (writerMemoryArea.equals(readerMemoryArea)))
	    {
		allowableMemory = writerMemoryArea;
	    }
	else
	    {
		allowableMemory = ImmortalMemory.instance();
	    }

	if (memory == null)
	    {
		memArea = ImmortalMemory.instance();
	    }
	else
	    memArea = memory;


	/* Allocate the queue */

	theQueue = (java.lang.Object[])memArea.newArray(Class.forName("java.lang.Object"),maximum + 1);

	ReaderWaitingLock = memArea.newInstance(Class.forName("java.lang.Object"));
	MonitorControl.setMonitorControl(ReaderWaitingLock,new PriorityInheritance());
	WriterWaitingLock = memArea.newInstance(Class.forName("java.lang.Object"));
	MonitorControl.setMonitorControl(WriterWaitingLock,new PriorityInheritance());

	if (theQueue != null)
	    {
		queueSize = maximum + 1;
	    }
    }
    
    /**
      Set this to empty.
     */
    public void clear()
    {
	head = tail = 0;
    }
   
   /**
      Used to determine if this is empty.
      @return True if this is empty and false if this is not empty.
     */
    public boolean isEmpty()
    {
	return (head == tail);
    }

    /**
      Used to determine if this is full.
      @return True if this is full and false if this is not full.
     */
    public boolean isFull()
    {
	/* if next insertion will result in a clash with
	 * the head we consider the queue full
	 */

	return (( (tail+1)%queueSize) == head );
    }

     /**
      Returns the next element in the queue unless the queue is empty.  If the
      queue is empty null is returned.
      */
    /* note that because there can be only one reader at any time
     * and only the reader can modify the head, then we cannot expect
     * the head value to move from the test "isEmpty()" to the actual
     * update of it. On the other hand we expect tail to be modified
     * between the test and the modification. Two cases are expected:
     *
     * 1. isEmpty() return true and then tail is modified to turn
     *    the condition into false. In this case the behavior is not 
     *    to retrieve any object from the queue. We consider that this 
     *    is a minor miss-behavior worth trading off for the gain in
     *    concurrency.
     *
     * 2. isEmpty() returned false and tail is modified to turn the 
     *    condition into false. THIS CANNOT HAPPEN given that turning
     *    the condition from false to true is an effect of this same
     *    function, read().
     */

    public java.lang.Object read()
    {
	java.lang.Object obj = null;
	if (!isEmpty())
	    {
		obj = theQueue[head];
		head = (head + 1) % queueSize;

		
		/*
		if (WriterWaiting)
		    synchronized(WriterWaitingLock)
			{
			    WriterWaiting = false;
			    WriterWaitingLock.notify();
			}
		*/
	    }
	return obj;
    }

    /**
      Used to determine the number of elements in this.
      @return An integer which is the number of empty positions in this.
     */
    public int size()
    {
	if (head <= tail)
	    return (queueSize - (tail - head) -1);
	return (head - tail + 1);
    }
  
  /**
    If this is empty <code>waitForData()</code> waits on the event until the writer 
    inserts data. Note that true priority inversion does not occur since
    the writer locks a different object and the notify is executed by
    the {@link AsyncEventHandler} which has <code>noHeap</code> characteristics.
   */
    public void waitForData()
    {
		if (addedEvent != null) 
		{
			synchronized (addedEvent) 
			{
				while(isEmpty())
				{
					try 
					{
						addedEvent.wait();
					} 
					catch(InterruptedException ie)
					{
					}
				}
			}
		}

	/*
	if (isEmpty())
	    {
		synchronized(ReaderWaitingLock)
		    {
			ReaderWaiting = true;
			try
			    {
				ReaderWaitingLock.wait();
			    }
			catch(Exception e)
			    {
			    }
		    }
	    }
	*/
    }
    
    /**
      The synchronized and blocking write.  This call blocks on queue full and
      will wait until there is space in the queue.
      @param object The {@link Object} that is placed in this.
      */

    public synchronized boolean write (java.lang.Object object)
           throws MemoryScopeException
    {
	MemoryArea area = MemoryArea.getMemoryArea(object);
	if (!(area instanceof ImmortalMemory) && 
	    (!area.equals(allowableMemory)))
	    {
		throw new MemoryScopeException();
	    }

	/* Please read the arguments for the synchronization
	 * between reader and writer for the read() method
	 */
     
	while (isFull())
	    {
		try
		    {
			Thread.sleep(10);
		    }
		catch(Exception e)
		    {
		    }
	    }

	/*
	if (isFull())
	    {
		synchronized(WriterWaitingLock)
		    {
			WriterWaiting = true;
			try
			    {
				WriterWaitingLock.wait();
			    }
			catch(Exception e)
			    {
			    }
		    }
	    }
	*/

	boolean empty = isEmpty();

	theQueue[tail] = object;
	tail = (tail+1)%queueSize;
	
	if (addedEvent != null)
	    {
		addedEvent.fire();
	    }

	return true;
    }
}
