/*
 * -----------------------------------------------------------------
 * WaitFreeWriteQueue.java :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying Common Public License in file CPLicense.  Your use
 * of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */

package javax.realtime;

/**
  The wait-free queue classes facilitate communication and synchronization
  between instances of {@link RealtimeThread} and {@link Thread}.  
  The problem is that synchronized access
  objects shared between real-time threads and threads might cause the real-time threads to
  incur delays due to execution of the garbage collector.  

  <p>The <code>write</code> method of
  this class does not block on an imagined queue-full condition variable.  If the
  <code>write()</code> method is called on a full queue false is returned.  If two
  real-time threads intend to read from this queue they must provide their own
  synchronization. 
  
  <p>The <code>read()</code> method of this queue is synchronized and may be
  called by more than one writer and will block on queue empty.
  */

public class WaitFreeWriteQueue
{
    Object [] theQueue;
    int head=0, tail=0, queueSize=0;
    MemoryArea memArea, allowableMemory;
    Object ReaderWaitingLock;
    boolean ReaderWaiting = false;
    Object WriterWaitingLock;
    boolean WriterWaiting = false;

    /** A queue with an unsynchronized and nonblocking <code>write()</code> method and a
      synchronized and blocking <code>read()</code> method.
      @param writer An instance of {@link Thread}.
      @param reader An instance of {@link Thread}.
      @param maximum The maximum number of elements in the queue.
      @param memory The {@link MemoryArea} in which this object and internal elements are
      allocated.
      */
    public WaitFreeWriteQueue(java.lang.Thread writer,
			     java.lang.Thread reader,
			     int maximum,
			     MemoryArea memory) throws 
			          IllegalArgumentException, 
				  InstantiationException, 
				  ClassNotFoundException, 
				  IllegalAccessException
    {
	MemoryArea writerMemoryArea=null, readerMemoryArea=null;
	/* find the memoryArea where to create it */

	/* should check that object being enqueued are from
	 * this area or from
	 * immortal memory
	 */

	allowableMemory = null;

	if (writer instanceof RealtimeThread)
	    {
		writerMemoryArea = ((RealtimeThread)writer).getMemoryArea();
	    }
	else
	    {
		// simple Thread
		writerMemoryArea = HeapMemory.instance();
	    }		
	
	if (reader instanceof RealtimeThread)
	    {
		readerMemoryArea = ((RealtimeThread)reader).getMemoryArea();
	    }
	else
	    {
		// simple Thread
		readerMemoryArea = HeapMemory.instance();
	    }
	
	/* Common area? */
	if ((writerMemoryArea != null) && (writerMemoryArea.equals(readerMemoryArea)))
	    {
		allowableMemory = writerMemoryArea;
	    }
	else
	    {
		allowableMemory = ImmortalMemory.instance();
	    }

	if (memory == null)
	    {
		memArea = ImmortalMemory.instance();
	    }
	else
	    memArea = memory;

	/* Allocate the queue */

	theQueue = (java.lang.Object[])memArea.newArray(Class.forName("java.lang.Object"),maximum + 1);

	ReaderWaitingLock = memArea.newInstance(Class.forName("java.lang.Object"));
	MonitorControl.setMonitorControl(ReaderWaitingLock,new PriorityInheritance());
	WriterWaitingLock = memArea.newInstance(Class.forName("java.lang.Object"));
	MonitorControl.setMonitorControl(WriterWaitingLock,new PriorityInheritance());

	if (theQueue != null)
	    {
		queueSize = maximum + 1;
	    }
    }
    
    /**
      Set this to empty.
     */
    public void clear()
    {
	head = tail = 0;
    }
    
    /**
      Used to determine if this is empty.
      @return True if this is empty and false if this is not empty.
     */
    public boolean isEmpty()
    {
	return (head == tail);
    }
    
    /**
      Used to determine if this is full.
      @return True if this is full and false if this is not full.
     */
    public boolean isFull()
    {
	/* if next insertion will result in a clash with
	 * the head we consider the queue full
	 */

	return (( (tail+1)%queueSize) == head );
    }

    
    /**
      A synchronized read on the queue.
      @return The {@link Object} read or null if this is empty.
      */
    /* note that because there can be only one reader at any time
     * and only the reader can modify the head, then we cannot expect
     * the head value to move from the test "isEmpty()" to the actual
     * update of it. On the other hand we expect tail to be modified
     * between the test and the modification. Two cases are expected:
     *
     * 1. isEmpty() return true and then tail is modified to turn
     *    the condition into false. In this case the behavior is not 
     *    to retrieve any object from the queue. We consider that this 
     *    is a minor miss-behavior worth trading off for the gain in
     *    concurrency.
     *
     * 2. isEmpty() returned false and tail is modified to turn the 
     *    condition into false. THIS CANNOT HAPPEN given that turning
     *    the condition from false to true is an effect of this same
     *    function, read().
     */

    public synchronized java.lang.Object read()
    {
	java.lang.Object obj = null;

	while (isEmpty())
	    {
		/* Minimum sleep */
		try
		    {
			Thread.sleep(10);
		    }
		catch(Exception e)
		    {
		    }
	    }
	/*
	if (isEmpty())
	    {
		synchronized(ReaderWaitingLock)
		    {
			ReaderWaiting = true;
			try
			    {
				ReaderWaitingLock.wait();
			    }
			catch(Exception e)
			    {
			    }
		    }		
	    }
	*/
	obj = theQueue[head];
	head = (head + 1) % queueSize;
	return obj;
    }
    
    /**
      Used to determine the number of elements in this.
      @return An integer which is the number of empty positions in this.
     */
    public int size()
    {
	if (head <= tail)
	    return (queueSize - (tail - head) -1);
	return (head - tail + 1);
    }
    
    /**
      Force this {@link Object} to replace the last one.  
      If the reader should happen to have just removed the other {@link Object} just as we were
      updating it, we will return false.  False may mean that it just saw what we put in there.
      Either way, the best thing to do is to just
      write again -- which will succeed, and check on the readers side for consecutive identical
      read values.
     */
    public boolean force(java.lang.Object object) throws MemoryScopeException
    {
	MemoryArea area = MemoryArea.getMemoryArea(object);

	if (!(area instanceof ImmortalMemory) && 
	    (!area.equals(allowableMemory)))
	    {
		throw new MemoryScopeException();
	    }

	if (!isFull())
	    write(object);
	else
	    {
		int i = tail;
		if (i==0)
		    i = queueSize - 1;
		else
		    i--;
		theQueue[i] = object;
	    }

	if (tail == head)
	    return false;

	return true;
    }

    /**
      Try to insert an element into the queue.
      @param object The {@link Object} to insert.
      @return True if the insert succeeded, false if not.
      */
    public boolean write (java.lang.Object object)
           throws MemoryScopeException
    {
	MemoryArea area = MemoryArea.getMemoryArea(object);

	if (!(area instanceof ImmortalMemory) && 
	    (!area.equals(allowableMemory)))
	    {
		throw new MemoryScopeException("DAMN write");
	    }


	/* Please read the arguments for the synchronization
	 * between reader and writer for the read() method
	 */
     
	if (!isFull())
	    {
		theQueue[tail] = object;
		tail = (tail+1)%queueSize;
	
		/* For this synchronization only two threads are
		 * considered, one reader (by definition of the queue)
		 * and one writer (by the synchronized method).
		 * The reader can only set the ReaderWaiting variable to
		 * true inside the synchronized blocked just before the
		 * wait() call. This scheme rules out any other thread. GC
		 * is ruled out because this object is allocated into
		 * the same memory area as the queue (so they share the fate).
		 * The only other preemption we need to take into account is
		 * the priority preemption, i.e., higher priority threads.
		 * Therefore we use priority inheritance in this object.
		 */
	
		/*
		if (ReaderWaiting)
		    synchronized(ReaderWaitingLock)
			{
			    ReaderWaiting = false;
			    ReaderWaitingLock.notify();
			}
		*/
		return true;
	    }

	return false;
    }
}
