/*
 * -----------------------------------------------------------------
 * interpreter.h : The header definitions for per-thread data 
 * structures.
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying TimeSys Community Source License Agreement from TimeSys
 * in file TCSL.
 *
 * This file is derived from software distributed under the following terms:
 *
 * Sun Community Source License
 * Sun Microsystems
 * For details, see http://www.sun.com/981208/scsl/principles.html
 *
 * Your use of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


/*
 * @(#)interpreter.h	1.153 00/11/02
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#ifndef _INCLUDED_INTERPRETER_H
#define _INCLUDED_INTERPRETER_H

/* ROBBINS SEZ: added to make rempote exceptions work */
#define CVM_REMOTE_EXCEPTIONS_SUPPORTED

#include "javavm/include/defs.h"
#include "javavm/include/stacks.h"
#include "javavm/include/sync.h"
#include "javavm/include/cstates.h"
#include "javavm/include/jni_impl.h"
#include "javavm/include/gc_common.h"
#include "javavm/include/porting/threads.h"
#include "javavm/include/porting/vm-defs.h"
#include "javavm/include/mem_registry.h"

#define CVMJAVAPKG "java/lang/"

typedef enum {
    CVM_THREAD_RUNNING = 0x0,
    CVM_THREAD_SUSPENDED = 0x100
} CVMThreadState;

CVM_STRUCT_TYPEDEF(CVMPrivilegeItem);

struct CVMPrivilegeItem {
    CVMFrame *frame;		/* doPrivileged0 frame */
    CVMClassBlock *clazz;	/* class this item refers to */
    CVMObjectICell *context;	/* for doPrivileged(context) */
    CVMPrivilegeItem *next;	/* linked list of privileged item's */
};

/*
 * Per-thread C stack global pool operations and macros
 */
#define CVM_CSTACK_BUF_SIZE     8192    /* Fixed size of the EE buffer used by
                                         * ZipFile and RandomAccess IO read 
				 	 * and write operation. */

#define CVMCstackBuffer(ee)              ((ee)->cstackBuffer)
#define CVMCstackBufferFlag(ee)          ((ee)->cstackBufferFlag)
#define CVMCstackBufferIsNull(ee)        ((ee)->cstackBuffer == NULL)
#define CVMCstackBufferIsSet(ee)         ((ee)->cstackBufferFlag)

/* 
 * The flag that tells the status of the EE-buffer is initially unset 
 * when a thread is created 
 */
#define CVMCstackBufferInit(ee) {					\
   (ee)->cstackBufferFlag = CVM_FALSE;					\
 }		

/* 
 * Return C stack global buffer in ee if exists.
 * Otherwise, allocate C stack global buffer in ee.
 * Assert that the flag is not set before using the EE-buffer.
 * The flag is set when the buffer allocated successfully. Otherwise,
 * it remains unset.
 */
#define CVMCstackGetBuffer(ee, buf) {					\
    CVMassert(!CVMCstackBufferIsSet(ee));				\
    if (CVMCstackBufferIsNull(ee)) {					\
        CVMCstackBuffer(ee) = (char *)malloc(CVM_CSTACK_BUF_SIZE); 	\
    }									\
    buf = CVMCstackBuffer(ee);						\
    if (buf != NULL) {							\
    	CVMCstackBufferFlag(ee) = CVM_TRUE;				\
    }									\
 }

/* 
 * Deallocate the EE-buffer when a thread is killed.
 * Assert that the flag is not set before freeing the EE-buffer.
 */
#define CVMCstackFreeBuffer(ee) {   					\
    CVMassert(!CVMCstackBufferIsSet(ee));				\
    if (!CVMCstackBufferIsNull(ee)) 					\
        free(CVMCstackBuffer(ee));  					\
 }

/* 
 * Release the EE-buffer at the end of the buffer operation.
 * Assert that the flag is set before it's released.
 */
#define CVMCstackReleaseBuffer(ee) {   					\
    CVMassert(CVMCstackBufferIsSet(ee));				\
    CVMCstackBufferFlag(ee) = CVM_FALSE;				\
 }



/****************************************************************************
 * CVMExecEnv
 *
 * CVMExecEnv is the per thread execution context.
 ****************************************************************************/

struct CVMExecEnv {
    /* per-thread data for consistent states */
    CVMTCState tcstate[CVM_NUM_CONSISTENT_STATES];

    CVMUint16          remoteExceptionsDisabledCount;
    CVMThrowableICell* localExceptionICell;  /* local exception object */
    CVMThrowableICell* remoteExceptionICell; /* remote exception object */
    CVMThrowableICell* currentExceptionICell;/* exception being processed */
    union {
	struct {
	    CVMUint8 remote;    /* async, like from Thread.stop() */
	    CVMUint8 local;     /* local to the interpreter */
	} oneflag;
	CVMUint16 bothflags;    /* for checking both flags at one time */
    } exceptionFlags;

    CVMThreadICell* threadICell;/* back-pointer to thread object */
    CVMObjectICell* miscICell;  /* Per-thread root for miscellaneous use */

    /*
     * CVMgcAllocNewInstance() needs a root to hold the newly allocated object
     * in if it has to call Finalizer.register() on that object.
     */
    CVMObjectICell* finalizerRegisterICell;

    CVMJNIEnv jniEnv;		/* per-thread JNI data */

    /* linked-list of EE's */
    CVMExecEnv **prevEEPtr;
    CVMExecEnv *nextEE;

    CVMStack  interpreterStack; /* stack for interpreter frames */
    CVMStack  localRootsStack;  /* stack for local root frames */

    CVMBool cstackBufferFlag;   /* flag to tell the EE buffer is in use */
    char   *cstackBuffer;	/* buffer for reading bytes of data from
				 * input stream into a global pool for
				 * Java_java_io_FileInputStream_readBytes
				 * Java_java_io_RandomAccessFile_readBytes */

    void * nativeRunInfo;	/* for "system" threads */

    CVMThreadID threadInfo;	/* platform-specific thread info */
    CVMThreadState threadState;

#ifdef CVM_CLASSLOADING
    CVMSeenClass* seenclasses; /* used to detect class circularities */
#endif

    CVMUint32 threadID;

#ifdef CVM_DEBUG


    CVMSysMutex *sysLocks;	/* list of all CVMSysMutex locks held */
    int microLock;		/* micro lock depth */
#endif

    /* NOTE: objLockCurrent could be replaced by an atomically-updated
       integer reference count */
    CVMObjMonitor *objLockCurrent; /* being referenced, set while unsafe */
    CVMObjMonitor *objLocksFreeUnlocked;
    CVMOwnedMonitor *objLocksOwned; /* list of all CVMObjMonitor locks owned */
    CVMOwnedMonitor *objLocksFreeOwned;

    CVMBool userThread;		/* true if not a daemon thread */

    CVMPrivilegeItem *privilege;
#ifdef CVM_TEST_GC
    CVMBarrierType barrier;
#endif

    /* Added by Dio */
    int canChangeScheduler;
    jlong startNextPeriodMicros;

  /********** ISHIN **********/
  jlong deadline;  /* Added by ISHIN */
  /********** ISHIN **********/

    /* Added by Pratik */
    CVMObjectICell* memoryArea;
    PerThreadMemoryInfo* perThreadStack;

    /* Added by Kunal for GC Stopping -- to identify NoHeapRealTimeThread*/
    CVMBool isNoHeapRealtimeThread;
    
    /* Added by Pratik to enable/disable allocation checking */
    CVMBool checkConstraints;
        
    /* Added by Kalpesh */
    int currentStackDepth;
    CVMBool isIOBlocked;
    short dirtyAie;

    /* ROBBINS SEZ:
     * This is the ATC and AIE data, when this thread is in a defered
     * section, we store the exception object here and wait until we leave
     * the defered section to handle it.
     */

    /*pending AIE object */
    CVMThrowableICell* pendingAIExceptionICell;
    /* true if exception is held pending */
    /* basically it's true when AIExceptionPendingICell is valid */
    CVMBool pendingAIExceptionFlag;

};



#define CVM_TCSTATE(ee, x)	(&(ee)->tcstate[(x)])

#define CVMjniEnv2ExecEnv(je)					\
	((CVMExecEnv *)((char *)(je) - 				\
	    (CVMUint32)CVMexecEnv2JniEnv((CVMExecEnv *)0)))

#define CVMexecEnv2JniEnv(ee)	CVMjniPrivEnv2PubEnv(&(ee)->jniEnv)

#define CVMthreadID2ExecEnv(tid)					\
	((CVMExecEnv *)((char *)((tid)) - 				\
	    (CVMUint32)CVMexecEnv2threadID((CVMExecEnv *)0)))

#define CVMexecEnv2threadID(ee)	(&(ee)->threadInfo)

#define CVMeeGetCurrentFrame(ee) \
     (ee->interpreterStack.currentFrame)

/*
 * Initialize an ExecEnv before first use, and destroy after use.
 * If CVMinitExecEnv fails, it returns CVM_FALSE. An exception is
 * NOT thrown.
 */
extern CVMBool
CVMinitExecEnv(CVMExecEnv* ee, CVMBool userThread);

extern void
CVMdestroyExecEnv(CVMExecEnv* ee);

extern CVMExecEnv * CVMgetEE();

extern void CVMaddThread(CVMExecEnv *ee, CVMBool userThread);
extern void CVMremoveThread(CVMExecEnv *ee, CVMBool userThread);

/*
 * Macros for iterating over all threads. For efficiency, the caller is
 * responsible for seizing CVMglobals.threadLock; this macro checks to
 * ensure that it is seized. **WARNING**: the action may NOT cause
 * allocation of Java objects (which might trigger a GC). GC will
 * attempt to seize the threads lock, which could result in a
 * deadlock. NOTE: maybe this should go in a separate threads.h, but it
 * doesn't seem to belong in src/share/javavm/include/porting/threads.h. 
 */

#define CVM_WALK_ALL_THREADS_START(ee, eeNameInAction) {   \
   CVMExecEnv* eeNameInAction = CVMglobals.threadList;	   \
   CVMassert(CVMsysMutexIAmOwner(ee, &CVMglobals.threadLock)); \
   while (eeNameInAction != NULL) {                        \

#define CVM_WALK_ALL_THREADS_END(ee, eeNameInAction) 	   \
     eeNameInAction = eeNameInAction->nextEE;              \
   }                                                       \
 }

#define CVM_WALK_ALL_THREADS(ee, eeNameInAction, action)   \
    CVM_WALK_ALL_THREADS_START((ee), (eeNameInAction))	   \
    action;						   \
    CVM_WALK_ALL_THREADS_END((ee), (eeNameInAction))

/* Special call, which prevents NoHeapRealtimeThread 
 *		--Kunal
 **/
#define CVM_WALK_ALL_THREADS_START_EXCEPT_NOHEAP(ee, eeNameInAction) {   \
   CVMExecEnv* eeNameInAction = CVMglobals.threadList;	   \
   CVMassert(CVMsysMutexIAmOwner(ee, &CVMglobals.threadLock)); \
   while (eeNameInAction != NULL) {                        \
     if(!CVMisNoHeapRealtimeThread(eeNameInAction)){

#define CVM_WALK_ALL_THREADS_END_EXCEPT_NOHEAP(ee, eeNameInAction) 	   \
	 }													   \
     eeNameInAction = eeNameInAction->nextEE;              \
   }                                                       \
 }

#define CVM_WALK_ALL_THREADS_EXCEPT_NOHEAP(ee, eeNameInAction, action)   \
    CVM_WALK_ALL_THREADS_START_EXCEPT_NOHEAP((ee), (eeNameInAction))	   \
    action;						   \
    CVM_WALK_ALL_THREADS_END_EXCEPT_NOHEAP((ee), (eeNameInAction))

/*
 * Per-thread root cell for miscellaneous use
 */
#define CVMmiscICell(ee)   ((ee)->miscICell)

/*
 * Per-thread memory area cell (created to keep consistency with 
 * other icell accessors
 * - Pratik.
 */
#define CVMmemoryAreaICell(ee) ((ee)->memoryArea)

/* Macro for checking if the thread is a NoHeapRealtimeThread
 * - Pratik.
 */
#define CVMisNoHeapRealtimeThread(ee) ((ee)->isNoHeapRealtimeThread)

/* Marks/unmarks the ee as a NoHeapRealtimeThread ee.
 * - Pratik.
 */
#define CVMsetNoHeapRealtimeThread(ee) ((ee)->isNoHeapRealtimeThread = CVM_TRUE)
#define CVMunsetNoHeapRealtimeThread(ee) ((ee)->isNoHeapRealtimeThread = CVM_FALSE)

/* Macros for manipulating the checkConstrains flag
 * -Pratik
 */
#define CVMcheckForConstraints(ee_) ((ee_)->checkConstraints)
#define CVMdisableCheckForConstraints(ee_) (CVMcheckForConstraints(ee_) = CVM_FALSE)
#define CVMenableCheckForConstraints(ee_) (CVMcheckForConstraints(ee_) = CVM_TRUE)
#define CVMsetCheckForConstraints(ee_, value_) \
    if(value_) \
        { CVMenableCheckForConstraints(ee_); } \
    else \
        { CVMdisableCheckForConstraints(ee_); } 
        
    
/* ROBBINS SEZ:
 * set the AI exception objs in the target exec env's 
 * and set the flags accordingly.
 * ACCESS TO THESE MACROS MUST BE PROPERLY SYNCHRONIZED
 * it is assumed that the programmer knows what he/she's doing.
 */

/* ROBBINS SEZ:
 * ATC info accessor macros
 */
#define CVMpendingAIException(ee) ((ee)->pendingAIExceptionFlag == CVM_TRUE)

#define CVMsetPendingAIExceptionFlag(ee) ((ee)->pendingAIExceptionFlag = CVM_TRUE)

#define CVMclearPendingAIExceptionFlag(ee) ((ee)->pendingAIExceptionFlag = CVM_FALSE)

#define CVMpendingAIExceptionICell(ee) ((ee)->pendingAIExceptionICell)

#define CVMgetPendingAIExceptionObj(ee) \
     (CVMID_icellDirect(ee, CVMpendingAIExceptionICell(ee)))

#define CVMsetPendingAIExceptionObj(ee, targetEE, obj_){                   \
     CVMID_icellSetDirect(ee, CVMpendingAIExceptionICell(targetEE), obj_); \
}

#define CVMclearPendingAIException(ee) {                \
    CVMID_icellSetNull(CVMpendingAIExceptionICell(ee));	\
    CVMclearPendingAIExceptionFlag(ee);		        \
}

/* This is how AI exceptions actually get posted this macro */
/* unconditionally sets targets exc env pending object to obj */ 
#define CVMgcUnsafeSetPendingAIException(ee, targetEE, obj) { \
    CVMsetPendingAIExceptionObj(ee, targetEE, obj);           \
    CVMsetPendingAIExceptionFlag(targetEE);		      \
}

/* ROBBINS SEZ: 
 * I didn't make up this 'dirtyAIE' name. I think it's just as
 * bad as you do. Believe me, if I knew where Kalpesh disappeared
 * to, I'd make him fix it at gunpoint. But, since I don't, and
 * I don't have the time to fix it myself right now, you'll just
 * have to do with my apologizing on his behalf
*/
 /* -Kalpesh
 *
 * Macros for accessing ATC flags in ExeEnv.
 * 
 **/	
#define CVMATCdirty(_ee) ((_ee)->dirtyAie)
#define CVMATCsetDirty(_ee,_depth){	\
	CVMATCdirty(_ee) = (_depth);	\
}

/* This macros should be in stacks.h but its related to ATC so i've put it here -Kalpesh */

#define CVMstackDepth(_frame) ((_frame)->ATCflags.val.depth)

#define CVMATCPermission(_frame) ((_frame)->ATCflags.val.permission)
#define CVMATCPermissionIs(_state) ((_frame)->ATCflags.val.permission==(_state))	
	
#define CVMATCSetPermission(_frame,_state){	\
						\
	CVMATCPermission(_frame) = (_state);	\
}						

#define CVMATCOldPermission(_frame) ((_frame)->ATCflags.val.old_permission)	

#define CVMATCSavePermission(_frame) {				\
	CVMATCOldPermission(_frame) = CVMATCPermission(_frame);	\
}								

#define CVMATCRestorePermission(_frame) {				\
	        CVMATCPermission(_frame) = CVMATCOldPermission(_frame);	\
}									
	
#define CVMATCSynchCounter(_frame) ((_frame)->synch_counter)
#define CVMATCSynchCounterIsZero(_frame) ((_frame)->synch_counter==0)
#define CVMATCSynchCounterIncreament(_frame) {		\
							\
	(_frame)->synch_counter += 1;       		\
}							

#define CVMATCSynchCounterDecreament(_frame) {          \
		                                        \
	(_frame)->synch_counter -= 1; 	       	        \
}                                                       

/*
 * Per-thread root cell used when calling Finalizer.register().
 */
#define CVMfinalizerRegisterICell(ee)   ((ee)->finalizerRegisterICell)

/*
 * Be sure to use these macros to access the exception fields.  Do
 * not access the fields directly.
 */

#define CVMlocalExceptionICell(ee)   ((ee)->localExceptionICell)
#define CVMremoteExceptionICell(ee)  ((ee)->remoteExceptionICell)
#define CVMcurrentExceptionICell(ee) ((ee)->currentExceptionICell)

#define CVMgetLocalExceptionObj(ee)   \
     (CVMID_icellDirect(ee, CVMlocalExceptionICell(ee)))
#define CVMgetRemoteExceptionObj(ee)  \
     (CVMID_icellDirect(ee, CVMremoteExceptionICell(ee)))
#define CVMgetCurrentExceptionObj(ee) \
     (CVMID_icellDirect(ee, CVMcurrentExceptionICell(ee)))

#define CVMsetLocalExceptionObj(ee, obj_)   \
     CVMID_icellSetDirect(ee, CVMlocalExceptionICell(ee), obj_);
#define CVMsetRemoteExceptionObj(ee, targetEE, obj_)  \
     CVMID_icellSetDirect(ee, CVMremoteExceptionICell(targetEE), obj_);
#define CVMsetCurrentExceptionObj(ee, obj_) \
     CVMID_icellSetDirect(ee, CVMcurrentExceptionICell(ee), obj_);

#define CVMlocalExceptionFlag(ee)  ((ee)->exceptionFlags.oneflag.local)
#define CVMremoteExceptionFlag(ee) ((ee)->exceptionFlags.oneflag.remote)

#define CVMgcSafeThrowLocalExceptionNoDumpStack(ee, objICell)	\
    CVMD_gcUnsafeExec(ee, {					\
	CVMgcUnsafeThrowLocalExceptionNoDumpStack(		\
	    ee, CVMID_icellDirect(ee, objICell)); 		\
    });

#define CVMgcUnsafeThrowLocalExceptionNoDumpStack(ee, obj) {	\
    CVMassert(!CVMlocalExceptionOccurred(ee));			\
    CVMsetLocalExceptionObj(ee, obj);				\
    CVMlocalExceptionFlag(ee) = 1;				\
}

#define CVMgcSafeThrowLocalException(ee, objICell)			\
    CVMD_gcUnsafeExec(ee, {						\
	CVMgcUnsafeThrowLocalException(ee, CVMID_icellDirect(ee, objICell)); \
    });

#ifdef CVM_DEBUG
#define CVMgcUnsafeThrowLocalException(ee, obj)	{			\
    CVMtraceExceptions(("[<%d> Exception thrown: %O]\n",		\
			ee->threadID, obj));				\
    if (CVMcheckDebugFlags(CVM_DEBUGFLAG(TRACE_EXCEPTIONS))) {		\
	CVMdumpStack(&ee->interpreterStack, CVM_FALSE, CVM_FALSE, 100);	\
    }									\
    CVMassert(!CVMlocalExceptionOccurred(ee));				\
    CVMsetLocalExceptionObj(ee, obj);					\
    CVMlocalExceptionFlag(ee) = 1;					\
}
#else
#define CVMgcUnsafeThrowLocalException(ee, obj)	{		\
    CVMassert(!CVMlocalExceptionOccurred(ee));			\
    CVMsetLocalExceptionObj(ee, obj);				\
    CVMlocalExceptionFlag(ee) = 1;				\
}
#endif

#ifdef CVM_REMOTE_EXCEPTIONS_SUPPORTED
#define CVMgcSafeThrowRemoteException(ee, targetEE, objICell)	\
    CVMD_gcUnsafeExec(ee, {					\
	CVMgcUnsafeThrowRemoteException(ee, targetEE,		\
	    CVMID_icellDirect(ee, objICell));			\
    });
#else
#define CVMgcSafeThrowRemoteException(ee, targetEE, objICell) \
    CVMassert(CVM_FALSE);
#endif

#ifdef CVM_REMOTE_EXCEPTIONS_SUPPORTED
#ifdef CVM_DEBUG
#define CVMgcUnsafeThrowRemoteException(ee, targetEE, obj) {		      \
    CVMtraceExceptions(("[<%d> Remote Exception thrown from <%d>: %O]\n",     \
			targetEE->threadID, ee->threadID, obj));	      \
    if (CVMcheckDebugFlags(CVM_DEBUGFLAG(TRACE_EXCEPTIONS))) {		      \
        CVMtraceExceptions(("Thread <%d> stack dump follows:\n",	      \
			    targetEE->threadID));			      \
	CVMdumpStack(&targetEE->interpreterStack, CVM_FALSE, CVM_FALSE, 100); \
        CVMtraceExceptions(("Thread <%d> stack dump follows:\n",	      \
			    ee->threadID));				      \
	CVMdumpStack(&ee->interpreterStack, CVM_FALSE, CVM_FALSE, 100);	      \
    }									      \
    CVMsetRemoteExceptionObj(ee, targetEE, obj);			      \
    CVMremoteExceptionFlag(targetEE) = 1;				      \
}
#else
#define CVMgcUnsafeThrowRemoteException(ee, targetEE, obj) {	\
    CVMsetRemoteExceptionObj(ee, targetEE, obj);		\
    CVMremoteExceptionFlag(targetEE) = 1;			\
}
#endif
#else
#define CVMgcUnsafeThrowRemoteException(ee, targetEE, obj) \
    CVMassert(CVM_FALSE);
#endif

#define CVMclearLocalException(ee) {			\
    CVMID_icellSetNull(CVMlocalExceptionICell(ee));	\
    CVMlocalExceptionFlag(ee) = 0; 			\
}
#ifdef CVM_REMOTE_EXCEPTIONS_SUPPORTED
#define CVMclearRemoteException(ee) {			\
    CVMID_icellSetNull(CVMremoteExceptionICell(ee));	\
    CVMremoteExceptionFlag(ee) = 0;			\
}
#else
#define CVMclearRemoteException(ee)
#endif

#define CVMexceptionOccurred(ee)					 \
    ((ee)->exceptionFlags.bothflags != 0 &&				 \
     (CVMlocalExceptionOccurred(ee) || CVMremoteExceptionOccurred(ee)))

#define CVMlocalExceptionOccurred(ee)		\
     (CVMlocalExceptionFlag(ee) != 0)

#ifdef CVM_REMOTE_EXCEPTIONS_SUPPORTED
#define CVMremoteExceptionOccurred(ee)		\
    (CVMremoteExceptionFlag(ee) != 0 && !CVMremoteExceptionsDisabled(ee))
#else
#define CVMremoteExceptionOccurred(ee) (CVM_FALSE)
#endif

#define CVMcurrentThreadICell(ee) ((ee)->threadICell)


/* ROBBINS SEZ:
 * while not explicitly ATC related, we often need to lock and unlock the 
 * mutex associated with the current executing thread. The code that does this
 * in ungainly, so we have a macro to do this for us. Now, look carefully, these
 * macro immitates second order functions (other CVM macros do this also), the
 * failure actions are intended to be code blocks that execute when lock actions
 * fail. NOTE: They are both GC unsafe
 */

#define CVMgcUnsafeLockThreadMutex(ee, quick_lock_failure, lock_failure)        \
{                                                                               \
  if(!CVMobjectTryLock(ee,CVMID_icellDirect(ee,CVMcurrentThreadICell(ee)))){    \
    quick_lock_failure;                                                         \
    if(!CVMobjectLock(ee, CVMcurrentThreadICell(ee))){                          \
      CVMdebugPrintf(("blocking mutex lock failed\n"));                         \
      lock_failure;                                                             \
    }                                                                           \
  }                                                                             \
}

#define CVMgcUnsafeUnlockThreadMutex(ee, quick_unlock_failure, unlock_failure)  \
{                                                                               \
  if(!CVMobjectTryUnlock(ee,CVMID_icellDirect(ee,CVMcurrentThreadICell(ee)))){  \
    quick_unlock_failure;                                                       \
    if(!CVMobjectUnlock(ee,CVMcurrentThreadICell(ee))){                         \
      CVMdebugPrintf(("blocking mutex unlock failed\n"));                       \
      unlock_failure;                                                           \
    }                                                                           \
  }                                                                             \
}

/**********************************************************************
 * Interpreter Frames:
 **********************************************************************/

/*
 * The CVMInterpreterFrame is used to keep track of java method calls. Note
 * that local variables (including arguments) are stored right in front of the
 * frame, so locals points to (frame - CVMjmdMaxLocals(frame->mb). However,
 * a possible optimization might allow the locals to be in the previous
 * CVMStackChunk if they will fit there but the entire frame will not,
 * but this would also make gc scans more difficult.
 */
struct CVMInterpreterFrame {
    CVMFrame         frameX;
    CVMUint8*        pcX;     /* pointer to the current instruction */

    /* The following fields are not used by transition frames, but making
     * them common to both CVMJavaFrame and CVMTransitionFrame allows us
     * to avoid checking what type of frame we are returning to when
     * an opc_return is executed. In other words, we waste 8 bytes in
     * transition frames in order to make all returns faster.
     *
     * If you really want to save memory, constantPool and
     * locals can be eliminated at the cost of more instructions
     * to recalculate when returning from a method invocation.
     * constantPool can be accessed via the CVMMethodBlock. locals can
     * be calculated using the maxLocals of the CVMMethodBlock of the
     * previous frame.
     */
    CVMConstantPool* cpX;     /* our constant pool */
    CVMSlotVal32*    localsX; /* pointer to the local variables. */
};

/* CVMJavaFrame - used for real java methods */
struct CVMJavaFrame {
    CVMInterpreterFrame	frameX;

    CVMObjectICell    receiverObjX; /* the object we are dispatching on or
				     * the Class object for static calls */
    CVMStackVal32     opstackX[1];  /* the operand stack */
};

/* CVMTransitionFrame - used for "stub" java methods */
struct CVMTransitionFrame {
    CVMInterpreterFrame	frameX;
    CVMBool       incrementPcFlagX; /* flag that indicates if the pc should
				     * be incremented by opc_exittransition */
    CVMStackVal32 opstackX[1];      /* the operand stack */
};

#define CVM_JAVAFRAME_SIZE 	  (CVMoffsetof(CVMJavaFrame, opstackX))
#define CVM_TRANSITIONFRAME_SIZE  (CVMoffsetof(CVMTransitionFrame, opstackX))

#define CVMframePc(frame)      (CVMgetInterpreterFrame(frame)->pcX)
#define CVMframeCp(frame)      (CVMgetInterpreterFrame(frame)->cpX)
#define CVMframeLocals(frame)  (CVMgetInterpreterFrame(frame)->localsX)
#define CVMframeReceiverObj(frame)	\
                               (CVMgetJavaFrame(frame)->receiverObjX)
#define CVMframeIncrementPcFlag(frame) \
                               (CVMgetTransitionFrame(frame)->incrementPcFlagX)
#define CVMframeOpstack(frame, frameType) \
                               (CVMget##frameType##Frame(frame)->opstackX)


#define CVMgetInterpreterFrame(frame)					\
    (CVMassert(CVMframeIsJava(frame) || CVMframeIsTransition(frame)),	\
     (CVMInterpreterFrame*)(frame))
#define CVMgetJavaFrame(frame) \
    (CVMassert(CVMframeIsJava(frame)), (CVMJavaFrame*)(frame))
#define CVMgetTransitionFrame(frame) \
    (CVMassert(CVMframeIsTransition(frame)), (CVMTransitionFrame*)(frame))
#define CVMgetFreelistFrame(frame) \
    (CVMassert(CVMframeIsFreelist(frame)), (CVMFreelistFrame*)(frame))

#define CVMframeIsInterpreter(frame)					\
    (CVMframeIsJava(frame) || CVMframeIsTransition(frame))
#define CVMframeIsJava(frame) \
    (((CVMFrame*)(frame))->scanner == CVMjavaFrameScanner)
#define CVMframeIsTransition(frame)               \
    (((CVMFrame*)(frame))->scanner == CVMtransitionFrameScanner)
#define CVMframeIsFreelist(frame)               \
    (((CVMFrame*)(frame))->scanner == CVMfreelistFrameScanner)

/*
 * Reset the topOfStack to point to the beginning of the operand stack.
 */
#define CVM_RESET_INTERPRETER_TOS(topOfStack, frame)	\
{							\
    if (CVMframeIsJava(frame)) {			\
	CVM_RESET_JAVA_TOS(topOfStack, frame);		\
    } else {						\
	CVM_RESET_TRANSITION_TOS(topOfStack, frame);	\
    }							\
}
#define CVM_RESET_JAVA_TOS(topOfStack, frame) \
    (topOfStack) = CVMframeOpstack(frame, Java);
#define CVM_RESET_TRANSITION_TOS(topOfStack, frame) \
    (topOfStack) = CVMframeOpstack(frame, Transition);

/**********
 * APIs
 *********/

/*
 * CVMgcUnsafeExecuteJavaMethod - the main entry point for interpreting
 * java bytecodes.
 */
extern void
CVMgcUnsafeExecuteJavaMethod(CVMExecEnv* ee, CVMMethodBlock* mb,
			     CVMBool isStatic, CVMBool isVirtual);


/*
 * Functions used to display information about a given pc for debug builds.
 */
#ifdef CVM_DEBUG_CLASSINFO
extern CVMInt32
CVMpc2lineno(CVMMethodBlock *mb, CVMUint16 pc_offset);
#endif
#ifdef CVM_DEBUG_STACKTRACES
extern CVMMethodBlock*
CVMpc2method(CVMClassBlock* cb, CVMUint8* pc);
#endif
#if defined(CVM_DEBUG) || defined(CVM_DEBUG_DUMPSTACK) || defined(CVM_DEBUG_STACKTRACES)
extern void
CVMpc2string(CVMFrame* frame, CVMMethodBlock* mb, CVMUint8* pc, 
	     char *buf, char* limit);
#endif

/*
 * Return the pc of the exception handler for the specified exception.
 */
extern CVMUint8*
CVMgcSafeFindPCForException(CVMExecEnv* ee, CVMFrame* frame, 
			    CVMClassBlock* exceptionClass, CVMUint8* pc);

/*
 * CVMgcUnsafeHandleException - handle an outstanding exception.
 */
extern CVMFrame*
CVMgcUnsafeHandleException(CVMExecEnv* ee, CVMFrame* frame, 
			   CVMFrame* initialframe);

/*
 * CVMsignalError - signal an exception. The exception message can
 * be a format string followed by arguments just like printf.
 */
extern void
CVMsignalError(CVMExecEnv* ee, CVMClassBlock* exceptionCb,
	       const char *format, ...);

extern void
CVMsignalErrorVaList(CVMExecEnv* ee, CVMClassBlock* exceptionCb, 
		     const char* format, va_list ap);

/*
 * CVMpushTransitionFrame - push a transition frame.
 */
extern CVMTransitionFrame*
CVMpushTransitionFrame(CVMExecEnv* ee, CVMMethodBlock* mb);

/*
 * Ensure space for and perform stackmap allocations for a given frame
 *
 * CVM_TRUE  if stackmaps successfully created
 * CVM_FALSE on error
 */
extern CVMBool CVMjavaFrameEnsureStackmaps(CVMExecEnv* ee, CVMFrame* frame);

/*
 * Scan java frame roots for GC, calling 'callback' on each
 */
extern CVMFrameGCScannerFunc CVMjavaFrameScanner;

/*
 * Scan transition frame for GC
 */
extern CVMFrameGCScannerFunc CVMtransitionFrameScanner;

/*
 * Scan free-list frame for GC
 */
extern CVMFrameGCScannerFunc CVMfreelistFrameScanner;

/*
 * Get the classblock of the caller of the current method in the
 * execution environment. If SKIP is true then the current method is a
 * constructor and should be ignored; that is, we should go back to
 * the previous frame. Automatically skips transition,
 * Method.invoke(), Constructor.newInstance(), and Class.newInstance()
 * frames; see CVMgetCallerFrame, below. This is equivalent to
 * getCallerClass (jvm.c) in JDK 1.2 but was moved because it was
 * inappropriately placed there.
 */
CVMClassBlock* CVMgetCallerClass(CVMExecEnv* ee, int skip);

/*
 * Step back n frames, skipping Method.invoke(),
 * Constructor.newInstance(), Class.newInstance(), and transition
 * frames in between. This function is used in Class.forName,
 * Class.newInstance, Method.Invoke, AccessController.doPrivileged.
 * This is equivalent to getCallerFrame (jvm.c) in JDK 1.2 but was
 * moved because it was inappropriately placed there. NOTE that the
 * JDK 1.2 code and the Exact VM code have a third argument which is
 * needed in the case of compiled code, but this has been temporarily
 * removed.
 */
#define CVMgetCallerFrame(f, n) \
	CVMgetCallerFrameSpecial((f), (n), CVM_TRUE)

/*
 * Like CVMgetCallerFrame() above, but only skips special frames
 * if the "skipSpecial" flag is set.
 */
CVMFrame*
CVMgetCallerFrameSpecial(CVMFrame* frame, int n, CVMBool skipSpecial);

/*
 * Check if type srcCb is assignable to type dstCb.  */
extern CVMBool
CVMisAssignable(CVMExecEnv* ee, CVMClassBlock* srcCb, 
		CVMClassBlock* dstCb);

/*
 * CVMgcUnsafeIsInstanceOf - Check if obj is an instance of cb. 
 */
extern CVMBool
CVMgcUnsafeIsInstanceOf(CVMExecEnv* ee, CVMObject* obj, 
			CVMClassBlock* cb);

/*
 * CVMisSubclassOf - Check if subclasscb is an subclass of cb 
 */
extern CVMBool
CVMisSubclassOf(CVMExecEnv* ee, CVMClassBlock* subclasscb, 
		CVMClassBlock* cb);

/*
 * CVMimplementsInterface - Check if 'cb' implements 'interfacecb'
 */
extern CVMBool
CVMimplementsInterface(CVMExecEnv* ee, CVMClassBlock* cb, 
		       CVMClassBlock* interfacecb);

/*
 * Brought over from JDK 1.2: originally named "VerifyClassAccess".
 *
 * Verify that currentClass can access newClass. 
 *
 * resolverAccess is set true when called from constant pool resolution
 * code. It fixes some problems with buggy 1.1.x compilers by being
 * a bit more loose with verification requirements.
 *
 * Does not throw an exception under any circumstances.
 */
CVMBool
CVMverifyClassAccess(CVMExecEnv* ee,
		     CVMClassBlock* currentClass, CVMClassBlock* newClass, 
		     CVMBool resolverAccess);

/*
 * Brought over from JDK 1.2: originally named "VerifyFieldAccess2".
 *
 * Verify that currentClass can access a field or method of newClass,
 * where that field's or method's access bits are "access".  We assume
 * that we've already verified that class can access memberClass.
 *
 * resolverAccess is set true when called from constant pool resolution
 * code. It fixes some problems with buggy 1.1.x compilers by being
 * a bit more loose with verification requirements.
 *
 * The protectedRestriction flag appears to implement the restriction
 * in the JLS, section 6.6.2, in which a protected constructor may not
 * be called by invocation of Class.newInstance(). In the JDK 1.2
 * sources the only place where VerifyFieldAccess2 was called with
 * TRUE as the final argument was JVM_NewInstance.
 *
 * Does not throw an exception under any circumstances.
 */
extern CVMBool
CVMverifyMemberAccess2(CVMExecEnv* ee,
		       CVMClassBlock* currentClass,
		       CVMClassBlock* memberClass, 
		       CVMUint32 access, CVMBool resolverAccess,
		       CVMBool protectedRestriction);

/*
 * Brought over from JDK 1.2: originally named "VerifyFieldAccess".
 *
 * This is equivalent to a call to CVMverifyMemberAccess2(currentClass,
 * memberClass, access, classloaderOnly, CVM_FALSE);
 */
CVMBool
CVMverifyMemberAccess(CVMExecEnv* ee,
		      CVMClassBlock* currentClass, 
		      CVMClassBlock* memberClass, 
		      int access, CVMBool resolverAccess);

/*
 * Brought over from JDK 1.2: originally named "IsTrustedClassLoader".
 */
#ifdef CVM_CLASSLOADING
extern CVMBool
CVMisTrustedClassLoader(CVMExecEnv* ee, CVMClassLoaderICell* loader);
#endif

/*
 * Multi-dimensional array creation
 */
extern void
CVMmultiArrayAlloc(CVMExecEnv*     ee,
		   CVMInt32        nDimensions,
		   CVMInt32*       dimensions,
		   CVMClassBlock*  arrayCb,
		   CVMObjectICell* resultCell);

/*
 * Functions for enabling and disabling remote exceptions. If a remote
 * exception is thrown while remote exceptions are disabled, then the
 * exception will be ignored until remote exceptions are enabled. In other
 * words, CVMremoteExceptionOccurred() will return false while remote
 * exceptions are disabled. This prevents remote exceptions from blowing
 * us out of things like static initializers and class loading.
 */
extern void
CVMdisableRemoteExceptions(CVMExecEnv* ee);
extern void
CVMenableRemoteExceptions(CVMExecEnv* ee);
extern CVMBool
CVMremoteExceptionsDisabled(CVMExecEnv* ee);

#ifndef CVM_REMOTE_EXCEPTIONS_SUPPORTED
#define CVMdisableRemoteExceptions(ee)
#define CVMenableRemoteExceptions(ee)
#define CVMremoteExceptionsDisabled(ee) (CVM_FALSE)
#endif
/*
 * Fill in the specified Throwable object with the current backtrace,
 */
extern void 
CVMfillInStackTrace(CVMExecEnv *ee, CVMThrowableICell* objICell);
extern void 
CVMprintStackTrace(CVMExecEnv *ee, CVMThrowableICell* throwableICell,
		   CVMObjectICell* printableICell);

/*
 * Quickening opcodes. 
 * On entry quickening[0] contains the opcode to be quickened. 
 * On exit  quickening[0] contains the new opcode.
 *          quickening[1,2] contain the new operands if they have changed.
 *
 * Possible return values are:
 *
 *    CVM_QUICKEN_SUCCESS_OPCODE_ONLY:
 *        quickening succeeded, only opcode changed.
 *    CVM_QUICKEN_SUCCESS_OPCODE_AND_OPERANDS
 *        quickening succeeded, opcode and operands changed.
 *    CVM_QUICKEN_NEED_TO_RUN_STATIC_INITIALIZERS
 *        need to run static initializers, p_cb updated
 *    CVM_QUICKEN_ALREADY_QUICKENED
 *        someone else already raced us to it
 *            OR 
 *        in the JVMDI case, the instruction stream is already
 *        updated in CVMquickenOpcode().
 *    CVM_QUICKEN_ERROR
 *        an error occurred, exception pending
 *
 * Upon success, either quickening[] holds the instruction
 * sequence to replace the current one (max. length 3),
 * or we need to run static initializers, in which case
 * "*p_cb" will contain the CVMClassBlock* to initialize.
 *
 * If an error occurred, the interpreter handles the exception.
 */
typedef enum {
    CVM_QUICKEN_SUCCESS_OPCODE_ONLY,
    CVM_QUICKEN_SUCCESS_OPCODE_AND_OPERANDS,
    CVM_QUICKEN_NEED_TO_RUN_STATIC_INITIALIZERS,
    CVM_QUICKEN_ALREADY_QUICKENED,
    CVM_QUICKEN_ERROR
} CVMQuickenReturnCode;

/* 
 * Macro to make sure that no register caching is done when we
 * write a new byte code.
 */
#undef  CVM_WRITE_CODE_BYTE
#define CVM_WRITE_CODE_BYTE(ptr, offset, newvalue) \
    ((volatile CVMUint8*) (ptr))[(offset)] = (newvalue);

/*
 * Lock around writing _quick instruction when new instruction operands
 * have to be written. If just the instruction opcode is changed, then
 * no locking is needed. Contention for this lock is very low, so a global
 * lock is appropriate.
 */
#ifdef CVM_CLASSLOADING
#ifdef CVM_JVMDI
/* Use the GC-safe sysMutex abstraction */
#define CVM_CODE_LOCK(ee)     CVMsysMutexLock(ee, &CVMglobals.codeLock)
#define CVM_CODE_UNLOCK(ee)   CVMsysMutexUnlock(ee, &CVMglobals.codeLock)
#else
/* Use the global micro-lock. It is forbidden to block under this */
#define CVM_CODE_LOCK(ee)     CVMsysMicroLock(ee, CVM_CODE_MICROLOCK)
#define CVM_CODE_UNLOCK(ee)   CVMsysMicroUnlock(ee, CVM_CODE_MICROLOCK)
#endif
#else
/* No code lock required in the non-classloading build. jvmdi is the only
 * code that will try to use CVM_CODE_LOCK in a non-classloading build.
 * Since it always synchronizes on the DEBUGGER_LOCK, there is no need
 * for it to synchronize on the CVM_CODE_LOCK too.
 */
#define CVM_CODE_LOCK(ee)
#define CVM_CODE_UNLOCK(ee)
#endif

extern CVMQuickenReturnCode
CVMquickenOpcode(CVMExecEnv* ee, CVMUint8* quickening, CVMUint8* pc, 
		 CVMConstantPool* cp, CVMClassBlock** p_cb,
		 CVMBool clobbersCpIndex);

/*
 * Byte-code statistics gathering
 */
extern void
CVMinitStats();

extern void
CVMdumpStats();

/*
 * Routines which acquire and release locks needed for GC and thread
 * suspension. The latter was called "lock_for_debugger" in JDK 1.2,
 * even though it was also called from the native code for
 * Thread.suspend().
 */

extern void
CVMlocksForGCAcquire(CVMExecEnv* ee);

extern void
CVMlocksForGCRelease(CVMExecEnv* ee);

/*
 * NOTE: The forcing of all other threads to become GC-safe has been
 * separated out from these routines because of the race conditions it
 * caused.
 */

extern void
CVMlocksForThreadSuspendAcquire(CVMExecEnv* ee);

extern void
CVMlocksForThreadSuspendRelease(CVMExecEnv* ee);

/*
 * These have been separated out from the lock acquisition for thread
 * suspension because of race conditions caused by combining the two.
 * See jvm.c, JVM_SuspendThread, for proper use. They should be called
 * just before calling CVMthreadSuspend. Once a thread makes a request
 * and other threads have become consistent, the requesting thread may
 * not become inconsistent (i.e., GC-unsafe) until the request has
 * been cleared. For example, this implies that once a thread has
 * called CVMthreadSuspendGCSafeRequest, it can not read fields from
 * Java objects until it calls CVMthreadSuspendGCSafeRelease. It is
 * expected that CVMlocksForThreadSuspendAcquire will be called before
 * CVMthreadSuspendGCSafeRequest.
 */

extern void
CVMthreadSuspendConsistentRequest(CVMExecEnv* ee, CVMExecEnv* targetEE);

extern void
CVMthreadSuspendConsistentRelease(CVMExecEnv* ee, CVMExecEnv* targetEE);

/*
 * CVMmangleMethodName: create a mangled method name for a method.
 * It is the caller's responsiblity to free the mangled name returned.
 */
#ifdef CVM_CLASSLOADING

typedef enum {
    CVM_MangleMethodName_JNI_SHORT,
    CVM_MangleMethodName_JNI_LONG
} CVMMangleType;

extern char* 
CVMmangleMethodName(CVMExecEnv* ee, CVMMethodBlock* mb, 
		    CVMMangleType mangleType);

#endif /* CVM_CLASSLOADING */
		    

/*
 * CVMlookupNativeMethodCode - look up the native code for a dynamically
 * loaded JNI method.
 */

#ifdef CVM_CLASSLOADING
extern CVMBool
CVMlookupNativeMethodCode(CVMExecEnv* ee, CVMMethodBlock* mb);
#endif /* CVM_CLASSLOADING */

/*****************************************************************
 * Support for CNI native methods.  cjp07/13/99
 *
 * Using the CNI native method interface rather than JNI provides a few
 * advantages. In general they are much faster and allow you to
 * eliminate C recursion back into the interpreter loop (like you get in
 * JDK when you call Method.invoke() or when static initializers are run).
 * Specifically, they have the following advantages:
 *
 * 1. No stack frame has to be pushed or popped for the method call.
 * 2. No argument wrapping/marshalling has to be done. Arguments are
 *    accessed directly from the java stack.
 * 3. No glue code has to be used to invoke the native method. It is
 *    called directly from the interpreter loop.
 * 4. CNI methods execute gcUnsafe, so they save the overhead of having
 *    to become gcSafe first.
 * 5. CNI methods can be used to invoke a method without the
 *    need for any C recursion back into the interpreter.
 *
 * Of course, all these advantages come at a price. Mostly the price you
 * pay is that you need to be very careful about gcSafety, so closely
 * follow the rules mentioned below.
 *
 * When an CNI method is entered, frame->topOfStack points above the
 * arguments so they will be scanned. There is no need to touch
 * frame->topOfStack, since the interpreter will handle updating it
 * after the CNI method returns.
 *
 ******************************************************************/

/* Result codes for CNI native methods. */

typedef enum {
    /* The following 3 CNIResultCodes are used for simple CNI
     * methods that act just like a regular methods that process the 
     * arguments and return a result. The CNIResultCode indicates the
     * size of the result. The result should be stored at *arguments
     * (see CNINativeMethod prototype below). 
     *
     * WARNING: Do not become gcSafe after storing the result, because
     * the GC will end up scanning the result with the wrong stackmap.
     */
    CNI_VOID = 0,     /* no result stored on stack. */
    CNI_SINGLE = 1,   /* a single word result was stored on the stack. */
    CNI_DOUBLE = 2,   /* a double word result was stored on the stack. */

    /* CNI_NEW_TRANSITION_FRAME is used to indicate that the CNI method
     * pushed a transition frame on the stack. This is used by CNI methods
     * that want the interpreter to call another method. The method could
     * have been called directly by the CNI method, but this introduces
     * C recursion. This facility is used by Method.invoke(), which calls
     * an CNI method to push a transition frame. The transition frame will
     * invoke the method passed to Method.invoke().
     *
     * Arguments are pushed on the transition frame just like they are
     * for any other transition frame, such as when jni uses a transition
     * frame to invoke a method.
     *
     * WARNING: The mb in the transition frame for non static methods
     * is assummed to be fully resolved. In other words, an interface
     * or superclass mb must be preresolved to the mb in the object's
     * methodtable, unless you really want the superclass method to be
     * invoked.
     */
    CNI_NEW_TRANSITION_FRAME = -1, /* method pushed a transition frame. */

    /* CNI_NEW_JAVA_FRAME is used when an CNI method pushes a JaveFrame
     * and it wants the interpreter to continue execution at the pc indicated
     * in the JavaFrame. Currently this isn't used because dealing with
     * arguments can get ugly, so CNI_NEW_TRANSITION_FRAME usually gets
     * used instead. 
     */
#if 0 /* currently no one uses this feature */    
    CNI_NEW_JAVA_FRAME = -2,       /* method pushed a java frame. */
#endif

    /* CNI_NEW_MB is used when an CNI method wants the interpreter to start
     * execution in a specific mb. The mb to start executing in is stored
     * in the p_mb argument (see CNINativeMethod prototype below), which is
     * a pointer to the mb local variable in the interpreter loop. 
     *
     * CNI_NEW_MB is useful when you want to call to a method that you 
     * aren't allowed to call directly from Java. <clinit> methods are an 
     * example of this. Class.runStaticInitializers is written in Java and
     * wants to directly call <clinit> methods, but the compiler won't 
     * allow this. Instead it calls the CVM.executeClinit CNI method,
     * which simply returns the mb of the <clinit> method so the interpeter 
     * can call it.
     *
     * WARNING: There are limitations on the type of mb that can be returned
     * with CNI_NEW_MB. You can't return an mb that will cause the stackmap
     * used by gc to be wrong. This means the arguments must meet the
     * following criteria:
     * 1. The CNI method must have at least as many arguments as the mb
     *    that it returns.
     * 2. Arguments that are refs for the CNI method must be refs for the
     *    returned mb and vice-versa.
     * 3. The results for the CNI method and the returned mb must be 
     *    compatible.
     */
    CNI_NEW_MB = -3,               /* method returned a new mb to invoke. */
    
    /* CNI_EXCEPTION is used to indicate that an exception was thrown
     * by the CNI method. This would normally be done by calling
     * CVMsignalError().
     */
    CNI_EXCEPTION = -4             /* an exception was thrown. */
} CNIResultCode;

/* All CNI native methods have the CNINativeMethod prototype. "arguments"
 * points to the beginning of the arguments on the java stack and is also
 * the location where any result should be stored. p_mb is a pointer
 * to the CVMMethodBlock* local variable in the interpreter loop. You
 * can use it to cause the interpreter to start execution in a new
 * method (see the CNI_NEW_MB description above.)
 */

typedef CNIResultCode CNINativeMethod(CVMExecEnv* ee,
				      CVMStackVal32* arguments,
				      CVMMethodBlock** p_mb);

/*
 * Shutdown support
 */

extern void CVMwaitForUserThreads(CVMExecEnv *ee);
extern void CVMwaitForAllThreads(CVMExecEnv *ee);
extern int CVMprepareToExit(void);
extern int CVMatExit(void (*func)(void));
extern void CVMexit(int);
extern void CVMabort(void);
extern CVMBool CVMsafeExit(CVMExecEnv *ee, CVMInt32 status);

#if defined(CVM_CLASSLOADING)
extern void
CVMunloadApplicationclasses(CVMExecEnv* ee);
#endif

#endif /* _INCLUDED_INTERPRETER_H */
