/*
 * @(#)threads.h	1.20 00/10/26
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

/*
 * Porting layer for threads.
 */

#ifndef _INCLUDED_PORTING_THREADS_H
#define _INCLUDED_PORTING_THREADS_H

#include "javavm/include/porting/defs.h"
#include "javavm/include/porting/vm-defs.h"

/*
 * Platform must provide definitions for the following opaque struct(s):
 *
 * struct CVMThreadID		Uniquely identifies a thread.
 *
 * Macros:
 *
 * CVM_THREAD_MIN_C_STACK_SIZE	The minimum supported C stack size.
 * 
 * Macros used in C stack check codes for various recursive functions:
 *
 * CVM_REDZONE_ILOOP		Stack space required for interpreter loop
 *				in CVMgcUnsafeExecuteJavaMethod().
 *
 * CVM_REDZONE_CVMclassLookupFromClassLoader
 *				Stack space required for class loading in
 *				a deep class lookup hierarchy in 
 *				classlookup:CVMclassLookupFromClassLoader().
 *
 * CVM_REDZONE_CVMclassLink	Stack space required for class linking in
 *			        CVMclassLink().	
 *
 * CVM_REDZONE_CVMclassScan	Stack space required for traversing the scannable
 *				state of one class by gc in CVMclassScan(). 
 *
 * CVM_REDZONE_CVMimplementsInterface	
 *				Stack space required for checking if a 
 *				non-array class is an interface type class in
 *				CVMimplementsInterface(). 
 *
 * CVM_REDZONE_CVMCstackCVMpc2string
 *				Stack space required for building a formatted 
 *				string %P format of PC information for console I/O
 *				in CVMformatStringVaList().
 *
 * CVM_REDZONE_CVMCstackCVMID_objectGetClassAndHashSafe
 *				Stack space required for building a formatted 
 *				string %I format of class name information for 
 *				console I/O in CVMformatStringVaList().
 *
 * CVM_REDZONE_CVMCstackmerge_fullinfo_to_types
 *                              Stack space required for class verifier to do
 *				type merging between two object types or two
 *				array of object types in
 *				verifycode:merge_fullinfo_types().	
 *
 * CVM_REDZONE_CVMsignalErrorVaList
 *                              Stack space required for signaling an exception.  *
 * Recursive functions that have a C stack check are listed as below:
 *
 * CVMgcUnsafeExecuteJavaMethod
 * classlookup:CVMclassLookupFromClassLoader
 * CVMclassLink
 * CVMclassScan
 * CVMimplementsInterface
 * CVMformatStringVaList
 * CVMsignalErrorVaList
 * verifycode:merge_fullinfo_types
 *
 * C stack redzones used in the recursive functions are defined as below:
 *
 * #define CVM_REDZONE_ILOOP                                       11592 
 * #define CVM_REDZONE_CVMimplementsInterface                       4752
 * #define CVM_REDZONE_CVMCstackCVMpc2string                        4752
 * #define CVM_REDZONE_CVMCstackCVMID_objectGetClassAndHashSafe     5392
 * #define CVM_REDZONE_CVMclassLookupFromClassLoader                8744 
 * #define CVM_REDZONE_CVMclassLink                                 7456
 * #define CVM_REDZONE_CVMCstackmerge_fullinfo_to_types             5800
 * #define CVM_REDZONE_CVMsignalErrorVaList                         8400 
 *
 * A documentation that describes the worst stack size usage required
 * by each recursive function will be written at doc/internal_doc/cstack-notes.html.
 *
 */

/*
 * Create a new, runnable thread with the given Java priority
 * and C stacksize.  The new thread must call func(arg) upon
 * creation and clean up after itself and exit when func()
 * returns.
 */
extern CVMInt32 CVMthreadCreate(CVMSize stackSize, CVMInt16 priority,
    void (*func)(void *), void *arg);

/*
 * Yield to other threads, as defined by Thread.yield().
 */
extern void CVMthreadYield(void);

/*
 * Set the priority of a thread to the given value, mapping the
 * Java priority to an appropriate value.
 */
extern void CVMthreadSetPriority(CVMThreadID *thread, CVMInt32 prio);

/*
 * Suspend a thread, according to the semantics of Thread.suspend()
 * (which has been deprecated) or JVMDI SuspendThread().
 */
extern void CVMthreadSuspend(CVMThreadID *thread);

/*
 * Resume a suspended thread, according to the semantics of Thread.resume()
 * (which has been deprecated) or JVMDI ResumeThread().
 */
extern void CVMthreadResume(CVMThreadID *thread);

/*
 * Associated a CVMThreadID * with the current thread.
 */
extern void CVMthreadAttach(CVMThreadID *self);

/*
 * Remove the CVMThreadID * from the current thread.
 */
extern void CVMthreadDetach(CVMThreadID *self);

/*
 * Return the CVMThreadID * previously associated with
 * the current thread by CVMthreadAttach(), or NULL if
 * CVMthreadAttach() has not been called yet or
 * CVMthreadDetach() has been called.
 */
extern CVMThreadID * CVMthreadSelf(void);

/*
 * Break the given thread out of a CVMcondvarWait() call,
 * causing that call to return CVM_FALSE, signaling that
 * the wait was interrupted.  If the given thread is not
 * currently in CVMcondvarWait(), the requested is
 * remembered until CVMcondvarWait() is called, which
 * should return immediately with CVM_FALSE.
 */
extern void CVMthreadInterruptWait(CVMThreadID *thread);

/*
 * Return the current interrupted state for the thread.
 * If "clearInterrupted" is true, cancel any previous
 * interrupt requests made by CVMthreadInterruptWait().
 * (Currently, we only set "clearInterrupted" when "thread"
 * is the current thread.)
 */
extern CVMBool CVMthreadIsInterrupted(CVMThreadID *thread,
    CVMBool clearInterrupted);

/*
 * Check for a stack overflow.  Return CVM_TRUE if the stack
 * of the current thread has at least redZone bytes free,
 * CVM_FALSE otherwise.
 */
extern CVMBool CVMthreadStackCheck(CVMThreadID *self, CVMUint32 redZone);

#include CVM_HDR_THREADS_H

#if CVM_THREAD_MIN_C_STACK_SIZE < CVM_REDZONE_ILOOP
#error CVM_THREAD_MIN_C_STACK_SIZE < CVM_REDZONE_ILOOP
#endif

#endif /* _INCLUDED_PORTING_THREADS_H */
