/*
 * -----------------------------------------------------------------
 *  classlookup.c :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying TimeSys Community Source License Agreement from TimeSys
 * in file TCSL.
 *
 * This file is derived from software distributed under the following terms:
 *
 * Sun Community Source License
 * Sun Microsystems
 * For details, see http://www.sun.com/981208/scsl/principles.html
 *
 * Your use of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


/*
 * @(#)classlookup.c	1.78 00/09/22
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

/*
 * Contains functions used for looking up a class.
 */

#include "javavm/include/interpreter.h"
#include "javavm/include/classes.h"
#include "javavm/include/utils.h"
#include "javavm/include/common_exceptions.h"
#include "javavm/include/globals.h"
#include "javavm/include/preloader.h"
#include "javavm/include/indirectmem.h"
#include "javavm/include/directmem.h"
#include "javavm/include/stacks.h"
#include "javavm/include/stackwalk.h"
#include "javavm/include/globalroots.h"
#include "javavm/include/localroots.h"

/* *** RTSJ *** */
// Need this file for ExecuteInImmortal() macro. This is used to ensure
// THat class loading takes place in ImmortalMemory.
#include "javavm/include/mem_registry.h"
/* *** JSTR *** */

#ifdef CVM_CLASSLOADING
#include "generated/offsets/java_lang_ClassLoader.h"
#endif

#ifdef CVM_CLASSLOADING
#define CVM_NULL_CLASSLOADER_LOCK(ee)   \
  CVMsysMutexLock(ee, &CVMglobals.nullClassLoaderLock)
#define CVM_NULL_CLASSLOADER_UNLOCK(ee) \
  CVMsysMutexUnlock(ee, &CVMglobals.nullClassLoaderLock)
#else
#define CVM_NULL_CLASSLOADER_LOCK(ee)
#define CVM_NULL_CLASSLOADER_UNLOCK(ee)
#endif

/* static functions used by this compilation unit. */

static CVMClassBlock*
CVMclassLookupFromClassLoader(CVMExecEnv* ee, 
			      CVMClassTypeID typeID, const char* name,
			      CVMBool init,
			      CVMClassLoaderICell* loader,
			      CVMObjectICell* pd,
			      CVMBool throwError);


/*
 * Class lookup by name or typeID. 
 *   -The "init" flag indicates if static initializers should be run.
 *   -The "throwError" flag will cause NoClassDefFoundError to be thrown
 *    instead of ClassNotFoundException.
 *   -An exception os thrown if a NULL result is returned.
 *
 * All of these functions will cause an attempt to load the class if it
 * is not already loaded.
 */

CVMClassBlock*
CVMclassLookupByNameFromClass(CVMExecEnv* ee, const char* name,
			      CVMBool init, CVMClassBlock* fromClass)
{
    CVMClassLoaderICell* loader;
    CVMObjectICell*      pd;
    CVMClassBlock*       cb;
    CVMClassTypeID       typeID = 
	CVMtypeidNewClassID(ee, name, strlen(name));

    if (typeID == CVM_TYPEID_ERROR) {
	return NULL; /* exception already thrown */
    }

    if (fromClass != NULL) {
	loader = CVMcbClassLoader(fromClass);
	pd = CVMcbProtectionDomain(fromClass);
    } else {
	loader = NULL;
	pd = NULL;
    }
    
    // - Pratik
    ExecuteInImmortal(ee, {
    cb = CVMclassLookupFromClassLoader(ee, typeID, name, init,
				       loader, pd, CVM_TRUE);
    });
    
    CVMtypeidDisposeClassID(ee, typeID);
    return cb;
}
					       
CVMClassBlock*
CVMclassLookupByTypeFromClass(CVMExecEnv* ee, CVMClassTypeID typeID, 
			      CVMBool init, CVMClassBlock* fromClass)
{
    CVMClassLoaderICell* loader;
    CVMObjectICell*      pd;
    CVMClassBlock*       cb;  // -Pratik
    if (fromClass != NULL) {
	loader = CVMcbClassLoader(fromClass);
	pd = CVMcbProtectionDomain(fromClass);
    } else {
	loader = NULL;
	pd = NULL;
    }
    // - Pratik
    ExecuteInImmortal(ee, {
    cb = CVMclassLookupFromClassLoader(ee, typeID, NULL, init, 
					 loader, pd, CVM_TRUE);
    });
    return cb;
}

CVMClassBlock*
CVMclassLookupByNameFromClassLoader(CVMExecEnv* ee, const char* name, 
				    CVMBool init,
				    CVMClassLoaderICell* loader, 
				    CVMObjectICell* pd,
				    CVMBool throwError)
{
    CVMClassBlock* cb;
    CVMClassTypeID typeID = 
	CVMtypeidNewClassID(ee, name, strlen(name));

    if (typeID == CVM_TYPEID_ERROR) {
	return NULL; /* exception already thrown */
    }
    // - Pratik.
    ExecuteInImmortal(ee, {
    cb = CVMclassLookupFromClassLoader(ee, typeID, name, init,
				       loader, pd, throwError);
    });
    CVMtypeidDisposeClassID(ee, typeID);
    return cb;
}

CVMClassBlock*
CVMclassLookupByTypeFromClassLoader(CVMExecEnv* ee, CVMClassTypeID typeID, 
				    CVMBool init,
				    CVMClassLoaderICell* loader, 
				    CVMObjectICell* pd,
				    CVMBool throwError)
{
    CVMClassBlock* cb;
    ExecuteInImmortal(ee, {
    cb = CVMclassLookupFromClassLoader(ee, typeID, NULL, init,
					 loader, pd, throwError);
    });
    return cb;
}


/*
 * CVMclassLookupFromClassLoader - The main lookup function. The class
 * typeID is passed in, but the name may be NULL. If both are passed in
 * then it can save some time by avoiding memory allocations for the name.
 */

static CVMClassBlock*
CVMclassLookupFromClassLoader(CVMExecEnv* ee, 
			      CVMClassTypeID typeID, const char* name,
			      CVMBool init,
			      CVMClassLoaderICell* loader,
			      CVMObjectICell* pd,
			      CVMBool throwError)
{
    CVMClassBlock* cb = NULL;
    char* namebuf = NULL;

    if (!CVMCstackCheckSize(ee, CVM_REDZONE_CVMclassLookupFromClassLoader,
			    "CVMclassLookupFromClassLoader", CVM_TRUE)) {
	return NULL;
    }

    CVMassert(!CVMlocalExceptionOccurred(ee));

    if (typeID == CVM_TYPEID_ERROR) {
	/*
	 * This should only happen if the typeid system ran out of memory.
	 * Note that passing CVM_TYPEID_ERROR is not allowed unless you also
	 * pass in the class name.
	 */
	CVMassert(name != NULL);
	CVMthrowNoClassDefFoundError(ee, "%s", name);
	return NULL;
    }

    /* No CVMFieldTypeIDs allowed. Upper bits must never be set. */
    CVMassert(CVMtypeidGetType(typeID) == typeID);

#ifndef CVM_CLASSLOADING
    CVMassert(loader == NULL);
    loader = NULL; /* should cause compiler to optimize away some code */
#endif

    /* 
     * See if the class is a romized one. If it is, then just return
     * it. Otherwise we'll need to do further lookups.
     */
    if (loader == NULL || CVMtypeidIsPrimitive(typeID) || 
	CVMtypeidIsArray(typeID)) {
	cb = CVMpreloaderLookupFromType(typeID);
	if (cb != NULL) {
	    /* 
	     * If the class was preloaded then make sure it
	     * has been initialized if requested.
	     */
	    if (init) {
		if (!CVMclassInit(ee, cb)) {
		    cb = NULL;
		}
	    }
	    return cb;
	}
    }

    if (loader == NULL) {
	CVM_NULL_CLASSLOADER_LOCK(ee);
    } else {
         /* 
	  * Although class loaders are supposed to define loadClass as
	  * a synchronized method, the VM does not trust the class loader
	  * to do so. It enters the loader monitor just in case.
	  */
	if (!CVMgcSafeObjectLock(ee, loader)) {
	    CVMthrowOutOfMemoryError(ee, NULL);
	    return NULL;
	}
   }

    /* 
     * See if the class is already in the loader cache.
     */
    CVM_LOADERCACHE_LOCK(ee);
    cb = CVMloaderCacheLookup(ee, typeID, loader);
    CVM_LOADERCACHE_UNLOCK(ee);

    /*
     * If the class isn't already in the loader cache then load it if
     * necessary. If may already be loaded with a different initiating
     * class loader. If so, then we'll eventually recurse back into this
     * function with a parent class loader as the initiating class loader.
     */
    if (cb == NULL) {
	/*
	 * We need to disable remote exceptions, because we don't want
	 * to allow them to interrupt any java code we may end up
	 * executing or cause us to abort any C code while class loading.
	 */
	CVMdisableRemoteExceptions(ee);
	CVMassert(!CVMlocalExceptionOccurred(ee));
	if (CVMtypeidIsArray(typeID)) {
#ifdef CVM_CLASSLOADING
	    /*
	     * Before trying to create the array class, we first need to
	     * see if it was already loaded using one of the loader's parents
	     * as the initiating class loader. This mimics the delegation
	     * that is done when ClassLoader.loadClass() is called for
	     * a non-array class.
	     */
	    if (loader != NULL) {
		CVMID_localrootBegin(ee) {
		    CVMID_localrootDeclare(CVMClassLoaderICell, parentLoader);
		    CVMID_icellAssign(ee, parentLoader, loader);
		    do {
			CVMID_fieldReadRef(
                            ee, parentLoader, 
			    CVMoffsetOfjava_lang_ClassLoader_parent,
			    parentLoader);
			CVM_LOADERCACHE_LOCK(ee);
			if (CVMID_icellIsNull(parentLoader)) {
			    cb = CVMloaderCacheLookup(ee, typeID, NULL);
			} else {
			    cb = CVMloaderCacheLookup(ee, typeID,
						      parentLoader);
			}
			CVM_LOADERCACHE_UNLOCK(ee);
			if (cb != NULL) {
			    break;
			}
		    } while (cb == NULL && !CVMID_icellIsNull(parentLoader));
		} CVMID_localrootEnd();
	    }
#endif
	    if (cb == NULL) {
		cb = CVMclassCreateArrayClass(ee, typeID, loader, pd);
	    }
	} else {
	    /*
	     * The API supports passing in a valid typeID and a NULL class
	     * name. This way we only need to convert the typeID to a name
	     * string if necessary. At this point the conversion is necessary.
	     */
	    if (name == NULL) {
		namebuf = CVMtypeidClassNameToAllocatedCString(typeID);
		name = namebuf;
	    }
	    if (name == NULL) {
		CVMthrowOutOfMemoryError(ee, NULL);
	    } else {
#ifdef CVM_CLASSLOADING
		CVMassert(!CVMexceptionOccurred(ee));
		cb = CVMclassLoadClass(ee, loader, name, typeID);
#else
		CVMdebugPrintf(("CVMclassLoadClass: class loading"
				" not supported\n"));
		/*
		 * Class loading not supported. Throw a ClassNotFoundException.
		 * If throwError is set, this will be converted to
		 * a NoClassDefFoundError below.
		 */
		CVMthrowClassNotFoundException(ee, "%s", name);
#endif
	    }
	}
	CVMenableRemoteExceptions(ee);
	if (CVMexceptionOccurred(ee)) {
	    cb = NULL;
	}

	/* Update the loader cache */
	if (cb != NULL) {
	    if (!CVMloaderCacheAdd(ee, cb, loader)) {
		cb = NULL; /* exception already thrown */
	    }
	}
    }

#ifdef CVM_CLASSLOADING
    if (cb != NULL) {
	if (!CVMloaderCacheCheckPackageAccess(ee, loader, cb, pd)) {
	    cb = NULL;
	}
    }
#endif

    /* unlock the class loader we used. */
    if (loader == NULL) {
	CVM_NULL_CLASSLOADER_UNLOCK(ee);
    } else {
	CVMBool success = CVMgcSafeObjectUnlock(ee, loader);
	CVMassert(success); (void) success;
    }

    if (cb != NULL) {
	/* 
	 * If there were no errors then make sure the class
	 * has been initialized if requested.
	 */
    if (init) {
	    if (!CVMclassInit(ee, cb)) {
		cb = NULL;
	    }
	}
    } else {
	/*
	 * Handle errors: Convert ClassNotFoundException to 
	 * NoClassDefFoundError if requested.
	 */
	CVMassert(CVMlocalExceptionOccurred(ee));
	if (throwError) {
	    CVMD_gcUnsafeExec(ee, {
		CVMClassBlock* classNotFoundCb =
		    CVMsystemClass(java_lang_ClassNotFoundException);
		CVMObject* exceptionObj = 
		    CVMID_icellDirect(ee, CVMlocalExceptionICell(ee));
		if (CVMgcUnsafeIsInstanceOf(ee,
					    exceptionObj,
					    classNotFoundCb)) {
		    CVMclearLocalException(ee);
		    CVMthrowNoClassDefFoundError(ee, "%!C", typeID);
		}
	    });
	}
    }
    
    if (namebuf != NULL) {
	free(namebuf);
    }
    return cb;	
}
