/*
 * -----------------------------------------------------------------
 *  objsync.c :
 *
 * Copyright (C) 2001 TimeSys Corporation, All Rights Reserved.
 *
 * This software is subject to the terms and conditions of the
 * accompanying TimeSys Community Source License Agreement from TimeSys
 * in file TCSL.
 *
 * This file is derived from software distributed under the following terms:
 *
 * Sun Community Source License
 * Sun Microsystems
 * For details, see http://www.sun.com/981208/scsl/principles.html
 *
 * Your use of this software indicates your acceptance of these terms.
 * -----------------------------------------------------------------
 */


/*
 * @(#)objsync.c	1.43 00/11/27
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#include "javavm/include/defs.h"
#include "javavm/include/sync.h"
#include "javavm/include/indirectmem.h"
#include "javavm/include/localroots.h"
#include "javavm/include/common_exceptions.h"
#include "javavm/include/clib.h"

#include "javavm/export/jvmcei.h"

/* Dio: global lock accouting */
extern int concurrentLocks;
extern int maximumConcurrentLocks;
extern int reservedExceptionsLocks;

/*** !!!!!! FORCING monitor debug ****/
// #define CVM_DEBUG_MONITOR_1 true


/* NOTES: About CVM object monitors and their life-cycles
   ======================================================

   CVM object monitors are implemented using 2 data structures:
     1. The CVMOwnedMonitor record.
     2. The CVMObjMonitor monitor implementation.

   The Fast Locking mechanism:
   ==========================
   CVM implements a fast locking mechanism which attempts to speed up monitor
   locking/unlocking operations for the monitors which meet the following
   criteria:
     1. The object monitor is locked and unlocked only by one thread at a time
        (i.e. no thread contention occurs while trying to lock or unlock the
        monitor).
     2. The object monitor isn't used for the Object.wait(), Object.notify(),
        Object.notifyAll(), and hash code calculation operations.

   If the above criteria is met, then the lock is implemented by simply
   setting the low 2 bits of a header bit flag word in each object using some
   form of atomic operation.  The fast lock mechanism basically provides some
   speed optimization benefits due to:
     1. Not having to actually invoke OS system calls to lock/unlock a monitor.
     2. Reduce the amount of memory/resource allocations necessary to implement
        the monitor.  Monitors are allocated per thread as well as globally and
        are reclaimed for reuse through various scavenging mechanism.  This
        reuse speeds up monitor operations if a previously reclaimed monitor is
        available for reuse.

   At any time that the criteria for fast locking/unlocking are not met, the
   monitor implementation will inflate itself into a heavy weight monitor which
   relies on the platform-specific mutex and condvar implementations defined by
   the porting layer to do its locking/unlocking.

   The Monitor Life-Cycles:
   =======================

    Initial state:
        . The object header bits indicate CVM_LOCKSTATE_UNLOCKED.

    Locked without contention (fast lock):
        . The object header bits indicate CVM_LOCKSTATE_LOCKED and holds a
          pointer to a CVMOwnedMonitor instance.
        . The thread who owns the lock on that object also holds a pointer to
          the same CVMOwnedMonitor in its owned list.

    Locked with contention (heavy weight lock):
        . The object header bits indicate CVM_LOCKSTATE_MONITOR and holds a
          pointer to a CVMObjMonitor instance.
        . The CVMObjMonitor is marked as CVM_OBJMON_OWNED and put on the
          global bound list.
        . The thread who owns the lock still have a pointer to the
          CVMOwnedMonitor record.  The CVMOwnedMonitor record has pointer to
          to the CVMObjMonitor for the lock.

    Unlocked while holding a fast lock:
        . The object header bits are changed back to CVM_LOCKSTATE_UNLOCKED.
        . The CVMOwnedMonitor is moved from the owner thread's owned list to
          its free list.

    Unlocked while holding a heavy weight lock:
        . The object header bits remains CVM_LOCKSTATE_MONITOR and the object
          retains its pointer to the CVMObjMonitor instance.
        . The CVMObjMonitor is marked as CVM_OBJMON_BOUND and remains on the
          global bound list.

    Each thread always keep at least one free CVMObjMonitor around to
    accelerate impending monitor inflation operations.  After a monitor has
    been inflated, the thread will attempt to replenish its supply of free
    CVMObjMonitors.  It does this by either:
        1. Getting one from the global free list.
        2. Scavenging for bound monitors which are not locked.
        3. Allocating new ones.

    During CVMObjMonitor scavenging:
        . CVMObjMonitors which are bound (on the global bound list) but not
          locked are released into the thread specific free list if it is
          empty.  Else, it is released into the global free list.
        . CVMObjMonitors which are on the global unbound list (see notes on
          monitor GC below) are unlocked and released into the thread specific
          free list if it is empty.  Else, it is released into the global free
          list.

    CVMOwnedMonitor scavenging:
        . CVMOwnedMonitor needed by most monitor operations and attained before
          the operation commences (at the top of the functions).
        . Usually, it is attained from the thread's free list.  If the free
          list is empty, then the thread will attempt to replenish the free
          CVMOwnedMonitor supply by one of the following ways:
            1. Run the scavenger which searches the thread's owned list for
               monitors which have fast locks but are no longer associated
               with a live object (see notes on monitor GC below).  These
               monitor records are unlocked and released back into the thread's
               free list.
            2. Allocate a new one.

    During a monitor GC (invoked by the GC):
        . CVMObjMonitors which are locked but no longer associated with a live
          object (because the object has been GC'ed but the lock hasn't been
          released prior) are moved from the global bound list to the global
          unbound list.  See notes on monitor scavenging above to see how these
          monitors get released back into the free lists for reuse.

    During thread destruction:
        . When a thread shuts down, it will unlock all monitors whose lock it
          owns if those monitors are still associated with live objects.
        . While attempting to unlock monitors, if the thread encounters any
          monitors which are no longer associated with any live objects, it
          will invoke the appropriate scavengers to reclaim the monitors into
          the free lists.
        . After reclaiming or unlocking all the monitors on its owned list, the
          thread will release/free up all the free monitors on its free list
          before exiting.

    Deflation of heavy weight monitors:
    ==================================
    Heavy weight monitors are deflated in the scavenging process.

    Synchronization between various Monitor Methods:
    ===============================================
    The following methods must be protected from themselves and each other by
    sync'ing on the same lock.  The lock currently used for this purpose is
    CVMglobals.syncLock.  The methods are as follows:
        1. CVMobjectInflate()
        2. CVMobjectSetHashCode()
        3. CVMfastUnlock()
*/

static void CVMsyncMonitorScavenge(CVMExecEnv *ee);
static void CVMownedMonitorInit(CVMOwnedMonitor *m, CVMExecEnv *owner);
static void CVMownedMonitorDestroy(CVMOwnedMonitor *m);

static CVMTryLockFunc   CVMfastTryLock, CVMdetTryLock;
static CVMTryLockFunc   CVMfastTryUnlock, CVMdetTryUnlock;
static CVMLockFunc      CVMfastLock, CVMdetLock;
static CVMLockFunc      CVMfastUnlock, CVMdetUnlock;
static CVMNotifyFunc    CVMfastNotify, CVMdetNotify;
static CVMNotifyAllFunc CVMfastNotifyAll, CVMdetNotifyAll;
static CVMWaitFunc      CVMfastWait, CVMdetWait;

/*
 * The locking vector.
 */
const CVMSyncVector CVMsyncKinds[2] = {
    {CVMfastTryLock, CVMfastLock, CVMfastTryUnlock, CVMfastUnlock,
     CVMfastWait, CVMfastNotify, CVMfastNotifyAll},
    {CVMdetTryLock,  CVMdetLock,  CVMdetTryUnlock, CVMdetUnlock,
     CVMdetWait,  CVMdetNotify,  CVMdetNotifyAll}
};

/*
 * The fast lockers.
 * All functions are called while gc-unsafe, but Lock is
 * allowed to become safe because it can block.
 */

static CVMBool
CVMfastTryLock(CVMExecEnv* ee, CVMObject* obj)
{
    CVMOwnedMonitor *o = ee->objLocksFreeOwned;

    /* return FALSE on:
     * - exception
     * - contention
     * - monitor not allocated
     */

    /* %comment lamm005 */
    CVMtraceFastLock(("fastTryLock(%x,%x)\n", ee, obj));

    if (o == NULL) {
	return CVM_FALSE;
    }

    /* %comment lamm006 */
#ifdef CVM_DEBUG
    CVMassert(o->state == CVM_OWNEDMON_FREE);
#endif
    CVMassert(o->type == CVM_OWNEDMON_FAST);
    o->object = obj;
#ifdef CVM_DEBUG
    o->state = CVM_OWNEDMON_OWNED;
#endif

#if CVM_FASTLOCK_TYPE != CVM_FASTLOCK_NONE
    {
	CVMUint32 nbits = (CVMUint32)(o) | CVM_LOCKSTATE_LOCKED;

	CVMassert(o->owner == ee);
#if CVM_FASTLOCK_TYPE == CVM_FASTLOCK_ATOMICOPS
	{
	    CVMUint32 obits0;
        /* %comment lamm004 */
	    CVMUint32 obits =
		CVMhdrBitsPtr(obj->hdr.various32) | CVM_LOCKSTATE_UNLOCKED;

	    o->u.fast.bits = obits; /* be optimistic */
	    obits0 = CVMatomicCompareAndSwap(&obj->hdr.various32,
		nbits, obits);
	    if (obits0 != obits) {
		goto fast_failed;
	    }
	}

#elif CVM_FASTLOCK_TYPE == CVM_FASTLOCK_MICROLOCK

	/* NOTE: All changes to bits must be made under micro lock */
	CVMobjectMicroLock(ee, obj);
	{
	    CVMUint32 obits = obj->hdr.various32;
	    if (CVMhdrBitsSync(obits) == CVM_LOCKSTATE_UNLOCKED) {
		o->u.fast.bits = obits;
		obj->hdr.various32 = nbits;
	    } else {
		CVMobjectMicroUnlock(ee, obj);
		goto fast_failed;
	    }
	}
	CVMobjectMicroUnlock(ee, obj);
#else
#error Unknown value for CVM_FASTLOCK_TYPE
#endif
	/*
	 * After this point, the monitor could have been inflated
	 * by another thread...
	 */
	CVMassert(ee->objLocksFreeOwned == o);

	/* remove from free list */
	ee->objLocksFreeOwned = o->next;

	o->next = ee->objLocksOwned;
	ee->objLocksOwned = o;

	return CVM_TRUE;
    }

fast_failed:

#endif

#ifdef CVM_DEBUG
    o->state = CVM_OWNEDMON_FREE;
#endif

    if (CVMobjMonitorState(obj) == CVM_LOCKSTATE_MONITOR) {
	CVMObjMonitor *mon = CVMobjMonitor(obj);

	    if (mon->mon.monitorControl != 0)
	      {
		//int i = 0;
	      }

#ifdef CVM_DEBUG
	CVMassert(mon->magic == CVM_OBJMON_MAGIC);
#endif
	CVMassert(mon->obj == obj);
	if (!CVMsysMonitorTryEnter(ee, &mon->mon)) {
	    return CVM_FALSE;
	}

	if (CVMsysMonitorCount(&mon->mon) == 1) {
	    mon->state = CVM_OBJMON_OWNED;
	    if (mon->mon.monitorControl != 0)
	      {
		//int i = 0;
	      }

	    CVMassert(ee->objLocksFreeOwned == o);
	    /* remove from free list */
	    ee->objLocksFreeOwned = o->next;

	    o->type = CVM_OWNEDMON_HEAVY;
#ifdef CVM_DEBUG
	    o->state = CVM_OWNEDMON_OWNED;
	    mon->owner = o;
#endif
            o->object = mon->obj;
	    o->u.heavy.mon = mon;
	    o->next = ee->objLocksOwned;
	    ee->objLocksOwned = o;
	}
	return CVM_TRUE;
    }
    return CVM_FALSE;
}

/* Purpose: Replenish the thread specific free list of CVMObjMonitors by
            tapping into the global free list. */
static void
CVMreplenishFreeListsNonBlocking(CVMExecEnv *ee)
{
    CVMObjMonitor *f = ee->objLocksFreeUnlocked;
    /*
     * Grab a free lock from the global pool, if available
     * CVMglobals.objLocksFree is protected by syncLock
     */
    if (f == NULL) {
	CVMObjMonitor *m = CVMglobals.objLocksFree;
	if (m != NULL) {
	    CVMassert(m->state == CVM_OBJMON_FREE);
	    if (CVMsysMonitorOwner(&m->mon) == NULL) {
		CVMglobals.objLocksFree = m->next;
		m->next = ee->objLocksFreeUnlocked;
		ee->objLocksFreeUnlocked = m;
	    }
	}
    }
}

/* Purpose: Replenish the thread specific free list of CVMObjMonitors by
            scavenging for unused CVMObjMonitors or creating a new one. */
static void
CVMreplenishFreeListsBlocking(CVMExecEnv *ee)
{
    if (ee->objLocksFreeUnlocked == NULL) {
        /* NOTE: ee->objLocksOwned, and ee->objLocksFreeOwned may be changed
            by CVMsyncMonitorScavenge().  Do not cache these values. */
        CVMsyncMonitorScavenge(ee);
        if (CVMglobals.objLocksFree != NULL) {
            /* It appears that after scavenging, we have replenished the global
               free list.  So try to replenish the thread specific free list
               for CVMObjMonitors by tapping into the global free list. 
               NOTE: This is not guaranteed to succeed because some other
                 thread might come along and consume the monitor our scavenger
                 has just freed. */
            CVMsysMutexLock(ee, &CVMglobals.syncLock);
            CVMreplenishFreeListsNonBlocking(ee);
            CVMsysMutexUnlock(ee, &CVMglobals.syncLock);
        }
        /* If all else fails, let's create a new one: */
        if (ee->objLocksFreeUnlocked == NULL) {
            CVMObjMonitor *f;
#ifdef CVM_DEBUG
	    //printf("[OBJSYNC.C] replenishFreeList allocating concurrentlocks(%d)\n",concurrentLocks);
#endif
	    if ((maximumConcurrentLocks == -1) || ((concurrentLocks +1) <= maximumConcurrentLocks))
	      {
		f = (CVMObjMonitor *)calloc(1, sizeof (CVMObjMonitor));
		if (f != NULL) {
		  concurrentLocks++;
		  CVMobjMonitorInit(f, NULL, 0);
		  ee->objLocksFreeUnlocked = f;
		  CVMassert(CVMsysMonitorOwner(&f->mon) == NULL);
		}
	      }
	    else
	      {
		/* Dio: I allocate the lock to be able to 
		 * throw the exception given that it needs additional 
		 * locks to throw it 
		 */

		if (reservedExceptionsLocks < 5)
		  {
		    f = (CVMObjMonitor *)calloc(1, sizeof (CVMObjMonitor));
		    reservedExceptionsLocks++;
		    if (f != NULL) {
		      // concurrentLocks++;
		      CVMobjMonitorInit(f, NULL, 0);

		      /* We set the sticky bit so they will not be taken off
		       * from the objects needed for the exception
		       */

		      f->sticky = CVM_TRUE;
		      ee->objLocksFreeUnlocked = f;
		      CVMassert(CVMsysMonitorOwner(&f->mon) == NULL);
		    }
		  }
		else
		  {
		    f = NULL;	   
#ifdef CVM_DEBUG
		    //printf("[resorcelimit] overreserved\n");
#endif
		  }
		CVMthrowResourceLimitError(ee, "Maximum Concurrent Locks Exceeded");
		//f = NULL;	   
	      }
        }
    }
}

/* Purpose: Replenish the thread specific CVMOwnedMonitor free list either by
            scavenging for it or creating a new one.  */
static void
CVMreplenishLockRecordUnsafe(CVMExecEnv *ee)
{
    CVMassert(CVMD_isgcUnsafe(ee));

    /* Scavenge for unused CVMOwnedMonitors: */
    CVMD_gcSafeExec(ee, {
        /* NOTE: ee->objLocksOwned, and ee->objLocksFreeOwned may be changed
             by CVMsyncMonitorScavenge().  Do not cache these values.*/
        CVMsyncMonitorScavenge(ee);

        /* If the scavenger was not successful in reclaiming one, then create
           a new one: */
        if (ee->objLocksFreeOwned == NULL) {
            CVMOwnedMonitor *o;
            o = (CVMOwnedMonitor *)calloc(1, sizeof (CVMOwnedMonitor));
            if (o != NULL) {
                CVMownedMonitorInit(o, ee);
                ee->objLocksFreeOwned = o;
            }
        }
    });
}

static CVMObjMonitor *
CVMobjectInflate(CVMExecEnv *ee, CVMObjectICell* indirectObj)
{
    CVMObjMonitor *mon=NULL, *omon=NULL;
    CVMObjMonitor **fromListPtr=NULL;
    CVMUint32 bits;
    
    /* Dio: I promoted the obj variable to the function scope, we will see if it works */
    CVMObject *obj;
    int JVMCeiContinue = JVMCEI_CONTINUE_CURRENT_CALL;

    {
        // CVMObject *obj;
        obj = CVMID_icellDirect(ee, indirectObj);
	if (CVMobjMonitorState(obj) == CVM_LOCKSTATE_MONITOR) {
	    CVMObjMonitor *mon = CVMobjMonitor(obj);
	    CVMassert(mon->obj == obj);
	    return mon;
	}
    }

    /* %comment lamm008 */
    /* See notes above on "Synchronization between various Monitor Methods": */
    CVMD_gcSafeExec(ee, {
	CVMsysMutexLock(ee, &CVMglobals.syncLock);
    });

    {
        // CVMObject *obj 
        // obj = CVMID_icellDirect(ee, indirectObj);
	bits = obj->hdr.various32;
    }
    if (CVMhdrBitsSync(bits) == CVM_LOCKSTATE_MONITOR) {
	CVMObjMonitor *mon = (CVMObjMonitor *)CVMhdrBitsPtr(bits);

	CVMassert(mon->state != CVM_OBJMON_FREE);
#ifdef CVM_DEBUG
	CVMassert(mon->magic == CVM_OBJMON_MAGIC);
#endif
	CVMassert(mon->obj == CVMID_icellDirect(ee, indirectObj));

	/* Another thread beat us to it */

#ifdef CVM_AGGRESSIVE_REPLINISH
	/* replenish free lists */
	goto done;
#else
	CVMsysMutexUnlock(ee, &CVMglobals.syncLock);

	return mon;
#endif
    }

    /* unlocked --> monitor or locked (fast) --> monitor */

    {

      /* Dio: if the monitor control is not 0 then we need to allocate the
       * monitor and initialize it ourselves
       */

    {
      extern intFuncPtr JVMCeiFilters[];
      extern int JVMCeiInitialized;
      JVMCeiContinue = JVMCEI_CONTINUE_CURRENT_CALL;
      
      if (JVMCeiInitialized &&
	  (JVMCeiFilters[JVMCEF_BEFORE_CREATE_MONITOR] != NULL))
	{
	  struct arg_t { JNIEnv *env; CVMObjMonitor **pmon; CVMObject *obj;} arg;
	  //arg.env = CVMexecEnv2JniEnv(ee);
	  arg.pmon = &mon;
	  arg.obj = obj;
	  
	  JVMCeiContinue = JVMCeiFilters[JVMCEF_BEFORE_CREATE_MONITOR]((void *) &arg);
	}
      
      if (JVMCeiContinue == JVMCEI_CONTINUE_CURRENT_CALL)
	{
	  fromListPtr = &ee->objLocksFreeUnlocked;
	  mon = *fromListPtr;
          if (mon == NULL) {
            CVMsysMutexUnlock(ee, &CVMglobals.syncLock);
            return NULL;
          }
	  CVMassert(CVMsysMonitorOwner(&mon->mon) == NULL);
          /* %comment lamm009 */
	}
#if CVM_DEBUG_MONITOR_1
      else
	{
	  int i;
	}
#endif
    }
#if CVM_FASTLOCK_TYPE == CVM_FASTLOCK_NONE
	{
	    CVMBool success = CVMsysMonitorTryEnter(ee, &mon->mon);
	    CVMassert(success);
	}
#endif
    }

    if (JVMCeiContinue == JVMCEI_CONTINUE_CURRENT_CALL)
      {
      omon = mon;

      CVMassert(mon->state == CVM_OBJMON_FREE);
      mon->state = CVM_OBJMON_BOUND;
      }

    {
	CVMObject *obj = CVMID_icellDirect(ee, indirectObj);
#if CVM_FASTLOCK_TYPE != CVM_FASTLOCK_NONE
	CVMUint32 obits;

#if CVM_FASTLOCK_TYPE == CVM_FASTLOCK_ATOMICOPS
	obits = CVMatomicSwap(&(obj)->hdr.various32, CVM_LOCKSTATE_LOCKED);
#elif CVM_FASTLOCK_TYPE == CVM_FASTLOCK_MICROLOCK
	CVMobjectMicroLock(ee, obj);
	obits = obj->hdr.various32;

#endif

	if (CVMhdrBitsSync(obits) == CVM_LOCKSTATE_LOCKED) {
	    CVMOwnedMonitor *o = (CVMOwnedMonitor *)CVMhdrBitsPtr(obits);
	    CVMExecEnv *owner = o->owner;
#ifdef CVM_DEBUG
	    CVMassert(o->magic == CVM_OWNEDMON_MAGIC);
#endif
	    CVMassert(o->type == CVM_OWNEDMON_FAST);
	    if (owner != ee) {
		/* set mutex ownership */
		CVMsysMonitorCount(&mon->mon) = 1;
		CVMsysMonitorOwner(&mon->mon) = owner;
		CVMsysMonitorSetMutexOwner(ee, &mon->mon, owner);
#ifdef CVM_DEBUG
		{
		  CVMBool success;
		    success = CVMsysMonitorTryEnter(ee, &mon->mon);
		    CVMassert(!success);
		}
#endif
	    } else {
	      CVMBool success;
		success = CVMsysMonitorTryEnter(ee, &mon->mon);
		CVMassert(success); (void) success;
	    }
	    mon->bits = o->u.fast.bits;
	    mon->obj = obj;
#ifdef CVM_DEBUG
	    mon->owner = o;
#endif
	    o->type = CVM_OWNEDMON_HEAVY;
            o->object = mon->obj;
	    o->u.heavy.mon = mon;
	    mon->state = CVM_OBJMON_OWNED;

#ifdef CVM_DEBUG_MONITOR_1
	    if (mon->mon.monitorControl != 0)
	      {
		int i =0;
	      }
#endif

	    CVMassert(CVMsysMonitorOwner(&mon->mon) == o->owner);
	    CVMobjMonitorSet(obj, mon, CVM_LOCKSTATE_MONITOR);
	} else if (CVMhdrBitsSync(obits) == CVM_LOCKSTATE_UNLOCKED) {
	    CVMassert(CVMsysMonitorOwner(&mon->mon) == NULL);
	    mon->bits = obits;
	    mon->obj = obj;
#ifdef CVM_DEBUG
	    mon->owner = NULL;
#endif
	    CVMobjMonitorSet(obj, mon, CVM_LOCKSTATE_MONITOR);
	} else /* CVM_LOCKSTATE_MONITOR */ {
	    mon = (CVMObjMonitor *)CVMhdrBitsPtr(obits);
	    CVMassert(mon->obj == obj);
#if CVM_FASTLOCK_TYPE == CVM_FASTLOCK_ATOMICOPS
	    obj->hdr.various32 = obits;
#endif
	}

#if CVM_FASTLOCK_TYPE == CVM_FASTLOCK_MICROLOCK
	CVMobjectMicroUnlock(ee, obj);
#endif

#else /* !FAST */

#ifdef CVM_DEBUG
	CVMassert(mon->magic == CVM_OBJMON_MAGIC);
#endif

	CVMassert(CVMsysMonitorCount(&mon->mon) == 1);
	CVMassert(CVMsysMonitorOwner(&mon->mon) == ee);
	{
	    CVMUint32 bits = obj->hdr.various32;
	    CVMassert(CVMhdrBitsSync(bits) == CVM_LOCKSTATE_UNLOCKED);
	    mon->bits = bits;
	}
	mon->obj = obj;
	CVMobjMonitorSet(obj, mon, CVM_LOCKSTATE_MONITOR);
#endif
    }

    /* From a free list or object header? */
    if (omon == mon) {
	*fromListPtr = mon->next;
	mon->next = CVMglobals.objLocksBound;
	CVMglobals.objLocksBound = mon;
    }

    if (JVMCeiContinue != JVMCEI_CONTINUE_CURRENT_CALL)
      {
	mon->next = CVMglobals.objLocksBound;
	CVMglobals.objLocksBound = mon;
      }


#ifdef CVM_AGGRESSIVE_REPLINISH
done:
#endif

    CVMassert(CVMobjMonitor(CVMID_icellDirect(ee, indirectObj)) == mon);
    CVMassert(mon->state != CVM_OBJMON_FREE);

    /* This is like a reference count, needed if count==0 */
    ee->objLockCurrent = mon;

    /* %comment lamm010 */
    CVMD_gcSafeExec(ee, {
	CVMreplenishFreeListsNonBlocking(ee);
	CVMsysMutexUnlock(ee, &CVMglobals.syncLock);
	CVMreplenishFreeListsBlocking(ee);
    });

    ee->objLockCurrent = NULL;
    return mon;
}

/* Returns: CVM_TRUE if successful.  CVM_FALSE if we're out of memory, and
            hence not able to lock the monitor. */
static CVMBool
CVMfastLock(CVMExecEnv* ee, CVMObjectICell* indirectObj)
{
    CVMObjMonitor *mon;

    CVMassert(CVMD_isgcUnsafe(ee));

    /* %comment lamm011 */
    CVMtraceFastLock(("fastLock(%x,%x)\n", ee,
	CVMID_icellDirect(ee, indirectObj)));

    if (ee->objLocksFreeOwned == NULL) {
	/* becomes safe */
        CVMreplenishLockRecordUnsafe(ee);
        if (ee->objLocksFreeOwned == NULL) {
            /* out of memory */
            return CVM_FALSE;
        }
    }

    mon = CVMobjectInflate(ee, indirectObj);
    if (mon == NULL) {
	/* out of memory */
	return CVM_FALSE;
    }

#if CVM_FASTLOCK_TYPE != CVM_FASTLOCK_NONE
    {
	CVMassert(mon->state != CVM_OBJMON_FREE);
	/* This is like a reference count, needed if count==0 */
        /* %comment lamm012 */
	ee->objLockCurrent = mon;
	CVMsysMonitorEnterUnsafe(ee, &mon->mon);
	ee->objLockCurrent = NULL;
	CVMassert(CVMsysMonitorOwner(&mon->mon) == ee);
	CVMassert(mon->state != CVM_OBJMON_FREE);
#ifdef CVM_DEBUG
	CVMassert(mon->magic == CVM_OBJMON_MAGIC);
#endif
	CVMassert(mon->obj == CVMID_icellDirect(ee, indirectObj));
	if (CVMsysMonitorCount(&mon->mon) == 1) {
	    CVMOwnedMonitor *o = ee->objLocksFreeOwned;
	    ee->objLocksFreeOwned = ee->objLocksFreeOwned->next;
	    CVMassert(o != NULL);

	    mon->state = CVM_OBJMON_OWNED;
	    if (mon->mon.monitorControl != 0)
	      {
		//int i=0;
	      }
	    o->type = CVM_OWNEDMON_HEAVY;
#ifdef CVM_DEBUG
	    o->state = CVM_OWNEDMON_OWNED;
	    mon->owner = o;
#endif
            o->object = mon->obj;
	    o->u.heavy.mon = mon;
	    o->next = ee->objLocksOwned;
	    ee->objLocksOwned = o;
	} else {
	    CVMassert(mon->state == CVM_OBJMON_OWNED);
	}
    }
#else /* !FAST */
    {
	CVMassert(mon->state == CVM_OBJMON_BOUND);
	CVMassert(CVMsysMonitorOwner(&mon->mon) == ee);
	CVMassert(CVMsysMonitorCount(&mon->mon) == 1);

	{
	    CVMOwnedMonitor *o = ee->objLocksFreeOwned;
	    ee->objLocksFreeOwned = ee->objLocksFreeOwned->next;
	    CVMassert(o != NULL);

	    mon->state = CVM_OBJMON_OWNED;

#ifdef CVM_DEBUG_MONITOR_1 
	    if (mon->mon.monitorControl != 0)
	      {
		int i=0;
	      }
#endif

#ifdef CVM_DEBUG
	    o->state = CVM_OWNEDMON_OWNED;
	    mon->owner = o;
#endif
	    o->type = CVM_OWNEDMON_HEAVY;
            o->object = mon->obj;
	    o->u.heavy.mon = mon;
	    o->next = ee->objLocksOwned;
	    ee->objLocksOwned = o;
	}
    }
#endif

    return CVM_TRUE;
}

/* Returns: CVM_TRUE if successful.  CVM_FALSE if the current thread if not
            the owner of the specified lock, and hence not able to unlock the
            monitor. */
/* Note: This function is ONLY to be executed under GC unsafe conditions,
         hence we don't have to worry about GC compacting memory and
         changing the value of pointers. */
static CVMBool
CVMprivateUnlock(CVMExecEnv *ee, CVMObjMonitor *mon)
{
    CVMUint32 count = CVMsysMonitorCount(&mon->mon);
    CVMassert(mon->state != CVM_OBJMON_FREE);

#ifdef CVM_DEBUG_MONITOR_1
    if (mon->mon.monitorControl != 0)
      {
	printf("privateunlock: unlocking pid(%d) mon count(%d)\n",getpid(),count);
      }
#endif

#ifdef CVM_DEBUG
    CVMassert(mon->magic == CVM_OBJMON_MAGIC);
#endif
    if (ee->objLocksOwned == NULL) {
        /* Fail because this ee (i.e thread) does not own the lock on this
           monitor.  Hence it is not allowed to unlock it. */
	return CVM_FALSE;
    }

    if (count > 0 && CVMsysMonitorOwner(&mon->mon) == ee) {
	if(mon->state != CVM_OBJMON_OWNED){
	   //printf("CVMprivateUnlock() thr-type %d,  mon-type %d\n",
		//	   CVMisNoHeapRealtimeThread(ee), mon->state);
	}
	CVMassert(mon->state == CVM_OBJMON_OWNED);
	if (count == 1) {
            CVMOwnedMonitor **prev = &ee->objLocksOwned;
            CVMOwnedMonitor *o = ee->objLocksOwned;

            /* Assume the most frequent case i.e. the last monitor locked is
               the one being unlocked and the appropriate CVMOwnedMonitor is
               the first on the owned list.  If this turns out to be untrue,
               then we'll have to search for the appropriate CVMOwnedMonitor:
            */
            if (o->u.heavy.mon != mon) {
                /* We use a do-while loop here instead of a while loop because
                   we've already eliminated the first element in the list as a
                   candidate and will not have to check it again. */
                do {
                    prev = &o->next;
                    o = o->next;
                } while ((o != NULL) && (o->u.heavy.mon != mon));

                /* If we've successfully found the appropriate CVMOwnedMonitor,
                   the ptr should not be NULL: */
                CVMassert (o != NULL);
            }
            /* By now, we should have attained the appropriate CVMOwnedMonitor
               as pointed to by the o pointer. */

	    CVMassert(o->owner == ee);
	    CVMassert(o->type == CVM_OWNEDMON_HEAVY);

            /* NOTE: The operation of moving the released CVMOwnedMonitor from
                the owned list to the free list is thread safe because the
                possible thread contentions are handled as follows:
                1. GC, who may want to compact memory, is held off because
                   this function is GC unsafe.
                2. Other threads who may want to inflate the monitor won't be
                   contending because the monitor has already been inflated. */

            mon->state = CVM_OBJMON_BOUND;

            /* Remove the released CVMOwnedMonitor from the owned list: */
            *prev = o->next;

	    o->type = CVM_OWNEDMON_FAST;
#ifdef CVM_DEBUG
	    o->state = CVM_OWNEDMON_FREE;
	    o->u.fast.bits = 0;
	    o->object = NULL;
	    mon->owner = NULL;
#endif

            /* Add the released CVMOwnedMonitor back on to the free list: */
	    o->next = ee->objLocksFreeOwned;
	    ee->objLocksFreeOwned = o;
	}

	CVMreentrantMutexUnlock(ee, &mon->mon.rmutex);
	return CVM_TRUE;
    }
    return CVM_FALSE;
}

/* Returns: CVM_TRUE if successful.  CVM_FALSE if the current thread if not
            the owner of the specified lock, and hence not able to unlock the
            monitor. */
/* Note: This function is ONLY to be executed under GC unsafe conditions,
         hence we don't have to worry about GC compacting memory and
         changing the value of pointers. */
static CVMBool
CVMfastTryUnlock(CVMExecEnv* ee, CVMObject* obj)
{
    CVMtraceFastLock(("fastUnlock(%x,%x)\n", ee, obj));

#if CVM_FASTLOCK_TYPE != CVM_FASTLOCK_NONE

    {
        CVMOwnedMonitor **prev;
	CVMOwnedMonitor *o;
	CVMBool result;
	CVMUint32 obits;

	/* Check to ensure we own some locks: */
        o = ee->objLocksOwned;
        if (o == NULL) {
            goto fast_failed;
        }

        /* Assume the most frequent case i.e. the last monitor locked is the
           one being unlocked and the appropriate CVMOwnedMonitor is the first
           on the owned list.  If this turns out to be untrue, then we'll have
           to search for the appropriate CVMOwnedMonitor:
        */
        prev = &ee->objLocksOwned;
        if (o->object != obj) {
            /* We use a do-while loop here instead of a while loop because
               we've already eliminated the first element in the list as a
               candidate and will not have to check it again. */
            do {
                prev = &o->next;
                o = o->next;
            } while ((o != NULL) && (o->object != obj));

            /* If we've successfully found the appropriate CVMOwnedMonitor,
               the ptr should not be NULL: */
            if (o == NULL) {
                goto fast_failed;
            }
        }
        /* By now, we should have attained the appropriate CVMOwnedMonitor as
           pointed to by the o pointer. */

        /* %comment lamm013 */
	obits = (CVMUint32)o | CVM_LOCKSTATE_LOCKED;

#if CVM_FASTLOCK_TYPE == CVM_FASTLOCK_ATOMICOPS
	{
	    CVMUint32 obits0;
	    CVMUint32 nbits = o->u.fast.bits;
	    obits0 = CVMatomicCompareAndSwap(&obj->hdr.various32,
		    nbits, obits);
	    result = (obits0 == obits);
	}

#elif CVM_FASTLOCK_TYPE == CVM_FASTLOCK_MICROLOCK

	CVMobjectMicroLock(ee, obj);
	{
	    CVMUint32 obits0 = obj->hdr.various32;
	    if (obits0 == obits) {
		obj->hdr.various32 = o->u.fast.bits;
		result = CVM_TRUE;
	    } else {
		result = CVM_FALSE;
	    }
	}
	CVMobjectMicroUnlock(ee, obj);
#else
#error
#endif

        /* If we've failed at this point, it would be because the monitor was
           inflated before we got to changing out the value of
           obj->hdr.various32 above.  No problem with that, just let the heavy
           TryUnlock() take care of it below through CVMprivateUnlock().
           If we're successful, we won't have to worry about this monitor
           being inflated because the CVMOwnedMonitor pointer is no longer
           attainable from the object being unlocked.
        */
	if (result) {

            /* NOTE: The operation of moving the released CVMOwnedMonitor from
                the owned list to the free list is thread safe because the
                possible thread contentions are handled as follows:
                1. GC, who may want to compact memory, is held off because
                   this function is GC unsafe.
                2. Other threads who may want to inflate the monitor won't be
                   contending because they won't attempt to inflate a monitor
                   that has already been unlocked (hence, they can't get to
                   the CVMOwnedMonitor pointer we're operating on). */

	    /* Remove the released CVMOwnedMonitor record from the owned list: */
	    *prev = o->next;

	    CVMassert(o->type == CVM_OWNEDMON_FAST);
#ifdef CVM_DEBUG
	    CVMassert(o->state == CVM_OWNEDMON_OWNED);
	    /* recycle mon */
	    o->state = CVM_OWNEDMON_FREE;
	    o->u.fast.bits = 0;
	    o->object = NULL;
#endif
            /* Add the released CVMOwnedMonitor record back on to the free list: */
	    o->next = ee->objLocksFreeOwned;
	    ee->objLocksFreeOwned = o;
	    CVMassert(o->owner == ee);
	    return CVM_TRUE;
	}
    }

fast_failed:

#endif

    if (CVMobjMonitorState(obj) == CVM_LOCKSTATE_MONITOR) {
	//TODO: delete this
	//printf("CVMfastTryUnlock()\n");
	return CVMprivateUnlock(ee, CVMobjMonitor(obj));
    }
    return CVM_FALSE;
}

/* Returns: CVM_TRUE if successful.  CVM_FALSE if the current thread if not
            the owner of the specified lock, and hence not able to unlock the
            monitor. */
/* NOTE: Called while GC unsafe.  Can become GC safe. */
static CVMBool
CVMfastUnlock(CVMExecEnv* ee, CVMObjectICell* indirectObj)
{
    CVMBool success;
    /* %comment lamm014 */

    /* If the monitor has already been inflated by another thread, just unlock
       it: */
    {
        CVMObject *obj = CVMID_icellDirect(ee, indirectObj);
        if (CVMobjMonitorState(obj) == CVM_LOCKSTATE_MONITOR) {
			//TODO: delete this
			//printf("CVMfastUnlock()\n");
            return CVMprivateUnlock(ee, CVMobjMonitor(obj));
        }
    }

    /* See notes above on "Synchronization between various Monitor Methods": */
    CVMD_gcSafeExec(ee, {
        CVMsysMutexLock(ee, &CVMglobals.syncLock);
    });

    /* The only thing that would cause TryUnlock() to fail are as follows:
        1. The current thread is not the owner of the monitor.
        2. The process of TryUnlock() is interrupted by the inflation of this
           monitor or the setting of the object hash value by another thread.

       While the above are all valid reasons for why TryUnlock() to fail, the
       same does not hold true for Unlock().  The only valid reason for
       Unlock() to fail is reason 1.

       Acquiring CVMglobals.syncLock ensures that we cannot be interrupted by
       CVMobjectInflate() or CVMobjectSetHashCode().  This eliminates reason 2
       from interfering with TryUnlock().  Hence, after acquiring
       CVMglobals.syncLock, we can call upon TryUnlock() to do the work of
       Unlock().
    */
    success = CVMobjectTryUnlock(ee, CVMID_icellDirect(ee, indirectObj));

    CVMsysMutexUnlock(ee, &CVMglobals.syncLock);

    return success;
}

static CVMObjMonitor *
CVMprivateGetMonitor(CVMExecEnv* ee, CVMObjectICell* indirectObj, CVMBool fast)
{
    CVMObjMonitor *mon;
    {
	CVMUint32 bits;
	{
	    CVMObject *obj = CVMID_icellDirect(ee, indirectObj);
	    bits = obj->hdr.various32;
	}
	if (fast) {
	    if (CVMhdrBitsSync(bits) == CVM_LOCKSTATE_UNLOCKED) {
		CVMthrowIllegalMonitorStateException(ee,
		    "no monitor associated with object");
		return NULL;
	    } else {
		mon = CVMobjectInflate(ee, indirectObj);
                if (mon == NULL) {
                    CVMthrowOutOfMemoryError(ee, NULL);
                    return NULL;
                }
	    }
	} else {
	    CVMassert(CVMhdrBitsSync(bits) == CVM_LOCKSTATE_MONITOR);
	    mon = (CVMObjMonitor *)CVMhdrBitsPtr(bits);
#ifdef CVM_DEBUG
	    CVMassert(mon->magic == CVM_OBJMON_MAGIC);
#endif
	}
    }

    CVMassert(mon->obj == CVMID_icellDirect(ee, indirectObj));

    if (CVMobjMonitorOwner(mon) != ee) {
	CVMthrowIllegalMonitorStateException(ee,
	    "thread does not own monitor");
	return NULL;
    }
    return mon;
}

/* These should be inlined by the compiler */
static CVMBool
CVMprivateWait(CVMExecEnv* ee, CVMObjectICell* indirectObj,
    CVMJavaLong millis, CVMBool fast)
{
    CVMWaitStatus waitStatus;
    CVMObjMonitor *mon = CVMprivateGetMonitor(ee, indirectObj, fast);

    if (mon == NULL) {
        /* CVM_WAIT_NOT_OWNER */
        CVMassert (CVMlocalExceptionOccurred(ee));
	return CVM_FALSE;
    }

    CVMassert(mon->state == CVM_OBJMON_OWNED);
#ifdef CVM_DEBUG
    CVMassert(mon->magic == CVM_OBJMON_MAGIC);
#endif
    {
#ifdef CVM_DEBUG
	CVMOwnedMonitor *o = mon->owner;
	CVMassert(mon->owner->type == CVM_OWNEDMON_HEAVY);
	mon->owner = NULL;
#endif
	mon->state = CVM_OBJMON_BOUND;

	/* NOTE: the owned list is not modified */
	/* We don't allow the monitor to be recycled here */
	ee->objLockCurrent = mon;

	/* %comment lamm015 */
	CVMD_gcSafeExec(ee, {
	    waitStatus = CVMsysMonitorWait(ee, &mon->mon, millis);
	});

	ee->objLockCurrent = NULL;
#ifdef CVM_DEBUG
	mon->owner = o;
	CVMassert(mon->owner->type == CVM_OWNEDMON_HEAVY);
#endif

	CVMassert(mon->state == CVM_OBJMON_BOUND);
#ifdef CVM_DEBUG
	CVMassert(mon->magic == CVM_OBJMON_MAGIC);
#endif
	CVMassert(CVMobjMonitorOwner(mon) == ee);
	mon->state = CVM_OBJMON_OWNED;
	if (mon->mon.monitorControl != 0)
	  {
	    //int i=0;
	  }
	if (waitStatus != CVM_WAIT_OK) {
	    CVMassert(waitStatus == CVM_WAIT_INTERRUPTED);
	    CVMthrowInterruptedException(ee, "operation interrupted");
	}
    }

    return waitStatus == CVM_WAIT_OK;
}

static CVMBool
CVMprivateNotify(CVMExecEnv* ee, CVMObjectICell* indirectObj, CVMBool fast)
{
    CVMObjMonitor *mon = CVMprivateGetMonitor(ee, indirectObj, fast);
    if (mon != NULL) {
	CVMcondvarNotify(&mon->mon.condvar);
	return CVM_TRUE;
    } else {
        CVMassert (CVMlocalExceptionOccurred(ee));
	return CVM_FALSE;
    }
}

static CVMBool
CVMprivateNotifyAll(CVMExecEnv* ee, CVMObjectICell* indirectObj, CVMBool fast)
{
    CVMObjMonitor *mon = CVMprivateGetMonitor(ee, indirectObj, fast);
    if (mon != NULL) {
	CVMcondvarNotifyAll(&mon->mon.condvar);
	return CVM_TRUE;
    } else {
        CVMassert (CVMlocalExceptionOccurred(ee));
	return CVM_FALSE;
    }
}


static CVMBool
CVMfastWait(CVMExecEnv* ee, CVMObjectICell* indirectObj, CVMJavaLong millis)
{
    CVMtraceFastLock(("fastWait(%x,%x)\n", ee, 
	CVMID_icellDirect(ee, indirectObj)));
    return CVMprivateWait(ee, indirectObj, millis, CVM_TRUE);
}

static CVMBool
CVMfastNotify(CVMExecEnv* ee, CVMObjectICell* indirectObj)
{
    /* %comment lamm016 */
    CVMtraceFastLock(("fastNotify(%x,%x)\n", ee, 
	CVMID_icellDirect(ee, indirectObj)));
    return CVMprivateNotify(ee, indirectObj, CVM_TRUE);
}

static CVMBool
CVMfastNotifyAll(CVMExecEnv* ee, CVMObjectICell* indirectObj)
{
    /* %comment lamm017 */
    CVMtraceFastLock(("fastNotifyAll(%x,%x)\n", ee, 
	CVMID_icellDirect(ee, indirectObj)));
    return CVMprivateNotifyAll(ee, indirectObj, CVM_TRUE);
}


/*
 * The deterministic lockers
 * Object header should have forwarding pointer to
 * monitor structure, with "evacuated" bits set
 */
static CVMBool
CVMdetTryLock(CVMExecEnv* ee, CVMObject* obj)
{
    if (ee->objLocksFreeOwned != NULL) {
	CVMObjMonitor *mon = CVMobjMonitor(obj);

#ifdef CVM_DEBUG_MONITOR_1
	if (mon->mon.monitorControl != 0)
	  {
	    int i =0;
	    printf("detTrylock: trying to lock pid(%d)\n", getpid());
	  }
#endif

	CVMtraceDetLock(("detTryLock(%x,%x)\n", ee, obj));
	CVMassert(CVMobjMonitorState(obj) == CVM_LOCKSTATE_MONITOR);

	if (!CVMsysMonitorTryEnter(ee, &mon->mon)) {

#ifdef CVM_DEBUG_MONITOR_1
	  printf("dettrylock: returning false pid(%d)\n",getpid());
#endif
	    return CVM_FALSE;
	}

	if (CVMsysMonitorCount(&mon->mon) == 1) {
	    CVMOwnedMonitor *o = ee->objLocksFreeOwned;
	    /* remove from free list */
	    ee->objLocksFreeOwned = o->next;
	    mon->state = CVM_OBJMON_OWNED;

#ifdef CVM_DEBUG_MONITOR_1
	    if (mon->mon.monitorControl != 0)
	      {
		int i=0;
		printf("dettrylock: owned! pid(%d)\n", getpid());
	      }
#endif 

	    o->type = CVM_OWNEDMON_HEAVY;
#ifdef CVM_DEBUG
	    o->state = CVM_OWNEDMON_OWNED;
	    mon->owner = o;
#endif
            o->object = mon->obj;
	    o->u.heavy.mon = mon;
	    o->next = ee->objLocksOwned;
	    ee->objLocksOwned = o;
	}

	return CVM_TRUE;
    }
    return CVM_FALSE;
}

/* Returns: CVM_TRUE if successful.  CVM_FALSE if we're out of memory, and
            hence not able to lock the monitor. */
static CVMBool
CVMdetLock(CVMExecEnv* ee, CVMObjectICell* indirectObj)
{
    CVMObjMonitor *mon;

    CVMtraceDetLock(("detLock(%x,%x)\n", ee,
	CVMID_icellDirect(ee, indirectObj)));

    if (ee->objLocksFreeOwned == NULL) {
	/*
	 * The following call becomes GC-safe, so it is possible that
	 * when we return, the monitor is no longer allocated.  Therefore,
	 * we redispatch through the vector.
	 */
        CVMreplenishLockRecordUnsafe(ee);
	if (ee->objLocksFreeOwned == NULL) {
	    /* out of memory */
	    return CVM_FALSE;
	}
	return CVMobjectLock(ee, indirectObj);
    }

    {
	CVMObject *obj = CVMID_icellDirect(ee, indirectObj);
	mon = CVMobjMonitor(obj);

#ifdef CVM_DEBUG_MONITOR_1
	if (mon->mon.monitorControl != 0)
	  {
	    printf("detlock: locking pid(%d)\n",getpid());
	  }
#endif

	CVMtraceDetLock(("detLock(%x,%x)\n", ee, obj));
	CVMassert(CVMobjMonitorState(obj) == CVM_LOCKSTATE_MONITOR);
	CVMassert(mon->obj == obj);
    }

    CVMassert(mon->state != CVM_OBJMON_FREE);
#ifdef CVM_DEBUG
    CVMassert(mon->magic == CVM_OBJMON_MAGIC);
#endif
    /* This is like a reference count, needed if count==0 */
    ee->objLockCurrent = mon;

    /* %comment lamm018 */

#ifdef CVM_DEBUG_MONITOR_1
    if (mon->mon.monitorControl !=0)
      {
	printf("detLock: about to call lock owner(%p), ee(%p)\n",CVMsysMonitorOwner(&mon->mon), ee);
      }
#endif
    
    CVMsysMonitorEnterUnsafe(ee, &mon->mon);

#ifdef CVM_DEBUG_MONITOR_1
    if (mon->mon.monitorControl !=0)
      {
	printf("inside CS pid(%d)\n",getpid());
	/* because this monitor is not part of the cache we
	 * are assigning it to the thread that locks it by
	 * hand
	 */
	//CVMsysMonitorOwner(&mon->mon) = ee;
      }
#endif

    ee->objLockCurrent = NULL;

    CVMassert(CVMsysMonitorOwner(&mon->mon) == ee);
#ifdef CVM_DEBUG
    CVMassert(mon->magic == CVM_OBJMON_MAGIC);
#endif
    if (CVMsysMonitorCount(&mon->mon) == 1) {
	CVMOwnedMonitor *o = ee->objLocksFreeOwned;
	/* remove from free list */
	ee->objLocksFreeOwned = o->next;
	CVMassert(o != NULL);

	/* Debugging */
	if (mon->state != CVM_OBJMON_BOUND)

#ifdef CVM_DEBUG_MONITOR_1
	  printf("*** mon %p not bounded state(%d), pid(%d)\n", mon, mon->state,getpid());
#endif

	CVMassert(mon->state == CVM_OBJMON_BOUND);
	mon->state = CVM_OBJMON_OWNED;

#ifdef CVM_DEBUG_MONITOR_1
	if (mon->mon.monitorControl != 0)
	  {
	    printf("privatelock: mon <-CVM_OBJMON_OWNED pid(%d)\n",getpid());
	  }
#endif

	o->type = CVM_OWNEDMON_HEAVY;
#ifdef CVM_DEBUG
	mon->owner = o;
	o->state = CVM_OWNEDMON_OWNED;
#endif
        o->object = mon->obj;
	o->u.heavy.mon = mon;
	o->next = ee->objLocksOwned;
	ee->objLocksOwned = o;
    } else {
	CVMassert(mon->state == CVM_OBJMON_OWNED);
    }

    return CVM_TRUE;
}

/* Returns: CVM_TRUE if successful.  CVM_FALSE if the current thread if not
            the owner of the specified lock, and hence not able to unlock the
            monitor. */
static CVMBool
CVMdetTryUnlock(CVMExecEnv* ee, CVMObject* obj)
{
    CVMtraceDetLock(("detTryUnlock(%x,%x)\n", ee, obj));

    CVMassert(CVMobjMonitorState(obj) == CVM_LOCKSTATE_MONITOR);

	//TODO: delete this
	//printf("CVMdetTryUnlock()\n");
    return CVMprivateUnlock(ee, CVMobjMonitor(obj));
}

/* Returns: CVM_TRUE if successful.  CVM_FALSE if the current thread if not
            the owner of the specified lock, and hence not able to unlock the
            monitor. */
static CVMBool
CVMdetUnlock(CVMExecEnv* ee, CVMObjectICell* indirectObj)
{
    CVMObject *obj = CVMID_icellDirect(ee, indirectObj);
    CVMtraceDetLock(("detUnlock(%x,%x)\n", ee, obj));

    CVMassert(CVMobjMonitorState(obj) == CVM_LOCKSTATE_MONITOR);

    /* %comment lamm019 */
	//TODO: delete this
	//printf("CVMdetUnlock()\n");
    return CVMprivateUnlock(ee, CVMobjMonitor(obj));
}

static CVMBool
CVMdetWait(CVMExecEnv* ee, CVMObjectICell* indirectObj, CVMJavaLong millis)
{
    CVMtraceDetLock(("detWait(%x,%x)\n", ee, 
	CVMID_icellDirect(ee, indirectObj)));
    return CVMprivateWait(ee, indirectObj, millis, CVM_FALSE);
}

static CVMBool
CVMdetNotify(CVMExecEnv* ee, CVMObjectICell* indirectObj)
{
    CVMtraceDetLock(("detNotify(%x,%x)\n", ee, 
	CVMID_icellDirect(ee, indirectObj)));
    return CVMprivateNotify(ee, indirectObj, CVM_FALSE);
}

static CVMBool
CVMdetNotifyAll(CVMExecEnv* ee, CVMObjectICell* indirectObj)
{
    CVMtraceDetLock(("detNotifyAll(%x,%x)\n", ee, 
	CVMID_icellDirect(ee, indirectObj)));
    return CVMprivateNotifyAll(ee, indirectObj, CVM_FALSE);
}

void
CVMobjMonitorInit(CVMObjMonitor *m, CVMExecEnv *owner, CVMUint32 count)
{
    m->sticky = CVM_FALSE;
    m->next = NULL;
    m->obj = NULL;
    m->state = CVM_OBJMON_FREE;
    //    m->mon.monitorControl = protocol;
#ifdef CVM_DEBUG
    m->magic = CVM_OBJMON_MAGIC;
    m->owner = NULL;
#endif
    CVMsysMonitorInit(&m->mon, owner, count);
}

#if 0
void 
CVMobjMonitorInit(CVMObjMonitor *m, CVMExecEnv *owner, CVMUint32 count)
{
  CVMobjMonitorInitRK(m,owner,count,0);
}
#endif

void CVMobjMonitorDestroy(CVMObjMonitor *m)
{
    CVMsysMonitorDestroy(&m->mon);
}

static void
CVMownedMonitorInit(CVMOwnedMonitor *m, CVMExecEnv *owner)
{
    m->next = NULL;
    m->owner = owner;
    m->type = CVM_OWNEDMON_FAST;
    m->object = NULL;
#ifdef CVM_DEBUG
    m->state = CVM_OWNEDMON_FREE;
    m->magic = CVM_OWNEDMON_MAGIC;
#endif
}

static void
CVMownedMonitorDestroy(CVMOwnedMonitor *m)
{
}

/* Purpose: Nullifies the object reference in any bound CVMObjMonitors and
            moves them from the global bound list to the global unbound list
            if they are no longer associated with a live object. */
/* NOTE: The CVMObjMonitors (and their CVMOwnedMonitors) will actually be
         cleaned up in CVMmonitorScavengeUnbound(). */
/* NOTE: This method is only called during a GC cycle. */
static void
deleteUnreferencedBound(CVMExecEnv *ee,
			CVMRefLivenessQueryFunc isLive, void* isLiveData,
			CVMObjMonitor **fromListPtr,
			CVMObjMonitor **toListPtr) 
{
    int cont = JVMCEI_CONTINUE_CURRENT_CALL; 
    CVMObjMonitor **monPtr = fromListPtr;
    CVMObjMonitor *mon = *monPtr;
    while (mon != NULL) {
	CVMassert(mon->state != CVM_OBJMON_FREE);
	CVMassert(mon->obj != NULL);
	if (!(*isLive)(&mon->obj, isLiveData)) {
	    /* not referenced, remove */
	    mon->obj = NULL;
	    if (toListPtr != NULL) {
		*monPtr = mon->next;
		mon->sticky = CVM_FALSE;
		/* Dio: Call the filter to destroy PI monitors
		 */

		{
		  extern intFuncPtr JVMCeiFilters[];
		  extern int JVMCeiInitialized;
		  /* Dio: calling the filter */
		  if (JVMCeiInitialized &&
		      (JVMCeiFilters[JVMCEF_BEFORE_FREE_MONITOR] != NULL))
		    {
		      struct arg_t { CVMObjMonitor **pmon;} arg;
		      arg.pmon = &mon;
		      
		      cont = JVMCeiFilters[JVMCEF_BEFORE_FREE_MONITOR]((void *) &arg);
		    }
		}

		/* If it was not a PI Monitor then add it to the list */
		if (cont == JVMCEI_CONTINUE_CURRENT_CALL)
		  {
		    mon->next = *toListPtr;
		    *toListPtr = mon;
		  }

		mon = *monPtr;
		continue;
	    }
	}
	monPtr = &mon->next;
	mon = *monPtr;
    }
}

/* Purpose: Nullifies the object reference of any CVMOwnedMonitors if they are
            no longer associated with a live object. */
/* NOTE: The fast lock CVMOwnedMonitors will actually be cleaned up in
         CVMmonitorScavengeFast(). */
/* NOTE: This method is only called during a GC cycle. */
static void
deleteUnreferencedFast(CVMExecEnv *ee,
		       CVMRefLivenessQueryFunc isLive, void* isLiveData,
		       CVMOwnedMonitor **fromListPtr)
{
    CVMOwnedMonitor *mon = *fromListPtr;
    while (mon != NULL) {
        CVMObject **objPtr = &mon->object;
        if ((*objPtr != NULL) && !(*isLive)(objPtr, isLiveData)) {
            /* not referenced, remove reference to object */
            *objPtr = NULL;
#ifdef CVM_DEBUG
            CVMconsolePrintf("Warning! GC found "
                "unreachable *locked* object!\n");
#endif
        }
	mon = mon->next;
    }
}

/* Purpose: Delete all monitors which correspond to unreferenced objects. */
/* NOTE: This method is only called during a GC cycle. */
void
CVMdeleteUnreferencedMonitors(CVMExecEnv *ee,
			      CVMRefLivenessQueryFunc isLive, 
			      void* isLiveData,
			      CVMRefCallbackFunc transitiveScanner,
			      void* transitiveScannerData)
{
    deleteUnreferencedBound(ee, isLive, isLiveData,
	&CVMglobals.objLocksBound,
	&CVMglobals.objLocksUnbound);
    CVM_WALK_ALL_THREADS_EXCEPT_NOHEAP(ee, targetEE, {
	/* fast locks are not on bound list */
	deleteUnreferencedFast(ee, isLive, isLiveData,
	    &targetEE->objLocksOwned);
    });

    CVMscanMonitors(ee, transitiveScanner, transitiveScannerData);
}

static void
scanListBound(CVMExecEnv *targetEE, CVMRefCallbackFunc refCallback, void* data,
    CVMObjMonitor **fromListPtr, CVMObjMonitor **toListPtr)
{
    CVMObjMonitor *mon = *fromListPtr;
    while (mon != NULL) {
	CVMassert(mon->state != CVM_OBJMON_FREE);
	CVMassert(mon->obj != NULL);
	(*refCallback)(&mon->obj, data);
	mon = mon->next;
    }
}

static void
scanListFast(CVMExecEnv *targetEE, CVMRefCallbackFunc refCallback, void* data)
{
    CVMOwnedMonitor *mon = targetEE->objLocksOwned;
    while (mon != NULL) {
#ifdef CVM_DEBUG
	CVMassert(mon->state != CVM_OWNEDMON_FREE);
#endif
        if (mon->object != NULL) {
            (*refCallback)(&mon->object, data);
	}
	mon = mon->next;
    }
}

void
CVMscanMonitors(CVMExecEnv *ee, CVMRefCallbackFunc refCallback, void* data)
{
    /* Set busy state */
    CVM_WALK_ALL_THREADS_EXCEPT_NOHEAP(ee, targetEE, {
    //CVM_WALK_ALL_THREADS(ee, targetEE, {
	CVMObjMonitor *mon = targetEE->objLockCurrent;
	if (mon != NULL) {
	    CVMassert(mon->state != CVM_OBJMON_FREE);
	    mon->state |= CVM_OBJMON_BUSY;
	}
    });

    /* Scan global "bound" list */
    scanListBound(ee, refCallback, data,
	&CVMglobals.objLocksBound,
	&CVMglobals.objLocksFree);

    /* Scan per-thread owned lists */
    CVM_WALK_ALL_THREADS_EXCEPT_NOHEAP(ee, targetEE, {
    //CVM_WALK_ALL_THREADS(ee, targetEE, {
	/* fast locks are not on bound list */
	scanListFast(targetEE, refCallback, data);
    });

    /* Undo busy state */
    CVM_WALK_ALL_THREADS_EXCEPT_NOHEAP(ee, targetEE, {
    //CVM_WALK_ALL_THREADS(ee, targetEE, {
	CVMObjMonitor *mon = targetEE->objLockCurrent;
	if (mon != NULL) {
	    mon->state &= ~CVM_OBJMON_BUSY;
	}
    });
}

/* Purpose: Scavenge for CVMOwnedMonitor for fast lock monitors which are no
           longer associated with any live objects. */
/* NOTE: Scavenging of heavy weight monitors are done in
         CVMmonitorScavengeUnbound(). */
/* NOTE: This function assumes that it won't be interrupted by a GC cycle.
         This is either enforced by being in a GC unsafe state or by acquiring
         some of the GC locks before entering this function. */
/* NOTE: ee->objLocksOwned, and ee->objLocksFreeOwned may be changed by
         CVMmonitorScavengeFast().  Do not cache these values across a call to
         this function.*/
static void
CVMmonitorScavengeFast(CVMExecEnv *ee)
{
    CVMOwnedMonitor **prev;
    CVMOwnedMonitor *o;

    /* Scavenge for unused CVMOwnedMonitors: */
    prev = &ee->objLocksOwned;
    o = ee->objLocksOwned;
    while (o != NULL) {
        /* NOTE: This check is not in contention with monitor inflation
            because only CVMOwnedMonitors whose o->object is not NULL will
            have a chance of being inflated. */
        if ((o->object == NULL) && (o->type == CVM_OWNEDMON_FAST)) {
            /* Remove the released CVMOwnedMonitor record from the owned list: */
            *prev = o->next;
#ifdef CVM_DEBUG
            CVMassert(o->state == CVM_OWNEDMON_OWNED);
            /* recycle mon: */
            o->state = CVM_OWNEDMON_FREE;
            o->u.fast.bits = 0;
#endif
            /* Add the released CVMOwnedMonitor record back on to the free list: */
            o->next = ee->objLocksFreeOwned;
            ee->objLocksFreeOwned = o;
            CVMassert(o->owner == ee);

            o = *prev;
            continue;
        }
        prev = &o->next;
        o = o->next;
    }
}

/* Purpose: Scavenge for CVMObjMonitors which no longer have a lock on them
            (found on the global bound list) and release them to the thread
            specific or global free list. */
/* NOTE: This function is protected from GC cycles by CVMglobals.syncLock. */
static void
CVMmonitorScavengeBound(CVMExecEnv *ee, CVMObjMonitor **fromListPtr,
                   CVMObjMonitor **toListPtr)
{
    CVMObjMonitor **monPtr = fromListPtr;
    CVMObjMonitor *mon = *monPtr;

    while (mon != NULL) {

	CVMassert(mon->state != CVM_OBJMON_FREE);

	if (CVMobjMonitorCount(mon) == 0 && mon->state == CVM_OBJMON_BOUND) {

	    CVMassert(CVMsysMonitorOwner(&mon->mon) == NULL);

	    if (!mon->sticky) {
                CVMassert(fromListPtr == &CVMglobals.objLocksBound);
                mon->obj->hdr.various32 = mon->bits;
                mon->obj = NULL;
		mon->state = CVM_OBJMON_FREE;

                /* NOTE: There is no CVMOwnedMonitor associated with this
                    CVMObjMonitor to be released because the object's lock
                    count is 0.  It was dis-associated with its CVMOwnedMonitor
                    back when it was unlocked for the last time. */

                /* Move the CVMObjMonitor over to either the thread specific
                   or global free list: */
                *monPtr = mon->next;
                mon->next = CVMglobals.objLocksFree;
                CVMglobals.objLocksFree = mon;
		mon = *monPtr;
		continue;
	    }
	}
	monPtr = &mon->next;
	mon = *monPtr;
    }
}

/* Purpose: Scavenge for CVMObjMonitors which are no longer associated with a
            live object (found on the global unbound list) and release them to
            the thread specific or global free list. */
/* NOTE: This function is protected from GC cycles by CVMglobals.syncLock. */
static void
CVMmonitorScavengeUnbound(CVMExecEnv *ee)
{
    CVMObjMonitor **monPtr = &CVMglobals.objLocksUnbound;
    CVMObjMonitor *mon = CVMglobals.objLocksUnbound;

    while (mon != NULL) {
        CVMassert(mon->state != CVM_OBJMON_FREE);
        if (!mon->sticky) {
            if (CVMsysMonitorOwner(&mon->mon) == ee) {

                /* NOTE: If we get here, the following conditions must be true:
                   1. The object which this CVMObjMonitor is associated with
                      is no longer reachable by anyone (including both java and
                      JNI code) other than the GC.
                   2. That means that no other threads will be competing to
                      invoke trylock(), lock(), tryUnlock(), unlock(), wait(),
                      notify(), or notifyAll() on the monitor.
                   3. Also, we're in a GC safe region operating under the
                      directions of the GC.  When any thread wishes to invoke
                      trylock(), lock(), tryUnlock(), unlock(), wait(),
                      notify(), or notifyAll() on any monitor, it must first
                      become GC unsafe.  This means that there won't be any
                      other threads competing to modify the global bound and
                      unbound lists.
                   4. We have also acquired the CVMglobals.syncLock.  Hence,
                      GC won't be competing for the global bound and unbound
                      lists.
    
                   Hence, it is thread-safe to release this CVMObjMonitor (and
                   its corresponding CVMOwnedMonitor) which is no longer
                   associated with a live object: */

                CVMOwnedMonitor **prev = &ee->objLocksOwned;
                CVMOwnedMonitor *o = ee->objLocksOwned;

                /* Get the CVMOwnedMonitor for this CVMObjMonitor: */
                /* Assume the most frequent case i.e. the last monitor locked is
                   the one being unlocked and the appropriate CVMOwnedMonitor is
                   the first on the owned list.  If this turns out to be untrue,
                   then we'll have to search for the appropriate CVMOwnedMonitor:
                */
                if (o->u.heavy.mon != mon) {
                    /* We use a do-while loop here instead of a while loop because
                       we've already eliminated the first element in the list as a
                       candidate and will not have to check it again. */
                    do {
                        prev = &o->next;
                        o = o->next;
                    } while ((o != NULL) && (o->u.heavy.mon != mon));

                    /* If we've successfully found the appropriate CVMOwnedMonitor,
                       the ptr should not be NULL: */
                    CVMassert (o != NULL);
                }
                /* By now, we should have attained the appropriate CVMOwnedMonitor
                   as pointed to by the o pointer. */

                CVMassert(o->owner == ee);
                CVMassert(o->type == CVM_OWNEDMON_HEAVY);

                /* Remove the released CVMOwnedMonitor from the owned list: */
                *prev = o->next;

                o->type = CVM_OWNEDMON_FAST;
#ifdef CVM_DEBUG
                o->state = CVM_OWNEDMON_FREE;
                o->u.fast.bits = 0;
                o->object = NULL;
                mon->owner = NULL;
#endif
                /* Add the released CVMOwnedMonitor back on to the free list: */
                o->next = ee->objLocksFreeOwned;
                ee->objLocksFreeOwned = o;

                /* Release the lock: */
                CVMobjMonitorCount(mon) = 1;
                CVMreentrantMutexUnlock(ee, &mon->mon.rmutex);
            }
            mon->state = CVM_OBJMON_FREE;

            /* Move the CVMObjMonitor over to either the thread specific or
               global free list: */

	    *monPtr = mon->next;
	    mon->next = CVMglobals.objLocksFree;
	    CVMglobals.objLocksFree = mon;
	    mon = *monPtr;
	    continue;
        }
        monPtr = &mon->next;
        mon = *monPtr;
    }
}

/* %comment lamm020 */
/* NOTE: ee->objLocksOwned, and ee->objLocksFreeOwned may be changed by
         CVMsyncMonitorScavenge().  Do not cache these values across a call
         this function. */
static void
CVMsyncMonitorScavenge(CVMExecEnv *ee)
{
    /* NOTE: These locks prevents GC from interfering: */
    CVMsysMutexLock(ee, &CVMglobals.threadLock);
    CVMsysMutexLock(ee, &CVMglobals.syncLock);

    /* NOTE: Becoming GC safe prevents other threads from interfering: */
    CVMD_gcBecomeSafeAll(ee);

    /* Set busy state */
    /* %comment lamm021 */ 
    CVM_WALK_ALL_THREADS_START(ee, targetEE) {
	CVMObjMonitor *mon = targetEE->objLockCurrent;
	if (mon != NULL && (!mon->sticky)) {
	    CVMassert(mon->state != CVM_OBJMON_FREE);
	    mon->state |= CVM_OBJMON_BUSY;
	}
    } CVM_WALK_ALL_THREADS_END(ee, targetEE)

    CVMmonitorScavengeBound(ee, &CVMglobals.objLocksBound,
	&CVMglobals.objLocksFree);
    CVMmonitorScavengeUnbound(ee);
    CVMmonitorScavengeFast(ee);

    /* Undo busy state */
    CVM_WALK_ALL_THREADS_START(ee, targetEE) {
	CVMObjMonitor *mon = targetEE->objLockCurrent;
	if (mon != NULL && (!mon->sticky)) {
	    mon->state &= ~CVM_OBJMON_BUSY;
	}
    } CVM_WALK_ALL_THREADS_END(ee, targetEE)

    CVMD_gcAllowUnsafeAll(ee);

    CVMsysMutexUnlock(ee, &CVMglobals.syncLock);
    CVMsysMutexUnlock(ee, &CVMglobals.threadLock);
}

void
CVMeeSyncInit(CVMExecEnv *ee)
{
    /* create pre-allocated object monitor and add to free list */
    /* %comment lamm022 */
    ee->objLocksFreeOwned =
	(CVMOwnedMonitor *)calloc(1, sizeof (CVMOwnedMonitor));
    CVMassert(ee->objLocksFreeOwned != NULL);
    CVMownedMonitorInit(ee->objLocksFreeOwned, ee);
    ee->objLocksFreeUnlocked =
	(CVMObjMonitor *)calloc(1, sizeof (CVMObjMonitor));

    concurrentLocks++;
    CVMassert(ee->objLocksFreeUnlocked != NULL);
    CVMobjMonitorInit(ee->objLocksFreeUnlocked, NULL, 0);
}

void
CVMeeSyncDestroy(CVMExecEnv *ee)
{
    CVMOwnedMonitor **prev = &ee->objLocksOwned;
    CVMOwnedMonitor *o = ee->objLocksOwned;

    /*
     * Unlock owned monitors
     */
#ifdef CVM_DEBUG
    if (ee->objLocksOwned != NULL) {
        CVMconsolePrintf("Thread %d (%x) exited with locks held\n",
            ee->threadID, ee);
    }
#endif

    /* NOTE: The purpose of this loop is to reclaim any locks which are still
        held by the thread but should be released because the thread is coming
        to an end.  Reclamation can occur in one of 2 ways:
            1. If the monitor still corresponds to a live object, then simply
               unlocking the monitor will put it back on the free lists which
               will be cleaned up at the bottom of this function.
            2. If the monitor no longer corresponds to a live object, then the
               unlock mechanisms won't be able to handle them.  We skip these
               and let the scavengers release them once we're out of the loop.
    */

    while (o != NULL) {
        CVMOwnedMonitor *next;
	CVMassert(o->owner == ee);

	CVMID_localrootBegin(ee) {
	    CVMID_localrootDeclare(CVMObjectICell, objectICell);
	    CVMD_gcUnsafeExec(ee, {
                CVMID_icellSetDirect(ee, objectICell, o->object);
	    });

            next = o->next;
            /* If we have a monitor with a NULL object reference, then there
               must be some monitors which need to be cleaned up by the
               scavenger first: */
            if (!CVMID_icellIsNull(objectICell)) {
                /* If we have a monitor with a NULL object reference, then
                   there must be some monitors which need to be cleaned up by
                   the scavenger first: */

                /* unlock multiple lockings */

	      /* Dio: if not priority inheritance monitors 
	       */ 

	      if ((o->type == CVM_OWNEDMON_HEAVY) && (o->u.heavy.mon->mon.monitorControl == 0))
		{
		  if (o->type == CVM_OWNEDMON_HEAVY) {
		    CVMassert(CVMobjMonitorCount(o->u.heavy.mon) > 0);
		    CVMobjMonitorCount(o->u.heavy.mon) = 1;
		  }
		  {
		    CVMBool success = CVMgcSafeObjectUnlock(ee, objectICell);
		    CVMassert(success); (void) success;
		  }
		
		}

	      CVMassert(*prev != o);		
            } else {
	      prev = &o->next;
            }
            o = next;
	} CVMID_localrootEnd();
    }

    /* NOTE: ee->objLocksOwned, and ee->objLocksFreeOwned may be changed by
        CVMsyncMonitorScavenge().  Do not cache these values.*/
    CVMsyncMonitorScavenge(ee);

    /*
     * Release locks from free lists
     */
    /* %comment lamm023 */
    while (ee->objLocksFreeOwned != NULL) {
	CVMOwnedMonitor *m = ee->objLocksFreeOwned;
	ee->objLocksFreeOwned = m->next;
	CVMownedMonitorDestroy(m);
	free(m);
    }

    while (ee->objLocksFreeUnlocked != NULL) {
	CVMObjMonitor *m = ee->objLocksFreeUnlocked;
	ee->objLocksFreeUnlocked = m->next;
	CVMobjMonitorDestroy(m);
	free(m);
    }

}

void
CVMeeSyncDestroyGlobal(CVMExecEnv *ee)
{
    CVMGlobalState *gs = &CVMglobals;

    /*
     * Transfer (hopefully all) locks from Bound and Unbound lists
     * to Free list.
     * NOTE: ee->objLocksOwned, and ee->objLocksFreeOwned may be changed by
     *       CVMsyncMonitorScavenge().  Do not cache these values.
     */
    CVMsyncMonitorScavenge(ee);

#ifdef CVM_DEBUG
    if (gs->objLocksBound != NULL) {
        CVMconsolePrintf("VM exited with bound busy monitors!\n");
    }
    if (gs->objLocksUnbound != NULL) {
        CVMconsolePrintf("VM exited with unbound busy monitors!\n");
    }
#endif

    /*
     * Release locks from free lists
     */
    while (gs->objLocksFreeOwned != NULL) {
	CVMOwnedMonitor *m = gs->objLocksFreeOwned;
	gs->objLocksFreeOwned = m->next;
	CVMownedMonitorDestroy(m);
	free(m);
    }

    while (gs->objLocksFree != NULL) {
	CVMObjMonitor *m = gs->objLocksFree;
	gs->objLocksFree = m->next;
	CVMobjMonitorDestroy(m);
	free(m);
    }

}

/* Purpose: Sets the hash value in an inflated object monitor. */
/* NOTE: Only called by CVMobjectSetHashCode() while GC unsafe.  Must not
         become GC safe. */
static CVMInt32
CVMobjectSetHashCodeInInflatedMonitor(CVMExecEnv *ee, CVMObject *directObj,
                                      CVMUint32 hashValue)
{
    CVMObjMonitor *mon = CVMobjMonitor(directObj);
    CVMassert(CVMD_isgcUnsafe(ee));

    /* NOTE: Deflation only occurs while scavenging, but scavenging can
        only occur when all threads have become gc safe.  Hence, being gc
        unsafe protects us from deflation. */
    CVMobjectMicroLock(ee, directObj);
    {
        CVMUint32 h = CVMhdrBitsHashCode(mon->bits);
        if (h == CVM_OBJECT_NO_HASH) {
            CVMhdrSetHashCode(mon->bits, hashValue);
        } else {
            hashValue = h;
        }
    }
    CVMobjectMicroUnlock(ee, directObj);
    return hashValue;
}

/* NOTE: Only called while in a GC unsafe state. */
static CVMInt32
CVMobjectSetHashCode(CVMExecEnv *ee, CVMObjectICell* indirectObj,
		     CVMUint32 hashValue)
{
    CVMUint32 bits;

    CVMassert(CVMD_isgcUnsafe(ee));
    CVMassert((hashValue & ~CVM_OBJECT_HASH_MASK) == 0);

    /* Check to see if the monitor has already been inflated by anyone else: */
    {
        CVMObject *obj = CVMID_icellDirect(ee, indirectObj);
        if (CVMobjMonitorState(obj) == CVM_LOCKSTATE_MONITOR) {
            hashValue = CVMobjectSetHashCodeInInflatedMonitor(ee, obj,
                            hashValue);
            return hashValue;
        }
    }

    /* See notes above on "Synchronization between various Monitor Methods": */
    CVMD_gcSafeExec(ee, {
        CVMsysMutexLock(ee, &CVMglobals.syncLock);
    });

    {
        CVMObject *obj = CVMID_icellDirect(ee, indirectObj);
        bits = obj->hdr.various32;
    }

    if (CVMhdrBitsSync(bits) == CVM_LOCKSTATE_MONITOR) {
        /* If we're here, then another thread has inflated the monitor: */

        CVMObject *obj = CVMID_icellDirect(ee, indirectObj);
        hashValue = CVMobjectSetHashCodeInInflatedMonitor(ee, obj,
                        hashValue);
    } else {

        CVMObject *obj = CVMID_icellDirect(ee, indirectObj);

#if CVM_FASTLOCK_TYPE != CVM_FASTLOCK_NONE
        CVMUint32 obits;

        /* Prevent any fast lock/unlock operations: */
#if CVM_FASTLOCK_TYPE == CVM_FASTLOCK_ATOMICOPS
        obits = CVMatomicSwap(&(obj)->hdr.various32, CVM_LOCKSTATE_LOCKED);
#elif CVM_FASTLOCK_TYPE == CVM_FASTLOCK_MICROLOCK
        CVMobjectMicroLock(ee, obj);
        obits = obj->hdr.various32;
#endif

        if (CVMhdrBitsSync(obits) == CVM_LOCKSTATE_LOCKED) {
            /* If we're here, then the object has been locked using the fast
               lock mechanism.  Hence, the hash value will have to be set in
               the corresponding CVMOwnedMonitor: */
            CVMOwnedMonitor *o = (CVMOwnedMonitor *)CVMhdrBitsPtr(obits);
            CVMUint32 h = CVMhdrBitsHashCode(o->u.fast.bits);

#ifdef CVM_DEBUG
            CVMassert(o->magic == CVM_OWNEDMON_MAGIC);
#endif
            CVMassert(o->type == CVM_OWNEDMON_FAST);

            if (h == CVM_OBJECT_NO_HASH) {
                CVMhdrSetHashCode(o->u.fast.bits, hashValue);
            } else {
                hashValue = h;
            }
        } else if (CVMhdrBitsSync(obits) == CVM_LOCKSTATE_UNLOCKED) {
            /* If we're here, then the object is not locked.  Hence, the hash
               value will have to be set in the object's header bits: */

            CVMUint32 h = CVMhdrBitsHashCode(obits);
            if (h == CVM_OBJECT_NO_HASH) {
                CVMhdrSetHashCode(obits, hashValue);
            } else {
                hashValue = h;
            }
            /* At this point the object header needs to be updated, but this
               done in common code below: */
        }

        /* Re-allow fast lock/unlock operations: */
        obj->hdr.various32 = obits;
#if CVM_FASTLOCK_TYPE == CVM_FASTLOCK_MICROLOCK
        CVMobjectMicroUnlock(ee, obj);
#endif

#else   /* CVM_FASTLOCK_TYPE == CVM_FASTLOCK_NONE */

        /* If we're here, then the monitor is not locked.  Hence, set the hash
           value in the object header: */
        CVMUint32 h = CVMhdrBitsHashCode(obj->hdr.various32);
        if (h == CVM_OBJECT_NO_HASH) {
            CVMhdrSetHashCode(obj->hdr.various32, hashValue);
        } else {
            hashValue = h;
        }
#endif
    }

    /* %comment lamm010 */
    CVMsysMutexUnlock(ee, &CVMglobals.syncLock);
    return hashValue;
}

/*
 * Compute the hash code for an object, store it in its header, and return
 * the hash code.
 */
static CVMInt32
CVMobjectComputeHash(CVMExecEnv *ee, CVMObjectICell* indirectObj)
{
    /*
     * If this is a ROM object, just return its address as its hash value.
     * That is guaranteed to never change because the GC will never relocate
     * a ROMized object.
     *
     * Normally, "ROMized objects" are writable, as they are exposed
     * to synchronization code which requires you to write into object
     * headers. The one notable exception is the shared char[] that
     * comprises the body of ROMized String's. This char[] is
     * immutable and not visible to application code, and so cannot be
     * locked by anybody, and so does not have to be made
     * writable. Except when we need a hash code for the char[], say
     * when tracing instructions. So return the address of ROMized objects as
     * their hashcode, without touching their header.
     */
    {
	CVMInt32 hashVal = CVM_OBJECT_NO_HASH;
	CVMD_gcUnsafeExec(ee, {
	    CVMObject *directObj = CVMID_icellDirect(ee, indirectObj);
	    if (CVMobjectIsInROM(directObj)) {
		hashVal = (CVMInt32)directObj;
	    }
	});
	if (hashVal != CVM_OBJECT_NO_HASH) {
	    return hashVal;
	}
    }
    
    {
	CVMInt32 hashVal;
	/*
	 * Get the next random number that doesn't have an alternate meaning.
	 */
	do {
	    hashVal = CVMrandomNext();
	    hashVal &= CVM_OBJECT_HASH_MASK;
	} while (hashVal == CVM_OBJECT_NO_HASH);
	CVMD_gcUnsafeExec(ee, {
	    hashVal = CVMobjectSetHashCode(ee, indirectObj, hashVal);
	});
	return hashVal;
    }
}

CVMInt32
CVMobjectGetHashSafe(CVMExecEnv *ee, CVMObjectICell* indirectObj)
{
    CVMInt32 hashValue;
    {
	CVMD_gcUnsafeExec(ee, {
	    CVMObject *directObj = CVMID_icellDirect(ee, indirectObj);
	    hashValue = CVMobjectGetHashNoSet(ee, directObj);
	});
    }
    if (hashValue == CVM_OBJECT_NO_HASH) {
	return CVMobjectComputeHash(ee, indirectObj);
    } else {
	return hashValue;
    }
}

/* NOTE: Only executes in a GC unsafe state. */
CVMInt32
CVMobjectGetHashNoSet(CVMExecEnv *ee, CVMObject* directObj)
{
#if CVM_FASTLOCK_TYPE != CVM_FASTLOCK_NONE
#if CVM_FASTLOCK_TYPE == CVM_FASTLOCK_ATOMICOPS
    {
	CVMUint32 bits = CVMobjectVariousWord(directObj);
	if (CVMhdrBitsSync(bits) == CVM_LOCKSTATE_MONITOR) {
	    CVMObjMonitor *mon = (CVMObjMonitor *)CVMhdrBitsPtr(bits);
#ifdef CVM_DEBUG
	    CVMassert(mon->magic == CVM_OBJMON_MAGIC);
#endif
	    return CVMhdrBitsHashCode(mon->bits);
	} else if (CVMhdrBitsSync(bits) == CVM_LOCKSTATE_UNLOCKED) {
	    return CVMhdrBitsHashCode(bits);
	} else /* CVM_LOCKSTATE_LOCKED */ {
            /* With the fast lock bit set, it is difficult to get a hold of
               the CVMOwnedMonitor in a consistent state (not in the midst of
               being changed by another thread) without first acquiring some
               synchronization locks.  We'll need to get a hold of the
               CVMOwnedMonitor because the object header bits are in it.
               Hence, force the slow path in order to get the hash value. */
	    return CVM_OBJECT_NO_HASH;
	}
    }
#elif CVM_FASTLOCK_TYPE == CVM_FASTLOCK_MICROLOCK
    {
	CVMInt32 hash;
	CVMUint32 bits;
	CVMobjectMicroLock(ee, directObj);
	bits = CVMobjectVariousWord(directObj);
	if (CVMhdrBitsSync(bits) == CVM_LOCKSTATE_MONITOR) {
	    CVMObjMonitor *mon = (CVMObjMonitor *)CVMhdrBitsPtr(bits);
#ifdef CVM_DEBUG
	    CVMassert(mon->magic == CVM_OBJMON_MAGIC);
#endif
	    hash = CVMhdrBitsHashCode(mon->bits);
	} else if (CVMhdrBitsSync(bits) == CVM_LOCKSTATE_LOCKED) {
	    CVMOwnedMonitor *o = (CVMOwnedMonitor *)CVMhdrBitsPtr(bits);
#ifdef CVM_DEBUG
	    CVMassert(o->magic == CVM_OWNEDMON_MAGIC);
#endif
	    CVMassert(o->type == CVM_OWNEDMON_FAST);
	    hash = CVMhdrBitsHashCode(o->u.fast.bits);
	} else {
	    hash = CVMhdrBitsHashCode(CVMobjectVariousWord(directObj));
	}
	CVMobjectMicroUnlock(ee, directObj);
	return hash;
    }
#else
#error
#endif
#else   /* CVM_FASTLOCK_TYPE == CVM_FASTLOCK_NONE */

    /* The following "if" is based on a sample of the object header bits.  In
       the else clause below, it is important to get the hash value from the
       previously sampled bits (as opposed to resampling) because if the monitor
       was inflated by another thread after the initial sampling, we would end
       up in the else clause but a resample of the bits could get us the
       address of the inflated monitor instead of the proper hash value.

           If the bits indicate that the monitor has been inflated, then we're
       safe in accessing the monitor bits without a lock because we're
       currently GC unsafe, and the monitor will not be deflated until this
       thread becomes GC safe.  Hence, the monitor bits are stable.

           If the bits indicate otherwise, then it may already have the hash
       value in it.  In this case, the proper hash value will be returned.  If
       it doesn't already have the hash value (even if the hash value was set
       by another thread after we've sampled the bits), then effectively,
       CVM_OBJECT_NO_HASH will be returned, and we'll just have to get the get
       the hash value the slow way (because we cannot reliably compete with a
       race condition):
    */
    CVMUint32 bits = CVMobjectVariousWord(directObj);
    if (CVMhdrBitsSync(bits) == CVM_LOCKSTATE_MONITOR) {
	return CVMhdrBitsHashCode(CVMobjMonitor(directObj)->bits);
    } else {
        return CVMhdrBitsHashCode(bits);
    }
#endif
}

CVMBool
CVMgcSafeObjectLock(CVMExecEnv *ee, CVMObjectICell *o)
{
    CVMBool success;
    CVMD_gcUnsafeExec(ee, {
        /* %comment lamm002 */
        success = CVMobjectTryLock(ee, CVMID_icellDirect(ee, o));
        if (!success) {
            success = CVMobjectLock(ee, o);
        }
    });
    return success;
}

CVMBool
CVMgcSafeObjectUnlock(CVMExecEnv *ee, CVMObjectICell *o)
{
    CVMBool success;
    CVMD_gcUnsafeExec(ee, {
        /* %comment lamm003 */
        success = CVMobjectTryUnlock(ee, CVMID_icellDirect(ee, o));
        if (!success) {
            success = CVMobjectUnlock(ee, o);
        }
    });
    return success;
}

CVMBool
CVMgcSafeObjectWait(CVMExecEnv *ee, CVMObjectICell *o, CVMInt64 millis)
{
    CVMBool status;
    CVMD_gcUnsafeExec(ee, {
	status = CVMobjectWait(ee, o, millis);
    });
    return status;
}

CVMBool
CVMgcSafeObjectNotify(CVMExecEnv *ee, CVMObjectICell *o)
{
    CVMBool status;
    CVMD_gcUnsafeExec(ee, {
        status = CVMobjectNotify(ee, o);
    });
    return status;
}

CVMBool
CVMgcSafeObjectNotifyAll(CVMExecEnv *ee, CVMObjectICell *o)
{
    CVMBool status;
    CVMD_gcUnsafeExec(ee, {
        status = CVMobjectNotifyAll(ee, o);
    });
    return status;
}


#ifdef CVM_DEBUG
void
CVMobjectMicroLock(CVMExecEnv *ee, CVMObject *obj)
{
    CVMassert(CVMD_isgcUnsafe(ee) && ++ee->microLock == 1);
    CVMobjectMicroLockImpl(obj);
}

void
CVMobjectMicroUnlock(CVMExecEnv *ee, CVMObject *obj)
{
    CVMassert(CVMD_isgcUnsafe(ee) && --ee->microLock == 0);
    CVMobjectMicroUnlockImpl(obj);
}
#endif
