/*
 * @(#)utils.c	1.102 00/11/02
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#include "javavm/include/porting/doubleword.h"
#include "javavm/include/porting/int.h"
#include "javavm/include/porting/time.h"
#include "javavm/include/clib.h"
#include "javavm/include/utils.h"
#include "javavm/include/indirectmem.h"
#include "javavm/include/preloader.h"
#include "javavm/include/localroots.h"
#include "javavm/include/common_exceptions.h"
#include "javavm/include/stacks.h"

#include "generated/offsets/java_lang_String.h"
#include "generated/offsets/java_lang_Class.h"
#include "generated/offsets/java_lang_Boolean.h"
#include "generated/offsets/java_lang_Character.h"
#include "generated/offsets/java_lang_Float.h"
#include "generated/offsets/java_lang_Double.h"
#include "generated/offsets/java_lang_Byte.h"
#include "generated/offsets/java_lang_Short.h"
#include "generated/offsets/java_lang_Integer.h"
#include "generated/offsets/java_lang_Long.h"
#ifdef CVM_DEBUG
#include "generated/offsets/java_lang_Thread.h"
#endif

#include "jni_util.h"

/*
 * Console I/O. Same as printf, but supports special formats like
 * %C, %P, and %F. See CVMformatStringVaList for details.
 *
 * WARNING: may crash if you pass bad arguments. See CVMformatStringVaList
 * for details on safety checking limitations.
 */
void
CVMconsolePrintf(const char* format, ...)
{
    va_list argList;
    char buf[256];

    va_start(argList, format);
    CVMformatStringVaList(buf, sizeof(buf), format, argList);
    va_end(argList);
    fprintf(stderr, "%s", buf);
    fflush(stderr);
}


/*
 * Random number generation.
 *
 * Standard, well-known linear congruential random generator with
 *
 *     s(n+1) = 16807 * s(n) mod(2^31 - 1)
 *
 * For references, see:
 *
 * (1) "Random Number Generators: Good Ones Are Hard to Find",
 *      S.K. Park and K.W. Miller, Communications of the ACM 31:10 
 *      (Oct 1988),
 *
 * (2) "Two Fast Implementations of the 'Minimal Standard' Random 
 *     Number Generator", David G. Carta, Comm. ACM 33, 1 (Jan 1990), 
 *     pp. 87-88. 
 */
CVMInt32
CVMrandomNext()
{
    const CVMInt32 a = 16807;
    const CVMInt32 m = 2147483647;
    CVMInt32 lastRandom = CVMglobals.lastRandom;
    
    /* compute az=2^31p+q */
    CVMUint32 lo = a * (CVMInt32)(lastRandom & 0xFFFF);
    CVMUint32 hi = a * (CVMInt32)((CVMUint32)lastRandom >> 16);
    lo += (hi & 0x7FFF) << 16;
    
    /* if q overflowed, ignore the overflow and increment q */
    if (lo > m) {
	lo &= m;
	++lo;
    }
    lo += hi >> 15;
    
    /* if (p+q) overflowed, ignore the overflow and increment (p+q) */
    if (lo > m) {
	lo &= m;
	++lo;
    }
    CVMglobals.lastRandom = lo;
    return lo;
}

void
CVMrandomInit()
{
    CVMInt32 seed;
    CVMInt64 millis = CVMtimeMillis();
    seed = CVMlong2Int(millis);
    if (seed == 0) {
	seed = 32461;
    }
    CVMglobals.lastRandom = seed;
}

/*
 * Functions for accessing the debugging flags. All of the following functions
 * return the previous state of the flag.
 *
 * You can pass in more than one flag at a time to any of the functions.
 */
#ifdef CVM_DEBUG

CVMUint32
CVMcheckDebugFlags(CVMUint32 flags)
{
    return CVMglobals.debugFlags & flags;
}

CVMUint32
CVMsetDebugFlags(CVMUint32 flags)
{
    CVMUint32 result = CVMcheckDebugFlags(flags);
    CVMglobals.debugFlags |= flags;
    return result;
}

CVMUint32
CVMclearDebugFlags(CVMUint32 flags)
{
    CVMUint32 result = CVMcheckDebugFlags(flags);
    CVMglobals.debugFlags &= ~flags;
    return result;
}

CVMUint32
CVMrestoreDebugFlags(CVMUint32 flags, CVMUint32 oldvalue)
{
    CVMUint32 result = CVMcheckDebugFlags(flags);
    CVMglobals.debugFlags = (CVMglobals.debugFlags & ~flags) | oldvalue;
    return result;
}

#endif /* CVM_DEBUG */

#ifdef CVM_DEBUG
void CVMlong2String(CVMInt64 number, char *s, char *limit)
{
#undef MAGIC
#define MAGIC    10000000       /* slightly more than (2^64) ^ (1/3) */
    int group1, group2, group3;
    CVMInt64 group1l, group2l, group3l;
    CVMInt64 magic;
    char *sign;

    if (CVMlongGez(number)) {
        sign = "";
	/* number = -number */
	number = CVMlongNeg(number);	/* this lets -2^63 still work! */
    } else {
        sign = "-";
    }

#if 0
    group3 = -(number % MAGIC); /* low 7 digits */
    number = (number + group3)/MAGIC;
    group2 = -(number % MAGIC); /* next 7 digits */
    number = (number + group2)/MAGIC;
    group1 = -number;           /* high 7 digits */
#endif

    magic = CVMint2Long(MAGIC);
    group3l = CVMlongRem(number, magic); /* low 7 digits */
    group3l = CVMlongNeg(group3l);
    number = CVMlongAdd(number, group3l);
    number = CVMlongDiv(number, magic); /* truncate middle 7 digits */
    group2l = CVMlongRem(number, magic); /* next 7 digits */
    group2l = CVMlongNeg(group2l);
    number = CVMlongAdd(number, group2l);
    number = CVMlongDiv(number, magic); /* truncate middle 7 digits */
    group1l = CVMlongNeg(number);         /* high 7 digits */

    group1 = CVMlong2Int(group1l);
    group2 = CVMlong2Int(group2l);
    group3 = CVMlong2Int(group3l);

    if (group1 != 0) {
	(void) sprintf(s, "%s%d%07d%07d", sign, group1, group2, group3);
    } else if (group2 != 0) {
	(void) sprintf(s, "%s%d%07d", sign, group2, group3);
    } else {
	(void) sprintf(s, "%s%d", sign, group3);
    }
}
#endif

/*
 * CVMdumpObject - dumps the object.
 */
#ifdef CVM_DEBUG
/* %comment cjp012 */
void
CVMdumpObject(CVMObject* directObj)
{
    CVMconsolePrintf("%C\n", CVMobjectGetClass(directObj)); 
}
#endif

/*
 * CVMdumpThread - prints thread name and stack trace.
 */
#ifdef CVM_DEBUG
void
CVMdumpThread(JNIEnv* env)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);

    CVMconsolePrintf("Thread \"");
    CVMprintThreadName(env, CVMcurrentThreadICell(ee));
    CVMconsolePrintf("\"\n");
    CVMdumpStack(&ee->interpreterStack, 0, 0, 0);
}

static void
CVMdumpThreadFromEENoAlloc(CVMExecEnv* ee)
{
    CVMconsolePrintf("Thread ee=0x%x, ID=%d\n", ee, ee->threadID);
    CVMdumpStack(&ee->interpreterStack, 0, 0, 0);
}

void
CVMdumpAllThreads()
{
    CVMExecEnv* ee = CVMgetEE();
    
    CVMsysMutexLock(ee, &CVMglobals.threadLock);
    CVM_WALK_ALL_THREADS(ee, currentEE, {
	/*
	 * Ideally, we should be calling CVMdumpThread() here. But we
	 * are not allowed to allocate within CVM_WALK_ALL_THREADS(), so
	 * we call a non-allocating version of a thread dump that does not
	 * print the thread's name.
	 */
	CVMdumpThreadFromEENoAlloc(currentEE);
	CVMconsolePrintf("\n");
    });
    CVMsysMutexUnlock(ee, &CVMglobals.threadLock);
}

void
CVMprintThreadName(JNIEnv* env, CVMObjectICell* threadICell)
{
    CVMExecEnv *ee = CVMjniEnv2ExecEnv(env);
    jchar *charArrayElements;
    jobject nameString;
    const char* name;

    /* Kind of hacky -- mixes local roots and JNI */
    CVMID_localrootBegin(ee) {
	CVMID_localrootDeclare(CVMObjectICell, nameICell);

	CVMID_fieldReadRef(ee, threadICell,
			   CVMoffsetOfjava_lang_Thread_name,
			   nameICell);
	if (!CVMID_icellIsNull(nameICell)) {
	    charArrayElements =
		(*env)->GetCharArrayElements(env, (jcharArray) nameICell,
					     NULL);
	    nameString =
		(*env)->NewString(env, charArrayElements,
				  (*env)->GetArrayLength(env,
							 (jarray) nameICell));
	    name = (*env)->GetStringUTFChars(env, nameString, NULL);
	    CVMconsolePrintf("%s", name);
	    (*env)->ReleaseStringUTFChars(env, nameString, name);
	    (*env)->ReleaseCharArrayElements(env, (jcharArray) nameICell,
					     charArrayElements, JNI_ABORT);
	} else {
	    CVMconsolePrintf("<unnamed thread>");
	}
    } CVMID_localrootEnd();
}

#endif

/*
 * CVMclassname2String - copies the class name in src to dst, converting
 * each '/' to '.' in the process.
 */

char*
CVMclassname2String(CVMClassTypeID typeID, char *dst, int size)
{
    char *buf = dst;
    CVMtypeidClassNameToCString(typeID, dst, size);
    for (; (--size > 0) && (*dst != '\0') ;  dst++) {
	if (*dst == '/') {
	    *dst = '.';
	}
    }
    return buf;
}

/*
 * Returns the size of the utf8 char.
 */
static int
CVMutfCharSize(const char* utfstring) {
    switch ((CVMUint8)utfstring[0] >> 4) {
    default:
	return 1;
	
    case 0xC: case 0xD:	
	/* 110xxxxx  10xxxxxx */
	return 2;
	break;
	
    case 0xE:
	/* 1110xxxx 10xxxxxx 10xxxxxx */
	return 3;
	break;
    } /* end of switch */
}

/*
 * Returns # of unicode chars needed to represent the utf8 string
 */
CVMSize
CVMutfLength(const char* utf8string)
{
    CVMSize length;
    for (length = 0; *utf8string != 0; length++) {
       utf8string += CVMutfCharSize(utf8string);
    }
    return length;
}

/*
 * Copy a utf8 array into a counted unicode array.
 * Return number of UTF characters moved from the source.
 * This is to allow a user to re-start the copying at a known
 * stopping point.
 */
int
CVMutfCountedCopyIntoCharArray(
    const char * utf8Bytes,
    CVMJavaChar* unicodeChars,
    CVMSize      unicodeLength
){
    CVMSize nutf = 0;

    while( unicodeLength-- > 0 ){
	char utfchar = *utf8Bytes++;
	CVMJavaChar t;
	switch ( utfchar & 0xf0 ){
	default:
	    *unicodeChars++ = utfchar;
	    nutf+= 1;
	    continue;

	case 0xC0: case 0xD0:
	    /* 110xxxxx  10xxxxxx */
	    *unicodeChars++ = ((utfchar&0x1f)<<6) | ((*utf8Bytes++)&0x3f);
	    nutf+= 2;
	    continue;
	    
	case 0xE0:
	    /* 1110xxxx 10xxxxxx 10xxxxxx */
	    t = ((utfchar&0xf)<<6) | ((*utf8Bytes++)&0x3f);
	    *unicodeChars++ = (t<<6) | ((*utf8Bytes++)&0x3f);
	    nutf+= 3;
	    continue;
	} /* end of switch */
    }
    return nutf;
}

/*
 * Copy a utf8 array into a unicode array.
 */
void 
CVMutfCopyIntoCharArray(const char* utf8Bytes, CVMJavaChar* unicodeChars)
{
    int i = 0;
    while (*utf8Bytes != 0) {
	unicodeChars[i] = CVMutfNextUnicodeChar(&utf8Bytes);
	i++;
    }
}

/*
 * Copy a unicode array into a utf8 array.  Returns pointer to end of array.
 *
 * Code stolen from JDK unicode2utf().
 */
char *
CVMutfCopyFromCharArray(const CVMJavaChar* unicodeChars,
                        char* utf8Bytes, CVMInt32 count)
{
    while (--count >= 0) {
	CVMJavaChar ch = *unicodeChars++;
        if ((ch != 0) && (ch <=0x7f)) {
            *utf8Bytes++ = (char)ch;
        } else if (ch <= 0x7FF) {
            /* 11 bits or less. */
            CVMUint8 high_five = ch >> 6;
            CVMUint8 low_six = ch & 0x3F;
            *utf8Bytes++ = high_five | 0xC0; /* 110xxxxx */
            *utf8Bytes++ = low_six | 0x80;   /* 10xxxxxx */
        } else {
            /* possibly full 16 bits. */
            CVMInt8 high_four = ch >> 12;
            CVMInt8 mid_six = (ch >> 6) & 0x3F;
            CVMInt8 low_six = ch & 0x3f;
            *utf8Bytes++ = high_four | 0xE0; /* 1110xxxx */
            *utf8Bytes++ = mid_six | 0x80;   /* 10xxxxxx */
            *utf8Bytes++ = low_six | 0x80;   /* 10xxxxxx*/
        }
    }
    return utf8Bytes;
}

/*
 * Return the next unicode character in the utf8 string and update
 * utf8String_p to point after the unicode characater return.
 *
 * This is the same as the JDK next_utf2unicode function.
 */

extern CVMJavaChar
CVMutfNextUnicodeChar(const char** utf8String_p)
{
    CVMUint8* ptr = (CVMUint8*)(*utf8String_p);
    CVMUint8 ch, ch2, ch3;
    int length = 1;		/* default length */
    CVMJavaChar result = 0x80;	/* default bad result; */
    switch ((ch = ptr[0]) >> 4) {
        default:
	    result = ch;
	    break;

	case 0x8: case 0x9: case 0xA: case 0xB: case 0xF:
	    /* Shouldn't happen. */
	    break;

	case 0xC: case 0xD:	
	    /* 110xxxxx  10xxxxxx */
	    if (((ch2 = ptr[1]) & 0xC0) == 0x80) {
		unsigned char high_five = ch & 0x1F;
		unsigned char low_six = ch2 & 0x3F;
		result = (high_five << 6) + low_six;
		length = 2;
	    } 
	    break;

	case 0xE:
	    /* 1110xxxx 10xxxxxx 10xxxxxx */
	    if (((ch2 = ptr[1]) & 0xC0) == 0x80) {
		if (((ch3 = ptr[2]) & 0xC0) == 0x80) {
		    unsigned char high_four = ch & 0x0f;
		    unsigned char mid_six = ch2 & 0x3f;
		    unsigned char low_six = ch3 & 0x3f;
		    result = (((high_four << 6) + mid_six) << 6) + low_six;
		    length = 3;
		} else {
		    length = 2;
		}
	    }
	    break;
	} /* end of switch */

    *utf8String_p = (char *)(ptr + length);
    return result;
}

/*
 * Allocate a java/lang/String and copy the specifed utf8 string into it.
 */
void
CVMnewStringUTF(CVMExecEnv* ee, CVMStringICell* stringICell,
		const char* utf8Bytes)
{
#undef STACK_CHAR_BUF_SIZE
#define STACK_CHAR_BUF_SIZE 128
    CVMJavaChar  buf[STACK_CHAR_BUF_SIZE];
    CVMJavaChar* unicodeChars;
    CVMSize len = CVMutfLength(utf8Bytes);

    if (len > STACK_CHAR_BUF_SIZE) {
	unicodeChars = malloc(len * sizeof(CVMJavaChar));
	if (unicodeChars == 0) {
	    CVMID_icellSetNull(stringICell);
	    return;
	}
    } else {
	unicodeChars = buf;
    }
    CVMutfCopyIntoCharArray(utf8Bytes, unicodeChars);
    CVMnewString(ee, stringICell, unicodeChars, len);
    if (unicodeChars != buf) {
	free(unicodeChars);
    }
}

/*
 * Allocate a java/lang/String and copied to specifed unicode string into it.
 */
void
CVMnewString(CVMExecEnv* ee, CVMStringICell* stringICell,
	     const CVMJavaChar* unicodeChars, CVMUint32 len)
{
    /* 
     * Allocate the java/lang/String object first, and store it away
     * in stringICell.
     */
    CVMID_allocNewInstance( 
	    ee, (CVMClassBlock*)CVMsystemClass(java_lang_String), stringICell);

    if (CVMID_icellIsNull(stringICell)) {
	return;
    }

    CVMID_localrootBegin(ee) {
	/*
	 * Allocate the char array. 
	 */
	CVMID_localrootDeclare(CVMArrayOfCharICell, charArrayICell);
	CVMID_allocNewArray(
		ee, CVM_T_CHAR, 
		(CVMClassBlock*)CVMbasicTypeArrayClassblocks[CVM_T_CHAR],
		len, (CVMObjectICell*)charArrayICell);

	if (CVMID_icellIsNull(charArrayICell)) {
	    /* null result if anything failed. */
	    CVMID_icellSetNull(stringICell);
	} else {
	    /*
	     * Fill in the char array.
	     */
	    CVMD_gcUnsafeExec(ee, {
		CVMArrayOfChar* charArray =
		    CVMID_icellDirect(ee, charArrayICell);
		CVMD_arrayWriteBodyChar(unicodeChars, charArray, 0, len);
	    });
	    /*
	     * Assign all the fields of the String object. Note that
	     * directObj may have moved during the array allocation,
	     * so we need to recache it. 
	     */
	    CVMID_fieldWriteRef(ee, stringICell, 
				CVMoffsetOfjava_lang_String_value,
				(CVMObjectICell*)charArrayICell);
	    CVMID_fieldWriteInt(ee, stringICell,
				CVMoffsetOfjava_lang_String_offset,
				0);
	    CVMID_fieldWriteInt(ee, stringICell, 
				CVMoffsetOfjava_lang_String_count,
				len);
	}
    } CVMID_localrootEnd();
}

/*
 * Build a formatted string.  Special formats include:
 *
 * code	     arg		      result
 *
 * %P    CVMFrame*	   The class name, method name, source file name
 *                         and line number of the pc in the frame.
 * %O    CVMObject*	   The class name and object hash.
 * %I    CVMObjectICell*   The class name and object hash.
 * %C    CVMClassBlock*	   CVMclassname2String(CVMcbClassName(cb))
 * %F	 CVMFieldBlock*	   CVMtypeidFieldNameToCString(CVMfbNameAndTypeID(fb))
 * %M	 CVMMethodBlock*   CVMtypeidMethodNameToCString(CVMmbNameAndTypeID(mb))
 *
 * -%C,%F, and %M support the '!' modifier if you want to pass the typeid
 *  rather than the cb, fb, or mb.
 * -Supports all ANSI formats except for %n.
 * -Does not support field modifiers on %C, %F, and %M fields.
 *
 * Returns total size of formatted result, untruncated.
 *
 * WARNING: This function is not bulletproof like sprintf. You must obey
 * the following rules to avoid a crash.
 * 1. No single expanded field, other than %C, %F, and %M fields, can
 *    expand bigger than bufSize. This means, for example, if you pass
 *    a format of "%20d" and a bufSize of 10, the buffer will still get
 *    20 bytes stored in, thus corrupting memory. It also means if you
 *    pass a format of "%s" with a string argument larger than bufsize,
 *    you will also corrupt memory.
 * 2. You cannot pass in a bufSize of 0 if you use %O, %C, %M, or %F.
 */

#undef MIN
#define MIN(x,y) ((x) < (y) ? (x) : (y))

size_t
CVMformatStringVaList(char* buf, size_t bufSize, const char* format,
		      va_list ap)
{
    CVMExecEnv *ee = CVMgetEE();
    const char* f = format;
    const char* s;
    char* tmp;
    size_t      length  = 0;
    int         result = 0;
    char format2[17]; /* must be bigger than the biggest known format.
		       * "%-+ #<num>.<num>wc" is the biggest. If we
		       * limit <num> to a resonable 4 digits, this gives
		       * us a total size of 17.
		       */

    if (buf == NULL) {
	/* Just sizing result */
	tmp = NULL;
	bufSize = 0;
    } else {
	tmp = (char*)malloc(bufSize);
	if (tmp == NULL) {
	    goto finish;
	}
	/* save room for terminating \0 */
	bufSize -= 1;
    }

    while (*f != '\0') {
	char* p;
	char c;
	size_t l;

	p = strchr(f, '%');
	l = (p == NULL ? strlen(f) : p - f);

	/* copy everything up to the % */
	if (bufSize > 0) {
	    l = MIN(bufSize, l);
	    strncpy(buf, f, l);
	    buf += l;
	    bufSize -= l;
	}
	length += l;

	if (p == NULL) {
	    /* There are no more %'s, so we are done */
	    break;
	}

	/* skip modifier fields */
	f = p;  /* f points to the % */
	p++; /* skip the % */
	while ((c = *p) != 0) {
	    /* if c is any alpha other than 'w', or c is a '%', then
	     * we are done.
	     */
	    if ((((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z'))
		 && (c != 'w')) || (c == '%')) {
		break;
	    }
	    p++;
	}

	/* copy everything from the % char up to and including the field
	 * type char into format2. We need to do this since we don't
	 * want the sprintf calls to look past the field type char.
	 */
	l = p - f + 1;
	CVMassert(l + 1 <= sizeof(format2));
	strncpy(format2, f, l);
	format2[l] = 0;

	switch (c) {
	case 'd':
	case 'i':
	case 'o':
	case 'p':
	case 'u':
	case 'x':
	case 'X': {
	    int i = va_arg(ap, int);
	    result = sprintf(tmp, format2, i);
	    goto copy_string;
	}
	case 'f':
	case 'e':
	case 'E':
	case 'g':
	case 'G': {
	    float f1 = va_arg(ap, double);
	    result = sprintf(tmp, format2, f1);
	    goto copy_string;
	}
	case 'c':  {
	    char c = va_arg(ap, int);
	    result = sprintf(tmp, format2, c);
	    goto copy_string;
	}
	case 's': {
	    const char* c = va_arg(ap, const char*);
	    /*
	     * If the format string is a simple "%s", then we just copy
	     * the argument string directly into buf and also check for
	     * overflow. Otherwise, we role the dice and let sprintf
	     * handle it. This means that if you ever use a format modifier
	     * with %s, you better make sure your buffer is big enough because
	     * we don't do any overflow checking here.
	     */
	    if (format2[1] == 's') {
		l = strlen(c);
		if (bufSize > 0) {
		    l = MIN(bufSize, l);
		    strncpy(buf, c, l);
		    buf += l;
		    bufSize -= l;
		}
		length += l;
		f = p + 1;
		continue;
	    } else {
		result = sprintf(tmp, format2, c);
		goto copy_string;
	    }
	}
	case 'n':
	    CVMassert(CVM_FALSE);
	    break;
	case '%':
	    result = sprintf(tmp, format2);
	    goto copy_string;
	case 'P': {
	    /*
	     * This format is only valid for DEBUG builds.
	     */
#ifdef CVM_DEBUG
	    CVMFrame*  frame = va_arg(ap, CVMFrame*);
	    /* C stack checks */
            if (!CVMCstackCheckSize(ee,
        	                    CVM_REDZONE_CVMCstackCVMpc2string,
        	                    "CVMformatStringVaList2CVMpc2string", 
				    CVM_TRUE)) {
		/* exception already thrown */ 
		CVMclearLocalException(ee);
   		result = sprintf(tmp, 
		    "<PC information unavailable due to C stack overflow>");
    	    }
            else {
		CVMpc2string(frame, NULL, NULL, tmp, tmp + bufSize);
		result = strlen(tmp);
	    }
#else
	    (void)va_arg(ap, CVMFrame*);
	    result = sprintf(tmp, "<%%P Unavailable>");
#endif
	    goto copy_string;
	}
	case 'O': {
	    CVMObject*           directObj = va_arg(ap, CVMObject*);
	    const CVMClassBlock* cb = CVMobjectGetClass(directObj);
	    CVMClassTypeID       tid = CVMcbClassName(cb);
	    /* "12" is to leave room for the "@%#x" used for the
	     * object address. */
	    CVMclassname2String(tid, tmp, bufSize - 12);
	    l = strlen(tmp);
	    result = sprintf(tmp+l, "@%#x",
		CVMobjectGetHashNoSet(ee, directObj));
	    goto copy_string;
	}
	case 'I': {
	    CVMObjectICell*      indirectObj = va_arg(ap, CVMObjectICell*);
	    const CVMClassBlock* cb;
	    CVMClassTypeID       tid;

	    if (indirectObj == NULL || CVMID_icellIsNull(indirectObj)) {
		result = sprintf(tmp, "<null>");
	    } else if (!CVMCstackCheckSize(ee,
                CVM_REDZONE_CVMCstackCVMID_objectGetClassAndHashSafe,
                "CVMformatStringVaList2CVMID_objectGetClassAndHashSafe",
                CVM_TRUE)) {
                /* C stack checks for a one level deep recursion in
	         * both CVMID_objectGetClass and CVMobjectGetHashSafe. 
                 * Exception is already thrown if fails. */
                CVMclearLocalException(ee);
   		result = sprintf(tmp, 
		    "<class name unavailable due to C stack overflow>");
    	    } else {
    	  	CVMID_objectGetClass(ee, indirectObj, cb);
		tid = CVMcbClassName(cb);
		/* "12" is to leave room for the "@%#x" used for the
		 * object address. */
		CVMclassname2String(tid, tmp, bufSize - 12);
		l = strlen(tmp);
		result = sprintf(tmp+l, "@%#x", 
		    		 CVMobjectGetHashSafe(ee, indirectObj));
	    }
	    goto copy_string;
	}
	case 'C': {
	    CVMClassTypeID       tid;
	    if (*(p-1) == '!') {
		tid = va_arg(ap, CVMClassTypeID);
	    } else {
		const CVMClassBlock* cb = va_arg(ap, const CVMClassBlock*);
		tid = CVMcbClassName(cb);
	    }
	    CVMclassname2String(tid, tmp, bufSize);
	    goto copy_string;
	}
	case 'F': {
	    CVMFieldTypeID       tid;
	    if (*(p-1) == '!') {
		tid = va_arg(ap, CVMFieldTypeID);
	    } else {
		const CVMFieldBlock* fb = va_arg(ap, const CVMFieldBlock*);
		tid = CVMfbNameAndTypeID(fb);
	    }
	    CVMtypeidFieldNameToCString(tid, tmp, bufSize);
	    l = strlen(tmp);
	    if (bufSize - l > 0) {
		tmp[l] = ':';
		l++;
	    }
	    CVMtypeidFieldTypeToCString(tid, tmp + l, bufSize - l);
	    goto copy_string;
	}
	case 'M': {
	    CVMMethodTypeID       tid;
	    if (*(p-1) == '!') {
		tid = va_arg(ap, CVMMethodTypeID);
	    } else {
		const CVMMethodBlock* mb = va_arg(ap, const CVMMethodBlock*);
		tid = CVMmbNameAndTypeID(mb);
	    }
	    CVMtypeidMethodNameToCString(tid, tmp, bufSize);
	    l = strlen(tmp);
	    CVMtypeidMethodTypeToCString(tid, tmp + l, bufSize - l);
	    goto copy_string;
	}
	default:
	    CVMassert(CVM_FALSE);
	}

    copy_string:
	if (result < 0) {
	    goto finish;
	}
	s = tmp;
	l = strlen(s);
	if (bufSize > 0) {
	    l = MIN(bufSize, l);
	    strncpy(buf, s, l);
	    buf += l;
	    bufSize -= l;
	}
	length += l;

	f = p + 1;
    }

    if (buf != NULL) {
	*buf = '\0';
    }
   
 finish:
    if (tmp != NULL) {
	free(tmp);
    }

    if (result < 0) {
	CVMassert(CVM_FALSE);
	return 0;
    } else {
	return length;
    }
}

size_t
CVMformatString(char *buf, size_t bufSize, const char *format, ...)
{
    size_t result;
    va_list ap;
    va_start(ap, format);
    result = CVMformatStringVaList(buf, bufSize, format, ap);
    va_end(ap);
    return result;
}

/*
 * Returns true if the two classes are in the same package. 
 */
CVMBool
CVMisSameClassPackage(CVMExecEnv* ee,
		      CVMClassBlock* class1, CVMClassBlock* class2) 
{
    /*
     * Determine if both classes have the same classloader.
     */
    if (CVMcbClassLoader(class1) != CVMcbClassLoader(class2)) {
	return CVM_FALSE; 
    } else {
	return CVMtypeidIsSameClassPackage(CVMcbClassName(class1),
					   CVMcbClassName(class2));
    }
}

/*
 * Convenience functions
 */

/* %comment cjp013 */

CVMClassBlock*
CVMgcSafeClassRef2ClassBlock(CVMExecEnv* ee, CVMClassICell *clazz)
{

    CVMJavaInt cbPtr;
    CVMID_fieldReadInt(ee, clazz,
		       CVMoffsetOfjava_lang_Class_classBlockPointer,
		       cbPtr);
    return (CVMClassBlock*)cbPtr;
}

CVMClassBlock *
CVMgcUnsafeClassRef2ClassBlock(CVMExecEnv *ee, CVMClassICell *clazz)
{
    CVMJavaInt cbPtr;
    CVMObject *directClass = CVMID_icellDirect(ee, clazz);
    CVMD_fieldReadInt(directClass,
		      CVMoffsetOfjava_lang_Class_classBlockPointer,
		      cbPtr);
    return (CVMClassBlock *)cbPtr;
}

CVMBool
CVMgcSafeJavaWrap(CVMExecEnv* ee, jvalue v,
		  CVMBasicType fromType,
		  CVMObjectICell* result)
{
    switch (fromType) {
    case CVM_T_BOOLEAN:
	/* NOTE: casting away const is safe */
	CVMID_allocNewInstance(ee,
			       (CVMClassBlock*)
			       CVMsystemClass(java_lang_Boolean),
			       result);
	if (CVMID_icellIsNull(result)) {
	    CVMthrowOutOfMemoryError(ee, "allocating Java wrapper object");
	    return CVM_FALSE;
	}
	CVMID_fieldWriteInt(ee, result,
			    CVMoffsetOfjava_lang_Boolean_value,
			    v.z);
	return CVM_TRUE;
    case CVM_T_CHAR:
	/* NOTE: casting away const is safe */
	CVMID_allocNewInstance(ee,
			       (CVMClassBlock*)
			       CVMsystemClass(java_lang_Character),
			       result);
	if (CVMID_icellIsNull(result)) {
	    CVMthrowOutOfMemoryError(ee, "allocating Java wrapper object");
	    return CVM_FALSE;
	}
	CVMID_fieldWriteInt(ee, result,
			    CVMoffsetOfjava_lang_Character_value,
			    v.c);
	return CVM_TRUE;
    case CVM_T_FLOAT:
	/* NOTE: casting away const is safe */
	CVMID_allocNewInstance(ee,
			       (CVMClassBlock*)
			       CVMsystemClass(java_lang_Float),
			       result);
	if (CVMID_icellIsNull(result)) {
	    CVMthrowOutOfMemoryError(ee, "allocating Java wrapper object");
	    return CVM_FALSE;
	}
	CVMID_fieldWriteFloat(ee, result,
			      CVMoffsetOfjava_lang_Float_value,
			      v.f);
	return CVM_TRUE;
    case CVM_T_DOUBLE:
	/* NOTE: casting away const is safe */
	CVMID_allocNewInstance(ee,
			       (CVMClassBlock*)
			       CVMsystemClass(java_lang_Double),
			       result);
	if (CVMID_icellIsNull(result)) {
	    CVMthrowOutOfMemoryError(ee, "allocating Java wrapper object");
	    return CVM_FALSE;
	}
	CVMID_fieldWriteDouble(ee, result,
			       CVMoffsetOfjava_lang_Double_value,
			       v.d);
	return CVM_TRUE;
    case CVM_T_BYTE:
	/* NOTE: casting away const is safe */
	CVMID_allocNewInstance(ee,
			       (CVMClassBlock*)
			       CVMsystemClass(java_lang_Byte),
			       result);
	if (CVMID_icellIsNull(result)) {
	    CVMthrowOutOfMemoryError(ee, "allocating Java wrapper object");
	    return CVM_FALSE;
	}
	CVMID_fieldWriteInt(ee, result,
			    CVMoffsetOfjava_lang_Byte_value,
			    v.b);
	return CVM_TRUE;
    case CVM_T_SHORT:
	/* NOTE: casting away const is safe */
	CVMID_allocNewInstance(ee,
			       (CVMClassBlock*)
			       CVMsystemClass(java_lang_Short),
			       result);
	if (CVMID_icellIsNull(result)) {
	    CVMthrowOutOfMemoryError(ee, "allocating Java wrapper object");
	    return CVM_FALSE;
	}
	CVMID_fieldWriteInt(ee, result,
			    CVMoffsetOfjava_lang_Short_value,
			    v.s);
	return CVM_TRUE;
    case CVM_T_INT:
	/* NOTE: casting away const is safe */
	CVMID_allocNewInstance(ee,
			       (CVMClassBlock*)
			       CVMsystemClass(java_lang_Integer),
			       result);
	if (CVMID_icellIsNull(result)) {
	    CVMthrowOutOfMemoryError(ee, "allocating Java wrapper object");
	    return CVM_FALSE;
	}
	CVMID_fieldWriteInt(ee, result,
			    CVMoffsetOfjava_lang_Integer_value,
			    v.i);
	return CVM_TRUE;
    case CVM_T_LONG:
	/* NOTE: casting away const is safe */
	CVMID_allocNewInstance(ee,
			       (CVMClassBlock*)
			       CVMsystemClass(java_lang_Long),
			       result);
	if (CVMID_icellIsNull(result)) {
	    CVMthrowOutOfMemoryError(ee, "allocating Java wrapper object");
	    return CVM_FALSE;
	}
	CVMID_fieldWriteLong(ee, result,
			     CVMoffsetOfjava_lang_Long_value,
			     v.j);
	return CVM_TRUE;
    case CVM_T_VOID:
	CVMID_icellSetNull(result);
	return CVM_TRUE;
    default:
	CVMassert(CVM_FALSE);
	return CVM_FALSE;
    }
}

CVMBool
CVMgcSafeJavaUnwrap(CVMExecEnv* ee, CVMObjectICell* obj,
		    jvalue* v, CVMBasicType* toType,
		    CVMClassBlock* exceptionCb)
{
    CVMBool result;

    CVMD_gcUnsafeExec(ee, {
	CVMObjectICell* nonnullObj = CVMID_NonNullICellPtrFor(obj);
	/*
	 * obj may be NULL. Leave the job of throwing the exception to
	 * CVMgcUnsafeJavaUnwrap().
	 */
	result = CVMgcUnsafeJavaUnwrap(ee, CVMID_icellDirect(ee, nonnullObj),
				       v, toType, exceptionCb);
    });

    return result;
}

CVMBool
CVMgcUnsafeJavaUnwrap(CVMExecEnv* ee, CVMObject* obj,
		      jvalue* v, CVMBasicType* toType,
		      CVMClassBlock* exceptionCb)
{
    CVMClassBlock* objectClassBlock;

    if (obj == NULL) {
	if (exceptionCb == NULL) {
	    /* NullPointerException or IllegalArgumentException?  What
	       do most callers want? NOTE: I think I tested this
	       against the JDK and found that it throws
	       NullPointerException here even though the specification
	       for Method.invoke says that if the unwrapping
	       conversion fails it should throw IllegalArgumentException. */
	    CVMthrowNullPointerException(ee, "unwrapping Java value");
	} else {
	    CVMsignalError(ee, exceptionCb, "unwrapping Java value");
	}
	return CVM_FALSE;
    }

    objectClassBlock = CVMobjectGetClass(obj);

    /* %comment rt008 */

    if (objectClassBlock == CVMsystemClass(java_lang_Boolean)) {
	CVMD_fieldReadInt(obj, CVMoffsetOfjava_lang_Boolean_value, v->z);
	*toType = CVM_T_BOOLEAN;
    } else if (objectClassBlock == CVMsystemClass(java_lang_Character)) {
	CVMD_fieldReadInt(obj, CVMoffsetOfjava_lang_Character_value, v->c);
	*toType = CVM_T_CHAR;
    } else if (objectClassBlock == CVMsystemClass(java_lang_Float)) {
	CVMD_fieldReadFloat(obj, CVMoffsetOfjava_lang_Float_value, v->f);
	*toType = CVM_T_FLOAT;
    } else if (objectClassBlock == CVMsystemClass(java_lang_Double)) {
	CVMD_fieldReadDouble(obj, CVMoffsetOfjava_lang_Double_value, v->d);
	*toType = CVM_T_DOUBLE;
    } else if (objectClassBlock == CVMsystemClass(java_lang_Byte)) {
	CVMD_fieldReadInt(obj, CVMoffsetOfjava_lang_Byte_value, v->b);
	*toType = CVM_T_BYTE;
    } else if (objectClassBlock == CVMsystemClass(java_lang_Short)) {
	CVMD_fieldReadInt(obj, CVMoffsetOfjava_lang_Short_value, v->s);
	*toType = CVM_T_SHORT;
    } else if (objectClassBlock == CVMsystemClass(java_lang_Integer)) {
	CVMD_fieldReadInt(obj, CVMoffsetOfjava_lang_Integer_value, v->i);
	*toType = CVM_T_INT;
    } else if (objectClassBlock == CVMsystemClass(java_lang_Long)) {
	CVMD_fieldReadLong(obj, CVMoffsetOfjava_lang_Long_value, v->j);
	*toType = CVM_T_LONG;
    } else {
	if (exceptionCb == NULL) {
	    CVMthrowIllegalArgumentException(ee, "unwrapping Java value");
	} else {
	    CVMsignalError(ee, exceptionCb, "unwrapping Java value");
	}
	return CVM_FALSE;
    }
    return CVM_TRUE;

    /* %comment rt009 */

}

/* JDK compatibility */

int
jio_vsnprintf(char *str, size_t count, const char *fmt, va_list args)
{
    return CVMformatStringVaList(str, count, fmt, args);
}

int
jio_snprintf(char *str, size_t count, const char *fmt, ...)
{
    va_list argList;
    size_t n;

    va_start(argList, fmt);
    n = CVMformatStringVaList(str, count, fmt, argList);
    va_end(argList);
    return n;
}

int
jio_fprintf(FILE *fp, const char *fmt, ...)
{
    va_list argList;
    char *buf = NULL; 
    size_t n;
    CVMExecEnv *ee = CVMgetEE();

    CVMCstackGetBuffer(ee, buf); /* Lock EE-buffer */
    if (buf == NULL) {
	printf("<io information unavailable due to low memory>\n");
        return 0; 
    }

    va_start(argList, fmt);
    n = CVMformatStringVaList(buf, CVM_CSTACK_BUF_SIZE, fmt, argList);
    va_end(argList);
    /* We only output to stdout for now */
    printf("%s", buf);
    CVMCstackReleaseBuffer(ee); /* Unlock EE-buffer */
    return n;
}

/*
 * Assign value 'val' to property 'key' in properties object 'props'.
 * Use method 'putID' to put the property in.
 *
 * return CVM_TRUE on success, CVM_FALSE on any failure.
 */
CVMBool
CVMputProp(JNIEnv* env, jmethodID putID, 
	   jobject props, const char* key, const char* val)
{
    if (val != NULL) {
        jstring jkey;
        jstring jval;
	jobject r;
        jkey = (*env)->NewStringUTF(env, key);
	if ((*env)->ExceptionOccurred(env)) return CVM_FALSE;
        jval = (*env)->NewStringUTF(env, val);
	if ((*env)->ExceptionOccurred(env)) return CVM_FALSE;
	r = (*env)->CallObjectMethod(env, props, putID, jkey, jval);
	if ((*env)->ExceptionOccurred(env)) return CVM_FALSE;
        (*env)->DeleteLocalRef(env, jkey);
        (*env)->DeleteLocalRef(env, jval);
        (*env)->DeleteLocalRef(env, r);
    }
    return CVM_TRUE;
}

/*
 * Assign value 'val' to property 'key' in properties object 'props'.
 * Use method 'putID' to put the property in.
 *
 * Do the conversion of 'key' and 'value' using the platform specific
 * C string conversion.
 *
 * return CVM_TRUE on success, CVM_FALSE on any failure.
 */
CVMBool
CVMputPropForPlatformCString(JNIEnv* env, jmethodID putID, 
			     jobject props, const char* key, const char* val)
{
    if (val != NULL) {
        jstring jkey;
        jstring jval;
	jobject r;
        jkey = JNU_NewStringPlatform(env, key);
	if ((*env)->ExceptionOccurred(env)) return CVM_FALSE;
        jval = JNU_NewStringPlatform(env, val);
	if ((*env)->ExceptionOccurred(env)) return CVM_FALSE;
	r = (*env)->CallObjectMethod(env, props, putID, jkey, jval);
	if ((*env)->ExceptionOccurred(env)) return CVM_FALSE;
        (*env)->DeleteLocalRef(env, jkey);
        (*env)->DeleteLocalRef(env, jval);
        (*env)->DeleteLocalRef(env, r);
    }
    return CVM_TRUE;
}

/*
 * Parse a size from options such as -Xms and -Xss.
 */
CVMInt32 
CVMatoml(const char *s)
{
    char* tmp = (char*) s;

    CVMUint32 n = strtol(tmp, &tmp, 10);
    switch (*tmp) {
    case 'M': case 'm':
	n *= 1024 * 1024;
	break;
    case 'K': case 'k':
	n *= 1024;
	break;
    case '\0':
	break;
    default:
	return -1;
    }
    return n;
}
