/*
 * @(#)ClassDependenceAnalyzer.java	1.2 00/05/19
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

package dependenceAnalyzer;
import java.io.FilenameFilter;
import util.ClassFileFinder;
import util.Set;
import java.util.Enumeration;
import java.util.Dictionary;
import java.util.Hashtable;
import components.*;
import ClassFile;
import java.io.InputStream;

/*
 * ClassDependenceAnalyzer is a concrete subclass of DependenceAnalyzer
 * that operates class-by-class. It is not used by JavaFilter or
 * by JavaCodeCompact. It is for use by other dependence analysis
 * tools, which might operate on a more course-grained basis.
 *
 * Following this class definition is an example main program,
 * in a comment, that shows a simple driver for this class.
 */

public class ClassDependenceAnalyzer extends DependenceAnalyzer {

    private FilenameFilter filter;
    private ClassFileFinder finder;
    private Dictionary	    nodeDictionary;

    public ClassDependenceAnalyzer( ClassFileFinder find, FilenameFilter filt ){
	super();
	filter = filt;
	finder = find;
	nodeDictionary = new Hashtable();
    }

    /*
     * This analyzer maintains a name-to-node dictionary of
     * its graph. This lets you lookup nodes by name. It will
     * return null if the named entity is not in the graph.
     */
    public DependenceNode nodeByName( Object name ){
	return (DependenceNode)(nodeDictionary.get( name ) );
    }

    /*
     * To conditionally add a node to the graph. It is not
     * an error to add something that's already in the graph.
     * The return value is a DependenceNode. It may be
     * in any state, but if it was not in the graph previous to
     * this call, it will certainly be DependenceNode.UNANALYZED.
     * In other words, this call doesn't cause analysis.
     */
    public DependenceNode addNodeByName( Object name ) {
	Object t;
	if ( (t = nodeDictionary.get( name ) ) != null )
	    return (DependenceNode)t;
	DependenceNode newNode = new ClassDependenceNode( name );
	nodeDictionary.put( name, newNode );
	return newNode;
    }

    public Enumeration allNodes() { return nodeDictionary.elements(); }

    public int analyzeDependences( DependenceNode n ){

	if ( n.state() != DependenceNode.UNANALYZED )
	    return n.state();
	if ( filter.accept( null, (String)(n.name()) ) ){
	    // oops. on our excluded list.
	    // mark as excluded and analyzed.
	    n.flags = ClassDependenceNode.EXCLUDED;
	    n.nodeState = DependenceNode.ANALYZED;
	    return DependenceNode.ANALYZED;
	}
	ClassDependenceNode classNode = (ClassDependenceNode) n;
	String className = (String)(classNode.name());
	ClassInfo ci = readClassFile( className );
	if ( ci == null ){
	    // sorry. cannot find.
	    classNode.nodeState = DependenceNode.ERROR;
	    return DependenceNode.ERROR;
	}
	if ( classNode.nodeDependsOn == null ){
	    classNode.nodeDependsOn = new Set();
	}
	DependenceNode other;
	// we depend on our super, if any.
	if ( ci.superClass != null ){
	    other = this.addNodeByName( ci.superClass.name.string );
	    classNode.superClass = (ClassDependenceNode)other;
	}

	// we might depend on our interfaces.
	// what do you think?

	//
	// find all the constants mentioned in the constant
	// pool. Call them dependences.
	//
	ConstantObject ctable[] = ci.constants;
	if ( ctable != null ){
	    int nconst = ctable.length;
	    for ( int i = 0; i < nconst ; i++ ){
		ConstantObject co = ctable[i];
		if ( co instanceof ClassConstant ){
		    other = this.addNodeByName( ((ClassConstant)co).name.string );
		    newDependence( n, other );
		}
	    }
	}
	n.nodeState = DependenceNode.ANALYZED;
	return DependenceNode.ANALYZED;
    }

    private ClassInfo readClassFile( String className )
    {
	InputStream classFileStream = finder.findClassFile( className );
	if ( classFileStream == null ){
	    return null;
	}
	ClassFile cf = new ClassFile( className, classFileStream, false );
	if ( ! cf.readClassFile( null ) ){
	    return null;
	}
	return cf.clas;
    }

    private void newDependence( DependenceNode here, DependenceNode there )
    {
	if ( here == there ) return; // don't depend on self.
	DependenceArc a = new DependenceArc( here, there, 0 );
	here.nodeDependsOn.addElement( a );
	there.nodeDependedOn.addElement( a );
    }

}

/*
 * Following is a simple driver program.
 * Here is an example of its use:
 *
 * java ClassDependenceDemo -classpath <your classes> -path java/lang/Object java/io/OptionalDataException
 *
 * And the example output (on Personal Java 1.0), showing two
 * dependence chains of equal length from java.lang.Object to
 * java.io.OptionalDataException:
 *
 * java/lang/Object => java/lang/Class => java/lang/ClassLoader => java/net/URL => java/io/ObjectInputStream => java/io/OptionalDataException
 * java/lang/Object => java/lang/Class => java/lang/ClassLoader => java/util/Hashtable => java/io/ObjectInputStream => java/io/OptionalDataException
 *
 */
// example of  class dependence analysis
// import util.*;
// import java.util.Enumeration;
// import dependenceAnalyzer.*;
// 
// class ClassDependenceDemo {
//     ClassFileFinder	    finder = new ClassFileFinder();
//     ClassnameFilter	    filter = new DummyFilter();
//     ClassDependenceAnalyzer cda = new ClassDependenceAnalyzer( finder, filter );
// 
//     private void addToSearchPath( String pathname ){
// 	finder.addToSearchPath( pathname );
//     }
// 
//     private void showOne( String classname ){
// 	DependenceNode node = cda.addNodeByName( classname );
// 	if ( node == null ){
// 	    System.out.println("Cannot find class "+classname);
// 	    return;
// 	}
// 	cda.analyzeDependences( node );
// 	printDependences( node );
//     }
// 
//     private void showAll(){
// 	cda.analyzeAllDependences();
// 	Enumeration e = cda.allNodes();
// 	int nClasses = 0;
// 	while ( e.hasMoreElements()){
// 	    DependenceNode node = (DependenceNode)(e.nextElement());
// 	    printDependences( node );
// 	    nClasses++;
// 	}
// 	System.out.println("TOTAL OF "+nClasses+" CLASSES");
//     }
// 
//     private void showPath( String srcclassname, String destclassname ){
// 	ClassDependenceNode src = (ClassDependenceNode)cda.addNodeByName( srcclassname );
// 	ClassDependenceNode dest = (ClassDependenceNode)cda.addNodeByName( destclassname );
// 	Set tipset = new Set();
// 	PathComponent r = new PathComponent( null, src );
// 	tipset.addElement( r );
// 	for ( int i = 0; i < 50; i++ ){
// 	    Set newtips = new Set();
// 	    Enumeration newGrowth = tipset.elements();
// 	    while ( newGrowth.hasMoreElements() ){
// 		PathComponent t = (PathComponent)(newGrowth.nextElement() );
// 		if ( t.link.state() == DependenceNode.UNANALYZED )
// 		    cda.analyzeDependences( t.link );
// 		t.grow( newtips );
// 	    }
// 	    if ( newtips.size() == 0 ){
// 		// exhaused the space without finding it.
// 		System.out.println("No path from "+srcclassname+" to "+destclassname );
// 		return;
// 	    }
// 	    // now see if we got to our destination.
// 	    newGrowth = newtips.elements();
// 	    boolean foundDest = false;
// 	    while ( newGrowth.hasMoreElements() ){
// 		PathComponent t = (PathComponent)(newGrowth.nextElement() );
// 		if ( t.link == dest ){
// 		    t.print( System.out );
// 		    System.out.println();
// 		    foundDest = true;
// 		}
// 		t.link.flags |= PathComponent.INSET;
// 	    }
// 	    if ( foundDest ){
// 		return;
// 	    }
// 	    // round and round.
// 	    tipset = newtips;
// 	}
// 	System.out.println("Path from "+srcclassname+" to "+destclassname+" is too long" );
//     }
// 
//     private void process( String args[] ){
// 	for ( int i = 0; i < args.length; i++ ){
// 	    String a = args[i];
// 	    if ( a.equals("-classpath") ){
// 		addToSearchPath( args[++i] );
// 		continue;
// 	    }
// 	    if ( a.equals("-path") ){
// 		showPath( args[++i], args[++i] );
// 		continue;
// 	    }
// 	    if ( a.equals("-show") ){
// 		showOne( args[++i] );
// 		continue;
// 	    }
// 	    if ( a.equals("-showAll") ){
// 		showAll();
// 		continue;
// 	    }
// 	    // else must be a class name!
// 	    cda.addNodeByName( a );
// 	}
//     }
// 
//     public static void main( String args[] ){
// 	new ClassDependenceDemo().process( args );
//     }
// 
//     static void printDependences( DependenceNode node ){
// 	if ( node.state() != DependenceNode.ANALYZED ){
// 	    System.out.println( node.name()+" unanalysed");
// 	    return;
// 	}
// 	System.out.println( node.name()+" depends on:");
// 	Enumeration e = node.dependsOn();
// 	while ( e.hasMoreElements() ){
// 	    DependenceArc a = (DependenceArc)(e.nextElement());
// 	    System.out.println("	"+a.to().name() );
// 	}
//     }
// 
// }
// 
// class DummyFilter extends util.ClassnameFilter{
//     public boolean accept( java.io.File dir, String className ){
// 	return false;
//     }
// }
// 
// class PathComponent {
//     PathComponent	root;
//     ClassDependenceNode link;
// 
//     public static int INSET = 8;
// 
//     PathComponent( PathComponent r, ClassDependenceNode l ){
// 	root = r;
// 	link = l;
//     }
// 
//     void print( java.io.PrintStream o ){
// 	if ( root != null ){
// 	    root.print( o );
// 	    o.print(" => ");
// 	}
// 	o.print( link.name() );
//     }
// 
//     void grow( Set tipset ){
// 	Enumeration t = link.dependsOn();
// 	while ( t.hasMoreElements() ){
// 	    DependenceArc arc = (DependenceArc)(t.nextElement());
// 	    ClassDependenceNode to = (ClassDependenceNode) arc.to();
// 	    if ( (to.flags&INSET) != 0 )
// 		continue;
// 	    tipset.addElement( new PathComponent( this, to ) );
// 	}
//     }
// }
