/*
 * @(#)linker_md.c	1.22 00/11/11
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#ifdef CVM_DYNAMIC_LINKING

/*
 * On VxWorks these routines require three libraries to be built into
 * your project if they are not already: loadLib, moduleLib, and
 * symLib.
 */

#include "javavm/include/porting/linker.h"
#include <ioLib.h>
#include <loadLib.h>
#include <moduleLib.h>
#include <symLib.h>
#include <sysSymTbl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* VxWorks doesn't have local (per-module) symbol tables. Therefore we
   have to search the entire system symbol table every
   time we want to do a symbol lookup, and ensure that the "group" of
   the returned symbol is equal to the module's group ID. This has
   been filed as an RFE with TSR# 143425. */

/*
 * The following library handle is used for the VxWorks CVMdynlink* calls.
 *
 * We differentiate between 'global' symbol lookup (built-in symbols), vs.
 * symbols looked up from shared libraries.
 */

typedef struct VxWorksLibraryHandle {
    MODULE_ID id;
    int isGlobal;
} VxWorksLibraryHandle;

void *
CVMdynlinkOpen(const void *absolutePathName)
{
    VxWorksLibraryHandle* handle;

    handle = malloc(sizeof(VxWorksLibraryHandle));
    if (handle == NULL) {
	return NULL; /* Out of memory */
    }
    if (absolutePathName == NULL) {
	handle->isGlobal = 1;
    } else {
	MODULE_ID mid;
	int fd;
	/*
	 * Not a global symbol. Load module.
	 */
	handle->isGlobal = 0;
	fd = open((char *) absolutePathName, O_RDONLY, 0);
	if (fd == ERROR) {
	    free(handle);
	    return NULL;
	}
	mid = loadModule(fd, LOAD_GLOBAL_SYMBOLS);
	if (mid == NULL) {
	    free(handle);
	    handle = NULL;
	    perror("CVMdynlinkOpen/loadModule");
	} else {
	    handle->id = mid;
	}
	close(fd);
    }
    return (void*)handle;
}

struct FindWellKnownSymbolCallbackInfo {
    char *name;
    UINT16 group;
};

static BOOL
findWellKnownSymbolCB(char *name,
		      int val,
		      SYM_TYPE type,
		      int arg,
		      UINT16 group)
{
    struct FindWellKnownSymbolCallbackInfo *info =
	(struct FindWellKnownSymbolCallbackInfo *) arg;

    if (!strcmp(name, info->name)) {
	info->group = group;
	return FALSE;
    }
    return TRUE;
}

/* Returns CVM_TRUE upon success, CVM_FALSE upon error */
static CVMBool
getWellKnownSymbolGroupId(UINT16* group)
{
    /* This finds the group ID associated with a well-known symbol
       ("CVMgcUnsafeExecuteJavaMethod") and returns it. Note that this
       is static and therefore persistent across Java VMs. This should
       not be a problem because they will all be using the same
       program code and symbol table. */
    static CVMBool gotId = CVM_FALSE;
    static UINT16 id;

    if (gotId == CVM_FALSE) {
	struct FindWellKnownSymbolCallbackInfo info;
	SYMBOL* sym;
#if defined(i386) || defined(__i386__) || defined(__i386)
	/*
	 * %comment: mam051
	 */
	info.name = "_CVMgcUnsafeExecuteJavaMethod";
#else
	info.name = "CVMgcUnsafeExecuteJavaMethod";
#endif
	
	sym = symEach(sysSymTbl, (FUNCPTR) &findWellKnownSymbolCB,
		      (int) &info);
	/* If sym is NULL, failed to find symbol -- should not happen. */
	if (sym == NULL) {
	    return CVM_FALSE;
	}

	id = info.group;
	gotId = CVM_TRUE;
    }

    *group = id;
    return gotId;
}

struct CallbackInfo {
    char *name;
    UINT16 group;
    void *val;
};

static BOOL
findSymbolCB(char *name,
	     int val,
	     SYM_TYPE type,
	     int arg,
	     UINT16 group)
{
    struct CallbackInfo *info = (struct CallbackInfo *) arg;

    if ((!strcmp(name, info->name)) &&
	(info->group == group)) {
	/* Found the symbol we're looking for. Returning FALSE from
	   this callback will cause symEach() to return the current
	   symbol to its caller. */
	info->val = (void *) val;
	return FALSE;
    }
    return TRUE;
}

/*
 * If dsoHandle is NULL, or if the handle indicates that we should do a
 * global, module-independent lookup, we do it. Otherwise, we have to match
 * on module and symbol name.
 */
void *
CVMdynlinkSym(void *dsoHandle, const void *nameArg)
{
    VxWorksLibraryHandle* handle = (VxWorksLibraryHandle*)dsoHandle;
    const char* name = (const char*)nameArg;
    struct CallbackInfo info;
    SYMBOL* sym;
    CVMBool isGlobal;
    void *value;

#if defined(i386) || defined(__i386__) || defined(__i386)
#define CVM_PREPEND_AOUT_UNDERSCORE
#endif
#ifdef CVM_PREPEND_AOUT_UNDERSCORE
    /*
     * %comment: mam051
     * It might be nice if we could modify nameArg in place.
     */
    char *buf = malloc(strlen(nameArg) + 1 + 1);
    if (buf == NULL) {
	return NULL;
    }
    strcpy(buf, "_");
    strcat(buf, nameArg);
    name = buf;
#endif

    if (handle == NULL || handle->isGlobal) {
	isGlobal = CVM_TRUE;
    } else {
	isGlobal = CVM_FALSE;
    }
    
    if (isGlobal) {
	/* Look up a well-known CVM symbol and match its group
	   ID. This protects us whether cvm.o is dynamically loaded,
	   or statically linked into the kernel. */
	if (getWellKnownSymbolGroupId(&info.group) == CVM_FALSE) {
	    value = NULL;
	    goto done;
	}
    } else {
	MODULE_ID mid;
	MODULE_INFO mInfo;

	mid = handle->id;
	if (moduleInfoGet(mid, &mInfo) == ERROR) {
	    value = NULL;
	    goto done;
	}
	info.group = (UINT16) mInfo.group;
    }
    info.name = (char*)name;
    sym = symEach(sysSymTbl, (FUNCPTR) &findSymbolCB, (int) &info);
    if (sym == NULL) {
	value = NULL;
    } else {
	value = info.val;
    }

done:
#ifdef CVM_PREPEND_AOUT_UNDERSCORE
    free(buf);
#endif
    return value;
}

void
CVMdynlinkClose(void *dsoHandle)
{
    VxWorksLibraryHandle* handle = (VxWorksLibraryHandle *)dsoHandle;
    unld(handle->id, 0);
    free(handle);
}

/* Basically copied from the JDK */
CVMBool
CVMdynlinkBuildLibName(void *holder, int holderlen,
		       void *pname, void *fname)
{
    const int pnamelen = pname ? strlen(pname) : 0;
    char *suffix;

#ifdef CVM_DEBUG
    suffix = "_g";
#else
    suffix = "";
#endif 

    /* Truncate on buffer overflow and return an error. */
    if (pnamelen + strlen(fname) + 10 > (size_t) holderlen) {
        *((char *) holder) = '\0';
        return CVM_FALSE;
    }

    if (pnamelen == 0) {
        sprintf((char *) holder, "lib%s%s.o",
		    (char *) fname, (char *) suffix);
    } else {
        sprintf((char *) holder, "%s/lib%s%s.o",
		    (char *) pname, (char *) fname, (char *) suffix);
    }
    return CVM_TRUE;
}

#endif /* #defined CVM_DYNAMIC_LINKING */
