/*
 * @(#)sync.h	1.9 00/05/19
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#ifndef _INCLUDED_PORTING_SYNC_H
#define _INCLUDED_PORTING_SYNC_H

#include "javavm/include/porting/defs.h"
#include "javavm/include/porting/vm-defs.h"

/*
 * The platform defines the CVM_MICROLOCK_TYPE variable to be
 * one of the following.
 */

#define CVM_MICROLOCK_DEFAULT		-1
#define CVM_MICROLOCK_SCHEDLOCK		1
/*
 * The platform defines the CVM_FASTLOCK_TYPE variable to be
 * one of the following.
 */

#define CVM_FASTLOCK_NONE		-1
#define CVM_FASTLOCK_ATOMICOPS		1
#define CVM_FASTLOCK_MICROLOCK		3

/*
 * Platform must provide definitions for the following opaque structs:
 *
 * struct CVMMutex
 * struct CVMCondVar
 *
 */

/*
 * Initialize the given CVMMutex, allocating additional resources if necessary.
 */
extern CVMBool CVMmutexInit   (CVMMutex* m);

/*
 * Destroy the given CVMMutex, releasing any resources that were
 * previously allocated by CVMmutexInit.
 */
extern void    CVMmutexDestroy(CVMMutex* m);

/*
 * Attempt to acquire a lock for the given mutex.  This call must return
 * immediately and never block. CVM_TRUE is returned if the lock was
 * acquired, otherwise CVM_FALSE is returned.
 */
extern CVMBool CVMmutexTryLock(CVMMutex* m);

/*
 * Acquire a lock for the given mutex.  This call may block.
 * A thread will never attempt to acquire the same mutex
 * more than once, so a counter to keep track of recursive
 * entry is not required.
 */
extern void    CVMmutexLock   (CVMMutex* m);

/*
 * Release a lock previously acquired by CVMmutexLock.
 */
extern void    CVMmutexUnlock (CVMMutex* m);

#ifdef CVM_ADV_MUTEX_SET_OWNER
/*
 * Lock an unlocked mutex on behalf of another thread.
 * The result will be identical to the thread having
 * called CVMmutexLock itself.
 */
extern void    CVMmutexSetOwner(CVMThreadID *self, CVMMutex* m,
				CVMThreadID *ti);
#endif

/*
 * Initialize the given CVMCondVar, allocating additional resources if
 * necessary.  The CVMMutex that will be used by any future calls
 * to CVMcondvarWait is also provided, for those platforms that
 * require the mutex for proper condition variable initialization.
 */
extern CVMBool CVMcondvarInit   (CVMCondVar *c, CVMMutex* m);
/*
 * Destroy the given CVMCondVar, releasing any resources that were
 * previously allocated by CVMcondvarInit.
 */
extern void    CVMcondvarDestroy(CVMCondVar *c);

/*
 * Perform a timed wait on a condition variable.  The mutex is
 * unlocked during while the thread blocks and locked again
 * when it wakes up.  CVM_FALSE is returned if the operation
 * is interrupted by CVMthreadInterruptWait.  This function
 * is used to implement Object.wait, therefore the atomicity of 
 * unlocking the lock and checking for notification or an
 * interrupt must follow the rules described by Object.wait.
 * A timeout of 0ms is treated as infinity, or "wait forever".
 */
extern CVMBool CVMcondvarWait     (CVMCondVar* c, CVMMutex* m,
				   CVMJavaLong millis);
/*
 * Wakeup a thread waiting on the given CVMCondVar, as defined
 * by the semantics of Object.notify.
 */
extern void    CVMcondvarNotify   (CVMCondVar* c);
/*
 * Wakeup all threads waiting on the given CVMCondVar, as defined
 * by the semantics of Object.notifyAll.
 */
extern void    CVMcondvarNotifyAll(CVMCondVar* c);

/*
 * Take advantage of other advanced features of the platform
 */

#ifdef CVM_ADV_SCHEDLOCK
/*
 * Lock the scheduler.  The current thread will remain running
 * and no other threads in this VM instance can become runnable
 * until the schedular lock is released.
 */
extern void CVMschedLock(void);
/*
 * Release the scheduler lock acquired by CVMschedLock.  Normal
 * thread scheduling will resume.
 */
extern void CVMschedUnlock(void);
#endif

#ifdef CVM_ADV_ATOMIC_CMPANDSWAP
/*
 * Conditionally set a word in memory, returning the previous
 * value.  The new value "new" is written to address "addr" only
 * if the current value matches the value given by "old".
 */
extern CVMUint32 CVMatomicCompareAndSwap(volatile CVMUint32 *addr,
	CVMUint32 new, CVMUint32 old);
#endif

#ifdef CVM_ADV_ATOMIC_SWAP
/*
 * Atomically set a word in memory, returning the previous
 * value.
 */
extern CVMUint32 CVMatomicSwap(volatile CVMUint32 *addr, CVMUint32 new);
#endif

#ifdef CVM_ADV_ATOMIC_INCDEC
/*
 * Atomically increment a word in memory, returning the new
 * value.
 */
extern CVMUint32 CVMatomicIncrement(CVMUint32 *addr);
/*
 * Atomically decrement a word in memory, returning the new
 * value.
 */
extern CVMUint32 CVMatomicDecrement(CVMUint32 *addr);
#endif

#include CVM_HDR_SYNC_H

/*
 * Sanity checks
 */
#ifndef CVM_MICROLOCK_TYPE
#define CVM_MICROLOCK_TYPE CVM_MICROLOCK_DEFAULT
#elif CVM_MICROLOCK_TYPE == CVM_MICROLOCK_SCHEDLOCK
#ifndef CVM_ADV_SCHEDLOCK
#error CVM_MICROLOCK_SCHEDLOCK but not CVM_ADV_SCHEDLOCK
#endif
#elif CVM_MICROLOCK_TYPE == CVM_MICROLOCK_DEFAULT
#else
#error Unknown CVM_MICROLOCK_TYPE
#endif

#ifndef CVM_FASTLOCK_TYPE
#define CVM_FASTLOCK_TYPE CVM_FASTLOCK_NONE
#elif CVM_FASTLOCK_TYPE == CVM_FASTLOCK_ATOMICOPS
#ifndef CVM_ADV_ATOMIC_CMPANDSWAP
#error CVM_FASTLOCK_ATOMICOPS but not CVM_ADV_ATOMIC_CMPANDSWAP
#endif
#ifndef CVM_ADV_ATOMIC_SWAP
#error CVM_FASTLOCK_ATOMICOPS but not CVM_ADV_ATOMIC_SWAP
#endif
#elif CVM_FASTLOCK_TYPE == CVM_FASTLOCK_MICROLOCK
#elif CVM_FASTLOCK_TYPE == CVM_FASTLOCK_NONE
#else
#error Unknown CVM_FASTLOCK_TYPE
#endif

#if CVM_FASTLOCK_TYPE != CVM_FASTLOCK_NONE && !defined(CVM_ADV_MUTEX_SET_OWNER)
#error Must define CVM_ADV_MUTEX_SET_OWNER for fast locking
#endif

#endif /* _INCLUDED_PORTING_SYNC_H */
