/*
 * @(#)InnerClassAttribute.java	1.5 00/11/09
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

package components;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Vector;

/*
 * A class to represent the innerclass attribute of a class. The innerclass
 * attribute is new in 1.2 and is documented in the 1.2 JVM Spec.
 */
public
class InnerClassAttribute extends Attribute {
//__________ CLASS METHODS
    public static Attribute
    readAttribute(DataInput in, int len,
                  UnicodeConstant name,
                  ConstantObject globals[]) throws IOException {
	return finishReadAttribute(in, len, name, globals);
    }

    public static Attribute
    finishReadAttribute(DataInput in, int len,
                        UnicodeConstant name,
                        ConstantObject globals[]) throws IOException {
	int num        = in.readUnsignedShort();
	Vector entries = new Vector(num);
	int i, next, anonNum;
	ClassConstant innerInfo;
	ClassConstant outerInfo;
	UnicodeConstant innerName;

	for (i=0, anonNum=1; i < num; i++) {
	    innerInfo = (ClassConstant)globals[in.readUnsignedShort()];
	    outerInfo = (ClassConstant)globals[in.readUnsignedShort()];

	    next = in.readUnsignedShort();
	    if (next == 0) { // A zero value means this is an anonymous inner class.
		// Note: We are guessing at the name of the anonymous class.
		// But the guess will only be the correct name if the ordering
		// of the inner class attributes is the same as the ordering
		// of the anonymous inner classes as they appear in the source
		// file. And that is totally dependent on how the javac compiler
		// wants to do it. There is nothing in the JVM spec that says it
		// has to be that way. This won't hurt anything since we are just
		// replacing an unused name field with a guess. The JVM doesn't
		// care what is in this field.
		innerName = new UnicodeConstant((new Integer(anonNum++)).toString());
	    } else {
		innerName = (UnicodeConstant)globals[next];
	    }
	    
	    entries.addElement(new InnerClassEntry(innerInfo,
	                                           outerInfo,
	                                           innerName,
	                                           in.readUnsignedShort()));
//	    entries.add(new InnerClassEntry(innerInfo,
//	                                    outerInfo,
//	                                    innerName,
//	                                    in.readUnsignedShort()));
	}
		
	return new InnerClassAttribute(name, len, entries);
    }

//__________ INSTANCE VARIABLES
    private Vector myInnerClassesVector;
    private InnerClassEntry myInnerClassesArray[];

//__________ CONSTRUCTORS
    public
    InnerClassAttribute(UnicodeConstant name,
                        int len,
                        Vector entries) {
	super(name, len);
	myInnerClassesVector = entries;

	// Save an array copy because some methods use an index value
	// to find individual entries. But don't forget to update this
	// array if the vector changes, like in the removeEntry method.
	myInnerClassesArray = new InnerClassEntry[myInnerClassesVector.size()];
//	myInnerClassesVector.toArray(myInnerClassesArray);
	myInnerClassesVector.copyInto(myInnerClassesArray);
    }

//__________ INSTANCE METHODS
    protected int
    writeData(DataOutput o) throws IOException {
	int numberOfInnerClasses = myInnerClassesVector.size();
	o.writeShort(numberOfInnerClasses);
	
	Enumeration e = myInnerClassesVector.elements();
	while (e.hasMoreElements()) {
	    ((InnerClassEntry)e.nextElement()).write(o);
	}
		
	//return total number of bytes written, which is equal to
	//2 (the number of bytes of the numberOfInnerClasses local
	//variable) + the size of the InnerClassEntry * the number
	//of inner classes.
	return 2 + InnerClassEntry.SIZE * numberOfInnerClasses;
    }

    /*
     * I am unsure exactly how this should operate. If innerInfo is null,
     * should I still check the values of outerInfo and innerName? I would
     * think that it wouldn't matter because if innerInfo is null, the other
     * two should also be null.
     */
    public void
    externalize(ConstantPool cp) {
	super.externalize(cp);
	/*
	 * This code doesn't work for CVM, but may be needed if we ever
	 * support a shared constant pool.
	 *
	InnerClassEntry ice;
	Enumeration e = myInnerClassesVector.elements();
	while (e.hasMoreElements()) {
	    ice = (InnerClassEntry)e.nextElement();

	    if (ice.innerInfo != null) {
		ice.innerInfo = (ClassConstant)cp.add(ice.innerInfo);
	    }
	    if (ice.outerInfo != null) {
		ice.outerInfo = (ClassConstant)cp.add(ice.outerInfo);
	    }
	    if (ice.innerName != null) {
		ice.innerName = (UnicodeConstant)cp.add(ice.innerName);
	    }
	}

	// Don't forget to update the array copy now that we've changed
	// the vector.
//	myInnerClassesVector.toArray(myInnerClassesArray);
	myInnerClassesVector.copyInto(myInnerClassesArray);
	*/
    }

    /*
     * See comment for externalize() method above.
     */
    public void
    countConstantReferences(boolean isRelocatable) {
	super.countConstantReferences(isRelocatable);

	InnerClassEntry ice;
	Enumeration e = myInnerClassesVector.elements();
	while (e.hasMoreElements()) {
	    ice = (InnerClassEntry)e.nextElement();

	    if (ice.innerInfo != null) {
		ice.innerInfo.incReference();
	    }
	    if (ice.outerInfo != null) {
		ice.outerInfo.incReference();
	    }
	    if (ice.innerName != null) {
		ice.innerName.incReference();
	    }
	}

	// Don't forget to update the array copy now that we've changed
	// the vector. 
//	myInnerClassesVector.toArray(myInnerClassesArray);
	myInnerClassesVector.copyInto(myInnerClassesArray);
    }

    public void
    removeEntry(int i) {
System.out.println("*** Removing inner class "+ myInnerClassesArray[i].getFullName());
//	myInnerClassesVector.remove(myInnerClassesArray[i]);
	myInnerClassesVector.removeElement(myInnerClassesArray[i]);
	
	// Don't forget to update the array copy now that we've changed
	// the vector.
	myInnerClassesArray = new InnerClassEntry[myInnerClassesVector.size()];
//	myInnerClassesVector.toArray(myInnerClassesArray);
	myInnerClassesVector.copyInto(myInnerClassesArray);
    }	
    
    public int
    getInnerClassCount() {
	return myInnerClassesVector.size();
    }
    
    /*
     * The following four methods return information about individual
     * inner class entries contained in this attribute. These methods
     * call corresponding methods in the InnerClassEntry class. They
     * are self-explanatory.
     */

    public int
    getInnerInfoIndex(int i) {	
	return myInnerClassesArray[i].getInnerInfoIndex();
    }

    public int
    getOuterInfoIndex(int i) {
	return myInnerClassesArray[i].getOuterInfoIndex();
    }

    public int
    getInnerNameIndex(int i) {	
	return myInnerClassesArray[i].getInnerNameIndex();
    }
    
    public int
    getAccessFlags(int i) {	
	return myInnerClassesArray[i].getAccessFlags();
    }
    
    /*
     * Like the above methods, these methods return information about
     * an individual inner class entry. The string returned will be
     * the name of the inner class, if it has one. If the inner class
     * is anonymous, it will not have a name and the string "anonymous"
     * will be returned.
     */
    public String
    getName(int i) {
	return myInnerClassesArray[i].getName();
    }
    
    /*
     * Like the getName method, but return the outer class name as well.
     */
    public String
    getFullName(int i) {
	return myInnerClassesArray[i].getFullName();
    }
}
