/*
 * @(#)classinitialize.c	1.37 00/09/21
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#include "javavm/include/interpreter.h"
#include "javavm/include/directmem.h"
#include "javavm/include/indirectmem.h"
#include "javavm/include/preloader.h"
#include "javavm/include/common_exceptions.h"

#include "javavm/export/jni.h"

/* *** RTSJ *** */
// For ExecuteInImmortal macro
#include "javavm/include/mem_registry.h"
/* *** JSTR *** */

static int
CVMprivateClassInit(CVMExecEnv* ee, CVMClassBlock* cb, CVMMethodBlock **p_mb);

/* 
 * Perform the class initialization as described in the VM spec,
 * 2nd Edition (2.17.5). Link the class if necessary.
 *
 * This is the public version that will use JNI to call 
 * Class.runStaticInitializers.
 */
CVMBool
CVMclassInit(CVMExecEnv* ee, CVMClassBlock* cb)
{
    CVMBool retVal;
    //TODO: Delete This
    //if(ee == ee->perThreadStack)
    //{
    //    // Crash deliberately
    //    char *c = NULL;
    //    *c = 'a';
    //}
    CheckAndExecuteInImmortal(ee, {
        retVal = (CVMprivateClassInit(ee, cb, 0) == 0);
        });
    return retVal;
}

/* 
 * Perform the class initialization as described in the VM spec,
 * 2nd Edition (2.17.5). Link the class if necessary.
 *
 * This is the version that interpreter loop will use to intialize
 * classes. It avoids C recursion by not using JNI to invoke
 * Class.runStaticInitializers. Instead it munges the interpreter
 * state so that upon return the interperer will start executiong 
 * in Class.runStaticInitializers.
 */
int
CVMclassInitNoCRecursion(CVMExecEnv* ee, CVMClassBlock* cb,
			 CVMMethodBlock **p_mb)
{
    int retVal;

    //TODO: Delete thsi
    /*if(ee == ee->perThreadStack)
    {
        // Crash deliberately
        char *c = NULL;
        *c = 'a';
    }*/
    CheckAndExecuteInImmortal(ee, {
        retVal = CVMprivateClassInit(ee, cb, p_mb);
    });
    return retVal;
}

/*
 * Perform the class initialization as described in the VM spec,
 * 2nd Edition (2.17.5). Link the class if necessary.
 *
 * This version supports both CVMclassInit and CVMclassInitNoCRecursion.
 * If p_mb == 0, then is uses JNI to invoke Class.runStaticInitializers.
 *
 * If p_mb != 0 then it avoids C recursion by pushing a transition
 * frame that knows how to invoke Class.runStaticInitializers and then
 * munges the interpreter state so that upon return the interpeter 
 * starts executing at the beginning of Class.runStaticInitializers.
 *
 * We could have avoided the transition frame and just setup a JavaFrame
 * for Class.runStaticInitializers, but this would require some
 * extra logic in handle_return, which would slow down all method returns.
 *
 * Please note that in one case Class.runStaticInitializers is actually
 * invoked from here and in other caes the interpreter state is modified
 * so that upon return the interpreter will start executing 
 * Class.runStaticInitializers.
 * 
 * RESULT if p_mb != NULL:
 *   0 - No action taken. Class is already initialized.
 *   1 - Transition frame successfully pushed.
 *  -1 - An error occured. Exception was thrown.
 * 
 * RESULT if p_mb == NULL:
 *   0 - Initialization successful or class is already initialized 
 *  -1 - An error occured. Exception was thrown.
 */
static int
CVMprivateClassInit(CVMExecEnv* ee, CVMClassBlock* cb, CVMMethodBlock **p_mb)
{
    if (CVMcbCheckRuntimeFlag(cb, INITIALIZED) || CVMcbClinitEE(cb) == ee) {
	/* nothing to do */
	return 0;
    }

    CVMtraceClinit(("[InitClass() called for %C]\n", cb));

    CVMassert(CVMD_isgcSafe(ee));

    /* Don't need to initialize or verify primitive or array classes */
    if (CVMcbIs(cb, PRIMITIVE) || CVMisArrayClass(cb)) {
	CVMcbSetRuntimeFlag(cb, ee, INITIALIZED);
	return 0;
    } else {
	/* Get the mb for Class.runStaticInitializers. */
	CVMMethodBlock* mb = CVMglobals.java_lang_Class_runStaticInitializers;

#ifdef CVM_CLASSLOADING
	    /*
	     * Link the class hierarchy.
	     */
	    if (!CVMclassLink(ee, cb)) {
		return -1;
	    }
#endif

	if (p_mb != NULL) {
	    /* 
	     * We need to do the special invocation of runStaticInitializers 
	     * that avoids C recursion. This section of code must run gcUnsafe
	     * because of the stack manipulation it does.
	     */
	    CVMD_gcUnsafeExec(ee, {
		CVMFrame* frame = (CVMFrame *)CVMpushTransitionFrame(ee, mb);
		if (frame == NULL) {
		    return -1;   /* exception already thrown */
		}
		/* Push Class object of the class that we want to run
		 * static initializers on. */
		CVMID_icellAssignDirect(ee, &frame->topOfStack->j.r,
					CVMcbJavaInstance(cb));
		++frame->topOfStack;
		*p_mb = mb;
	    });
	    return 1;
	} else {
	    /* This is a normal invocation of runStaticInitializers that
	     * just uses JNI to enter the interpreter.
	     */
	    JNIEnv*  env = CVMexecEnv2JniEnv(ee);
	    int retVal;

	    /* Call Class.runStaticInitializers. We need to disable remote
	     * exceptions during this call, since we will be executing
	     * java code that we don't want to abort if a remote
	     * exception occurrs.
	     */
	    CVMdisableRemoteExceptions(ee);
	    (*env)->CallVoidMethod(env, CVMcbJavaInstance(cb), mb);
	    if (CVMlocalExceptionOccurred(ee)) {
		retVal = -1;
	    } else {
		retVal = 0;
	    }
	    CVMenableRemoteExceptions(ee);
	    return retVal;
	}
    }
}
