/*
 * @(#)classload.c	1.57 00/09/22
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

/*
 * Contains functions that deal with loading a class.
 */

#ifdef CVM_CLASSLOADING

#include "javavm/include/interpreter.h"
#include "javavm/include/classes.h"
#include "javavm/include/utils.h"
#include "javavm/include/common_exceptions.h"
#include "javavm/include/globals.h"
#include "javavm/include/preloader.h"
#include "javavm/include/indirectmem.h"
#include "javavm/include/directmem.h"
#include "javavm/include/porting/io.h"
#include "javavm/include/porting/path.h"
#include "native/java/util/zip/zip_util.h"

/*
 * Class path element, which is either a directory or zip file.
 */
struct CVMClassPathEntry {
    enum { CVM_CPE_DIR, CVM_CPE_ZIP, CVM_CPE_INVALID } type;
    jzfile *zip;
    char *path;
};


static CVMClassBlock*
CVMclassLoadSystemClass(CVMExecEnv* ee, const char* name);

static CVMClassBlock*
CVMclassLoadClassWithClassLoader(CVMExecEnv* ee, CVMClassLoaderICell* loader, 
				 const char* classname, 
				 CVMClassTypeID classTypeID);

static CVMBool
CVMclassLoadSuperclasses0(CVMExecEnv* ee, CVMClassBlock* cb);

static CVMClassBlock* 
CVMclassLoadFromFile(CVMExecEnv* ee,
		     const char* filename,
		     const char* dirname,
		     const char* classname);

static CVMClassBlock* 
CVMclassLoadFromZipFile(CVMExecEnv* ee,
			char* filename,
			const char* classname,
			jzfile* zipfile);


/*
 * Detect class circularities from CVMclassLoadSuperclasses()
 */

struct CVMSeenClass {
    CVMSeenClass*  next;
    CVMClassBlock* cb;
};

static void
CVMpushSeen(CVMExecEnv* ee, CVMSeenClass* seen, CVMClassBlock* cb)
{
    seen->cb = cb;
    seen->next = ee->seenclasses;
    ee->seenclasses = seen;
}

static void
CVMpopSeen(CVMExecEnv* ee, CVMSeenClass* seen)
{
    CVMassert(seen == ee->seenclasses);
    ee->seenclasses = ee->seenclasses->next;
}

static CVMBool
CVMcheckSeen(CVMExecEnv* ee, CVMClassTypeID typeID,
	     CVMClassLoaderICell* loader)
{
    CVMSeenClass* seen = ee->seenclasses;
    while (seen != NULL) {
	if (CVMcbClassName(seen->cb) == typeID &&
	    CVMcbClassLoader(seen->cb) == loader) {
	    return CVM_TRUE;
	}
	seen = seen->next;
    }
    return CVM_FALSE;
}

/*
 * Load a class' hierarchy and resolve symbolic links to its superclass
 * and interfaces.
 */

CVMBool
CVMclassLoadSuperclasses(CVMExecEnv* ee, CVMClassBlock* cb) {
    CVMSeenClass thisSeenClass[1];
    CVMBool result;
    
    CVMtraceClassLoading(("CL: Loading superclasses of %C\n", cb));

    /* Push this class onto the list of classes in the hierarchy. */
    CVMpushSeen(ee, thisSeenClass, cb);

    /* Lock the class and load its hierarchy */
    result = CVMgcSafeObjectLock(ee, CVMcbJavaInstance(cb));
    if (!result) {
	CVMthrowOutOfMemoryError(ee, NULL);
    } else {
	if (!CVMcbCheckRuntimeFlag(cb, SUPERCLASS_LOADED)) {
	    result = CVMclassLoadSuperclasses0(ee, cb);
	}
	if (!CVMgcSafeObjectUnlock(ee, CVMcbJavaInstance(cb))) {
	    CVMassert(CVM_FALSE); /* should never happen */
	}
    }

    /* Pop the class from the list of seen classes. */
    CVMpopSeen(ee, thisSeenClass);

    CVMtraceClassLoading(("CL: %s loading superclasses of %C\n", 
			  result ? "Done" : "Failed", cb));

    if (!result) {
	CVMcbSetRuntimeFlag(cb, ee, ERROR);
    }
    return result;
}

static CVMBool
CVMclassLoadSuperclasses0(CVMExecEnv* ee, CVMClassBlock* cb)
{
    CVMClassTypeID superTypeID = CVMcbSuperclassTypeID(cb);
    CVMClassBlock* superCb;
    int i;

    /* 
     * Make sure we haven't already seen the superclass while loading
     * superclasses.
     */
    if (CVMcheckSeen(ee, superTypeID, CVMcbClassLoader(cb))) {
        CVMthrowClassCircularityError(ee, "%C", cb);
	return CVM_FALSE;
    }
   
    /* Make sure the superclass is loaded */
    /* %comment cjp014 */
    superCb = CVMclassLookupByTypeFromClass(ee, superTypeID, CVM_FALSE, cb);

    if (superCb != NULL) {
	CVMassert(CVMcbCheckRuntimeFlag(superCb, SUPERCLASS_LOADED));
	/* Don't allow a class to have a superclass it can't access */
	if (!CVMverifyClassAccess(ee, cb, superCb, CVM_FALSE)) {
	    CVMthrowIllegalAccessError(
                ee, "cannot access superclass %C from class %C", superCb, cb);
	    superCb = NULL;
	}
    }
    if (superCb == NULL) {
	return CVM_FALSE; /* exception already thrown */
    }

    /* 
     * Link the superclass in. This overwrites the CVMcbSuperclassTypeID() 
     * field, and we don't want to do this if there is an error, but we have
     * to do it now for the following code to work. Because of this, we
     * already saved the typeid and will restore it if there is an error.
     */
    CVMcbSuperclass(cb) = superCb;

    /* 
     * Make sure the interfaces are loaded and link them in. We only
     * look at the interfaces the class declares it implements.
     * Inherited interfaces will automatically be loaded recursively and
     * the inteface table slots will be filled in later by CVMclassLink. 
     * Note that CVMcbInterfaceCount() isn't even setup at this point.
     */
    for (i = 0; i < CVMcbImplementsCount(cb); i++) {
	CVMClassTypeID interfaceTypeID = CVMcbInterfaceTypeID(cb, i);
	CVMClassBlock* interfaceCb; 

	/* 
	 * Make sure we haven't already seen the interface while loading
	 * superclasses.
	 */
	if (CVMcheckSeen(ee, interfaceTypeID, CVMcbClassLoader(cb))) {
	    CVMthrowClassCircularityError(ee, "%C", cb);
	    goto error;
	}

	/* %comment cjp014 */
	interfaceCb = 
	    CVMclassLookupByTypeFromClass(ee, interfaceTypeID, CVM_FALSE, cb);
	if (interfaceCb == NULL) {
	    goto error; /* exception already thrown */
	}
	
	CVMassert(CVMcbCheckRuntimeFlag(interfaceCb, SUPERCLASS_LOADED));
	/* Make sure the class is really an interface. */
	if (!CVMcbIs(interfaceCb, INTERFACE)) {
	    CVMthrowIncompatibleClassChangeError(
                ee, "class %C is not an interface", interfaceCb);
	    goto error; /* exception already thrown */
	}
	
	/* Don't allow a class to have an interface it can't access */
	if (!CVMverifyClassAccess(ee, cb, interfaceCb, CVM_FALSE)) {
	    CVMthrowIllegalAccessError(
                ee, "cannot access interface %C from class %C",
		interfaceCb, cb);
	    goto error; /* exception already thrown */
	}
	
	/* If the following invariant isn't true, then someone has
	 * mucked with CVMInterfaceTable in a potentially bad way.
	 */
	CVMassert((CVMUint8*)&CVMcbInterfaceTypeID(cb, i) == 
		  (CVMUint8*)&CVMcbInterfacecb(cb, i) 
		  + sizeof(CVMClassBlock*));
	CVMcbInterfacecb(cb, i) = interfaceCb;
    }

    /*
     * java.lang.ref.Reference is already ROMized and has its
     * REFERENCE bit set. Same for all its ROMized subclasses. So the
     * bootstrapping is easy.
     */
    if (CVMcbIs(superCb, REFERENCE)) {
	CVMcbSetAccessFlag(cb, REFERENCE);
    }
    CVMcbSetRuntimeFlag(cb, ee, SUPERCLASS_LOADED);
    return CVM_TRUE;

 error:
    CVMcbSuperclassTypeID(cb) = superTypeID;
    return CVM_FALSE;
}

/*
 * Load a class using the specified class loader.
 */

CVMClassBlock*
CVMclassLoadClass(CVMExecEnv* ee, CVMClassLoaderICell* loader, 
		  const char* classname, CVMClassTypeID classTypeID)
{
    JNIEnv* env = CVMexecEnv2JniEnv(ee);
    CVMClassBlock* cb;

    CVMtraceClassLoading(("CL: trying to load class \"%s\" using "
			  "0x%p class loader\n", classname, loader));

    CVMassert(!CVMexceptionOccurred(ee));

    /* We need a JNI local frame so we can make method calls using JNI. */
    if (CVMjniPushLocalFrame(env, 4) != JNI_OK) {
	return NULL; /* excpeption already thrown */
    }

    if (loader == NULL) {
	cb = CVMclassLoadSystemClass(ee, classname);
    } else {
	cb = CVMclassLoadClassWithClassLoader(ee, loader, 
					      classname, classTypeID);
    }

    CVMjniPopLocalFrame(env, NULL);


    {
	CVMBool isSameClassLoader = CVM_TRUE;
	if (cb != NULL) {
	    CVMID_icellSameObjectNullCheck(ee, loader, CVMcbClassLoader(cb), 
					   isSameClassLoader);
	}
	if (!isSameClassLoader) {
	    CVMtraceClassLoading(("CL: class \"%s\" already loaded by "
				  "0x%p class loader.\n",
				  classname, CVMcbClassLoader(cb)));
	} else {
	    CVMtraceClassLoading(("CL: class \"%s\"%s loaded by "
				  "0x%p class loader.\n", classname,
				  cb == NULL ? " not" : "", loader));
	}
    }
    return cb;
}

/*
 * Load a class using a ClassLoader.
 */

static CVMClassBlock*
CVMclassLoadClassWithClassLoader(CVMExecEnv* ee, CVMClassLoaderICell* loader, 
				 const char* classname, 
				 CVMClassTypeID classTypeID)
{
    CVMClassBlock* cb = NULL;
    jclass         result;
    JNIEnv*        env = CVMexecEnv2JniEnv(ee);
    jstring        classString;
    char           dot_name_buf[256];
    char*          dot_name;
    int i;

    CVMassert(!CVMexceptionOccurred(ee));

    /* 
     * Allocate a buffer for the tranlated class name if the stack buffer
     * isn't big enough.
     */
    if (strlen(classname) >= sizeof(dot_name_buf)) {
        dot_name = malloc(strlen(classname) + 1);
	if (dot_name == NULL) {
	    CVMthrowOutOfMemoryError(ee, NULL);
	    return NULL;
	}
    } else {
        dot_name = dot_name_buf;
    }

    /* Copy the class name, translating slashes to dots in the process. */
    for (i = 0; classname[i]; i++) {
        if (classname[i] == '/') {
	    dot_name[i] = '.';
        } else {
	    dot_name[i] = classname[i];
	}
    }
    dot_name[i] = 0;

    /* Convert the class name to a String object. */
    classString = (*env)->NewStringUTF(env, dot_name);
    if (classString == 0) {
        goto end;
    }

    /* Call loader.loadClass() */
    result = 
	(*env)->CallObjectMethod(env, loader, 
				 CVMglobals.java_lang_ClassLoader_loadClass,
				 classString);
    if (result == NULL || CVMID_icellIsNull(result)) {
	cb = NULL;
    } else {
	cb = CVMgcSafeClassRef2ClassBlock(ee, result);
    }

    if (!CVMexceptionOccurred(ee) && cb == NULL) {
	/* Loader returned NULL without raising an exception. */
	CVMthrowClassNotFoundException(ee, "%s", classname);
	goto end;
    }
    
    /* 
     * We do not trust that the ClassLoader.loadClass method actually
     * returns the right class.
     */
    if (cb != NULL && classTypeID != CVMcbClassName(cb)) {
        CVMthrowNoClassDefFoundError(ee, 
				     "Bad class name (expect: %!C, get: %!C)",
				     classTypeID, CVMcbClassName(cb));
	cb = NULL;
    }

 end:
    if (dot_name != dot_name_buf) {
        free(dot_name);
    }
    return cb;
}

/*
 * Load a system class.
 */

static CVMClassBlock*
CVMclassLoadSystemClass(CVMExecEnv* ee, const char* classname)
{
    /* 
     * Iterate over each class path entry, trying to load the class file
     * from a directory or zip file in the class path.
     */
    CVMUint32 idx;
    char * filename = malloc(CVM_PATH_MAXLEN);

    if (filename == NULL) {
        CVMthrowOutOfMemoryError(ee, NULL);
	return NULL;
    }

    for (idx = 0; idx < CVMglobals.numBootClassPathEntries; idx++) {
	CVMClassBlock* cb = NULL;
	CVMClassPathEntry* currEntry = &CVMglobals.bootClassPathEntries[idx];

	if (currEntry->type == CVM_CPE_INVALID) {
	    continue;
	} else if (currEntry->type == CVM_CPE_DIR) {
	    /* Look for class file in specified bootclasspath directory. */
	    const char* dirname = currEntry->path;

	    if (strlen(dirname) + 1 + strlen(classname) + 1 +
		strlen(CVM_PATH_CLASSFILEEXT) > CVM_PATH_MAXLEN) {
		CVMtraceClassLoading(("CL: filename is too long: "
				      "%s%c%s.%s\n", 
				      dirname, CVM_PATH_LOCAL_DIR_SEPARATOR,
				      classname, CVM_PATH_CLASSFILEEXT));
		continue;
	    }
	
	    sprintf(filename, "%s%c%s.%s", dirname, 
			CVM_PATH_LOCAL_DIR_SEPARATOR,
			classname, CVM_PATH_CLASSFILEEXT);
	    cb = CVMclassLoadFromFile(ee, CVMioNativePath(filename), 
				      dirname, classname);
	} else if (currEntry->type == CVM_CPE_ZIP) {
	    /* Look for class file in specified bootclasspath zip file. */

	    if (strlen(classname) + 1 +
		strlen(CVM_PATH_CLASSFILEEXT) > CVM_PATH_MAXLEN) {
		CVMtraceClassLoading(("CL: filename is too long: "
				      "%s.%s\n", classname,
				      CVM_PATH_CLASSFILEEXT));
		continue;
	    }
	    sprintf(filename, "%s.%s", classname, CVM_PATH_CLASSFILEEXT);
	    cb = CVMclassLoadFromZipFile(ee, filename, classname,
					 currEntry->zip);
	} else {
#ifdef CVM_DEBUG
        printf("classload.c: Could not find currEntry %d. Asserting.\n", currEntry->type);
        printf("classload.c: class name is %s\n", classname);
#endif
	    CVMassert(CVM_FALSE);
	}

	if (cb != NULL || CVMexceptionOccurred(ee)) {
	    free(filename);
	    return cb;
	}
    }

    CVMassert(!CVMexceptionOccurred(ee));
    CVMthrowClassNotFoundException(ee, "%s", classname);
    free(filename);
    return NULL;
}

/*
 * CVMclassLoadFromFile: Creates and returns the cb for the specified
 * class file.
 *
 *   -filename:  full path name of the class file
 *   -classname: name of the class (including package part)
 *
 * Returns NULL if the file cannot be opened or if there are any errors
 * while trying to access the class file. An exception is thrown for
 * all errors other than failing to open the class file.
 */
static CVMClassBlock* 
CVMclassLoadFromFile(CVMExecEnv* ee,
		     const char* filename,
		     const char* dirname,
		     const char* classname)
{
    CVMClassBlock* cb = NULL;
    CVMInt32 codefd;
    CVMInt64 filesize64;
    CVMInt64 filesize64Truncated;
    CVMInt32 filesize32;
    CVMUint8* externalClass = NULL;

    CVMtraceClassLoading(("CL: trying to load class file \"%s\".\n",
			  filename));

    /* open the file */
    codefd = CVMioOpen(filename, O_RDONLY, 0);
    if (codefd <= 0) {
	/* If the file isn't found, then we don't throw an exception
	 * because the caller may still want to search in other directories.
	 */
	return NULL;
    }

    /* get the file size */
    if (CVMioFileSizeFD(codefd, &filesize64) != 0) {
	CVMthrowInternalError(ee, "could not get class file size");
	goto finish;
    }
    filesize32 = CVMlong2Int(filesize64);
    filesize64Truncated = CVMint2Long(filesize32);
    if (CVMlongCompare(filesize64Truncated, filesize64)) {
	CVMthrowIOException(ee, "%s", filename);
	goto finish; /* file is too big */
    }

    /* allocate space for the file */
    externalClass = malloc(filesize32);
    if (externalClass == NULL) {
	CVMthrowOutOfMemoryError(ee, "loading class file");
	goto finish;
    }

    /* read in the file */
    if (CVMioRead(codefd, externalClass, filesize32) != filesize32) {
	CVMthrowIOException(ee, "%s", filename);
	goto finish;
    }
    
    /*
     * Close the file now, or else we may end up with a bunch of open
     * files because CVMclassCreateInternalClass() is recursive.
     */
    if (codefd > 0) {
	CVMioClose(codefd);
	codefd = -1; /* so we don't try to close it again below. */
    }

    /* create the CVMClassBlock */
    cb = CVMclassCreateInternalClass(ee, externalClass, filesize32, NULL,
				     classname, dirname);

 finish:
    if (codefd > 0) {
	CVMioClose(codefd);
    }
    if (externalClass != NULL) {
	free(externalClass);
    }
  
    if (cb != NULL) {
	CVMtraceClassLoading(("CL: class \"%s\" loaded from file \"%s\".\n",
			      classname, filename));
    }
    return cb;
}

/*
 * CVMclassLoadFromZipFile: Creates and returns the cb for the specified
 * class file in the specified zip file..
 *
 *   -filename:  full path name of the class file
 *   -classname: name of the class (including package part). The buffer will
 *               get clobbered during the call to CVMclassLoadFromZipFile(),
 *   -zipfile:   the zip file.
 *
 * Returns NULL if the file cannot be opened or if there are any errors
 * while trying to access the class file. An exception is thrown for
 * all errors other than failing to open the class file.
 */

static CVMClassBlock* 
CVMclassLoadFromZipFile(CVMExecEnv* ee,
			char* filename,
			const char* classname,
			jzfile* zipfile)
{
    jzentry* zipEntry;
    CVMClassBlock* cb = NULL;
    CVMUint8* externalClass = NULL;
    jint fileSize;
    jint fileNameLen;

    CVMtraceClassLoading(("CL: trying to load class file \"%s\" "
			  "from zip file \"%s\"\n",
			  filename, zipfile->name));
 
    /* See if the class file is in this zip file. */
    zipEntry = ZIP_FindEntry(zipfile, filename, &fileSize, &fileNameLen);
    if (zipEntry == NULL) {
	return NULL;
    }

    /* Allocate space for the class file. */
    externalClass = malloc(fileSize);
    if (externalClass == NULL) {
	CVMthrowOutOfMemoryError(ee, "loading class file");
	goto finish;
    }

    /* 
     * Read in the class file. Note that filename will get overwritten
     * with itself. ZIP_ReadEntry requires a buffer to store the file name
     * in even though we already know the file name.
     */
    if (!ZIP_ReadEntry(zipfile, zipEntry, externalClass, filename)) {
	goto finish;
    }


    /* create the CVMClassBlock */
    cb = CVMclassCreateInternalClass(ee, externalClass, fileSize, NULL,
				     classname, zipfile->name);
    if (cb == NULL) {
	goto finish;
    }

 finish:
    if (externalClass != NULL) {
	free(externalClass);
    }

    if (cb != NULL) {
	CVMtraceClassLoading(("CL: class \"%s\" loaded from zip file "
			      "\"%s\".\n",
			      classname, zipfile->name));
    }
    
    return cb;
}

/* 
 * Set CVMglobals.bootClassPathEntries to the boot classpath stored
 * in sun.boot.class.path, which has already been setup by
 * parsing command line arguments, including -Xbootclasspath.
 */

CVMBool
CVMclassBootClassPathInit(JNIEnv *env)
{
    /*
     * Set CVMglobals.bootClassPathString pointing to the 
     * bootclasspath C string. Note that this string will get partly
     * corrupted later on, but will still be pointed to by entries in 
     * CVMglobals.bootClassPathEntries.
     *
     * All out of memory errors are fatal in this routine. We can't do
     * much in the VM unless we get the boot classpath right.
     */
    {
	jclass      systemClass;
	jmethodID   getPropID;
	jobject     path;
	const char* cpath;

	systemClass = CVMcbJavaInstance(CVMsystemClass(java_lang_System));
	getPropID = (*env)->GetStaticMethodID(
            env, systemClass, "getProperty",
	    "(Ljava/lang/String;)Ljava/lang/String;");
	{
	    jstring propertyName = 
		(*env)->NewStringUTF(env, "sun.boot.class.path");
	    if (propertyName == NULL) {
		return CVM_FALSE;
	    }
	    path = (*env)->CallStaticObjectMethod(env, systemClass, getPropID,
						  propertyName);
	    (*env)->DeleteLocalRef(env, propertyName);
	}
	if (path == NULL) {
	    return CVM_FALSE;
	}
	cpath = (*env)->GetStringUTFChars(env, path, NULL);
	if (cpath == NULL) {
	    (*env)->DeleteLocalRef(env, path);
	    return CVM_FALSE;
	}
	CVMassert(strlen(cpath) != 0);
	   
	CVMglobals.bootClassPathString = strdup(cpath);
	if (CVMglobals.bootClassPathString == NULL) {
	    (*env)->ReleaseStringUTFChars(env, path, cpath);
	    (*env)->DeleteLocalRef(env, path);
	    return CVM_FALSE;
	}
	(*env)->ReleaseStringUTFChars(env, path, cpath);
	(*env)->DeleteLocalRef(env, path);
    }
    
    /*
     * Count the number of entries in the class path string and allocate
     * a large enough CVMglobals.bootClassPathEntries for it.
     */
    {
	const char* pathSepPtr = CVMglobals.bootClassPathString;
	while (pathSepPtr != NULL) {
	    CVMglobals.numBootClassPathEntries++;
	    pathSepPtr = strstr(pathSepPtr, CVM_PATH_CLASSPATH_SEPARATOR);
	    if (pathSepPtr != NULL) {
		pathSepPtr += strlen(CVM_PATH_CLASSPATH_SEPARATOR);
	    }
	}
    
	/* allocate space for all of the classpath entries */
	CVMglobals.bootClassPathEntries = (CVMClassPathEntry*)
	    malloc(CVMglobals.numBootClassPathEntries
		       * sizeof(CVMClassPathEntry));
	if (CVMglobals.bootClassPathEntries == NULL) {
	    return CVM_FALSE;
	}
    }
    
    /* initialize the classpath entries */
    {
	char* currPath = CVMglobals.bootClassPathString;
	char* nextPath;
	int bootClassPathIdx = 0;
	while (currPath != NULL && *currPath != '\0') {
	    CVMInt32 fileType;
	    CVMClassPathEntry* currEntry =
		&CVMglobals.bootClassPathEntries[bootClassPathIdx];

	    /* 
	     * Find the start of the next class path item and null terminate
	     * the current class path item.
	     */
	    nextPath = strstr(currPath, CVM_PATH_CLASSPATH_SEPARATOR);
	    if (nextPath != NULL) {
		*nextPath = '\0';
		nextPath += strlen(CVM_PATH_CLASSPATH_SEPARATOR);
	    }

	    /* The empty string is the same as the current directory. */
	    if (*currPath == '\0') {
		currPath = CVM_PATH_CURDIR;
	    }

	    /* Find out what the file type of the class path item is. */
	    fileType = CVMioFileType(currPath);
	    currEntry->path = NULL; /* make error cleanup easier */
	    if (fileType < 0) {
		/* File or directory does not exist */
		CVMdebugPrintf(("\"%s\" is not a valid "
				"bootclasspath zip file or directory.\n",
				currPath));
		currEntry->type = CVM_CPE_INVALID;
	    } else if (fileType == CVM_IO_FILETYPE_DIRECTORY) {
		/* it's a directory */
		CVMtraceClassLoading(("CL: Added directory \"%s\" to the "
				      "bootclasspath\n", currPath));
		currEntry->type = CVM_CPE_DIR;
		currEntry->path = strdup(CVMioNativePath(currPath));
	    } else if (fileType == CVM_IO_FILETYPE_REGULAR) {
		/* it's a zip file */
                char canonicalPath[CVM_PATH_MAXLEN];
		char* msg = NULL;
		jzfile* zip;

		CVMioNativePath(currPath);
		if (CVMcanonicalize(currPath, canonicalPath, CVM_PATH_MAXLEN)
		    < 0) {
		    CVMdebugPrintf(("Bad bootclasspath name: \"%s\"\n",
				    currPath));
		    return CVM_FALSE;
		}
 
		zip = ZIP_Open(canonicalPath, &msg);
		if (zip == NULL) {
		    CVMdebugPrintf(("Opening bootclasspath zip file "
				    "\"%s\" failed.\n", currPath));
		    currEntry->type = CVM_CPE_INVALID;
		} else {
		    CVMtraceClassLoading(("CL: Added zip file \"%s\" to the "
					  "bootclasspath\n", currPath));
		    currEntry->type = CVM_CPE_ZIP;
		    currEntry->path = strdup(currPath); /* canonical form not needed */
		    currEntry->zip = zip;
		}
	    } else {
		/* Neither a file nor a directory, so just ignore it */
		CVMdebugPrintf(("\"%s\" is not a valid "
				"bootclasspath zip file or directory.\n",
				currPath));
		currEntry->type = CVM_CPE_INVALID;
	    }

	    /* on to the next class path item. */
	    bootClassPathIdx++;
	    currPath = nextPath;
	}
    }

    return CVM_TRUE;
}

static void
freeBootClassPathZips()
{
    int bootClassPathIdx;
    for (bootClassPathIdx = 0; 
	 bootClassPathIdx < CVMglobals.numBootClassPathEntries; 
	 bootClassPathIdx++) {
	CVMClassPathEntry* currEntry =
	    &CVMglobals.bootClassPathEntries[bootClassPathIdx];
	
	if (currEntry->type == CVM_CPE_ZIP) {
	    jzfile* zip = currEntry->zip;
	    CVMtraceClassLoading(("CL: Closing boot class path entry %s\n",
				  currEntry->path));
	    ZIP_Close(zip);
	}
	if (currEntry->path != NULL) {
	    free(currEntry->path);
	}
    }
}

void
CVMclassBootClassPathDestroy() 
{
    if (CVMglobals.bootClassPathString != NULL) {
	free(CVMglobals.bootClassPathString);
    }
    if (CVMglobals.bootClassPathEntries != NULL) {
	freeBootClassPathZips();
	free(CVMglobals.bootClassPathEntries);
    }
}

#endif /* CVM_CLASSLOADING */
