/*
 * @(#)verifyformat.c	1.40 00/11/07
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#ifdef JDK12
#include <setjmp.h>
#include <stdlib.h>
#include <string.h>
#else
#include "javavm/include/porting/ansi/setjmp.h"
#include "javavm/include/clib.h"
#endif

#include "jni.h"
#include "jni_util.h"
#include "jvm.h"
#include "javavm/include/verify.h"

typedef unsigned short unicode;

#if CVM_CLASSLOADING

#define JAVA_CLASSFILE_MAGIC                 0xCafeBabe
#define JAVA_MIN_SUPPORTED_VERSION           45
#define JAVA_MAX_SUPPORTED_VERSION           47
#define JAVA_MAX_SUPPORTED_MINOR_VERSION     0

/* align byte code */
#undef ALIGN_UP
#define ALIGN_UP(n,align_grain) (((n) + ((align_grain) - 1)) & ~((align_grain)-1))

/* represent a UTF-8 string as appeared in class files. */
typedef struct {
    unsigned char len_h; /* unaligned */
    unsigned char len_l;
    char bytes[1];
} utf_str;

typedef struct {
    unsigned	length;
    char *	bytes;
}UtfString;

typedef struct {
    UtfString	name;
    UtfString	sig;
}NameAndSig;

#define UTF_LEN(utf) \
  ((((unsigned int)(utf)->len_h) << 8) + (unsigned int)((utf)->len_l))

#define UTF_ASSIGN(stringStruct, utf) \
	( ((stringStruct).length = UTF_LEN(utf)) ,  \
	  ((stringStruct).bytes  = (utf)->bytes) )

typedef union {
    jint i;
    utf_str *s;
} cp_type;

typedef struct {
    const unsigned char *ptr;
    const unsigned char *end_ptr;
    class_size_info *size;
    jmp_buf jump_buffer;

    cp_type *constant_pool;
    unsigned char *type_table;
    int nconstants;

    char *cp_buffer;
    jint err_code;
    const char *class_name;
    char *msg;
    unsigned int msg_len;
    jboolean in_clinit;
   
    int major_version;
    int minor_version;

#undef MEASURE_ONLY
#undef CHECK_RELAXED
#ifdef CVM_TRUSTED_CLASSLOADERS
#define MEASURE_ONLY  (CVM_TRUE)
#define CHECK_RELAXED (CVM_TRUE)
#else
    jboolean measure_only;
    jboolean check_relaxed;
#define MEASURE_ONLY  (context->measure_only)
#define CHECK_RELAXED (context->check_relaxed)
#endif
} CFcontext;

static int utfcmp(utf_str *ustr, char *cstr);
static char * utfncpy(char *buf, int buf_len, utf_str *ustr);

/*
 * The following routine is used to find duplicate name/signature of
 * interfaces, fields, and methods.
 */
static CVMBool  FindDup(NameAndSig array[], int length);

static utf_str *getAsciz(CFcontext *);
static utf_str *getAscizFromCP(CFcontext *, unsigned int i);

static unsigned long get1byte(CFcontext *);
static unsigned long get2bytes(CFcontext *);
static unsigned long get4bytes(CFcontext *);

static void skip2bytes(CFcontext *);
static void skipNbytes(CFcontext *, size_t n);

static utf_str *getUTF8String(CFcontext *);

static void ParseConstantPool(CFcontext *);
static void ParseCode(CFcontext *, utf_str *mb_name, unsigned long args_size);
static void ParseLineTable(CFcontext *, unsigned long code_length);
static void ParseExceptions(CFcontext *);
static void ParseLocalVars(CFcontext *, unsigned long code_length);

static void CFerror(CFcontext *context, char *fmt, ...);
static void CNerror(CFcontext *context, char *name);
static void UVerror(CFcontext *context);
static void CFnomem(CFcontext *context);

static void 
verify_static_constant(CFcontext *context, utf_str *sig, unsigned int index);

static void
verify_constant_pool(CFcontext *context, cp_type* constant_pool,
		     unsigned char * type_table, unsigned int nconstants);

static void 
verify_legal_class_modifiers(CFcontext *context, unsigned int access);
static void 
verify_legal_field_modifiers(CFcontext *context, unsigned int access, 
			     jboolean isIntf);
static void
verify_legal_method_modifiers(CFcontext *context, unsigned int access, 
			      jboolean isIntf, utf_str *name);

static void 
verify_legal_utf8(CFcontext *context, utf_str *str);

/* Argument for verify_legal_name */
enum { LegalClass, LegalField, LegalMethod, LegalVariable };
static void 
verify_legal_name(CFcontext *context, utf_str *name, int type);

static unsigned int
verify_legal_method_signature(CFcontext *context, utf_str *name,
			      utf_str *signature);
static void
verify_legal_field_signature(CFcontext *context, utf_str *name, 
			     utf_str *signature);

static void
verify_constant_entry(CFcontext *context, unsigned int index, int type, 
		      char *msg);

static unsigned int
verify_innerclasses_attribute(CFcontext *context);

static char *
skip_over_fieldname(char *name, jboolean slash_okay, 
		    unsigned int len);
static char *
skip_over_field_signature(char *name, jboolean void_okay,
			  unsigned int len);


/* RETURNS 
 * 0: on success
 * -1: out of memory
 * -2: class format error
 * -3: unsupported version error
 * -4: bad class name error
 *
 * Also fills in the class_size_info structure which contains size 
 * information of verious class components.
 */
#define CF_OK 0
#define CF_NOMEM (-1)
#define CF_ERR (-2)
#define CF_UVERR (-3)
#define CF_BADNAME (-4)
#define LOCAL_BUFFER_SIZE 75

JNIEXPORT jint
CVMverifyClassFormat(const char *class_name,
		     const unsigned char *data,
		     unsigned int data_size,
		     class_size_info *size,
		     char *message_buffer,
		     jint buffer_length,
		     jboolean measure_only,
		     jboolean check_relaxed)
{
    CFcontext context_buf;
    CFcontext *context = &context_buf;
    unsigned int cb_access;
    unsigned int i;
    utf_str *cb_name;
    unsigned int attribute_count;

    unsigned int total;
    NameAndSig name_buffer[LOCAL_BUFFER_SIZE];
    NameAndSig * volatile utf_names = name_buffer;

    memset(context, 0, sizeof(CFcontext));
    memset(size, 0, sizeof(class_size_info));
    
    /* Set up the context */
    context->ptr = data;
    context->end_ptr = data + data_size;

    context->class_name = class_name;
    context->size = size;
    context->msg = message_buffer;
    context->msg_len = buffer_length;
    context->msg[0] = 0;
#ifndef CVM_TRUSTED_CLASSLOADERS
    context->measure_only = measure_only;
    context->check_relaxed = check_relaxed;
#endif

    if (setjmp(context->jump_buffer)) {
	if (context->cp_buffer  != NULL) {
	    free(context->cp_buffer);
	}
	if ((utf_names != name_buffer) && (utf_names != NULL)) {
	    free(utf_names);
	}
	return context->err_code;
    }

    if (get4bytes(context) != JAVA_CLASSFILE_MAGIC) 
	CFerror(context, "Bad magic number");

    context->minor_version = get2bytes(context);
    context->major_version = get2bytes(context);
    if ((context->major_version < JAVA_MIN_SUPPORTED_VERSION) 
	 || (context->major_version > JAVA_MAX_SUPPORTED_VERSION)
	 || ((context->major_version == JAVA_MAX_SUPPORTED_VERSION)
	      && (context->minor_version > JAVA_MAX_SUPPORTED_MINOR_VERSION))) {
	UVerror(context);
    }

    ParseConstantPool(context);

    cb_access = get2bytes(context) & JVM_RECOGNIZED_CLASS_MODIFIERS;

    verify_legal_class_modifiers(context, cb_access);

    /* Get the name of the class */
    i = get2bytes(context);	/* index in constant pool of class */
    cb_name = getAscizFromCP(context, i);
    verify_legal_name(context, cb_name, LegalClass);
    if (cb_name->bytes[0] == JVM_SIGNATURE_ARRAY) {
        CFerror(context, "Bad name");
    }
    if (context->class_name != NULL && 
	utfcmp(cb_name, (char *)(context->class_name)) != 0) {
        char buf[256];
        utfncpy(buf, sizeof(buf), cb_name);
        CNerror(context, buf);
    }

    /* Get the super class name. */
    i = get2bytes(context);	/* index in constant pool of class */
    if (i > 0) {
        utf_str *super_name;
        verify_constant_entry(context, i, JVM_CONSTANT_Class, 
			      "Superclass name");
	super_name = getAscizFromCP(context, i);
	verify_legal_name(context, super_name, LegalClass);
	if (super_name->bytes[0] == JVM_SIGNATURE_ARRAY) {
	    CFerror(context, "Bad superclass name");
	}
    } else {
        if (utfcmp(cb_name, "java/lang/Object") != 0) 
	    CFerror(context, "Bad superclass index");
    }

    total = context->size->interfaces = get2bytes(context);
    if (total > LOCAL_BUFFER_SIZE) {
	utf_names = malloc(total * sizeof(NameAndSig));
        if (!utf_names) {
	    CFnomem(context);
        }
    }
    for (i = 0; i < total; i ++) {
        utf_str *intf_name;
        int intf_index = get2bytes(context);
	verify_constant_entry(context, intf_index, JVM_CONSTANT_Class, 
			      "Interface name");
	intf_name = getAscizFromCP(context, intf_index);
	verify_legal_name(context, intf_name, LegalClass);
	if (intf_name->bytes[0] == JVM_SIGNATURE_ARRAY) {
	    CFerror(context, "Bad interface name");
	}
        UTF_ASSIGN(utf_names[i].name, intf_name);
	utf_names[i].sig.length = 0;
    }
    if ( total > 0 ){
	if (FindDup( utf_names, total) ){
	    CFerror(context, "Repeative interface name");
	}
    }
    if (utf_names != name_buffer) {
	free(utf_names);
	utf_names = name_buffer;
    }

    total = context->size->fields = get2bytes(context);
    if (total > LOCAL_BUFFER_SIZE) {
	utf_names = malloc(total * sizeof(NameAndSig));
        if (!utf_names) {
	    CFnomem(context);
        }
    }
    for (i = 0; i < total; i ++ ) {
        unsigned int j;
        jboolean hasSeenConstantValue = JNI_FALSE;
	unsigned int fb_access = 
	    get2bytes(context) & JVM_RECOGNIZED_FIELD_MODIFIERS;
	utf_str *fb_name = getAsciz(context);
	utf_str *fb_signature = getAsciz(context);

	verify_legal_field_modifiers(context, fb_access,
	(jboolean)((cb_access & JVM_ACC_INTERFACE) ? JNI_TRUE : JNI_FALSE));
	verify_legal_name(context, fb_name, LegalField);
	verify_legal_field_signature(context,  fb_name, fb_signature);

	attribute_count = get2bytes(context);
	for (j = 0; j < attribute_count; j++) {
	    utf_str *attr_name = getAsciz(context);
	    unsigned long length = get4bytes(context);
	    if ((fb_access & JVM_ACC_STATIC) 
		 && utfcmp(attr_name, "ConstantValue") == 0) {
		if (length != 2) 
		    CFerror(context, "Wrong size for VALUE attribute");
		if (hasSeenConstantValue)
		    CFerror(context, "Multiple ConstantValue attributes");
		else
		    hasSeenConstantValue = JNI_TRUE;
		if (fb_access & JVM_ACC_STATIC) {
		    verify_static_constant(context, 
					   fb_signature, 
					   get2bytes(context));
		}
	    } else {
		skipNbytes(context, length);
	    }
	}
	if (fb_access & JVM_ACC_STATIC) {
	    if (fb_signature->bytes[0] == 'D' || 
		fb_signature->bytes[0] == 'J') {
		context->size->staticFields += 2;
	    } else {
		context->size->staticFields += 1;
		if (fb_signature->bytes[0] == 'L' ||
		    fb_signature->bytes[0] == '[') {
		    context->size->staticRefFields += 1;
		}
	    }
	}

        UTF_ASSIGN(utf_names[i].name, fb_name);
        UTF_ASSIGN(utf_names[i].sig,  fb_signature);
    }
    if (FindDup(utf_names, total)) {
        CFerror(context, "Repeative field name/signature");
    }
    if (utf_names != name_buffer) {
	free(utf_names);
	utf_names = name_buffer;
    }

    total = context->size->methods = get2bytes(context);
    if (total > LOCAL_BUFFER_SIZE) {
	utf_names = malloc(total * sizeof(NameAndSig));
        if (!utf_names) {
	    CFnomem(context);
        }
    }
    for (i = 0; i < total; i ++ ) {
        unsigned int j;
        unsigned int args_size;
        jboolean hasSeenCodeAttribute = JNI_FALSE;
	jboolean hasSeenExceptionsAttribute = JNI_FALSE;
	unsigned int mb_access = 
	    get2bytes(context) & JVM_RECOGNIZED_METHOD_MODIFIERS;
	utf_str *mb_name = getAsciz(context);
	utf_str *mb_signature = getAsciz(context);

	if (utfcmp(mb_name, "<clinit>") == 0 &&
	    utfcmp(mb_signature, "()V") == 0) {
	    /* The VM ignores the access flags of <clinit>. We reset the
	     * access flags to avoid future errors in the VM */
	    mb_access = JVM_ACC_STATIC;
	    context->in_clinit = JNI_TRUE;
	} else {
	    verify_legal_method_modifiers(context, 
					  mb_access,
	    (jboolean)((cb_access & JVM_ACC_INTERFACE) ? JNI_TRUE : JNI_FALSE),
					  mb_name);
	}

	verify_legal_name(context, mb_name, LegalMethod);
	args_size = ((mb_access & JVM_ACC_STATIC) ? 0 : 1) + 
	    verify_legal_method_signature(context, mb_name, mb_signature);

	if (args_size > 255) {
	    CFerror(context, "Too many arguments in signature");
	}

        UTF_ASSIGN(utf_names[i].name, mb_name);
        UTF_ASSIGN(utf_names[i].sig,  mb_signature);

	attribute_count = get2bytes(context);
	for (j = 0; j < attribute_count; j++) {
	    utf_str *attr_name = getAsciz(context);
	    if (!utfcmp(attr_name, "Code")) {
		context->size->javamethods++;
	        if (mb_access & (JVM_ACC_NATIVE | JVM_ACC_ABSTRACT)) {
		    CFerror(context,
			    "Code attribute in native or abstract methods");
		}
		if (hasSeenCodeAttribute)
		    CFerror(context, "Multiple Code attributes");
		else
		    hasSeenCodeAttribute = JNI_TRUE;
		ParseCode(context, mb_name, args_size);
	    } else if (utfcmp(attr_name, "Exceptions") == 0) {
		if (hasSeenExceptionsAttribute)
		    CFerror(context, "Multiple Exceptions attributes");
		else
		    hasSeenExceptionsAttribute = JNI_TRUE;
	        ParseExceptions(context);
	    } else {
		unsigned long length = get4bytes(context);
		skipNbytes(context, length);
	    }
	}
	context->in_clinit = JNI_FALSE;
    }
    if (FindDup(utf_names, total)) {
        CFerror(context, "Repeative method name/signature");
    }
    if (utf_names != name_buffer) {
	free(utf_names);
	utf_names = name_buffer;
    }

    /* See if there are class attributes */
    attribute_count = get2bytes(context); 
    {
	jboolean seenSourceAttribute = JNI_FALSE;
	jboolean seenInnerClassesAttribute = JNI_FALSE;
	for (i = 0; i < attribute_count; i++) {
	    utf_str *attr_name = getAsciz(context);
	    unsigned long length = get4bytes(context);
	    if (utfcmp(attr_name, "SourceFile") == 0) {
	        utf_str *source_name;
		if (length != 2) {
		    CFerror(context, "Wrong size for VALUE attribute");
		}
		if (seenSourceAttribute)
		    CFerror(context, "Multiple SourceFile attributes");
		source_name = getAsciz(context);
		seenSourceAttribute = JNI_TRUE;
	    } else if (utfcmp(attr_name, "InnerClasses") == 0) {
		if (seenInnerClassesAttribute) {
		    CFerror(context, "Multiple InnerClasses attributes");
		}
		seenInnerClassesAttribute = JNI_TRUE;
		context->size->innerclasses = 
		    verify_innerclasses_attribute(context);
	    } else {
	        skipNbytes(context, length);
	    }
	}
    }
    if (context->ptr != context->end_ptr && !CHECK_RELAXED)
        CFerror(context, "Extra bytes at the end of the class file");
    free(context->cp_buffer);
    return CF_OK;
}

/*
 * Verify that the "InnerClasses" attribute is well formed, and return
 * the number of inner class records.
 */
static unsigned int
verify_innerclasses_attribute(CFcontext *context) 
{
    unsigned number_of_classes = get2bytes(context);
    unsigned int i = 0;

    if (MEASURE_ONLY) {
	skipNbytes(context, number_of_classes * 8);
	return number_of_classes;
    }

    for (; i < number_of_classes; i++) {
	unsigned short ioff  = get2bytes(context); /* inner_class_info_index */
	unsigned short ooff  = get2bytes(context); /* outer_class_info_index */
	unsigned short inoff = get2bytes(context); /* inner_name_index       */
	unsigned short iacc  = get2bytes(context); /* inner_class_access_flg */
    
	verify_constant_entry(context, ioff, JVM_CONSTANT_Class,
			      "inner_class_info_index");
	
	/* Can be zero for non-members. */
	if (ooff)
	    verify_constant_entry(context, ooff, JVM_CONSTANT_Class,
				  "outer_class_info_index");
	
	/* Can be zero for anonymous. */
	if (inoff)
	    verify_constant_entry(context, inoff, JVM_CONSTANT_Utf8,
				  "inner_name_index");
	
	if (ooff == ioff) 
	    CFerror(context, "Class is both outer and inner class");
	
	verify_legal_class_modifiers(context, iacc);
    }
    return number_of_classes;
}


/* Initialize the static variables in a class structure */
static void
verify_static_constant(CFcontext *context, utf_str *sig, unsigned int index)
{
    unsigned char *type_table;

    if (MEASURE_ONLY)
        return;

    type_table = context->type_table;

    if (index <= 0 || index >= context->size->constants) {
        CFerror(context, "Bad initial value");
    }
    switch (sig->bytes[0]) { 
    case JVM_SIGNATURE_DOUBLE: 
        if (type_table[index] != JVM_CONSTANT_Double) 
	    CFerror(context, "Bad index into constant pool");
	break;
    case JVM_SIGNATURE_LONG: 
        if (type_table[index] != JVM_CONSTANT_Long)
	    CFerror(context, "Bad index into constant pool");
	break;
    case JVM_SIGNATURE_CLASS:
        if (utfcmp(sig, "Ljava/lang/String;") != 0 ||
	    type_table[index] != JVM_CONSTANT_String) {
	    CFerror(context, "Bad string initial value");
	}
	break;
    case JVM_SIGNATURE_BYTE: 
    case JVM_SIGNATURE_CHAR:
    case JVM_SIGNATURE_SHORT:
    case JVM_SIGNATURE_BOOLEAN:
    case JVM_SIGNATURE_INT:  
        if (type_table[index] != JVM_CONSTANT_Integer)
	    CFerror(context, "Bad index into constant pool");
	break;
    case JVM_SIGNATURE_FLOAT:
        if (type_table[index] != JVM_CONSTANT_Float)
	    CFerror(context, "Bad index into constant pool");
	break;
			
    default: 
        CFerror(context, "Unable to set initial value");
    }
}

static void ParseConstantPool(CFcontext *context) 
{
    unsigned int nconstants = get2bytes(context);
    unsigned long cp_request;
    unsigned char *type_table;
    cp_type *constant_pool;
    unsigned int i;
    unsigned int lastNonUtf8Constant = 0;

    if (nconstants < 1) {
	CFerror(context, "Illegal constant pool size");
    }

    /*
     * Measure how much space we need for the constant pool
     * in memory: both the type table and the items.
     */
    cp_request = nconstants * (sizeof(char) + sizeof(cp_type));
    /* Allocate the space */
    context->cp_buffer = malloc(cp_request);
    if (context->cp_buffer == NULL) {
	CFnomem(context);
    }
    memset(context->cp_buffer, 0, cp_request);
    /* Divide up the space. */
    constant_pool = (cp_type *) context->cp_buffer;
    type_table = (unsigned char *)
        (context->cp_buffer + (nconstants * sizeof(cp_type)));

    context->constant_pool = constant_pool;
    context->type_table = type_table;
    context->size->constants = nconstants;

    /* Read in the constant pool */
    for (i = 1; i < nconstants; i++) {
	int type = get1byte(context);
	type_table[i] = type;
	if (type != JVM_CONSTANT_Utf8) {
	    lastNonUtf8Constant = i;
	}
	switch (type) {
	  case JVM_CONSTANT_Utf8: {
	      constant_pool[i].s = getUTF8String(context);
	      break;
	  }
	
	  case JVM_CONSTANT_Class:
	  case JVM_CONSTANT_String:
	      constant_pool[i].i = get2bytes(context);
	      break;

	  case JVM_CONSTANT_Fieldref:
	  case JVM_CONSTANT_Methodref:
	  case JVM_CONSTANT_InterfaceMethodref:
	  case JVM_CONSTANT_NameAndType:
	      constant_pool[i].i = get4bytes(context);
	      break;
	    
	  case JVM_CONSTANT_Integer:
	  case JVM_CONSTANT_Float:
	      get4bytes(context);
	      break;

	  case JVM_CONSTANT_Long:
	  case JVM_CONSTANT_Double: {
	      get4bytes(context);
	      get4bytes(context);
	      i++;		/* increment i for loop, too */
	      lastNonUtf8Constant++;
	      /* A mangled type might cause you to overrun allocated memory */
	      if (i >= nconstants) {
		  CFerror(context, "Invalid constant pool entry");
	      }
	      break;
	  }

	  default:
	      CFerror(context, "Illegal constant pool type");
	}
    }
    context->size->reducedConstants = lastNonUtf8Constant + 1;
    verify_constant_pool(context, constant_pool, type_table, nconstants);
}

static void ParseCode(CFcontext *context, utf_str *mb_name, 
		      unsigned long args_size)
{
    unsigned long attribute_length = get4bytes(context);
    const unsigned char *end_ptr = context->ptr + attribute_length;
    unsigned long attribute_count;
    unsigned long code_length;
    unsigned int i;
    unsigned int maxstack;
    unsigned int nlocals;

    if (context->major_version == 45 && context->minor_version <= 2) { 
	maxstack = get1byte(context);
	nlocals = get1byte(context);
	code_length = get2bytes(context);
    } else { 
	maxstack = get2bytes(context);
	nlocals = get2bytes(context);
	code_length = get4bytes(context);
    }

    if (context->in_clinit) {
        context->size->clinit.code += ALIGN_UP(code_length,sizeof(int));
    } else {
        context->size->main.code += ALIGN_UP(code_length,sizeof(int));
    }

    if (code_length > 65535) {
        CFerror(context, "Code of a method longer than 65535 bytes");
    }

    if (nlocals < args_size)
        CFerror(context, "Arguments can't fit into locals");

    skipNbytes(context, code_length);
    i = get2bytes(context);
    if (context->in_clinit) {
        context->size->clinit.etab += i;
    } else {
        context->size->main.etab += i;
    }
    for (; i > 0; i--) {
        get2bytes(context); /* start_pc */
	get2bytes(context); /* end_pc */	
	get2bytes(context); /* handler_pc */
	get2bytes(context); /* CatchType */
    }

    attribute_count = get2bytes(context);
    for (i = 0; i < attribute_count; i++) {
        utf_str *attr_name = getAsciz(context);
	if (utfcmp(attr_name, "LineNumberTable") == 0) {
	    ParseLineTable(context, code_length);
	} else if (utfcmp(attr_name, "LocalVariableTable") == 0) {
	    ParseLocalVars(context, code_length);
	} else {
	    unsigned long length = get4bytes(context);
	    skipNbytes(context, length);
	}
    }

    if (context->ptr != end_ptr) 
	CFerror(context, "Code segment has wrong length");
}

static void
ParseLineTable(CFcontext *context, unsigned long code_length)
{
    unsigned long attribute_length = get4bytes(context);
    const unsigned char *end_ptr = context->ptr  + attribute_length;
    unsigned long i, table_length;
    table_length = get2bytes(context);
    if (context->in_clinit) {
        context->size->clinit.lnum += table_length;
    } else {
        context->size->main.lnum += table_length;
    }
    for (i = 0; i < table_length; i++) {
        int pc = get2bytes(context);
	skip2bytes(context); /* skip line number */
	if (pc >= code_length)
	    CFerror(context, "Invalid pc in line number table");
    }
    if (context->ptr != end_ptr)
	CFerror(context, "Line number table has wrong length");
    return;
}	

static void
ParseLocalVars(CFcontext *context, unsigned long code_length)
{
    unsigned long attribute_length = get4bytes(context);
    const unsigned char *end_ptr = context->ptr  + attribute_length;
    unsigned long i, table_length;
    table_length = get2bytes(context);
    if (context->in_clinit) {
        context->size->clinit.lvar += table_length;
    } else {
        context->size->main.lvar += table_length;
    }
    for (i = 0; i < table_length; i++) {
        unsigned int pc0 = get2bytes(context);
	unsigned int length = get2bytes(context);
	unsigned int nameoff = get2bytes(context);
	unsigned int sigoff = get2bytes(context);
	unsigned long end_pc;
       	utf_str *name, *sig;
	skip2bytes(context); /* skip slot */
	/* Assign to a unsigned long to avoid overflow */
	end_pc = (unsigned long)pc0 + (unsigned long)length;

	/* We do not check that pc0 indeed points to the 
	 * beginning of an instruction, or slot in fact holds
	 * this variable. That would require changes in the byte 
	 * code verifier.
	 */
	if (pc0 >= code_length || end_pc > code_length)
	    CFerror(context, "Invalid start_pc/length in local var table");

	verify_constant_entry(context, nameoff, JVM_CONSTANT_Utf8, 
			      "Local variable name");
	verify_constant_entry(context, sigoff, JVM_CONSTANT_Utf8,
			      "Local variable signature");

	name = context->constant_pool[nameoff].s;
	sig = context->constant_pool[sigoff].s;
	verify_legal_name(context, name, LegalVariable);
	verify_legal_field_signature(context, name, sig);
    }
    if (context->ptr != end_ptr)
	CFerror(context, "Local variables table has wrong length");
    return;
}	

static void
ParseExceptions(CFcontext *context)
{
    unsigned long attribute_length = get4bytes(context);
    const unsigned char *end_ptr = context->ptr + attribute_length;
    unsigned long nexceptions = get2bytes(context);
    int i;

    if (context->in_clinit) {
        context->size->clinit.excs += nexceptions;
    } else {
        context->size->main.excs += nexceptions;
    }
    for (i = 0; i < nexceptions; i++) {
        unsigned int index = get2bytes(context);
	verify_constant_entry(context, index, JVM_CONSTANT_Class,
			      "Exception name");
    }
    if (context->ptr != end_ptr)
	CFerror(context, "Exceptions attribute has wrong length");
}
	
static unsigned long get1byte(CFcontext *context)
{
    const unsigned char *ptr = context->ptr;
    unsigned long value;
    if (context->end_ptr - ptr < 1) {
	CFerror(context, "Truncated class file");
    }
    value = ptr[0];
    (context->ptr) += 1;
    return value;
}

static unsigned long get2bytes(CFcontext *context)
{
    const unsigned char *ptr = context->ptr;
    unsigned long value;
    if (context->end_ptr - ptr < 2) {
	CFerror(context, "Truncated class file");
    }
    value = (ptr[0] << 8) + ptr[1];
    (context->ptr) += 2;
    return value;
}

static unsigned long get4bytes(CFcontext *context)
{
    const unsigned char *ptr = context->ptr;
    unsigned long value;
    if (context->end_ptr - ptr < 4) {
	CFerror(context, "Truncated class file");
    }
    value = (ptr[0] << 24) + (ptr[1] << 16) + (ptr[2] << 8) + ptr[3];
    (context->ptr) += 4;
    return value;
}

static utf_str *getAsciz(CFcontext *context) 
{
    cp_type *constant_pool = context->constant_pool;
    unsigned char *type_table = context->type_table;
    int nconstants = context->size->constants;

    int value = get2bytes(context);
    if ((value == 0) || 
	(value >= nconstants) || 
	(type_table[value] != JVM_CONSTANT_Utf8))
	CFerror(context, "Illegal constant pool index");
    return constant_pool[value].s;
}

static utf_str *getAscizFromCP(CFcontext *context, unsigned int value) 
{
    cp_type *constant_pool = context->constant_pool;
    unsigned char *type_table = context->type_table;
    int nconstants = context->size->constants;

    if ((value <= 0) || (value >= nconstants) 
	|| (type_table[value] != JVM_CONSTANT_Class))
	CFerror(context, "Illegal constant pool index");
    value = constant_pool[value].i;
    if ((value <= 0) || (value >= nconstants) ||
	 (type_table[value] != JVM_CONSTANT_Utf8))
	CFerror(context, "Illegal constant pool index");
    return constant_pool[value].s;
}

static void skip2bytes(CFcontext *context)
{
    if (context->end_ptr - context->ptr < 2) {
	CFerror(context, "Truncated class file");
    }
    (context)->ptr += 2;
}

static void skipNbytes(CFcontext *context, size_t n)
{
    if (context->end_ptr - context->ptr < n) {
	CFerror(context, "Truncated class file");
    }
    context->ptr += n;
}

static utf_str * getUTF8String(CFcontext* context)
{
    utf_str *result = (utf_str *)context->ptr;
    skip2bytes(context);
    skipNbytes(context, UTF_LEN(result));
    verify_legal_utf8(context, result);
    return result;
}

static void
verify_constant_pool(CFcontext *context, 
		     cp_type *cp, 
		     unsigned char *type_table,
		     unsigned int cp_count)
{
    unsigned int i, type;
    
    if (MEASURE_ONLY)
        return;

    /* Let's make two quick passes over the constant pool. The first one 
     * checks that everything is of the right type. 
     */
    for (i = 1; i < cp_count; i++) {
        switch(type = type_table[i]) {
	case JVM_CONSTANT_String:
	case JVM_CONSTANT_Class: {
	    unsigned int index = cp[i].i;
	    if ((index <= 0) ||
		(index >= cp_count) ||
		(type_table[index] != JVM_CONSTANT_Utf8)) {
	        CFerror(context, "Bad index in constant pool #%d", i);
	    }
	    break;
	}
	
	case JVM_CONSTANT_Fieldref:
	case JVM_CONSTANT_Methodref:
	case JVM_CONSTANT_InterfaceMethodref: 
	case JVM_CONSTANT_NameAndType: {
	    unsigned long index = (unsigned)(cp[i].i);
	    unsigned int key1 = index >> 16;
	    unsigned int key2 = index & 0xFFFF;
	    if (key1 <= 0 || key1 >= cp_count 
		|| key2 <= 0 || key2 >= cp_count) {
	        CFerror(context, "Bad index in constant pool #%d", i);
	    }
	    if (type == JVM_CONSTANT_NameAndType) {
	        if ((type_table[key1] != JVM_CONSTANT_Utf8) ||
		    (type_table[key2] != JVM_CONSTANT_Utf8)) {
		    CFerror(context, "Bad index in constant pool.");
		}
	    } else {
	        if ((type_table[key1] != JVM_CONSTANT_Class) ||
		    (type_table[key2] != JVM_CONSTANT_NameAndType)) {
		    CFerror(context, "Bad index in constant pool #%d", i);
		}
	    }
	    break;
	}
	case JVM_CONSTANT_Utf8:
	case JVM_CONSTANT_Integer:
	case JVM_CONSTANT_Float:
	    break;
	    
	case JVM_CONSTANT_Long:
	case JVM_CONSTANT_Double:
	    if ((i + 1 >= cp_count) || 
		(type_table[i + 1] != 0)) {
	        CFerror(context, 
			"Improper constant pool long/double #%d", i);
	    } else {
	        i++;	
		break;	    
	    }
	    
	default:
	    CFerror(context, "Illegal constant pool type at #%d", i);
	}
    }
    /* The second pass checks the strings are of the right format */
    for (i = 1; i < cp_count; i++) {
	switch(type = type_table[i]) {
	case JVM_CONSTANT_Class: {
	    unsigned int index = cp[i].i;
	    verify_legal_name(context, cp[index].s, LegalClass); 
	    break;
	}
	      
	case JVM_CONSTANT_Fieldref:
	case JVM_CONSTANT_Methodref:
	case JVM_CONSTANT_InterfaceMethodref: {
	    unsigned long index = (unsigned)(cp[i].i);
	    unsigned int name_type_index = index & 0xFFFF;
	    unsigned int name_type_key = cp[name_type_index].i;
	    unsigned int name_index = name_type_key >> 16;
	    unsigned int signature_index = name_type_key & 0xFFFF;
	    utf_str *name = cp[name_index].s;
	    utf_str *signature = cp[signature_index].s;

	    if (type == JVM_CONSTANT_Fieldref) {
	        verify_legal_name(context, name, LegalField);
		verify_legal_field_signature(context, name, signature);
	    } else {
	        verify_legal_name(context, name, LegalMethod);
		verify_legal_method_signature(context, name, signature);
	    }
	    break;
	}
	}
    }
}

static void 
verify_legal_class_modifiers(CFcontext *context, unsigned int access)
{
    if (MEASURE_ONLY)
        return;

    /* This is a workaround for a javac bug. Interfaces are not
     * marked as ABSTRACT.
     */
    if (access & JVM_ACC_INTERFACE) {
        access |= JVM_ACC_ABSTRACT;
    }

    if (access & JVM_ACC_INTERFACE) {
        if ((access & JVM_ACC_ABSTRACT) == 0)
	    goto failed;
	if (access & JVM_ACC_FINAL)
	    goto failed;
    } else {
        if ((access & JVM_ACC_FINAL) && (access & JVM_ACC_ABSTRACT))
	    goto failed;
    }
    return;

 failed:
    CFerror(context, "Illegal class modifiers: 0x%X", access);
}

static void 
verify_legal_field_modifiers(CFcontext *context, unsigned int access,
			     jboolean isIntf)
{
    if (MEASURE_ONLY)
        return;

    if (!isIntf) {
        /* class fields */
        if ((access & JVM_ACC_PUBLIC) && 
	    ((access & JVM_ACC_PROTECTED) || (access & JVM_ACC_PRIVATE)))
	    goto failed;
	if ((access & JVM_ACC_PROTECTED) && 
	    ((access & JVM_ACC_PUBLIC) || (access & JVM_ACC_PRIVATE)))
	    goto failed;
	if ((access & JVM_ACC_PRIVATE) && 
	    ((access & JVM_ACC_PROTECTED) || (access & JVM_ACC_PUBLIC)))
	    goto failed;
	if ((access & JVM_ACC_FINAL) && (access & JVM_ACC_VOLATILE))
	    goto failed;
    } else {
        /* interface fields */
        if (access & ~(JVM_ACC_STATIC | JVM_ACC_FINAL | JVM_ACC_PUBLIC))
	    goto failed;
        if (!(access & JVM_ACC_STATIC) ||
	    !(access & JVM_ACC_FINAL) ||
	    !(access & JVM_ACC_PUBLIC))
	    goto failed;
    }
    return;

 failed:
    CFerror(context, "Illegal field modifiers: 0x%X", access);
}

static void 
verify_legal_method_modifiers(CFcontext *context, unsigned int access, 
			      jboolean isIntf, utf_str *name)
{
    if (MEASURE_ONLY)
        return;

    if (utfcmp(name, "<init>") == 0) {
        /* The STRICT bit is new as of 1.2. */
        if (access & ~(JVM_ACC_PUBLIC
                            | JVM_ACC_PROTECTED
                            | JVM_ACC_PRIVATE
                            | JVM_ACC_STRICT))
	    goto failed;
    } else {
        if (!isIntf) {
	    /* class or instance methods */
	    if (access & JVM_ACC_ABSTRACT) {
	        if ((access & JVM_ACC_FINAL) ||
		    (access & JVM_ACC_NATIVE) /* || 
	     This is commented out until after javac is fixed so that it 
	     rejects abstract synchronized methods.
		    (access & JVM_ACC_SYNCHRONIZED)*/)
		    goto failed;
	    }
	    if ((access & JVM_ACC_PRIVATE) && (access & JVM_ACC_ABSTRACT))
	        goto failed;
	    if ((access & JVM_ACC_STATIC) && (access & JVM_ACC_ABSTRACT))
	        goto failed;		
	} else {
	    /* interface methods */
	    if (!(access & JVM_ACC_ABSTRACT) ||
		!(access & JVM_ACC_PUBLIC))
	        goto failed;	        
	}
    }
    if ((access & JVM_ACC_PUBLIC) && 
	((access & JVM_ACC_PROTECTED) || (access & JVM_ACC_PRIVATE)))
        goto failed;
    if ((access & JVM_ACC_PROTECTED) && 
	((access & JVM_ACC_PUBLIC) || (access & JVM_ACC_PRIVATE)))
        goto failed;
    if ((access & JVM_ACC_PRIVATE) && 
	((access & JVM_ACC_PROTECTED) || (access & JVM_ACC_PUBLIC)))
        goto failed;
    return;

 failed:
    CFerror(context, "Illegal method modifiers: 0x%X", access);
}

static void
verify_legal_utf8(CFcontext *context, utf_str *str)
{
    unsigned int i;
    unsigned int length;
    char *bytes;

    if (MEASURE_ONLY)
        return;

    length = UTF_LEN(str);
    bytes = str->bytes;

    for (i=0; i<length; i++) {
        unsigned int c = bytes[i];
        if (c == 0) /* no embedded zeros */
	    goto failed;
	if (c < 128)
	    continue;
        switch (c >> 4) {
        default:
	    break;

	case 0x8: case 0x9: case 0xA: case 0xB: case 0xF:
	    goto failed;

	case 0xC: case 0xD:	
	    /* 110xxxxx  10xxxxxx */
	    i++;
	    if (i >= length)
	        goto failed;
	    if ((bytes[i] & 0xC0) == 0x80)
	        break;
	    goto failed;

	case 0xE:
	    /* 1110xxxx 10xxxxxx 10xxxxxx */
	    i += 2;
	    if (i >= length)
	        goto failed;
	    if (((bytes[i-1] & 0xC0) == 0x80) &&
		((bytes[i] & 0xC0) == 0x80))
	        break;
	    goto failed;
	} /* end of switch */
    }
    return;
 failed:
    CFerror(context, "Illegal UTF8 string in constant pool");
}

/* Check if the entire second argument consists of a legal fieldname 
 * (or classname, if the third argument is LegalClass). 
 */

static void
verify_legal_name(CFcontext *context, utf_str *name, int type)
{
    jboolean result;
    unsigned int length;
    char *bytes;

    if (MEASURE_ONLY || CHECK_RELAXED)
        return;

    length = UTF_LEN(name);
    bytes = name->bytes;

    if (length > 0) {
        if (bytes[0] == '<') {
	    result = (type == LegalMethod) && 
	        ((utfcmp(name, "<init>") == 0) || 
		 (utfcmp(name, "<clinit>") == 0));
	} else {
	    char *p;
	    if (type == LegalClass && bytes[0] == JVM_SIGNATURE_ARRAY) {
	        p = skip_over_field_signature(bytes, JNI_FALSE, length);
	    } else {
	        p = skip_over_fieldname(bytes, 
		    (jboolean)(type == LegalClass ? JNI_TRUE : JNI_FALSE), 
					length);
	    }
	    result = (p != NULL) && (p - bytes == length);
	}
    } else {
        result = JNI_FALSE;
    }
    if (!result) {
        char buf[100];
	char *thing = (type == LegalField) ? "Field" 
	            : (type == LegalMethod) ? "Method" 
	            : (type == LegalVariable) ? "Variable" : "Class";
	utfncpy(buf, sizeof(buf), name);
	CFerror(context, "Illegal %s name \"%s\"", thing, buf);
    } 
}

/* Check if the entire string consists of a legal field signature */
static void
verify_legal_field_signature(CFcontext *context, utf_str *fieldname, 
			     utf_str *signature) 
{
    char *bytes;
    unsigned int length;
    char *p;

    if (MEASURE_ONLY)
        return;

    bytes = signature->bytes;
    length = UTF_LEN(signature);
    p = skip_over_field_signature(bytes, JNI_FALSE, length);

    if (p == NULL || p - bytes != length) {
        char buf1[100], buf2[100];
	utfncpy(buf1, sizeof(buf1), fieldname);
	utfncpy(buf2, sizeof(buf2), signature);
	CFerror(context, "Field \"%s\" has illegal signature \"%s\"", 
		buf1, buf2);
    }
}


static unsigned int
verify_legal_method_signature(CFcontext *context, utf_str *methodname, 
			      utf_str *signature)
{
    unsigned int args_size;
    char *p;
    unsigned length;
    char *next_p;

    if (MEASURE_ONLY)
        return 0;  /* We return 0 as the args_size when no checking
		    * is necessary. 
		    */

    args_size = 0;
    p = signature->bytes;
    length = UTF_LEN(signature);

    /* The first character must be a '(' */
    if ((length > 0) && (*p++ == JVM_SIGNATURE_FUNC)) {
        length--;
	/* Skip over however many legal field signatures there are */
	while ((length > 0) &&
	       (next_p = skip_over_field_signature(p, JNI_FALSE, length))) {
	    args_size++;
	    if (p[0] == 'J' || p[0] == 'D')
	        args_size++;
	    length -= (next_p - p);
	    p = next_p;
	}
	/* The first non-signature thing better be a ')' */
	if ((length > 0) && (*p++ == JVM_SIGNATURE_ENDFUNC)) {
	    length --;
	    if (UTF_LEN(methodname) > 0 && methodname->bytes[0] == '<') {
		/* All internal methods must return void */
		if ((length == 1) && (p[0] == JVM_SIGNATURE_VOID))
		    return args_size;
	    } else {
		/* Now, we better just have a return value. */
		next_p =  skip_over_field_signature(p, JNI_TRUE, length);
		if (next_p && (length == next_p - p))
		    return args_size;
	    }
	}
    }
    {
        char buf1[100], buf2[100];
	utfncpy(buf1, sizeof(buf1), methodname);
	utfncpy(buf2, sizeof(buf2), signature);
	CFerror(context, "Method \"%s\" has illegal signature \"%s\"", 
		buf1, buf2);
	return 0; /* never reached */
    }
}

static void
verify_constant_entry(CFcontext *context, unsigned int index, int type,
		      char *msg)
{
    cp_type *constant_pool;
    unsigned char *type_table;
    int nconstants;

    if (MEASURE_ONLY)
        return;

    constant_pool = context->constant_pool;
    type_table = context->type_table;
    nconstants = context->size->constants;

    if (index <= 0 || index >= nconstants) 
        CFerror(context, "%s has bad constant pool index", msg);

    if (type_table[index] != type)
        CFerror(context, "%s has bad constant type", msg);
}
#endif /* CVM_CLASSLOADING */

  /* The following tables and code generated using: */
  /* java GenerateCharacter -verbose -c -identifiers -spec UnicodeData-2.1.2.txt -template check_class.c.template -o check_class.c 8 4 4 */
  /* The X table has 256 entries for a total of 256 bytes. */

  static unsigned char X[256] = {
      0,   1,   2,   3,   4,   5,   6,   7,  /* 0x0000 */
      7,   8,   9,  10,  11,  12,  13,  14,  /* 0x0800 */
     15,  16,   7,   7,   7,   7,   7,   7,  /* 0x1000 */
      7,   7,   7,   7,   7,   7,  17,  18,  /* 0x1800 */
     19,  20,   7,   7,   7,   7,   7,   7,  /* 0x2000 */
      7,   7,   7,   7,   7,   7,   7,   7,  /* 0x2800 */
     21,  22,   7,   7,   7,   7,   7,   7,  /* 0x3000 */
      7,   7,   7,   7,   7,   7,   7,   7,  /* 0x3800 */
      7,   7,   7,   7,   7,   7,   7,   7,  /* 0x4000 */
      7,   7,   7,   7,   7,   7,  23,  23,  /* 0x4800 */
     23,  23,  23,  23,  23,  23,  23,  23,  /* 0x5000 */
     23,  23,  23,  23,  23,  23,  23,  23,  /* 0x5800 */
     23,  23,  23,  23,  23,  23,  23,  23,  /* 0x6000 */
     23,  23,  23,  23,  23,  23,  23,  23,  /* 0x6800 */
     23,  23,  23,  23,  23,  23,  23,  23,  /* 0x7000 */
     23,  23,  23,  23,  23,  23,  23,  23,  /* 0x7800 */
     23,  23,  23,  23,  23,  23,  23,  23,  /* 0x8000 */
     23,  23,  23,  23,  23,  23,  23,  23,  /* 0x8800 */
     23,  23,  23,  23,  23,  23,  23,  23,  /* 0x9000 */
     23,  23,  23,  23,  23,  23,  23,  24,  /* 0x9800 */
      7,   7,   7,   7,   7,   7,   7,   7,  /* 0xA000 */
      7,   7,   7,   7,  23,  23,  23,  23,  /* 0xA800 */
     23,  23,  23,  23,  23,  23,  23,  23,  /* 0xB000 */
     23,  23,  23,  23,  23,  23,  23,  23,  /* 0xB800 */
     23,  23,  23,  23,  23,  23,  23,  23,  /* 0xC000 */
     23,  23,  23,  23,  23,  23,  23,  23,  /* 0xC800 */
     23,  23,  23,  23,  23,  23,  23,  25,  /* 0xD000 */
      7,   7,   7,   7,   7,   7,   7,   7,  /* 0xD800 */
      7,   7,   7,   7,   7,   7,   7,   7,  /* 0xE000 */
      7,   7,   7,   7,   7,   7,   7,   7,  /* 0xE800 */
      7,   7,   7,   7,   7,   7,   7,   7,  /* 0xF000 */
      7,  23,  26,  27,  23,  28,  29,  30   /* 0xF800 */
  };

  /* The Y table has 496 entries for a total of 496 bytes. */

  static unsigned char Y[496] = {
      0,   0,   1,   2,   3,   4,   3,   5,  /*   0 */
      0,   0,   6,   7,   8,   9,   8,   9,  /*   0 */
      8,   8,   8,   8,   8,   8,   8,   8,  /*   1 */
      8,   8,   8,   8,   8,   8,   8,  10,  /*   1 */
      8,  11,   0,   0,   0,   8,   8,   8,  /*   2 */
      8,   8,  12,  13,  14,  14,  15,   0,  /*   2 */
     16,  16,  16,  16,  17,   0,  18,  19,  /*   3 */
     20,   8,  21,   8,  22,  23,  24,  25,  /*   3 */
     26,   8,   8,   8,   8,  26,   8,   8,  /*   4 */
     27,   8,   8,   8,  28,   8,  29,  30,  /*   4 */
      0,   0,   0,   3,   8,  31,   3,   8,  /*   5 */
     11,  32,  33,  34,  35,   8,   5,  36,  /*   5 */
      0,   0,   3,   5,  37,  38,   2,  39,  /*   6 */
      8,   8,   8,  40,  22,  41,  42,   2,  /*   6 */
      0,   0,   0,   0,   0,   0,   0,   0,  /*   7 */
      0,   0,   0,   0,   0,   0,   0,   0,  /*   7 */
     43,   8,   8,  44,  45,  46,  47,   0,  /*   8 */
     48,  49,  50,  51,  52,  53,  47,  25,  /*   8 */
     54,  49,  50,  55,  56,  57,  58,  59,  /*   9 */
     60,  21,  50,  61,  62,   0,  63,   0,  /*   9 */
     48,  49,  50,  64,  65,  66,  67,   0,  /*  10 */
     68,  69,  70,  71,  72,  73,  74,   0,  /*  10 */
     75,  24,  50,  76,  77,  78,  67,   0,  /*  11 */
     79,  24,  50,  76,  77,  80,  67,   0,  /*  11 */
     79,  24,  50,  81,  82,  73,  67,   0,  /*  12 */
      0,   0,   0,   0,   0,   0,   0,   0,  /*  12 */
      3,   8,  22,  83,  84,   2,   0,   0,  /*  13 */
     85,  86,  87,  88,  89,  90,   0,   0,  /*  13 */
      0,  91,   2,  92,  93,   8,  94,  32,  /*  14 */
     95,  96,  45,  97,   0,   0,   0,   0,  /*  14 */
      0,   0,   0,   0,   0,   0,   0,   0,  /*  15 */
      0,   0,   8,   8,  98,   8,   8,  99,  /*  15 */
      8,   8,   8,   8,   8, 100,   8,   8,  /*  16 */
      8,   8, 101,   8,   8,   8,   8,  94,  /*  16 */
      8,   8,   8,   8,   8,   8,   8,   8,  /*  17 */
      8, 102,   8,   8,   8,   8,   8,  94,  /*  17 */
      8, 103,   8,   8, 103, 104,   8, 105,  /*  18 */
      8,   8,   8, 106, 107, 108, 109, 107,  /*  18 */
      0,   0,   0, 110, 111,   0,   0, 110,  /*  19 */
      0,   0, 109,   0,   0, 112, 113,   0,  /*  19 */
    114, 115, 116, 117,   0,   0,   8,   8,  /*  20 */
     36,   0,   0,   0,   0,   0,   0,   0,  /*  20 */
    118,   0, 119, 120,   3,   8,   8,   8,  /*  21 */
      8, 121,   3,   8,   8,   8,   8, 122,  /*  21 */
    123,   8, 109,   3,   8,   8,   8,   8,  /*  22 */
     22,   0,   0,   0,   0,   0,   0,   0,  /*  22 */
      8,   8,   8,   8,   8,   8,   8,   8,  /*  23 */
      8,   8,   8,   8,   8,   8,   8,   8,  /*  23 */
      8,   8,   8,   8,   8,   8,   8,   8,  /*  24 */
      8,   8,  98,   0,   0,   0,   0,   0,  /*  24 */
      8,   8,   8,   8,   8,   8,   8,   8,  /*  25 */
      8,   8,  25,   0,   0,   0,   0,   0,  /*  25 */
      8,   8, 105,   0,   0,   0,   0,   0,  /*  26 */
      0,   0,   0,   0,   0,   0,   0,   0,  /*  26 */
     99, 124,  50, 125, 126,   8,   8,   8,  /*  27 */
      8,   8,   8,  14,   0, 127,   8,   8,  /*  27 */
      8,   8,   8, 105,   0,   8,   8,   8,  /*  28 */
      8, 128,   8,   8,  11,   0,   0, 102,  /*  28 */
      0,   0, 129, 130, 131,   0, 132, 133,  /*  29 */
      8,   8,   8,   8,   8,   8,   8, 109,  /*  29 */
      1,   2,   3,   4,   3,   5, 134,   8,  /*  30 */
      8,   8,   8,  22, 135, 136, 137,   0   /*  30 */
  };

  /* The A table has 2208 entries for a total of 552 bytes. */

  static unsigned long A[138] = {
    0x00000000,  /*   0 */
    0x00000300,  /*   1 */
    0x00055555,  /*   2 */
    0xFFFFFFFC,  /*   3 */
    0xC03FFFFF,  /*   4 */
    0x003FFFFF,  /*   5 */
    0x00300FF0,  /*   6 */
    0x00300C00,  /*   7 */
    0xFFFFFFFF,  /*   8 */
    0xFFFF3FFF,  /*   9 */
    0xFFF00FFF,  /*  10 */
    0x0000FFFF,  /*  11 */
    0x0003FFFF,  /*  12 */
    0xFFC3FFFF,  /*  13 */
    0x0000000F,  /*  14 */
    0x000003FF,  /*  15 */
    0x55555555,  /*  16 */
    0x00000555,  /*  17 */
    0x00000005,  /*  18 */
    0x00300000,  /*  19 */
    0xF33F3000,  /*  20 */
    0xFFFFFFCF,  /*  21 */
    0x3FFFFFFF,  /*  22 */
    0x33303FFF,  /*  23 */
    0xFFFFFFF3,  /*  24 */
    0x000000FF,  /*  25 */
    0xF3FFFFFC,  /*  26 */
    0x0000154F,  /*  27 */
    0x03C3C3FF,  /*  28 */
    0xF0FFFFFF,  /*  29 */
    0x000F0FFF,  /*  30 */
    0x000C3FFF,  /*  31 */
    0x55555554,  /*  32 */
    0x55555545,  /*  33 */
    0x45455555,  /*  34 */
    0x00000114,  /*  35 */
    0x0000003F,  /*  36 */
    0x557FFFFF,  /*  37 */
    0x00000015,  /*  38 */
    0xFFFFFFFD,  /*  39 */
    0x3FF0FFFF,  /*  40 */
    0x41555CFF,  /*  41 */
    0x05517D55,  /*  42 */
    0xFFFFFC54,  /*  43 */
    0x5D0FFFFF,  /*  44 */
    0x05555555,  /*  45 */
    0xFFFF0154,  /*  46 */
    0x5555505F,  /*  47 */
    0xC3FFFC54,  /*  48 */
    0xFFFFFFC3,  /*  49 */
    0xFFF3FFFF,  /*  50 */
    0x510FF033,  /*  51 */
    0x05414155,  /*  52 */
    0xCF004000,  /*  53 */
    0xC03FFC10,  /*  54 */
    0x510F3CF3,  /*  55 */
    0x05414015,  /*  56 */
    0x33FC0000,  /*  57 */
    0x55555000,  /*  58 */
    0x000003F5,  /*  59 */
    0xCCFFFC54,  /*  60 */
    0x5D0FFCF3,  /*  61 */
    0x05454555,  /*  62 */
    0x55555003,  /*  63 */
    0x5D0FF0F3,  /*  64 */
    0x05414055,  /*  65 */
    0xCF005000,  /*  66 */
    0x5555500F,  /*  67 */
    0xF03FFC50,  /*  68 */
    0xF33C0FF3,  /*  69 */
    0xF03F03C0,  /*  70 */
    0x500FCFFF,  /*  71 */
    0x05515015,  /*  72 */
    0x00004000,  /*  73 */
    0x55554000,  /*  74 */
    0xF3FFFC54,  /*  75 */
    0x500FFCFF,  /*  76 */
    0x05515155,  /*  77 */
    0x00001400,  /*  78 */
    0xF3FFFC50,  /*  79 */
    0x30001400,  /*  80 */
    0x500FFFFF,  /*  81 */
    0x05515055,  /*  82 */
    0xC01555F7,  /*  83 */
    0x15557FFF,  /*  84 */
    0x0C33C33C,  /*  85 */
    0xFFFCFF00,  /*  86 */
    0x3CF0CCFC,  /*  87 */
    0x0D4555F7,  /*  88 */
    0x055533FF,  /*  89 */
    0x0F055555,  /*  90 */
    0x00050000,  /*  91 */
    0x50044400,  /*  92 */
    0xFFFCFFFF,  /*  93 */
    0x000FFFFF,  /*  94 */
    0x00555155,  /*  95 */
    0x55544555,  /*  96 */
    0x00045554,  /*  97 */
    0x00000FFF,  /*  98 */
    0x00003FFF,  /*  99 */
    0xC00FFFFF,  /* 100 */
    0xFFFF003F,  /* 101 */
    0x00FFFFFF,  /* 102 */
    0x0FFF0FFF,  /* 103 */
    0xCCCCFFFF,  /* 104 */
    0x0FFFFFFF,  /* 105 */
    0x33FFF3FF,  /* 106 */
    0x03FFF3F0,  /* 107 */
    0x00FFF0FF,  /* 108 */
    0x03FFFFFF,  /* 109 */
    0xC0000000,  /* 110 */
    0x00000003,  /* 111 */
    0x01555555,  /* 112 */
    0x00000004,  /* 113 */
    0xFFF0C030,  /* 114 */
    0x0FFF0CFF,  /* 115 */
    0xFFF33300,  /* 116 */
    0x0003FFCF,  /* 117 */
    0x0000CC00,  /* 118 */
    0x555FFFFC,  /* 119 */
    0x00000FFC,  /* 120 */
    0x3FD403FF,  /* 121 */
    0x3F3FFFFF,  /* 122 */
    0xFFFFFC00,  /* 123 */
    0xD000FFC0,  /* 124 */
    0x33FF3FFF,  /* 125 */
    0xFFFFF3CF,  /* 126 */
    0xFFFFFFC0,  /* 127 */
    0xFFFFFFF0,  /* 128 */
    0x00000055,  /* 129 */
    0x000003C0,  /* 130 */
    0xFC000000,  /* 131 */
    0x000C0000,  /* 132 */
    0xFFFFF33F,  /* 133 */
    0xFFFFF000,  /* 134 */
    0xFFF0FFF0,  /* 135 */
    0x03F0FFF0,  /* 136 */
    0x00003C0F   /* 137 */
  };

  /* In all, the character property tables require 1304 bytes. */

/*
 * This code mirrors Character.isJavaIdentifierStart.  It determines whether
 * the specified character is a legal start of a Java identifier as per JLS.
 *
 * The parameter ch is the character to be tested; return 1 if the 
 * character is a letter, 0 otherwise.
 */
#define isJavaIdentifierStart(ch) (((A[Y[(X[ch>>8]<<4)|((ch>>4)&0xF)]]>>((ch&0xF)<<1))&3) & 0x2)

/*
 * This code mirrors Character.isJavaIdentifierPart.  It determines whether
 * the specified character is a legal part of a Java identifier as per JLS.
 * 
 * The parameter ch is the character to be tested; return 1 if the
 * character is a digit, 0 otherwise.
 */  
#define isJavaIdentifierPart(ch) (((A[Y[(X[ch>>8]<<4)|((ch>>4)&0xF)]]>>((ch&0xF)<<1))&3) & 0x1)


/* Take pointer to a string.  Skip over the longest part of the string that
 * could be taken as a fieldname.  Allow '/' if slash_okay is JNI_TRUE.
 *
 * Return a pointer to just past the fieldname.  Return NULL if no fieldname
 * at all was found, or in the case of slash_okay being true, we saw
 * consecutive slashes (meaning we were looking for a qualified path but
 * found something that was badly-formed). 
 */
static char *
skip_over_fieldname(char *name, jboolean slash_okay, 
		    unsigned int length)
{
    char *p;
    unicode ch;
    unicode last_ch = 0;
    /* last_ch == 0 implies we are looking at the first char. */
    for (p = name; p != name + length; last_ch = ch) {
	char *old_p = p;
	ch = *p;
	if (ch < 128) {
            p++;
	    /* quick check for ascii */
	    if ((ch >= 'a' && ch <= 'z') ||
		(ch >= 'A' && ch <= 'Z') ||
		(last_ch && ch >= '0' && ch <= '9')) {
	        continue;
	    }
	} else {
	    char *tmp_p = p;
	    ch = next_utf2unicode((const char**)&tmp_p);
	    p = tmp_p;
	    if (isJavaIdentifierStart(ch) || (last_ch && isJavaIdentifierPart(ch))) {
		continue;
	    }
	}

	if (slash_okay && ch == '/' && last_ch) {
	    if (last_ch == '/') {
		return 0;	/* Don't permit consecutive slashes */
	    }
	} else if (ch == '_' || ch == '$') {
	} else {
	    return last_ch ? old_p : 0;
	}
    }
    return last_ch ? p : 0;
}

/* Take pointer to a string.  Skip over the longest part of the string that
 * could be taken as a field signature.  Allow "void" if void_okay.
 *
 * Return a pointer to just past the signature.  Return NULL if no legal
 * signature is found.
 */

static char *
skip_over_field_signature(char *name, jboolean void_okay, 
			  unsigned int length)
{
    for (;length > 0;) {
	switch (name[0]) {
            case JVM_SIGNATURE_VOID:
		if (!void_okay) return 0;
		/* FALL THROUGH */
            case JVM_SIGNATURE_BOOLEAN:
            case JVM_SIGNATURE_BYTE:
            case JVM_SIGNATURE_CHAR:
            case JVM_SIGNATURE_SHORT:
            case JVM_SIGNATURE_INT:
            case JVM_SIGNATURE_FLOAT:
            case JVM_SIGNATURE_LONG:
            case JVM_SIGNATURE_DOUBLE:
		return name + 1;

	    case JVM_SIGNATURE_CLASS: {
		/* Skip over the classname, if one is there. */
		char *p;
		length--;
		p = skip_over_fieldname(name + 1, JNI_TRUE, length);
		/* The next character better be a semicolon. */
		if (p && p - name - 1 > 0 && p[0] == ';') 
		    return p + 1;
		return 0;
	    }
	    
	    case JVM_SIGNATURE_ARRAY: 
		/* The rest of what's there better be a legal signature.  */
		name++;
		length--;
		void_okay = JNI_FALSE;
		break;

	    default:
		return 0;
	}
    }
    return 0;
}

#ifdef CVM_CLASSLOADING

static void
CFerror(CFcontext *context, char *format, ...)
{
    va_list args;
    int n = 0;
    va_start(args, format);
    if (context->class_name)
        n = jio_snprintf(context->msg, context->msg_len, "%s (", 
			 context->class_name);
    n += jio_vsnprintf(context->msg + n, context->msg_len - n, format, args);
    if (context->class_name)
        jio_snprintf(context->msg + n, context->msg_len - n, ")");
    va_end(args);
    context->err_code = CF_ERR;
    longjmp(context->jump_buffer, 1);
}

static void
CNerror(CFcontext *context, char *name)
{
    jio_snprintf(context->msg, context->msg_len, "%s (wrong name: %s)", 
		 context->class_name, name);
    context->err_code = CF_BADNAME;
    longjmp(context->jump_buffer, 1);
}

static void
UVerror(CFcontext *context)
{
    int n = 0;
    if (context->class_name)
        n = jio_snprintf(context->msg, context->msg_len, "%s (", 
			 context->class_name);
    n += jio_snprintf(context->msg + n, context->msg_len - n, "Unsupported major.minor version %d.%d", context->major_version, context->minor_version);
    if (context->class_name)
        jio_snprintf(context->msg + n, context->msg_len - n, ")");
    context->err_code = CF_UVERR;
    longjmp(context->jump_buffer, 1);
}

static void
CFnomem(CFcontext *context)
{
    context->err_code = CF_NOMEM;
    longjmp(context->jump_buffer, 1);
}

#endif /* CVM_CLASSLOADING */

/* Used in java/lang/Class.c */
/* Determine if the specified name is legal
 * UTF name for a classname.
 *
 * Note that this routine expects the internal form of qualified classes:
 * the dots should have been replaced by slashes.
 */
JNIEXPORT jboolean
VerifyClassname(char *name, jboolean allowArrayClass) 
{ 
    unsigned int length = strlen(name);
    char *p;
    if (length > 0 && name[0] == JVM_SIGNATURE_ARRAY) {
	if (!allowArrayClass) {
	    return JNI_FALSE;
	} else { 
	    /* Everything that's left better be a field signature */
	    p = skip_over_field_signature(name, JNI_FALSE, length);
	}
    } else {
	/* skip over the fieldname.  Slashes are okay */
	p = skip_over_fieldname(name, JNI_TRUE, length);
    }
    return (p != 0 && p - name == length);
}

/*
 * Translates '.' to '/'.  Returns JNI_TRUE is any / were present.
 */
JNIEXPORT jboolean
VerifyFixClassname(char *name)
{
    char *p = name;
    jboolean slashesFound = JNI_FALSE;
    
    while (*p != '\0') {
	if (*p == '/') {
	    slashesFound = JNI_TRUE;
	    p++;
	} else if (*p == '.') {
	    *p++ = '/';
	} else {
            next_utf2unicode((const char **)&p);
        }
    }
    return slashesFound;
}

/*
 * Translates '/' to '.'.  Returns JNI_TRUE is any . were present.
 */
jboolean
TranslateFromVMClassName(char *name)
{
    char *p = name;
    jboolean dotsFound = JNI_FALSE;
    
    while (*p != '\0') {
	if (*p == '.') {
	    dotsFound = JNI_TRUE;
	    p++;
	} else if (*p == '/') {
	    *p++ = '.';
	} else {
            next_utf2unicode((const char **)&p);
        }
    }
    return dotsFound;
}

#ifdef CVM_CLASSLOADING

static int utfcmp(utf_str *ustr, char *cstr)
{
    int i;
    int length = UTF_LEN(ustr);
    char *bytes = ustr->bytes;
    for (i=0; i < length; i++) {
        int c1 = bytes[i];
	int c2 = cstr[i];
	if (c1 > c2) /* this handles the case when c2 is zero. */
	    return 1;
	if (c1 < c2)
	    return -1;
    }
    if (cstr[i] == 0)
        return 0;
    return -1;
}

static char * utfncpy(char *buf, int buf_len, utf_str *ustr)
{
    int i;
    int length = UTF_LEN(ustr);
    char *bytes = ustr->bytes;
    for (i = 0; i < length && i < buf_len - 1; i++) {
	buf[i] = bytes[i];
    }
    buf[i] = 0;
    return buf;
}

/*
 * this routine compares two names
 */
static int uucmpDoWork( NameAndSig *name1, NameAndSig *name2 )
{
    int nameLength1 = name1->name.length;
    int nameLength2 = name2->name.length;
    int sigLength1;
    int sigLength2;
    int result;

    if (nameLength1 < nameLength2) {
        return -1;
    }
    if (nameLength1 > nameLength2) {
        return 1;
    }
    /* 
     * The lengths of the names are equal. Compare.
     */
    result = strncmp(name1->name.bytes, name2->name.bytes, nameLength1);
    if (result != 0) {
	return result;
    }
    /* Names are equal. Now check for signatures. */
    sigLength1  = name1->sig.length;
    sigLength2  = name2->sig.length;

    if (sigLength1 < sigLength2){
        return -1;
    } else if (sigLength1 > sigLength2){
        return 1;
    } else if (sigLength1 == 0) {
	return 0; /* No signature, just names. Equal. */
    }
    /* Names are equal and signatures of the same length.
     * Compare signatures 
     */
    return strncmp(name1->sig.bytes, name2->sig.bytes, sigLength1);
}

static int uucmp(const void* p1, const void* p2)
{
    return uucmpDoWork((NameAndSig*)p1, (NameAndSig*)p2);
}

/*
 * The function returns CVM_TRUE if it finds any duplicates,
 * CVM_FALSE otherwise. 
 * After sorting the name/signature array, it will do
 * comparison of adjacent entries looking for duplicates.
 * The sorting is done in place, and complexity is O(n log n);
 * Note:
 *   For interfaces, the signature will just be a null pointer.
 */
static CVMBool
FindDup(NameAndSig array[], int length)
{
    int i;

    qsort(array, length, sizeof(NameAndSig), uucmp);
    for (i = 0; i < length - 1; i++) {
        if (uucmpDoWork(&array[i], &array[i+1]) == 0) {
            return CVM_TRUE;
        }
    }
    return CVM_FALSE;
}

#endif /* CVM_CLASSLOADING */
