/*
 * @(#)weakrefs.c	1.19 00/11/09
 *
 * Copyright 1994-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * This software is the proprietary information of Sun Microsystems, Inc.
 * Use is subject to license terms.
 *
 */

#include "javavm/include/defs.h"
#include "javavm/include/objects.h"
#include "javavm/include/directmem.h"
#include "javavm/include/indirectmem.h"
#include "javavm/include/interpreter.h"
#include "javavm/include/stackwalk.h"
#include "javavm/include/utils.h"
#include "javavm/include/string.h"

#include "javavm/include/preloader_impl.h"

/*
 * Four queues for the four strengths of weak references
 */
static CVMObject* discoveredSoftRefs;
static CVMObject* discoveredWeakRefs;
static CVMObject* discoveredFinalRefs;
static CVMObject* discoveredPhantomRefs;

/*
 * Clear all the discovered weak reference queues.
 */
static void
CVMweakrefClearDiscoveredQueues()
{
    discoveredSoftRefs = NULL;
    discoveredWeakRefs = NULL;
    discoveredFinalRefs = NULL;
    discoveredPhantomRefs = NULL;
}

/*
 * Get a pointer to the Reference.pending static variable.
 */
static CVMObject** 
CVMweakrefGetPendingQueue()
{
    return (CVMObject**)(&CVMglobals.java_lang_ref_Reference_pending->r);
}

#ifdef CVM_DEBUG
static char*
CVMweakrefQueueName(CVMObject** queue)
{
    if (queue == &discoveredSoftRefs) {
	return "discovered SoftReference";
    } else if (queue == &discoveredWeakRefs) {
	return "discovered WeakReference";
    } else if (queue == &discoveredFinalRefs) {
	return "discovered FinalReference";
    } else if (queue == &discoveredPhantomRefs) {
	return "discovered PhantomReference";
    } else if (queue == CVMweakrefGetPendingQueue()) {
	return "Reference.pending";
    } else {
	return "<UNKNOWN QUEUE>";
    }
}
#endif

/*
 * Enqueue an item in a queue (discovered or pending). These queues are
 * to preserve the invariant that a weak reference object is active
 * iff its next field is NULL. So they make sure the list is terminated
 * with a self-referring weakref object.
 */
static void
CVMweakrefEnqueue(CVMObject** queue, CVMObject* weakRef)
{
    CVMObject* nextItem;
    if (*queue == NULL) {
	/*
	 * Don't allow for NULL next fields. Otherwise the weak
	 * reference may be discovered multiple times by
	 * CVMobjectWalkRefsWithSpecialHandling().
	 */
	nextItem = weakRef;
    } else {
	nextItem = *queue;
    }
    *queue = weakRef;
    CVMweakrefField(weakRef, next) = nextItem;
    CVMtraceWeakrefs(("WR: Enqueued weak object 0x%x in the %s queue\n",
		      weakRef, CVMweakrefQueueName(queue)));
}

/*
 * Discover and enqueue new weak reference object.
 */
void
CVMweakrefDiscover(CVMExecEnv* ee, CVMObject* weakRef)
{
    CVMObject* referent = CVMweakrefField(weakRef, referent);
    CVMClassBlock* weakCb;
    /*
     * Nothing to do. We must have encountered this weak reference object
     * before its constructor has had a chance to run.
     */
    if (referent == NULL) {
	return;
    }

    CVMtraceWeakrefs(("WR: Discovered new weakref object 0x%x "
		      "of class %C, "
		      "referent 0x%x\n",
		      weakRef,
		      CVMobjectGetClass(weakRef),
		      referent));
    weakCb = CVMobjectGetClass(weakRef);
    /*
     * Enqueue the discovered ref in the right queue for its strength.
     *
     * Time-space tradeoff: If we had a 'reference strength'
     * field in the class block, we can make the following subclass
     * checks unnecessary. 
     */
    if (CVMisSubclassOf(ee, weakCb,
			CVMsystemClass(java_lang_ref_SoftReference))) {
	CVMweakrefEnqueue(&discoveredSoftRefs, weakRef);
    } else if (!CVMlocalExceptionOccurred(ee) && 
               CVMisSubclassOf(ee, weakCb,
			       CVMsystemClass(java_lang_ref_WeakReference))) {
	CVMweakrefEnqueue(&discoveredWeakRefs, weakRef);
    } else if (!CVMlocalExceptionOccurred(ee) && 
               CVMisSubclassOf(ee, weakCb,
			       CVMsystemClass(java_lang_ref_FinalReference))) {
	CVMweakrefEnqueue(&discoveredFinalRefs, weakRef);
    } else if (!CVMlocalExceptionOccurred(ee) && 
               CVMisSubclassOf(ee, weakCb,
			       CVMsystemClass(java_lang_ref_PhantomReference))){
	CVMweakrefEnqueue(&discoveredPhantomRefs, weakRef);
    }
}

/*
 * Enqueue a weak reference in the Reference.pending queue.
 */
static void
CVMweakrefEnqueuePending(CVMObject* thisRef)
{
    /*
     * One optimization here would be to check whether
     * the queue element of this guy is ReferenceQueue.NULL. If it is
     * the ReferenceHandler thread will do nothing with this reference
     * anyway, so we might as well discard it instead of enqueueing it.
     */
    CVMweakrefEnqueue(CVMweakrefGetPendingQueue(), thisRef);
    /* So that CVMgcStopTheWorldAndGCSafe() does notify the ReferenceHandler
     * thread */
    CVMglobals.referenceWorkTODO = CVM_TRUE;
}

/*
 * A function to handle a weak reference instance with a dying referent.
 *
 * An instance of this decides whether to clear the referent or not,
 * and whether to enqueue the weak reference in the pending queue or not.
 *
 * Return values:
 *
 *     CVM_TRUE:  The referent has to be marked.
 *         This happens with dying Finalizer and PhantomReference instances,
 *         as well as SoftReference instances we elect not to clear.
 *
 *     CVM_FALSE: The referent is cleared.
 *         This happens with dying WeakReference instances,
 *         as well as SoftReference instances we elect to clear.
 *
 */
typedef CVMBool (*CVMWeakrefHandleDying)(CVMObject* refAddr);

static CVMBool
CVMweakrefClearConditional(CVMObject* thisRef)
{
    /*
     * There should be some criteria here that determines
     * whether the soft ref is dying.  
     * But there isn't.
     */
    CVMBool clearReferent = CVM_TRUE;
    if (clearReferent) {
	CVMtraceWeakrefs(("WR: CVMweakrefClearConditional: "
			  "Clearing ref=%x (class %C)\n",
			  thisRef, CVMobjectGetClass(thisRef)));
	CVMweakrefField(thisRef, referent) = NULL;
	CVMweakrefEnqueuePending(thisRef);
	return CVM_FALSE; /* Really dead */
    } else {
	CVMtraceWeakrefs(("WR: CVMweakrefClearConditional: "
			  "NOT clearing ref=%x (class %C)\n",
			  thisRef, CVMobjectGetClass(thisRef)));
	/* We elected not to clear this weak reference. Make it active
	   again. */
	CVMweakrefField(thisRef, next) = NULL;
	return CVM_TRUE; /* Scan referent */
    }
}

static CVMBool
CVMweakrefClearUnconditional(CVMObject* thisRef)
{
    /* The weakref is dead, and we want to clear the referent
     * unconditionally.
     */
    CVMtraceWeakrefs(("WR: CVMweakrefClearUnonditional: "
		      "Clearing ref=%x (class %C)\n",
		      thisRef, CVMobjectGetClass(thisRef)));
    CVMweakrefField(thisRef, referent) = NULL;
    CVMweakrefEnqueuePending(thisRef);
    return CVM_FALSE; /* No referent to scan */
}

static CVMBool
CVMweakrefReferentKeep(CVMObject* thisRef)
{
    /* The weakref is dead, but we want to keep the referent alive
     * until the weakref object is cleared from the pending queue by
     * the ReferenceHandler thread. 
     */
    CVMtraceWeakrefs(("WR: CVMweakrefReferentKeep: "
		      "Enqueueing ref=%x (class %C), keeping referent\n",
		      thisRef, CVMobjectGetClass(thisRef)));
    CVMweakrefEnqueuePending(thisRef);
    return CVM_TRUE; /* Scan referent */
}


/*
 * A function to handle a weakref queue element. A weakref queue may
 * be one of the discovered references queues, or the pending
 * references queue.
 */
typedef void (*CVMWeakrefQueueCallback)(CVMObject* refAddr, void* data);

static void
CVMweakrefIterateQueue(CVMObject** queue,
		       CVMWeakrefQueueCallback qcallback,
		       void* data)
{
    CVMObject* thisRef = *queue;
    CVMObject* nextRef;
    if (thisRef == NULL) {
	return; /* Nothing to do */
    }
    while (CVM_TRUE) {
	nextRef = CVMweakrefField(thisRef, next);
	(*qcallback)(thisRef, data);
	if (thisRef == nextRef) {
	    break;
	} else {
	    thisRef = nextRef;
	}
    }
}

typedef struct CVMWeakrefHandlingParams {
    CVMRefLivenessQueryFunc isLive;
    void*                   isLiveData;
    CVMWeakrefHandleDying   handleDying;
    CVMRefCallbackFunc      transitiveScanner;
    void*                   transitiveScannerData;
} CVMWeakrefHandlingParams;

static void
CVMweakrefDiscoveredQueueCallback(CVMObject* thisRef, void* qdata)
{
    CVMWeakrefHandlingParams* params =
	(CVMWeakrefHandlingParams*)qdata;
    CVMRefLivenessQueryFunc isLive = params->isLive;
    CVMWeakrefHandleDying   handleDying = params->handleDying;
    CVMRefCallbackFunc      transitiveScanner = params->transitiveScanner;
    void* isLiveData = params->isLiveData;
    void* transitiveScannerData = params->transitiveScannerData;

    if ((*isLive)(CVMweakrefFieldAddr(thisRef, referent), isLiveData)) {
	/* Out of the queue, into the active state again */
	CVMassert(CVMweakrefField(thisRef, referent) != NULL);
	CVMweakrefField(thisRef, next) = NULL;
	/*
	 * Keep this referent and its children alive
	 */
	CVMtraceWeakrefs(("WR: Referent 0x%x of 0x%x, "
			  "is NOT dying\n",
			  CVMweakrefField(thisRef, referent),
			  thisRef));
	(*transitiveScanner)(CVMweakrefFieldAddr(thisRef, referent),
			     transitiveScannerData);
    } else {
	/*
	 * GC says this referent is dying. Now let's see if
	 * handling this weak reference object causes us to keep
	 * the referent alive or not.  
	 */
	if ((*handleDying)(thisRef)) {
	    CVMassert(CVMweakrefField(thisRef, referent) != NULL);
	    (*transitiveScanner)(CVMweakrefFieldAddr(thisRef, referent),
				 transitiveScannerData);
	    CVMtraceWeakrefs(("WR: Kept referent 0x%x of 0x%x, "
			      "class %C at address 0x%x\n",
			      CVMweakrefField(thisRef, referent),
			      thisRef, CVMobjectGetClass(thisRef),
			      CVMweakrefFieldAddr(thisRef, referent)));
	} else {
	    CVMassert(CVMweakrefField(thisRef, referent) == NULL);
	    CVMassert(CVMweakrefField(thisRef, next) != NULL);
	}
    }
}

/*
 * Scan through and process weak references with non-strongly referenced
 * referents.
 */

/* Forward declaration */
static void
CVMweakrefHandlePendingQueue(CVMRefCallbackFunc callback, void* data);

/*
 * Clear JNI weak references, and transitively mark remaining live
 * ones.
 */
static void
CVMweakrefHandleJNIWeakRef(CVMObject** refPtr, void* qdata)
{
    CVMWeakrefHandlingParams* params =
	(CVMWeakrefHandlingParams*)qdata;
    CVMRefLivenessQueryFunc isLive = params->isLive;
    void* isLiveData = params->isLiveData;

    /*
     * No transitive scan required here.
     * Just delete non-live entries.
     */
    if (!(*isLive)(refPtr, isLiveData)) {
	*refPtr = NULL; /* Clear JNI weak ref */
	CVMtraceWeakrefs(("WR: Cleared JNI weak global ref "
			  "at address 0x%x\n", refPtr));
    }
}

static void
CVMweakrefHandleJNIWeakRefs(CVMWeakrefHandlingParams* params)
{
    CVMstackScanGCRootFrameRoots(&CVMglobals.weakGlobalRoots,
				 CVMweakrefHandleJNIWeakRef,
				 params);
}

void
CVMweakrefProcessNonStrong(CVMExecEnv* ee,
			   CVMRefLivenessQueryFunc isLive, void* isLiveData,
			   CVMRefCallbackFunc transitiveScanner,
			   void* transitiveScannerData,
			   CVMGCOptions* gcOpts)
{
    CVMWeakrefHandlingParams params;
    params.isLive = isLive;
    params.isLiveData = isLiveData;
    params.transitiveScanner = transitiveScanner;
    params.transitiveScannerData = transitiveScannerData;

    params.handleDying = CVMweakrefClearConditional;
    CVMweakrefIterateQueue(&discoveredSoftRefs,
			   CVMweakrefDiscoveredQueueCallback,
			   &params);
    params.handleDying = CVMweakrefClearUnconditional;
    CVMweakrefIterateQueue(&discoveredWeakRefs,
			   CVMweakrefDiscoveredQueueCallback,
			   &params);
    params.handleDying = CVMweakrefReferentKeep;
    CVMweakrefIterateQueue(&discoveredFinalRefs,
			   CVMweakrefDiscoveredQueueCallback,
			   &params);
    params.handleDying = CVMweakrefReferentKeep;
    CVMweakrefIterateQueue(&discoveredPhantomRefs,
			   CVMweakrefDiscoveredQueueCallback,
			   &params);

    /*
     * Doing the JNI pass here makes the JNI weak refs the fifth weak
     * ref strength.
     *
     * See bug #4126360 for details.
     */
    params.handleDying = NULL; /* Special case. No java.lang.ref.*
                                  objects here */
    CVMweakrefHandleJNIWeakRefs(&params);

    /*
     * Done classifying discovered references. Clear for next
     * discovery phase.
     */
    CVMweakrefClearDiscoveredQueues();

    /*
     * Now that we've processed all the discovered queues, it is time
     * for marking the pending queue, which might include items from the
     * previous iteration, as well as the currently enqueued items.
     * 
     * The 'next' and 'referent' fields need special scanning, since
     * they were skipped in the discovery phase.
     */
    CVMweakrefHandlePendingQueue(transitiveScanner, transitiveScannerData);

    CVMweakrefUpdate(ee, transitiveScanner, transitiveScannerData, gcOpts);
}

/*
 * Make sure that the 'next' fields of uncleared weak references are
 * properly scanned.  All the ones we care about are in the pending
 * queue.  
 *
 * There are two possibilities here:
 *
 * 1) The weak reference objects were discovered in their final
 * location. This would typically be the case in a semispace copying
 * GC. In that case, there is no need to update the next fields, since
 * these already point to the right objects.
 *
 * 2) The weak reference objects were discovered in their old
 * locations. This would typically be the case in a mark-compact like
 * GC. In that case, we need to update the next fields in the pending
 * queue to point to the new addresses.
 *
 * How can we avoid calling this for case #1?
 */
typedef struct CVMWeakrefUpdateParams {
    CVMRefCallbackFunc      callback;
    void*                   data;
} CVMWeakrefUpdateParams;

static void
CVMweakrefUpdateCallback(CVMObject* thisRef, void* qdata)
{
    CVMWeakrefUpdateParams* params = (CVMWeakrefUpdateParams*)qdata;
    CVMRefCallbackFunc callback = params->callback;
    void*              data = params->data;
    CVMObject**        refPtr;   
    refPtr = CVMweakrefFieldAddr(thisRef, next);
    CVMassert(*refPtr != NULL);
    (*callback)(refPtr, data);
    refPtr = CVMweakrefFieldAddr(thisRef, referent);
    if (*refPtr != NULL) {
	(*callback)(refPtr, data);
    }
}

static void
CVMweakrefHandlePendingQueue(CVMRefCallbackFunc callback, void* data)
{
    CVMWeakrefUpdateParams params;
    CVMObject** pendingQueue = CVMweakrefGetPendingQueue();
    if (*pendingQueue != NULL) {
	/* First handle the static Reference.pending, in case it just
	   got filled in */
	(*callback)(pendingQueue, data);
	/* And now the rest of the items */
	params.callback = callback;
	params.data = data;
	CVMweakrefIterateQueue(pendingQueue,
			       CVMweakrefUpdateCallback, &params);
    }
}

void
CVMweakrefUpdate(CVMExecEnv* ee,
		 CVMRefCallbackFunc refUpdate, void* updateData,
		 CVMGCOptions* gcOpts)
{
    /*
     * Update live JNI weak global refs
     */
    CVMstackScanRoots(ee, &CVMglobals.weakGlobalRoots,
		      refUpdate, updateData, gcOpts);
}

void 
CVMweakrefInit()
{
    CVMweakrefClearDiscoveredQueues();
    /*
     * Make sure that referent and next are the first two fields of
     * Reference.java. That makes the "map-shifting" possible
     * in CVMobjectWalkRefsAux().
     */
    CVMassert(CVMoffsetOfjava_lang_ref_Reference_referent == FIELD_OFFSET(0));
    CVMassert(CVMoffsetOfjava_lang_ref_Reference_next == FIELD_OFFSET(1));
}
